import os
from numpy import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.optimize import leastsq
from scipy.interpolate import interp1d
from matplotlib.backends.backend_pdf import PdfPages


"""

Inputs:  	Call fit() with a list of fequencies, magnitudes and phases along with a fit function and a guess.
Returns:	Fitted values for the guess parameters.

"""


def amplitude_fit(fData, mData, varsGuess, func):
    e = lambda v, f, y: abs(abs(func(f,v))-y)
    popt, pcov = leastsq(e, varsGuess, args = (fData, mData), maxfev=1000)
    
    #plt.plot(fData, mData, 'b.')
    #plt.plot(fData, abs(func(fData, popt)), 'r')
    #plt.plot(fData, abs(func(fData, varsGuess)), 'g')
    #plt.show()
    
    return popt
    
def phase_fit(fData, pData, varsGuess, func):
    e = lambda v, f, y: abs(np.angle(func(f,v))-y)
    popt, pcov = leastsq(e, varsGuess, args = (fData,pData), maxfev=1000)
    
    #plt.plot(fData, pData, 'b.')
    #plt.plot(fData, np.angle(func(fData, popt)), 'r')
    #plt.plot(fData, np.angle(func(fData, varsGuess)), 'g')
    #plt.show()
    
    return popt
    
def complex_fit(fData, mData, pData, varsGuess, func):
    e = lambda v, f, y: abs(func(f,v)-y)
    y = [mData[i]*np.exp(1j*pData[i]) for i in range(len(mData))]
    popt, pcov, infodict, mesg, ier = leastsq(e, varsGuess, args = (fData,y), maxfev=1000, full_output=True)
    if (len(y) > len(varsGuess)) and pcov is not None:
        s_sq = (e(popt, fData, y)**2).sum()/(len(y)-len(varsGuess))
        pcov = pcov * s_sq
    else:
        pcov = eye(len(varsGuess))*0
    return popt, pcov

def fit(fData, mData, pData, func, Guess):
    varsPhase = phase_fit(fData, pData, Guess, func)
    varsAmp = amplitude_fit(fData, mData, Guess, func)
    newGuess = [(varsPhase[0]+varsAmp[0])/2, (varsPhase[1]+varsAmp[1])/2, varsAmp[2], varsPhase[3]]
    varsComplex, pcov = complex_fit(fData, mData, pData, newGuess, func)
    return varsComplex, pcov
	
"""

Inputs:  	Call get_parms() with a list of fequencies, magnitudes and phases.
Returns:	Array with guesses for:  Q, f0, resonancePhase, diameter, and globalPhase

"""

def findIndex(array, target):
    diff = 1e9
    index1 = 0
    index2 = 0
    for i in range(len(array)):
        if (np.abs(target - array[i]) < diff):
            index2 = index1
            index1 = i
            diff = np.abs(target - array[i])
    return [index1, index2]
    
def meanedF(fdata, indexes):
    fs = [fdata[i] for i in indexes]
    return np.mean(fs)
    
def findAmplitude(mdata, index):
    numpoints = 21
    return np.median([
        mdata[index + i - int(numpoints/2)] for i in range(numpoints)])

def david_get_params(fData, mData, pData):
    cDataU     = [mData[i]*np.cos(pData[i]) + mData[i]*np.sin(pData[i]) * 1j
                 for i in range(len(mData))]
    cData     = np.array(cDataU)
    f0Index   = mData.index(np.max(mData))
    ampF0     = findAmplitude(mData, f0Index)
    f0Index   = findIndex(mData, ampF0)[0]
    f0        = fData[f0Index]
    phi       = np.angle(cData[f0Index])
    
    
    cData     = np.subtract(cData, 1)
    cData     = np.subtract(cData, .5 * ampF0)
    
    angleToF0 = np.angle(cData[f0Index])
    cDataRot  = np.multiply(cData, np.exp(-1 * 1j * angleToF0) )
    
    cDataRotP = np.angle(cDataRot)
    
    piOn2Index = findIndex(cDataRotP, np.pi * ( 1.0/2.0))
    threePiOn2 = findIndex(cDataRotP, np.pi * (-1.0/2.0))
    
    deltaF     = np.abs(meanedF(fData, piOn2Index) - meanedF(fData, threePiOn2))
    
    Q  = 0
    if deltaF != 0:
        Q = f0/ deltaF
    
    r = .5 * ampF0
    if r > 1e4:
        r = 1
        
    print cDataRotP[f0Index]
    
    fitc = r * np.exp(1j * np.linspace(0,2*np.pi,len(fData)))
    
    # fig = plt.figure()
    # fig.subplots_adjust(wspace = .4)
    # fig.subplots_adjust(wspace = .4)
    
    print (piOn2Index[0], threePiOn2[0])
    
    # fig.add_subplot(221)
    # plt.plot(fData, mData, 'b.')
    ## plt.plot(fData, np.abs(fitc), 'r')
    # plt.plot(fData[piOn2Index[0]], mData[piOn2Index[0]], 'g^')
    # plt.plot(fData[threePiOn2[0]], mData[threePiOn2[0]], 'm^')
    # plt.plot([fData[piOn2Index[0]], fData[threePiOn2[0]]], 
             # [mData[piOn2Index[0]], mData[threePiOn2[0]]], 'g')
    # plt.plot(fData[f0Index], mData[f0Index], 'ks')

    # fig.add_subplot(222)
    # plt.plot(fData, pData, 'b.')
    ## plt.plot(fData, np.angle(fitc), 'r')
    # plt.plot(fData[piOn2Index[0]], pData[piOn2Index[0]], 'g^')
    # plt.plot(fData[threePiOn2[0]], pData[threePiOn2[0]], 'm^')
    
    # fig.add_subplot(223, aspect='equal')
    # plt.plot(np.real(cDataRot), np.imag(cDataRot), 'b.')
    # plt.plot(np.real(cDataRot[piOn2Index[0]]),
             # np.imag(cDataRot[piOn2Index[0]]), 'g^') 
    # plt.plot(np.real(cDataRot[threePiOn2[0]]),
             # np.imag(cDataRot[threePiOn2[0]]), 'm^')
    # plt.plot(np.real(cDataRot[f0Index]), np.imag(cDataRot[f0Index]), 'ks')
    # plt.plot(np.real(fitc), np.imag(fitc), 'r')
    
    # fig.add_subplot(224, aspect='equal')
    # plt.plot(np.real(cDataU), np.imag(cDataU), 'b.')
    # plt.plot(np.real(cDataU[piOn2Index[0]]),
             # np.imag(cDataU[piOn2Index[0]]), 'g^') 
    # plt.plot(np.real(cDataU[threePiOn2[0]]),
             # np.imag(cDataU[threePiOn2[0]]), 'm^')
    # plt.plot(np.real(cDataU[f0Index]), np.imag(cDataU[f0Index]), 'ks')
    
    # fitc = np.add(fitc, 1 + .5 * ampF0)
    # fitc = np.multiply(fitc, np.exp(1j * angleToF0)) 
    
    # plt.plot(np.real(fitc), np.imag(fitc), 'r')
    
    ## plt.show()
    
    return [Q, f0, Q/(2*r), phi]
    
    

def get_parms(fData, mData, pData, func):
	x = [mData[i]*np.cos(pData[i]) for i in range(len(mData))]
	y = [mData[i]*np.sin(pData[i]) for i in range(len(mData))]
	xc, yc, r = fit_circle(x,y)
	offSet = np.arctan2(yc,xc-1)
	
	f0 = fData[mData.index(np.max(mData))]
	
	x = np.subtract(x,xc)
	y = np.subtract(y,yc)
	p = [np.arctan2(y[n],x[n]) for n in range(len(x))]
	p = phase_correction(fData, p)
	p = list(np.subtract(p, offSet))
	
	find = lambda a,l:min(l,key=lambda x:abs(x-a))
	Pos3DbF = fData[p.index(find(np.pi/2,p))]
	Neg3DbF = fData[p.index(find(-1*np.pi/2,p))]
	deltaF = abs(Pos3DbF - Neg3DbF)

	if deltaF != 0:
		Q = f0 / deltaF
	else:
		Q = 0
		
	if r > 1e4:
		r = 1

	return [Q, f0, Q/(2*r), offSet]
	
def centered_phase_fit(fData, pData, qGuess, fGuess, phiGuess):
	func = lambda f, Q, f0, phi: phi + 2*np.arctan(2*Q*(1-f/f0))
	popt, pcov = curve_fit(func, fData, pData, [qGuess, fGuess, phiGuess], maxfev = 1000)
	vals = [func(x, popt[0], popt[1], popt[2]) for x in fData]
	#plt.plot(fData, vals)
	#plt.plot(fData, pData)
	#plt.show()
	#offset = np.mean(np.subtract(pData, vals))
	return popt[0], popt[1]

def phase_correction(fData, pData):
    for i in range(len(pData))[:-1]:
        if abs(pData[i] - pData[i+1]) >= np.pi and pData[i] < pData[i+1]:
            pData[i+1] = pData[i+1] - 2*np.pi
        if abs(pData[i] - pData[i+1]) >= np.pi and pData[i] > pData[i+1]:
            pData[i+1] = pData[i+1] + 2*np.pi
    
    find = lambda a,l:min(l,key=lambda x:abs(x-a))
    
    delta = find(np.mean(pData),[-2*np.pi, -1*np.pi, 0, 1*np.pi,2*np.pi])
    pData = np.subtract(pData,delta)

    return list(pData)

def fit_circle(x, y):
	x = np.r_[array(x)]
	y = np.r_[array(y)]
	x_m = mean(x)
	y_m = mean(y)
	def calc_R(xc, yc):
		return sqrt((x-xc)**2 + (y-yc)**2)
	def f_2b(c):
		Ri = calc_R(*c)
		return Ri - Ri.mean()
	def Df_2b(c):
		xc, yc     = c
		df2b_dc    = empty((len(c), x.size))
		Ri = calc_R(xc, yc)
		df2b_dc[ 0] = (xc - x)/Ri
		df2b_dc[ 1] = (yc - y)/Ri
		df2b_dc       = df2b_dc - df2b_dc.mean(axis=1)[:, newaxis]
		return df2b_dc
	center_estimate = x_m, y_m
	center_2b, ier = leastsq(f_2b, center_estimate, Dfun=Df_2b, col_deriv=True)
	xc_2b, yc_2b = center_2b
	Ri_2b        = calc_R(xc_2b, yc_2b)
	R_2b         = Ri_2b.mean()
	residu_2b    = sum((Ri_2b - R_2b)**2)
	residu2_2b   = sum((Ri_2b**2-R_2b**2)**2)
	return xc_2b, yc_2b, R_2b 	
	
	
def getFits_dv(fData, mData, pData, fitFunc, parameters):
    ''' mData NOT in dB
    '''
    deltaF = np.subtract(fData, parameters[1])
    deltaF = np.divide(deltaF, 1000)
    
    reData  = [mData[i]*np.cos(pData[i]) for i in range(len(mData))]
    imData  = [mData[i]*np.sin(pData[i]) for i in range(len(mData))]
    mDatadB = 20*np.log10(mData)
    
    # print "getFits' parameters: " + str(parameters)
    
    fitValues = fitFunc(fData, parameters)
    reFit  = real(fitValues)
    imFit  = imag(fitValues)
    mFitdB = 20*log10(abs(fitValues))
    pFit = angle(fitValues)
    
    if 0:
        fig = plt.figure()
        fig.subplots_adjust(wspace = 0.4)
        fig.subplots_adjust(hspace = 0.4)
        
        fig.add_subplot(221)
        plt.plot(deltaF, mDatadB, 'b.')
        plt.plot(deltaF, mFitdB, 'r')
        plt.xlabel('Delta F (KHz)')
        plt.ylabel('Magnitude of in dB')
        
        fig.add_subplot(222)
        plt.plot(deltaF, pData, 'b.')
        plt.plot(deltaF, pFit, 'r')
        plt.xlabel('Delta F (KHz)')
        plt.ylabel('Phase of in radians')
        
        fig.add_subplot(223, aspect='equal')
        plt.plot(reData, imData, 'b.')
        plt.plot(reFit, imFit, 'r')
        plt.xlabel('Real of ')
        plt.ylabel('Imag of ')
    
    plt.show()
    
    return [deltaF, mDatadB, mFitdB, pFit, reData, imData, reFit, imFit ]
"""

Inputs:  	Call fit_plot with lists of fequencies, magnitudes and phases to be plotted.
			Additionally include a fit function and parameters.
			
Returns:	None

"""
	
	
def fit_plot(fData, mData, pData, fitFunc, parameters, dataName, dataDir, dataFile, show, save):
    fig = plt.figure()
    fig.subplots_adjust(wspace = 0.4)
    fig.subplots_adjust(hspace = 0.4)
    #freqRange = np.divide(fData,1000)
    deltaF = np.subtract(fData, parameters[1])
    deltaF = np.divide(deltaF, 1000)
    
    reData = [mData[i]*np.cos(pData[i]) for i in range(len(mData))]
    imData = [mData[i]*np.sin(pData[i]) for i in range(len(mData))]
    ampData = 20*np.log10(mData)
    phaseData = pData
    
    # reFit = [np.real(fitFunc(x, parameters)) for x in fData]
    # imFit = [np.imag(fitFunc(x, parameters)) for x in fData]
    # ampFit = [20*np.log10(abs(fitFunc(x, parameters))) for x in fData]
    # phaseFit = [np.angle(fitFunc(x, parameters)) for x in fData]
    
    fitValues = fitFunc(fData, parameters)
    reFit = real(fitValues)
    imFit = imag(fitValues)
    ampFit = 20*log10(abs(fitValues))
    phaseFit = angle(fitValues)
    
    a = fig.add_subplot(221)	
    plt.plot(deltaF, ampData, 'b.')
    plt.plot(deltaF, ampFit, 'r')
    plt.xlabel('Delta F (KHz)')
    plt.ylabel('Magnitude of '+dataName+' in dB')
    
    b = fig.add_subplot(222)	
    plt.plot(deltaF, phaseData, 'b.')
    plt.plot(deltaF, phaseFit, 'r')
    plt.xlabel('Delta F (KHz)')
    plt.ylabel('Phase of '+dataName+' in radians')
    
    c = fig.add_subplot(223, aspect='equal')	
    plt.plot(reData, imData, 'b.')
    plt.plot(reFit, imFit, 'r')
    plt.xlabel('Real of '+dataName)
    plt.ylabel('Imag of '+dataName)
    
    a = fig.add_subplot(224)
    plt.text(-0.2,0.9,'Res Freq = ' + str(parameters[1]*1e-9)[:6] + 'GHz')
    plt.gca().axison = False
    
    if show:
        plt.show()
        plt.close()
        
    if save:
        if not os.path.exists("Data\\"+dataDir):
            os.makedirs("Data\\"+dataDir)

        pp = PdfPages("Data\\"+dataDir+"\\"+dataFile+".pdf")
        pp.savefig(fig)
        pp.close()	
    
    plt.close()
    
def data_plot(fData, mData, pData, dataName):
	fig = plt.figure()
	fig.subplots_adjust(wspace = 0.4)
	fig.subplots_adjust(hspace = 0.4)
	freqRange = np.divide(fData,1000)
	
	reData = [mData[i]*np.cos(pData[i]) for i in range(len(mData))]
	imData = [mData[i]*np.sin(pData[i]) for i in range(len(mData))]
	ampData = 20*np.log10(mData)
	phaseData = pData
	
	a = fig.add_subplot(221)	
	plt.plot(freqRange, ampData, 'b.')
	plt.xlabel('f (Hz)')
	plt.ylabel('Magnitude of '+dataName+' in dB')
	
	b = fig.add_subplot(222)	
	plt.plot(freqRange, phaseData, 'b.')
	plt.xlabel('f (Hz)')
	plt.ylabel('Phase of '+dataName+' in radians')
	
	c = fig.add_subplot(223, aspect='equal')	
	plt.plot(reData, imData, 'b.')
	plt.xlabel('Real of '+dataName)
	plt.ylabel('Imag of '+dataName)
	
	a = fig.add_subplot(224)
	plt.text(-0.2,0.9,'Note:  All Plot Frequencies Are 1e-3')
	plt.gca().axison = False
	
	plt.show()
	plt.close()        