# PowerSweep
# Performs powersweep of resonators on adr with pna


# Origional Author:    Charles Neill, Anthony Megrant 
# Revisions:           David Low

# Note:  All functions with lots of comments and triple quote documentation
#        were written from stratch by the revisional author.

DEBUG    = False # use test_PNAScan?
DEBUG01  = False

import fitS21v4_2       as fit
import resonatorModule  as rm
import davidutil        as util
import powerSweepUtil   as psu
import registryhandle   as rh
import datavaultSaver   as dvs
reload(psu)

if DEBUG:
    import test_PNAScan as PNAScan
else:
    import PNAScan

reload(PNAScan)
reload(fit)
reload(rm)
reload(util)
reload(psu)
reload(rh)
reload(dvs)
    
####
import os
import sys
import time
util.safePyle()

import labrad
import numpy as np
#from numpy import *
#from pylab import *
from scipy.interpolate  import interp1d

from labrad.units       import V, mV, us, ns, GHz, MHz, Hz, rad, dBm
from labrad.units       import Value

from pyle.workflow      import switchSession
from pyle.util          import sweeptools as st

import matplotlib.pyplot as plt
import traceback
####

# Error handling
LOGFILE      = 'powersweep'
MAXPOWER     = 1 * dBm

# Recursive calls to powersweep handling


def plotraw(fdata, mdata, pdata, redata, imdata, fitparams, func):
    ''' for Debug ONLY '''
    
    
    fitvalues = func(np.array(fdata), fitparams)
    refit = np.real(fitvalues)
    imfit = np.imag(fitvalues)
    mfit  = np.abs(fitvalues)
    pfit  = np.angle(fitvalues)
    
    fig = plt.figure()
    fig.subplots_adjust(wspace = 0.4)
    fig.subplots_adjust(hspace = 0.4)
    
    fig.add_subplot(221)
    plt.plot(fdata, mdata, 'b.')
    plt.plot(fdata, mfit, 'r')
    plt.xlabel('Frequency')
    plt.ylabel('Magnitude')

    fig.add_subplot(222)
    plt.plot(fdata,pdata, 'b.')
    plt.plot(fdata,pfit, 'r')
    plt.xlabel('Frequency')
    plt.ylabel('Phase in Radians')

    fig.add_subplot(223, aspect='equal' )
    plt.plot(redata, imdata, 'b.')
    plt.plot(refit, imfit, 'r')
    plt.xlabel('real')
    plt.ylabel('imag')
    
    plt.show()

def get_cold_calibration(fData, mData, pData, crop, cropOffset, showCal):
    
    changed = 1
    while changed:
        pData, changed = rm.phase_correction(pData)
    
    crop = 1
    indices = np.concatenate((np.arange(0, len(fData)/2 - crop - cropOffset), 
                np.arange(len(fData)/2 + crop - cropOffset, len(fData))))
    indices = [0,2]
    fCal = [fData[i] for i in indices]
    mCal = [mData[i] for i in indices]
    pCal = [pData[i] for i in indices]
    
    if 0: # old code
        # magParms = polyfit(fCal[len(mCal)/2-5:len(mCal)/2+5], 
        #                       mCal[len(mCal)/2-5:len(mCal)/2+5], 1)
        # a = magParms[0]
        # A = magParms[1]
        # mFit = [A+a*x for x in fCal]
        
        # N = 20
        # magParms = polyfit(fCal, mCal, N)
        # magPoly = lambda x: sum([magParms[N - i]*x**i for i in range(N+1)])
        # mFit = [magPoly(x) for x in fCal]

        # phaseParms = polyfit(fCal, pCal, 1)
        # p = phaseParms[0]
        # P = phaseParms[1]
        # pFit = [P+p*x for x in fCal]
        pass
    
    mCalInt = interp1d(fCal, mCal)
    pCalInt = interp1d(fCal, pCal)
    
    
    if showCal:
        py.plot(fCal, mCalInt(fCal), 'r')
        py.plot(fData, mData, 'b.')
        show()
        
        py.plot(fCal, pCalInt(fCal), 'r')
        py.plot(fData, pData, 'b.')
        show()
    
    return mCalInt, pCalInt
    
def scan(sample, center, span, numPoints, averages, currentPower, 
         bandwidth = 100 * Hz, name = 'resScan', 
         pna = '5242A', dir = ['ResXX', 'power'], numChunks = 1
        ):
    NUMSCAN = 5
    for i in range(NUMSCAN):
        try:
            [fdata, mdata, 
            pdata, setnum] = _scan(sample, center, span, numPoints, averages, currentPower,
                  bandwidth = bandwidth, name = name, pna = pna, dir = dir,
                  numChunks = numChunks)
        except Exception as e:
            util.printErr("Cannot scan!!!" + str(traceback.format_exc()), 
                          [LOGFILE], "powerSweep.scan")
            if i == NUMSCAN - 1:
                util.printErr("Cannot PNA Scan.  Error")
                raise
            time.sleep(20)
        else:
            i = NUMSCAN + 1
            return [fdata, mdata, pdata, setnum]
    
    return [-1, -1, -1, -1]
    
    
def _scan(sample, center, span, numPoints, averages, currentPower, 
         bandwidth = 100 * Hz, name = 'resScan', 
         pna = '5242A', dir = ['ResXX', 'power'], numChunks = 1
        ):
    #pna = '5230A' #from directly above
    #'PNA N5242A'
    pna = '5242A'
    #pna = '8720es'
    #pna = 'E8364B'
    
    print "SCAN ON!!!!!!"
    if dir == ['ResXX','power']:
        print "OH NOES!"
    #span = sample['span'] 
    start = center - span/2.0
    stop  = center + span/2.0
    print start
    print stop
    print 'span:  ',span
    print 'span2:  ',sample['span']

    PNAScan.pna_scan(sample, 
                     name, 
                     st.r[start:stop:(stop-start)/numChunks,GHz], 
                     numPoints, 
                     averages, 
                     currentPower, 
                     bandwidth,
                     pnaAverages = 1,
                     pnaname = pna,
                     scandir = dir
                    )
    dv  = sample._cxn.data_vault
    dv.cd(sample._dir[:-1] + dir)
    setnum = len(dv.dir()[-1])
    
    fData, mData, pData = rm.get_dataset(
                            sample._cxn, setnum, sample._dir[:-1] + dir
                            )
    if DEBUG01:
        print "setnum: " + str(setnum)
        redata, imdata = rm.ConvertMP(mData, pData)
        mDataInv, pDataInv = rm.invert_data(mData,pData)
        reDataInv, imDataInv = rm.ConvertMP(mDataInv, pDataInv)
        
        print "Plot S21 from powerSweep"
        plotraw(fData, 20*np.log10(mData), pData, redata, imdata)
        
        print "Plot 1/S21 from powerSweep"
        plotraw(fData, mDataInv, pDataInv, reDataInv, imDataInv)
    
    
    #fData in Hz, mData Not in db, all of S21
    return fData, mData, pData, setnum
    
def take_calibration(sample, fitdir):

    strout = ''.join(["Taking Calibration: ",psu.defaultMsg(sample)])
    util.printLog(strout, [LOGFILE], 'powersweep -> take_calibration')
    localpower = sample['power']
    ave = 1
    if   localpower in range( 20, -30, -1):
        bw = 1
    elif localpower in range(-30, -50, -1):
        bw = 1
        ave = 5
    elif localpower in range(-50, -65, -1):
        bw = 1
        ave = 10
    elif localpower in range(-65, -95, -1):
        bw = 1
        ave = 30  #This could be done much faster if ave by VNA 
                  #and then sent to us
    span                          = 0.0
    span                          = 30*sample['span'] #@#
    numPoints                     = 3
    sample['_datasetname']        = ''.join(['calibration, power=', 
                                              str(sample['power'])
                                            ])
    
    fData, mData, pData, calNum = scan (sample, 
                                        sample['f0'], 
                                        span, 
                                        numPoints, 
                                        ave, 
                                        sample['power'], 
                                        bw,
                                        name = 'calibration_scan',
                                        pna = psu.getpnaName(sample),
                                        dir = fitdir + [str(sample['power'])]
                                       )
    
    sample['calibrationScanList'] = sample['calibrationScanList'] + [calNum]
    cropSize                      = int(sample['crop'])
    cropOffset                    = int(sample['cropOffset'])
    showCal                       = sample['showCal']

    [mCalInt, pCalInt]\
     = get_cold_calibration(fData, mData, pData, cropSize, cropOffset, showCal)

    return mCalInt, pCalInt, sample

def scan_parameters(sample):
    '''Adjusts and sets span and bandwidth of scan'''
    reg = sample._cxn.registry()
    reg.cd(sample._dir)
    
    if sample['q0'] > 0:
        sample['span'] = sample['x_spanfactor']*sample['f0'] /\
                                    sample['q0']
    else:
        sample['span'] = sample['x_spanfactor'] * sample['f0'] /\
                                    sample['baddefaultq']
        strout         = ''.join(["Previous Q Error: ",
                                    psu.defaultMsg(sample),
                                    "lastQ = .  Setting Q to 1e6\n"
                                 ])
        print strout
        util.printErr(strout, [LOGFILE], "powerSweep -> scan_parameters")

    sample['bandwidth'] = 30
    if len(sample['averageScanList']) > 20:
        sample['bandwidth'] = 5
        
    return sample

def update_fits (sample, fds): 

    # If no new data, exit
    if(not fds.newData):
        return fds
    
    sample['+Status'] = 'Saving Fit to Data Vault'
    
    fitParamsLocalInv = list(fds.pwr_fitParams_Inv)
    # fitParamsLocalReg = list(fds.pwr_fitParams_Reg)
    dv                = sample._cxn.data_vault()
    
    dv.cd(sample._dir[:-1] + fds.fitdir)
    
    if fitParamsLocalInv[2] < 0:
        fitParamsLocalInv[3] = fitParamsLocalInv[3] + np.pi
    
    # if fitParamsLocalReg[2] < 0:
        # fitParamsLocalReg[3] = fitParamsLocalReg[3] + np.pi
    
    print "fitParams are: " + str(fitParamsLocalInv)
    
    fds.numPhoInv  = rm.photonNumber( 
                    int(fitParamsLocalInv[0]), 
                    abs(int(fitParamsLocalInv[2])), 
                    int(fitParamsLocalInv[1]), 
                    int(sample['power']) - np.abs(sample['physicalAttenuation'])
                    )
    # fds.numPhoReg  = rm.photonNumber( 
                    # int(fitParamsLocalReg[0]), 
                    # abs(int(fitParamsLocalReg[2])), 
                    # int(fitParamsLocalReg[1]), 
                    # int(sample['power']) - sample['physicalAttenuation']
                    # )                
    
    covarInv       = [np.sqrt(fds.covarInv[k,k]) 
                        for k in range(len(fitParamsLocalInv))]
    # covarReg       = [np.sqrt(fds.covarReg[k,k]) 
                        # for k in range(len(fitParamsLocalReg))]

    dvparams                                   = {}
    dvparams['Calibration_Scan(s)']            = sample['calibrationScanList']   
    dvparams['DataVault_Average_Scan(s)']      = sample['averageScanList']
    dvparams['Num_Avgs']                       = fds.currentAveCtr
    dvparams['VNA Power']                      = sample['power']
    dvparams['zz_randomnumber_ID']             = str(np.random.random()*1e17) # Cheap ID
    dvparams['Is_Complete']                    = str(fds.isComplete)
    
    dvparams['Inverse_Fit.covar']              = covarInv
    #dvparams['Regular_Fit.covar']              = covarReg
    dvparams['Inverse_Fit.Num_Photons']        = fds.numPhoInv
    #dvparams['Regular_Fit.Num_Photons']        = fds.numPhoReg
    dvparams['Inverse_Fit.phi']                = np.mod(fitParamsLocalInv[3], 
                                                         2*np.pi) * rad
    #dvparams['Regular_Fit.phi']                = np.mod(fitParamsLocalReg[3], 
                                                         #2*np.pi) * rad

    dvparams['Inverse_Fit.Qi']                 = int(fitParamsLocalInv[0])
    dvparams['Inverse_Fit.Qc']                 = abs(int(fitParamsLocalInv[2]))
    dvparams['Inverse_Fit.Resonance_Freq']     = int(fitParamsLocalInv[1]) *Hz
    #dvparams['Regular_Fit.Qi']                 = int(fitParamsLocalReg[0])
    #dvparams['Regular_Fit.Qc']                 = abs(int(fitParamsLocalReg[2]))
    #dvparams['Regular_Fit.Resonance_Freq']     = int(fitParamsLocalReg[1]) *Hz
    
    dvparams['Inverse_Fit.Initial_Guess.Qi']   = str(fds.guessInv[0])
    dvparams['Inverse_Fit.Initial_Guess.f0']   = str(fds.guessInv[1])
    dvparams['Inverse_Fit.Initial_Guess.Qc']   = str(fds.guessInv[2])
    dvparams['Inverse_Fit.Initial_Guess.phi']  = str(fds.guessInv[3])
    #dvparams['Regular_Fit.Initial_Guess.Qi']   = str(fds.guessReg[0])
    #dvparams['Regular_Fit.Initial_Guess.f0']   = str(fds.guessReg[1])
    #dvparams['Regular_Fit.Initial_Guess.Qc']   = str(fds.guessReg[2])
    #dvparams['Regular_Fit.Initial_Guess.phi']  = str(fds.guessReg[3])
    
    dvparams['Inverse_Fit.Percent_QiError']    = str(fds.QiErrorInv*100)[:5]
    #dvparams['Regular_Fit.Percent_QiError']    = str(fds.QiErrorReg*100)[:5]
    
    [deltaF, 
     fds.mDataCaledInvdB, 
     fds.mFitInvdB,
     fds.pFitInv, 
     reDataInv, 
     imDataInv, 
     reFitInv, 
     imFitInv
     ]              = fit.getFits_dv(
                        fds.fData,
                        fds.mDataCaledInv, 
                        fds.pDataCaledInv,
                        fds.func, 
                        fitParamsLocalInv) 
    # [deltaF, 
     # fds.mDataCaledRegdB, 
     # fds.mFitRegdB,
     # fds.pFitReg, 
     # reDataReg, 
     # imDataReg, 
     # reFitReg, 
     # imFitReg
     # ]              = fit.getFits_dv(
                        # fds.fData,
                        # fds.mDataCaled, 
                        # fds.pDataCaled,
                        # fds.func, 
                        # fitParamsLocalReg) 
    #The Reg's used a regular fit instead of the 1/S21 fit.
    
                        
    cDataInv        = [reDataInv[i] + imDataInv[i] * 1j 
                        for i in range(len(reDataInv))
                      ]
    # cDataReg        = [reDataReg[i] + imDataReg[i] * 1j 
                        # for i in range(len(reDataReg))
                      # ]

    filenameInvF    = ''.join (['Fit 1/S21, ',
                                str(fds.fitdir[0]),
                                ', power=', str(dvparams['VNA Power']),
                                ', DeltaF']
                            )
    filenameInvC    = ''.join (['Fit 1/S21, ',
                                str(fds.fitdir[0]),
                                ', power=', str(dvparams['VNA Power']),
                                ', complex']
                            )
    # filenameRegF    = ''.join (['Fit   S21, ',
                                # str(fds.fitdir[0]),
                                # ', power=', str(dvparams['VNA Power']),
                                # ', DeltaF']
                            # )
    # filenameRegC    = ''.join (['Fit   S21, ',
                                # str(fds.fitdir[0]),
                                # ', power=', str(dvparams['VNA Power']),
                                # ', complex']
                            # )

    independentsInvF = [ ('Delta F','KHz')]
    independentsInvC = [('Real','dB')]
    # independentsRegF = [ ('Delta F','KHz')]
    # independentsRegC = [('Real','dB')]
    
    dependentsInvF   = [ ('Magnitude 1/S21','data','dB'),
                         ('Magnitude 1/S21','fit' ,'dB'),
                         ('Phase 1/S21'    ,'data','rad'),
                         ('Phase 1/S21'    ,'fit' ,'rad'),
                       ]
    dependentsInvC   = [ ('Imaginary 1/S21','data','dB'),
                         ('Imaginary 1/S21','fit' ,'dB'),
                       ]
    # dependentsRegF   = [ ('Magnitude 1/S21'  ,'data','dB'),
                         # ('Magnitude 1/S21'  ,'fit' ,'dB'),
                         # ('Phase 1/S21'      ,'data','rad'),
                         # ('Phase 1/S21'      ,'fit' ,'rad'),
                       # ]
    # dependentsRegC   = [ ('Imaginary 1/S21'  ,'data','dB'),
                         # ('Imaginary 1/S21'  ,'fit' ,'dB'),
                       # ]

    dataInvF = []
    dataInvC = []
    # dataRegF = []
    # dataRegC = []
                    
    
    for i in range(len(deltaF)):
        dataInvF.append([ deltaF[i], 
                          fds.mDataCaledInvdB[i],
                          fds.mFitInvdB[i],
                          fds.pDataCaledInv[i],
                          fds.pFitInv[i]
                       ])
    for i in range(len(deltaF)):
        dataInvC.append([ reDataInv[i],
                          imDataInv[i],
                          imFitInv[i] #??? wonder if this will work
                        ])
    # for i in range(len(deltaF)):
        # dataRegF.append([ deltaF[i], 
                          # fds.mDataCaledInvdB[i], 
                          # fds.mFitRegdB[i], # I'm plotting 1/S21 w/ S21 Fit
                          # fds.pDataCaledInv[i],
                          # fds.pFitReg[i]
                       # ])
    # for i in range(len(deltaF)):
        # dataRegC.append([ reDataInv[i],
                          # imDataInv[i],
                          # imFitReg[i] #??? wonder if this will work
                        # ])
    
    psu.writeToVault(dv,dvparams,
                     filenameInvF, independentsInvF, dependentsInvF, 
                     dataInvF)
    psu.writeToVault(dv,dvparams,
                     filenameInvC, independentsInvC, dependentsInvC, 
                     dataInvC)
    # psu.writeToVault(dv,dvparams,
                     # filenameRegF, independentsRegF, dependentsRegF, 
                     # dataRegF)
    # psu.writeToVault(dv,dvparams,
                     # filenameRegC, independentsRegC, dependentsRegC, 
                     # dataRegC)
    return fds

def runPowerSweep(verbose=True):
    ''' Run THIS to start a powerSweep over a set of variables.
        void runPowerSweep(boolean verbose): 
        verbose -> If user is prompted to check the sample and sample location
                   at runtime
        It takes care of:
            1) Copying all the keys into the registry from a template directory
               set in registryhandle.py into the registry folder of your choice.
               Choice made by setting your sample variable
            2) Starts scanning
    '''

    # Before reading the code, a quick refresher on the "with ... as ..." 
    # construction:
    #
    # With ... as ... is a way make python call a specific function of an object
    # before running any code enclosed in the 'with' statement' and another 
    # specific function after the code in the 'with' has executed regardless 
    # of the success of the code in the with.
    # These two fuction are "__enter__" and "__exit__" respectively.
    # We use them to prevent a cxn object from EVER escaping the confines of 
    # this program.  Even if the code crashes and the error is not caught, the
    # with statement WILL call the "__exit__" method.  
    

    with rh.MakeSample() as sampleMaker:
        sample = sampleMaker.init()
        # sample = rh.makeSample()
    
        if sample == -1:
            return -1
        
        if(verbose):
            yesno = str( raw_input(''.join(['Sample = ', 
                                            str(sample),
                                            '\n(y/n)'
                                           ])))
            if str(yesno) == 'n':
                return 1
        
        psu.resetSig(sample)
        
        
        print str(psu.regget(sample,'deviceName'))
        if(psu.regget(sample,'deviceName') == 'quaid' or
           psu.regget(sample,'deviceName') == 'Quaid'):
            print "I am QUAID!!!"
            sample['lakeshore_gpib'] = 'Kimble GPIB Bus - GPIB0::12'
            sample['power_supply']   = 'Kimble GPIB Bus - GPIB0::5'
            
            if(psu.getpnaName(sample) == '5242A'):
                #sample['pna_gpib'] = 'Kimble GPIB Bus - TCPIP0::A-N5242A-21947::inst0'
                #sample['pna_gpib'] = 'Kimble GPIB Bus - GPIB0::5'
                sample['pna_gpib'] = 'hercules GPIB Bus - GPIB0::23'
            elif(psu.getpnaName(sample) == '8720es'):
                sample['pna_gpib'] = 'hercules GPIB Bus - GPIB0::16'
            elif(psu.getpnaName(sample) == 'E8364B'):
                sample['pna_gpib'] = 'hercules GPIB Bus - GPIB0::20'
            elif(psu.getpnaName(sample) == '5230A'):
                sample['pna_gpib'] = 'hercules GPIB Bus - GPIB0::20'
            
        
        if(verbose):
            print('Sample variables:')
            rh.printSample(sample)
            yesno = str(raw_input('Continue? (y/n)'))
            if str(yesno) == 'n':
                return 2
        
        reslist = psu.regget(sample,'listOfResIndexes')
        
        resindex = 0
        sample['resListIndex'] = resindex
        
        while resindex < len(reslist):
            sample['resListIndex'] = resindex
            sample['+Status']      = 'Preparing to Sweep'
            sample['resnum']       = reslist[resindex]  + 1                         # change to no plus
            sample['f0']           = psu.regget(sample,'f0s')[reslist[resindex]]    # change to -1
            sample['q0']           = psu.regget(sample,'q0s')[reslist[resindex]]    # change to -1
            
            code                   = powerSweepWrapper(sample)
            
            sample['+Status']      = 'Idle'
            if(code == -1):
                return -1
            reslist           = psu.regget(sample,'listOfResIndexes')
            resindex         += 1
            

    util.stopLog()
#} end of runPowerSweep

def powerSweepWrapper(sample):
    ''' int powerSweepWrapper(sample):
        Wraps the powerSweep class.  Its primary function is to pass in a bunch
        of dataVault Saver objects (home made in dataVaultSaver.py) incased in 
        with-as statements.
    '''
    
    resfolder = ''.join(['Res',str(sample['resnum']).zfill(2)])

    with dvs.dvsaver_temp(resfolder) as dvs_temp:
        with dvs.dvsaver_npho(resfolder) as dvs_npho:
            with dvs.dvsaver_time(resfolder) as dvs_time:
                with dvs.dvsaver_curr(resfolder) as dvs_curr:
                    with dvs.dvsaver_lognpho(resfolder) as dvs_Lognpho:
                        @util.freeze
                        class dvsholder(object):
                            def __init__(self, 
                                        dvs_temp, dvs_npho, 
                                        dvs_time, dvs_curr, dvs_Lognpho
                                        ):
                                self.dvs_temp = dvs_temp
                                self.dvs_npho = dvs_npho
                                self.dvs_time = dvs_time
                                self.dvs_curr = dvs_curr
                                self.dvs_Lnpo = dvs_Lognpho
                    
                        dvsh = dvsholder(dvs_temp, dvs_npho, dvs_time, dvs_curr,
                                         dvs_Lognpho)
                        
                        returncode = powerSweep(sample, dvsh)
    # } End of all with-as statements
    
    return returncode
        
def powerSweep(sample, dvsh):
    ''' int powerSweep(sample, dvsholder dvsh):
        dvsh -> a datavault saver holder defined in powerSweepWrapper
        Sweeps the sample for one resonator over many powers
    '''
    
    # TODO: add units!
    
    res_fitParameters_Inv     = []
    res_fitParameters_Reg     = []
    powers                    = psu.getpowerlist(sample)
    sample['power']           = powers[0]
    sample['y_powerindex']    = 0
    
    # Begin a dataVaultsaver data set.  THe current sample is used to populate
    dvsh.dvs_npho.startDataSet(sample)
    dvsh.dvs_Lnpo.startDataSet(sample)
    dvsh.dvs_temp.startDataSet(sample)
    dvsh.dvs_time.startDataSet(sample)
    dvsh.dvs_curr.startDataSet(sample)
    
    # While there are more powers to be scanned,
    while (sample['y_powerindex'] < len(powers)
        ):
        
        # Get the power and do some clean up
        sample['power'] = powers[sample['y_powerindex']]
        psu.clearSample_pwr(sample)
        
        if sample['power'] > MAXPOWER:
            raise Exception('TOO MUCH POWER TO VNA')
            return
        
        # IF SIGINT_gotopower, goto that power
        if (sample['z_SIGINT_gotopower'] != 0):
            if sample['z_SIGINT_gotopower'] > len(powers):
                util.printErr(
                        'Registry Error: z_SIGINT_gotopower > len(powers)',
                        [LOGFILE],
                    'powerSweep -> powerSweep'
                    );
            else:
                sample['y_powerindex'] = sample['z_SIGINT_gotopower']
                continue
        sample['bandwidth']=30
        # If first power, 3 times larger span NO!!
        spanFactorLoop = sample['x_spanfactor']
        if sample['y_powerindex'] == 0:
            spanFactorLoop *= 9
            sample['bandwidth']=10000 ## want higher bandwidth on the first pass to find the resonator more quickly
        
        [startTime, starTimeH] = util.getTimes()
        funct = lambda f, v: 1 + (v[0]/v[2]) * np.exp(1j * v[3]) /     \
                                    (1 + 2j*v[0]*(f/v[1]-1) )
        fds   = None
        fds   = util.Fds(sample, funct, startTime) 
        # Fds custom class. Fixed number & names of variables.
        
        sample['+Status'] = 'Calibrating'
        sample                            = scan_parameters(sample)
        [fds.mCalInt,fds.pCalInt, sample] = take_calibration(sample, fds.fitdir)
        fds.reAvg = np.zeros(spanFactorLoop*sample['x_points'])
        fds.imAvg = np.zeros(spanFactorLoop*sample['x_points'])
        
        # Tells the program if you are exiting with prejudice
        SIGinterrupt = False
        
        # Stop Scanning if QiError < some tolerance and current Average > min
        #       or if have been scanning for over timeout minutes
        while(  ( abs(fds.QiErrorInv) > sample['x_qtolerance'] or 
                 fds.currentAveCtr < sample['x_minnumave']       
                ) 
                and
                (time.time() - fds.startTime)/60 < sample['x_timeout'] 
            ):
            
            sample['+Status'] = ''.join([
                    'Scanning: Res = ', str(sample['resnum']), 
                                     '(', str(sample['resListIndex']), ')', 
                    ', VNA Power = ',   str(sample['power']),
                    ', for ', str(int((time.time() - fds.startTime) / 60)),
                    ' mins'])

            
            # The next code section helps in SIGINT or SIGnal INTerrupt
            # They are triggered from the sample variable folder in the registry.
            # These SHOULD be used. Ctrl-C ONLY at last resort.  Nothing will
            # break if you Ctrl-C, but fitted data may not be saved.  All fit
            # data can be found in the logs if need be, but SIGINT's give clean
            # save exits.
            
            # START Interrupt Signal handling
            
            if(sample['z_SIGINT_cleanexit'] != 0):
                sample['+Status'] = 'SIGINT Exiting'
                if (fds.newData):
                    fds.isComplete = False
                    fds = update_fits(sample,fds)
                sample['z_SIGINT_cleanexit']
                return -1
            # } end SIGINT cleanexit
            
            if(sample['z_SIGINT_skippower'] != 0):
                sample['+Status'] = 'SIGINT skipping power'
                psu.skippower(sample, [LOGFILE])
                sample['z_SIGINT_skippower'] = 0
                if(fds.newData):
                    fds.isComplete = False
                    fds = update_fits(sample,fds)
                SIGinterrupt = True
                break
            # } end SIGINT skippower
            
            if(sample['z_SIGINT_skipandsavepower'] != 0):
                sample['+Status'] = 'SIGINT skipping power'
                psu.skippower(sample, [LOGFILE])
                sample['z_SIGINT_skipandsavepower'] = 0
                if(fds.newData):
                    fds.isComplete = False
                    fds = update_fits(sample,fds)
                    dvsh = toDataVaultSaver(sample,fds,dvsh)
                SIGinterrupt = True
                break
            # } end SIGINT skipandsavepower
                
            if(sample['z_SIGINT_skipres'] != 0):
                sample['+Status'] = 'SIGINT skipping resonator'
                psu.skipres(sample, [LOGFILE])
                sample['z_SIGINT_skipres'] = 0
                if (fds.newData):
                    fds.isComplete = False
                    fds = update_fits(sample,fds)
                sample['y_powerindex'] = len(powers) + 1
                SIGinterrupt = True
                break
            # } end SIGINT skipres
            
            if(sample['z_SIGINT_gotores']  != 0):
                
                # sample['+Status'] = 'SIGINT going to resonator'
                # psu.skipres(sample, [LOGFILE])
                # sample['z_SIGINT_gotores'] = 0
                # fds.isComplete = False
                # fds                         = update_fits(sample,fds)
                # sample['y_powerindex']      = len(powers) + 1
                # if(sample['z_SIGINT_gotores'] in psu.regget(sample,'reslist')):
                    # allres           = psu.regget(sample,'reslist')
                    # sample['resnum'] = sample['z_SIGINT_gotores']
                    # sample['f0']     = psu.regget(sample,'f0s')[\
                                           # allres.index(sample['resnum'])]
                    # sample['q0']     = psu.regget(sample,'q0s')[\
                                           # allres.index(sample['resnum'])]
                    # powerSweep(sample, dvsh)
                    # sample['y_powerindex'] = len(powers) + 1
                    # SIGinterrupt = True
                    # break
                # else:
                    # sample['y_powerindex'] = len(powers) + 1
                    # SIGinterrupt = True
                    # break
                pass
            #} end SIGINT gotores
            
            # STOP Interrupt Signal handling:
            
            fds.currentAveCtr = fds.currentAveCtr + 1
            
            # Now we can finally do a powerScan
            fds = powerScan (sample, fds, 
                res_fitParameters_Inv, res_fitParameters_Reg, spanFactorLoop)
            
            
        #} end of qtolerance while loop
        endtime = time.time()
        
        
        # The values now stored in fds.pwr_fitParams were good enough! So we
        # store them for future powers.
        res_fitParameters_Inv.append(list(fds.pwr_fitParams_Inv))
        res_fitParameters_Reg.append(list(fds.pwr_fitParams_Reg))
        
        # Store the last values of f0 and q0 as the current values
        sample['f0'] = fds.pwr_fitParams_Inv[1]*1e-9 
        sample['q0'] = fds.pwr_fitParams_Inv[0]

        
        # IF we are not exiting the qtolerance while loop with prejudice,
        # save fits to their appropriate places and add data to the 
        # dataVaultSavers.  
        
        # We don't have an array of these data vault savers 
        # to emphasize how much overhead each one requires (each one takes 
        # a cxn).
        
        if SIGinterrupt == False:
            fds = update_fits(sample, fds)
            dvsh = toDataVaultSaver(sample,fds,dvsh)
                

        util.printLog(''.join(['powerSweep fitparams: ', 
                        str(fds.pwr_fitParams_Inv)]),
                        [LOGFILE], str(__file__) + ' -> powerSweep')
        
        # Ending conditions:
        # if ((endtime - fds.startTime)/60 >= sample['x_timeout'] and
             # fds.numPhoInv > 1 + float(sample['x_minphoton']) 
            # ):
            #  #we time'ed out.  What do you want to do about it?  
            #  #We exit.
            # sample['y_powerindex'] = len(powers) + 1
            # break
        
        if   (sample['y_powerindex'] + 1 >= len(powers)):
            util.printErr("Insufficient powers! No choice but to exit",
                          [LOGFILE], str(__file__) + "-> powersweep")
            break
        elif (fds.numPhoInv <= sample['x_minphoton']):
            break
        elif (fds.numPhoInv <= sample['x_maxphoton']):
            sample['y_powerindex'] += 1
            continue
        # elif(fds.QiErrorInv >= 1e10 and fds.numPho >= sample['x_maxphoton']): 
            # if bad fit only
            # remainingPho = sample['x_maxphoton'] - fds.numPho
            # indexChange  = int(np.log(fds.numPho / sample['x_maxphoton']) / \
                                # np.log(genphotonrule))
            # continue
        else:
            sample['y_powerindex'] += 1
            continue
                
    #} end of power list while loop, this resonator is now finished
    newcompleteres = psu.regget(sample,'completedres')
    newcompleteres.append(sample['resnum'])
    psu.regset(sample,'completedres', newcompleteres)
    
    return 0
#} end of powerSweep

def powerScan(sample, fds, 
            res_fitParameters_Inv, res_fitParameters_Reg, spanFactorLoop
    ):
    '''fds powerScan(sample, fds, fitparams for inverse, fitparams for reg, 
                     scanfactorloop):
       Scans a single power a single time and does a single batch of fits
       Fds provides a rough sense of averaging.
    '''
    sample['_datasetname'] = ''.join(['powerscan, res=', str(sample['resnum']), 
                                      ', power=', str(sample['power'])]) 
    firstscan = 1
    bwfactor = 1
    if sample['y_powerindex'] == 0:
            firstscan *= 3
            bwfactor *= 10
    print 'span3:  ',sample['span']        
    [fds.fData, fds.mData, fds.pData, fds.dataNum] \
        = scan(
            sample, 
            center       = sample['f0'], 
            span         = sample['span']*firstscan, 
            numPoints    = int(spanFactorLoop*sample['x_points']), 
            averages     = 1, 
            currentPower = sample['power'], 
            bandwidth    = sample['bandwidth']*bwfactor,
            name         = ''.join(['powerSweep_p=', str(sample['power']),
                                    'r=', str(sample['resnum']).zfill(2)
                                   ]),
            pna          = psu.getpnaName(sample),
            dir          = fds.fitdir + [str(sample['power'])],
            numChunks    = 1
            )

    fds.newData                = True
    
    sample['averageScanList']  = sample['averageScanList'] + [fds.dataNum]
    [reData, imData]           = rm.ConvertMP(fds.mData, fds.pData)
    
    # Average Scans
    fds.reAvg  = fds.reAvg + reData
    fds.imAvg  = fds.imAvg + imData
    reAvgN     = fds.reAvg/fds.currentAveCtr
    imAvgN     = fds.imAvg/fds.currentAveCtr

    fds.mData, fds.pData                 = rm.ConvertRI(reAvgN, imAvgN)
    
    #print fds.mCalInt(fds.fData[0]) + fds.mCalInt(fds.fData[len(fds.fData)-1])
    #print fds.pCalInt(fds.fData[0]) + fds.pCalInt(fds.fData[len(fds.fData)-1])
    
    fds.mDataCaled, fds.pDataCaled       = rm.calibrate_data(
                                                    fds.mCalInt, fds.pCalInt, 
                                                    fds.fData,   fds.mData, 
                                                    fds.pData
                                            )
    fds.mDataCaledInv, fds.pDataCaledInv = rm.invert_data(fds.mDataCaled, 
                                                    fds.pDataCaled
                                                   )
    reAvgN, imAvgN = rm.ConvertMP(fds.mDataCaledInv, fds.pDataCaledInv)
    
    # Fit Averaged Data
    if res_fitParameters_Inv == [] or fds.QiErrorInv == 1e10:
        guessInv = fit.david_get_params(fds.fData, 
                              fds.mDataCaledInv, 
                              fds.pDataCaledInv)
        fds.guessInv = guessInv
        
    else:
        guessInv = res_fitParameters_Inv[-1]
        fds.guessInv = guessInv
    
    # if res_fitParameters_Reg == [] or fds.QiErrorReg == 1e10:
        # guessReg = fit.get_parms(fds.fData, 
                              # fds.mDataCaled, 
                              # fds.pDataCaled, 
                              # fds.func)
        
        # fds.guessReg = guessReg
    # else:
        # guessReg = res_fitParameters_Reg[-1]
    
    fitParms, fds.covarInv     = fit.fit(fds.fData, fds.mDataCaledInv, 
                                     fds.pDataCaledInv, fds.func, fds.guessInv)
    fds.pwr_fitParams_Inv   = tuple(fitParms)

    # fitParms, fds.covarReg     = fit.fit(fds.fData, fds.mDataCaledInv, 
                                     # fds.pDataCaledInv, fds.func, guessReg)
    # fds.pwr_fitParams_Reg   = tuple(fitParms)
    
    kP = 0
    QiSigmaInv = np.sqrt(fds.covarInv[kP, kP])
    # QiSigmaReg = np.sqrt(fds.covarReg[kP, kP])
    
    util.printLog("Res={0}, Power={1}".format(sample['resnum'], sample['power']),
                  [LOGFILE],"powersweep.powerscan()")
    util.printLog("Time Elapsed = {0}".format( (time.time() - fds.startTime)/60 ),
                  [LOGFILE],"powersweep.powerscan()")
    
    if QiSigmaInv > 0:
        fds.QiErrorInv = QiSigmaInv/fds.pwr_fitParams_Inv[kP]
        util.printLog("Inverse QiError: " + str(fds.QiErrorInv*100)[:5] + " %",
                      [LOGFILE], "powersweep -> powerscan")
    else:
        fds.QiErrorInv = 1e10
        util.printLog("Inverse: Bad Fit", [LOGFILE], "powersweep -> powerscan")
    util.printLog("  Guess " + psu.paramsToString(fds.guessInv) + "\n",
                  [LOGFILE], "powersweep -> powerscan")
    util.printLog("  Fit   " + psu.paramsToString(fds.pwr_fitParams_Inv) + "\n",
                  [LOGFILE], "powersweep -> powerscan")
    
    
    # plotraw(fds.fData, fds.mDataCaledInv, fds.pDataCaledInv, reAvgN, imAvgN, 
        # fds.guessInv, fds.func)
    
    # if QiSigmaReg > 0:
        # fds.QiErrorReg = QiSigmaReg/fds.pwr_fitParams_Reg[kP]
        # util.printLog("Regular QiError: " + str(fds.QiErrorReg*100)[:5] + " %",
                      # [LOGFILE], "powersweep -> powerscan")
    # else:
        # fds.QiErrorReg = 1e10
        # util.printLog("Regular: Bad Fit", [LOGFILE], "powersweep -> powerscan")
    # util.printLog(psu.paramsToString(fds.guessReg) + "\n",
                  # [LOGFILE], "powersweep -> powerscan")
 
    currentPower = sample['power'] 
    # resName         = ''.join(['powerSweep_p=', str(sample['power']),'r=', str(sample['resnum']).zfill(2)])
    resName         = 'Res'+ str(sample['resnum'])
    #fit.fit_plot(fds.fData, fds.mDataCaledInv, fds.pDataCaledInv, fds.func, list(fitParms),'1/S21', str(mainDir[-3]) + "\\" + str(mainDir[-2]) +  "\\" + str(mainDir[-1]) + "\\" + resName, str(currentPower), False, True)	
    fit.fit_plot(fds.fData, fds.mDataCaledInv, fds.pDataCaledInv, fds.func, list(fitParms),'1/S21', 'u:\\ben\\data\\'+'ResonatorFits'+'\\'+sample._dir[3]+'\\'+sample._dir[4] ,'_'+resName +'_'+ str(currentPower), False, True)	
    
    return fds
#} end of powerScan





                    
###############################################################################
###############################################################################
###############################################################################

# old code:   
       
# def x_scan_parameters(cxn, mainDir, resName, spanFactor, points):	

    # sessionDir = list(mainDir)
    # sessionDir.append(resName)
    
    # reg = cxn.registry
    # reg.cd(sessionDir)
    # bw = 30
    # Avgs = [0]
    
    # if 'Qi' not in reg.dir()[1]:
        # reg.cd(mainDir)
        # vals = reg.get(resName)
        # lastF = float(vals[0])
        # lastQ = float(vals[1])
    # else:
        # lastQ = reg.get('Qi')[-1]
        # lastF = reg.get('resonance freq')[-1]
        # Avgs = reg.get('Num Avgs')
        
    
    # if lastQ != 0:
        # span = spanFactor*lastF/lastQ
    # else:
        # span = spanFactor * lastF/ baddefaultQ
        # util.printErr("Error! lastQ = 0.  Setting Q to 1e6", [LOGFILE], \
                      # "scan_parameters")
    
    
    # if max(Avgs) > 20:
        # bw = 5
        
    # return lastF, span, lastQ, bw
    
    
# def xxxpowerSweep(resName,
               # mainDir = ['','David','data','120727','4mTorr','powerSweep']
               # ):

    # Parameters To Be Set  
    # mainDir = ('', 'Charles', 'dataKing', '120529', 'Transmon', 'powerSweep')
        # Make this directory with the key resName = [f0, Qi]
    # mainDir = ('', 'David', 'data', '120727', 'Sputter', 'powerSweep')
    # spanFactor = 8      #total points = spanfactor * points
    # points = 60         # points per span, reduce the number of points, lower 
                        # bandwidth, time same
    # qTolerance = 0.025  # error tolerance in Q
    # powers = [-10,-15,-20, -25, -30, -35, -40, -45, -50, -55, 
            # -60, -65, -70, -75,-80,-85,-90,-95,-100,-105,-110,-115]
    # if (resName in resExceptions1):
        # powers = [-30,-35,-40,-45,-50,-55,-60,-65,-70,-75,-80,
                    # -85,-90,-95,-100,-105,-110]
    # if (resName in resExceptions2):
        # powers = [-80,-85,-90,-95,-100,-105,-110]    
    # if (resName in resExceptions3):
        # powers = [-90,-95,-100,-105,-110]
    # timeOut = 30
    # minAvgs = 3
    # End Set Parameters
   
    # fitParameters = []
    # with labrad.connect() as cxn:
        # for currentPower in powers:
        
            # if currentPower == powers[0]:
                # spanFactorLoop = spanFactor*3
            # else:
                # spanFactorLoop = spanFactor
            
            # sample = switch_session(cxn, currentPower, mainDir, resName)
            # lastF, span, lastQ, bw = scan_parameters(cxn, mainDir, resName, 
                                                # spanFactorLoop, points)
            # mCalInt, pCalInt, calNum = take_calibration(sample, lastF)
            
            # Scan until error tolerance is Reached
            # QiError = 1e10
            # QiSigma = 1e10
            # startTime = time.time()
            # reAvg = np.zeros(spanFactorLoop*points)
            # imAvg = np.zeros(spanFactorLoop*points)          
            # currentAvg = 0
            
            # while (abs(QiError) > qTolerance or currentAvg < minAvgs) and 
                    # (time.time() - startTime)/60 < timeOut:
                # currentAvg = currentAvg + 1
                
                # Perform a scan
                # fData, mData, pData, dataNum = scan(
                    # sample, lastF, span, 
                    # int(spanFactorLoop*points), 1, currentPower, bw)
                # reData, imData = ConvertMP(mData, pData)
                
                # Average Scans
                # reAvg = reAvg + reData
                # imAvg = imAvg + imData
                # reAvgN = reAvg/currentAvg
                # imAvgN = imAvg/currentAvg

                # mData, pData = ConvertRI(reAvgN, imAvgN)
                # mDataCal, pDataCal = calibrate_data(mCalInt, pCalInt, 
                                                    # fData, mData, pData)
                # mDataInv, pDataInv = invert_data(mDataCal, pDataCal)
                
                # Fit Averaged Data
                # func = lambda f, v:  1 + (v[0]/v[2]) * np.exp(1j * v[3]) /\
                                # ( 1 + 2j*v[0]*(f/v[1]-1) )
                # if fitParameters == [] or QiError == 1e10:
                    # guess = fit.get_parms(fData, mDataInv, pDataInv, func)
                    # print guess
                # else:
                    # guess = fitParameters[-1]
                # fitParms, covar = fit.fit(fData, mDataInv, 
                                        # pDataInv, func, guess)
                # fitParms = tuple(fitParms)

                # kP = 0
                # QiSigma = np.sqrt(covar[kP, kP])
                # if QiSigma > 0:
                    # QiError = QiSigma/fitParms[kP]
                    # print str(QiError*100)[:5], "%"
                # else:
                    # QiError = 1e10
                    # print "Bad Fit"
                    


            # fitParameters.append(list(fitParms))
            # numPho = update_fits(sample, fitParms, currentPower, covar, 
                            # currentAvg, calNum, dataNum)
            # fit.fit_plot(fData, mDataInv, pDataInv, func, list(fitParms), 
                    # '1/S21', str(mainDir[-3]) + "\\" + str(mainDir[-2]) + 
                    # "\\" + str(mainDir[-1]) + "\\" + resName, 
                    # str(currentPower), False, True)	
            # print fitParms
            # if(numPho < 1):
                # break;
                
                
# def getRegistryVals(sample,curres):
    # reg = sample._cxn.registry()
    # reg.cd(sample._dir[:-1])
    # regdict = {}
    # for key in REGCONT
        # rh.regCheck(reg,key,'Error')
        # regdict[key] = reg.get(key)
        # if regdict[key] == 'Error':
            # util.printErr('Registry key not found. Set correct key', [LOGFILE], 
                          # 'powerSweep -> getRegistryVals')
    # regdict['power'] = regdict['power_default']
    # if str(curres) in reg.dir()[1]:
        # regdict['power'] = reg.get(str(curres))
    
    # return regdict

# def bulkPowerSweep(resList = np.linspace(1,12,12),
                   # resListX = [],
                   # maindir = ['','David','data','120727','4mTorr','powerSweep']
                   # ): 
    # resList is linspace(1,12,12), resListX is other string names
    
    # allRes = []
    # for res in resList:
        # name = ''.join(['Res',str(res)])
        # allRes.append(name)
    # for res in resListX:
        # allRes.append(res)
    
    # print "Scanning Resonators " + str(allRes)
    
    
    # for res in allRes:
        # try:
            # powerSweep(res,maindir);
        # except Exception, e:
            # print "Error! Scan Failed. " + str(e)
            # continue
       
# def switch_session(cxn, currentPower, mainDir, resName):

    # reg = cxn.registry
    # reg.cd(usertemplate)
    # keys = reg.dir()
    # values = [reg.get(keys[1][i]) for i in range(len(keys[1]))]
    
    # reg.cd(mainDir)
    # if resName not in reg.dir()[0]:
        # reg.mkdir(resName)

    # reg.cd(resName)
    
    # sessionDir = list(mainDir)
    # sessionDir.append(resName)
    # sessionDir.append(str(currentPower))
    
    # if str(currentPower) not in reg.dir()[0]:
        # reg.mkdir(str(currentPower))
        # reg.cd(str(currentPower))
        # for i in range(len(keys[1])):
            # reg.set(keys[1][i], values[i])
        
    # reg.cd(userroot)
    # reg.set('sample', sessionDir[2:])
    # s = switchSession(cxn, user=theuser)
    
    # return s

    
def toDataVaultSaver(sample, fds, dvsh):
    datatosave = [fds.pwr_fitParams_Inv[2],
                  fds.pwr_fitParams_Inv[0],
                  fds.pwr_fitParams_Inv[1],
                  fds.pwr_fitParams_Inv[3]
                 ]
    formatresn = ''.join(['qi(',str(sample['power']),')'])
    try:
        dvsh.dvs_npho.addData([fds.numPhoInv] + datatosave)
        dvsh.dvs_temp.addData([PNAScan.get_temp(sample)] + datatosave)
        dvsh.dvs_curr.addData([sample['current']] + datatosave)
        dvsh.dvs_time.addData([time.time()] + datatosave)
        dvsh.dvs_Lnpo.addData(dvsh.dvs_Lnpo._logList(
                                [fds.numPhoInv] + datatosave))
        
        dvsh.dvs_npho.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_temp.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_curr.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_time.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_Lnpo.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        
    except Exception, e:
        util.printErr(str(e), [LOGFILE], "saving data to datavault")
        dvsh.dvs_npho.restartDataSet(sample, 
                                [fds.numPhoInv] + datatosave)
        dvsh.dvs_temp.restartDataSet(sample,
                                [PNAScan.get_temp(sample)] + datatosave)
        dvsh.dvs_time.restartDataSet(sample, 
                                [sample['current']] + datatosave)
        dvsh.dvs_curr.restartDataSet(sample, 
                                [time.time()] + datatosave)
        dvsh.dvs_Lnpo.restartDataSet(sample, dvsh.dvs_Lnpo._logList(
                                [fds.numPhoInv] + datatosave))
                                
        dvsh.dvs_npho.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_temp.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_curr.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_time.addParam(formatresn, fds.pwr_fitParams_Inv[0])
        dvsh.dvs_Lnpo.addParam(formatresn, fds.pwr_fitParams_Inv[0])
    
    return dvsh
