import labrad
from pyle.workflow import switchSession
import signal
import sys
import davidutil as util


SCANFOLDERNAME  = 'currentscan'
THEUSR          = 'Ben'
USRROOT         = ['',THEUSR]
TLOC            = ['','Ben','data','test','DavidsPS','template']
TESTLOC         = ['','David','tests','tests_powersweep','registryhandle']
LOG             = './log/registryhandle.log'
LOGFILE         = 'registryhandle'

class MakeSample(object):
    ''' MakeSample Class:
        Purpose: MakeSample returns a sample variable safely.
                 Provides good, safe handling of cxn connection objects.
                 See datavaultSaver.py for more details on safety.
    '''
    def __enter__(self):
        self.cxn = labrad.connect()
        self.reg = self.cxn.registry()
        return self
        
    def __exit__ (self, type, value, traceback):
        self.cxn.disconnect()
        
    def init(self):
        ''' sample init():
            Call this method to return a sample variable.  
            At the top of the file, there are a few variables you can change to 
            redirect where the program looks for the sample in the registry 
            (THEUSR, USRROOT).  Adjusting these to something like, 
            "THEUSR = 'Charles' ".  Adjust TLOC (Template LOCation) to change 
            where the program pulls the registry keys and values from.
            
            Note that we override anything within the sample variable directory
            (named as SCANFOLDERNAME) but we leave anything pre-defined in the 
            previous directory.
        '''
        
        # Goto user's root directory and store the sample name
        # Check if you are going to create an identical nested directory
        # structure.  If so, do not continue.
        self.reg.cd(USRROOT)
        powerSweepDir = self.reg.get('sample')
        if powerSweepDir[-1] == SCANFOLDERNAME:
            
            print "Nesting Directories: " + str(powerSweepDir)
            return -1
        
        try:
            [time,timeH] = util.getTimes()
            util.printLog("\nNew Sample Created on: " + timeH, [LOGFILE], 
                          str(__file__) + "-> MakeSample -> init")
            
            # Change to sample to template location
            # Point the sample to the directory 
            # (/the/directory/here/SCANFOLDERNAME)
            
            # self.reg.set('sample',TLOC[2:])
            s = switchSession(self.cxn,user=THEUSR)
            self.reg.cd(TLOC)
            
            # Copy every key here not already in the sample's folder
            for key in self.reg.dir()[1]:
                if key == SCANFOLDERNAME:
                    continue
                if(key in s.keys()):
                    continue
                s[key] = self.reg.get(key)
            
            # Go back home and put sample into template's SCANFOLDERNAME
            self.reg.cd(USRROOT)
            self.reg.set('sample',self.reg.get('sample') + [SCANFOLDERNAME])
            
            # Go the where the new sample will be
            self.reg.cd(powerSweepDir)
            
            # If the folder does not exist, create it.  Go to it
            if(SCANFOLDERNAME not in self.reg.dir()[0]):
                self.reg.mkdir(SCANFOLDERNAME)
            
            s = switchSession(self.cxn,user=THEUSR)
            
            self.reg.cd(TLOC + [SCANFOLDERNAME])
            
            # for all keys, set new sample to the template sample's values
            for key in self.reg.dir()[1]:
                s[key] = self.reg.get(key)
            
            # And reset sample
            self.reg.cd(USRROOT)
            self.reg.set('sample',powerSweepDir + [SCANFOLDERNAME])
            
        except Exception, e:
            # sys.stdout = actualstdout
            #file.close()
            util.printErr("Failed, exception " + str(e), [LOGFILE], 
                str(__file__) + " -> MakeSample -> init")
            self.reg.cd(USRROOT)
            self.reg.set('sample',powerSweepDir)
            return 0
         
        # } end of try-except
        s = switchSession(self.cxn,user=THEUSR)
        # sys.stdout = actualstdout
        
        return s
        
        
    def __init__(self):
        pass


def regCheck(reg, key, defaultval = ''):
    '''Safe check of registry'''
    if(key not in reg.dir()[1]):
        reg.set(key,defaultval)
    


def printSample(s):
    '''Prints all varialbes in the sample in ugly fashion'''
    for key in s.keys():
        print ''.join([str(key),': ',str(s.__getattr__(key))])
        
        
        
if 0:
    pass
    
    # def getRegUpdates(s,keylist):
    # '''Unused'''
    # ##to update: fridge, pna, 
    # reg = s._cxn.registry()
    # reg.cd(s._dir[:-1])
    # for key in keylist:
        # regCheck(reg,key)
        # s.__setattr__(key,reg.get(key))
    # reg.cd(s._dir)    
    # return s
    
    # def makeSample():#make for more device names
        # '''
        # 1) set sample in reg to the correct dir (/David/data/thischip/powerSweep1)
        # 2) run makeSample()
        # '''
        # actualstdout = sys.stdout
        # ##file =  open(LOG,'a')
        # ##sys.stdout = file
        
        # cxn = labrad.connect()
        
        # reg = cxn.registry()
        
        # reg.cd(USRROOT)
        # powerSweepDir = reg.get('sample')
        # if powerSweepDir[-1] == SCANFOLDERNAME:
            # print "Nesting Directories"
            # cxn.disconnect()
            # return -1
        
        # try:
            # [time,timeH] = util.getTimes()
            # print "\nNew Sample Created on: " + timeH 
            
            # reg.set('sample',TLOC[2:])
            
            # s = switchSession(cxn,user=THEUSR)
            
            # reg.cd(powerSweepDir)
            
            # for key in s.keys():
                # if key == SCANFOLDERNAME:
                    # continue
                # if(key in reg.dir()[1]):
                    # continue
                # reg.set(key,s[key])
            
            # reg.cd(USRROOT)
            # reg.set('sample',TLOC[2:] + [SCANFOLDERNAME])
            
            # s = switchSession(cxn,user=THEUSR)
            
            # reg.cd(powerSweepDir)
            
            # if(SCANFOLDERNAME not in reg.dir()[0]):
                # reg.mkdir(SCANFOLDERNAME)
            
            # reg.cd(SCANFOLDERNAME)
            
            # for key in s.keys():
                # reg.set(key,s[key])
            
            # reg.cd(USRROOT)
            # reg.set('sample',powerSweepDir + [SCANFOLDERNAME])
            
        # except Exception, e:
            # sys.stdout = actualstdout
            # ##file.close()
            # print "Failed, exception " + str(e)
            # reg.cd(USRROOT)
            # reg.set('sample',powerSweepDir)
            # return 0
        
        # s = switchSession(cxn,user=THEUSR)
        # sys.stdout = actualstdout
        # ##file.close()
        
        # return s
    pass
