'''

FluxNoiseAnalysis/calibraiton.py

Code relating to the analysis of calibration data (put a signal of known 
amplitude down the Z line).

'''

from __init__ import *

def calibration(a, bandwidth = 0.01, **kwargs):
    ac = a['RTO Calibration']
    ds = getDataset(a._cxn.data_vault, ac['dataset'], a['path'])
    #kwargs['plot'] = kwargs.get('plot', False)
    #kwargs['plotRawFFT'] = kwargs.get('plotRawFFT', True)
    kwargs['rawOnly'] = kwargs.get('rawOnly', True)
    r = doFFTstuff(ds, **kwargs)
    f = plt.figure()
    ax = f.add_subplot(111)
    ax.loglog(r['frequencies'], r['S'], '.')
    
    ax.set_xlabel("Frequency [Hz]")
    ax.set_ylabel("Power [GHz^2 / Hz]")
    #if 'powerFloor' not in ac.keys():
    #    pf = input("Power floor, GHz^2/Hz: ")
    #    ac['powerFloor'] = pf * (GHz*GHz/Hz)
    power, bins = _integrateRange(r['frequencies'], r['S'], ac['freqApplied']['Hz'], bandwidth, returnBins=True)
    #power -= ac['powerFloor']['GHz^2/Hz'] * bandwidth
    ax.loglog(r['frequencies'][bins], r['S'][bins], '.')
    f.show()
    print "Integrated power: %.2e GHz^2" % power
    print "Cosine power for %.2e V = %.2e V^2" % (ac['voltApplied']['V'], ac['voltApplied']['V']**2/2)
    print "df/dV = %.2e GHz/V" % np.sqrt(power / (ac['voltApplied']['V']**2/2))
    print "df/dV from Z spectroscopy = %.2e GHz/V" % np.sqrt(a['dfdDacAmp']/0.38)**2
    #print "Multiplied: %.2e V^2" % (power / (a['dfdDacAmp']*0.38)**2)
    #print "Expected: %.2e V => %.2e V^2" % (ac['voltApplied'][V], ac['voltApplied'][V]**2)

    return r

    
def calibrationRaw(a, plot=True, fit=True, **kwargs):
    ac = a['RTO Calibration']
    ds = getDataset(a._cxn.data_vault, ac['dataset'], a['path'])
    phases, holdTime = checkPhases(a, ds=ds, returnPhases=True, returnTime=True, plot=False)
    times = ds.data[:,0] * ds.parameters.timeStep
    freqs = phases / holdTime
    if fit:
        def cos(t, f, amp, phase, off):
            return amp*np.cos(2*pi*f * t + phase) + off
        popt = [ac['freqApplied']['Hz'], (freqs.max() - freqs.min())/2, np.pi/2, freqs.mean()]
        popt, pcov = curve_fit(cos, times, freqs, popt)
        print "Fit: f=%.4e Hz, amp=%.2e GHz" % (popt[0], popt[1])
    if plot:
        plt.figure()
        plt.plot(times, freqs, 'o', label='Raw Calibration Data')
        if fit:
            plt.plot(times, cos(times, *popt), '--', label="Fit: f=%.2e Hz, amp=%.2e GHz" % (popt[0], popt[1]))
        plt.xlabel('Time [s]', fontsize=19)
        plt.ylabel('$\Delta$ Frequency [GHz]', fontsize=19)
        l = 3.0/ac['freqApplied']['Hz']
        startTime = random.random() * (times.max()-l)
        plt.xlim([startTime, startTime+l])
        plt.legend(numpoints=1, loc='best')
    ampFreq = np.sqrt(2) * np.std(freqs)
    ampVolt = ac['voltApplied']['V']
    print "freq amp = %.2e GHz" % ampFreq
    print "volt amp = %.2e V" % ampVolt
    print "df/dV = %.2e GHz / V" % (ampFreq/ampVolt)
    print "compare to df/dV from Z spectroscopy = %.2e GHz/V" % (a['dfdDacAmp']*0.38)
    ac['frequency amplitude'] = ampFreq * GHz
    ac['df/dV'] = ampFreq/ampVolt * (GHz/V)
    
    return times, phases
    
    
def _integrateRange(x, y, center, bandwidth, returnBins=False):
    bins = np.logical_and(x > center-bandwidth/2.0, x < center+bandwidth/2.0)
    widths = np.gradient(x)
    if returnBins:
        return np.sum(y[bins]*widths[bins]), bins
    return np.sum(y[bins]*widths[bins])
