# This module contains functions used in analyzing data measured from transmon qubits
# Note that this should be compared with pyle/sim/transmon. Lots of shared functionality.
# Should probably merge those two files in some intelligent way


import numpy as np

#Transmon frequency is given by
# f_10 = \sqrt(8 f_J f_C) \sqrt(\cos(\pi \phi / \Phi_0)) - f_C
# where f_J and f_C are the characteristic frequencies of the junction
# and capacitance, ie f_J = E_J/h.

def frequencyVsFluxModel(phi, f_J, f_C):
    """Theoretical frequency vs flux of transmon
    phi - scalar: SQUID flux / Phi0
    """
    return np.sqrt(f_J*f_C)*np.sqrt(np.cos(np.pi * phi)) - f_C
    
def frequencyVsFluxModel_experiment(bias, fp, dFluxdBias, fluxOffset, fC):
    """Qubit frequency as a function of bias
    
    bias - scalar: bias (whatever units
    fp - scalar: plasma frequency of qubit at zero flux
    dFluxdBias - scalar: flux/Phi0 per unit bias
    fluxOffset - scalar: flux/Phi0 at zero bias
    fC - scalar: Ec/h. Also equal to -1 * anharmonicity
    """
    return fp*np.cos(np.pi * (dFluxdBias*bias + fluxOffset))**0.5 - fC
