import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import frequencyScan as fs
import powerScan as ps
import noiseScan as ns
import time

# Transmission and noise power as a function of frequency and flux for a fixed current bias. 
# It also record the Transmission as a function of power and flux, at the optimal frequency where maximum Transmission occurs.      
def SaturationPower(sample):
    print 'loading parameters'
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device()
    samp, dev = util.loadDeviceType(sample, 'PNA')
    SA = sample['noiseMeas']
    FS = sample['fluxScan']
    PNA = sample['PNA']
    dv = cxn.data_vault
    
    I_b= np.arange(0.33,0.34,0.01)
    flux = np.arange(0.671,0.684,0.001)
    
    for I_bs in I_b:
        print 'I_b=', I_bs
        
        dc.channel_set_voltage(3, 'B', 1, 0, I_bs)
        dv.new('Transmission at %f V CurrentBias (Forward)' %I_bs,['Flux Bias [V]','Frequency[GHz]'],['Transmission amplitude [dB]','Noise[dBm]'])
    
        p = dv.packet()
        for key in PNA.keys():
            p.add_parameter(key, PNA.__getitem__(key))
        for key in SA.keys():
            if key in PNA.keys():
                pass
            else:
                p.add_parameter(key, SA.__getitem__(key))
        p.send()         
        
        points=np.array([0,0,0])
        for fluxs in flux:  
            print 'flux=', fluxs

            dc.channel_set_voltage(3, 'A', 1, 0, fluxs)
            data = fs.freqScan(sample, save = False)
            
            freqs = np.asarray(data[:,0])
            freqs.shape = (len(freqs),1)
            FluxPoints = np.asarray([fluxs]*len(freqs))
            FluxPoints.shape = (len(freqs),1)
            mag = np.asarray(data[:,1])
            mag.shape = (len(mag),1)

            newData= np.hstack((FluxPoints, freqs, mag))

            Optfreq=freqs[mag.argmax()]
            Optfreqvalue=Optfreq[0]
            Optidata=np.hstack((fluxs,Optfreqvalue))
            
            Optmag=max(mag)
            Optidata=np.hstack((Optidata,Optmag))
            points = np.vstack((points,Optidata))
             
            #SA['name'] = 'Noise vs frequency for %f VFlux' %fluxs
            try:
                datan = ns.takeNoiseScan(sample, save = False)
            except:
                print 'initial noise scan failed'
                print 'trying again'
                time.sleep(5)
                try:
                    datan = ns.takeNoiseScan(sample, save = False)
                except:
                    print 'something may be wrong'
                    datan = np.vstack((freqs,[0]*SA['numberOfPoints'])).T

            noise = np.asarray(datan[:,1])   
            noise.shape = (len(noise),1)

            newData = np.hstack((newData, noise))
  
            dv.add(newData)
            
        dv.new('Extract maximum gain points',['Flux Bias [V]'],['Frequency[GHz]','Transmission amplitude [dB]'])
        
        Datapoint=np.delete(points,0,axis=0)        
        dv.add(Datapoint)
        
        dv.new('Power scan',['Flux Bias [V]','Power [dBm]'],['Saturation[dB]','Frequency[GHz]'])
        
        i=0        
        for fluxs in flux:
            print 'flux=', fluxs
                    
            dc.channel_set_voltage(3, 'A', 1, 0, fluxs)
            
            PNA['V_flux']=fluxs
            PNA['CWfrequency']=Datapoint[i,1]
            optfre=Datapoint[i,1]                    
            dataP=ps.powerScan(sample,save=False)
            
            power=np.asarray(dataP[:,0])
            power.shape = (len(power),1)
            FluxPoints = np.asarray([fluxs]*len(power))
            FluxPoints.shape = (len(power),1)
            optfrePoints = np.asarray([optfre]*len(power))
            optfrePoints.shape = (len(power),1)
            magP = np.asarray(dataP[:,1])
            magP.shape = (len(magP),1)
            
            newDataP = np.hstack((FluxPoints, power, magP))
            newDataP = np.hstack((newDataP,optfrePoints))
            i=i+1
                   
            dv.add(newDataP)        

# For a fixed bias current and fixed flux, it records the Transmission as a function of power and frequency.
def SaturationPowerInBand(sample):
    print 'loading parameters'
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device()
    samp, dev = util.loadDeviceType(sample, 'PNA')
    SA = sample['noiseMeas']
    FS = sample['fluxScan']
    PNA = sample['PNA']
    dv = cxn.data_vault
        
    I_bs= 0.41
    fluxs=-0.282
    Freq = np.arange(7.101,7.326,0.002)
    
    print 'I_b=', I_bs
    print 'flux=', fluxs
    
    dc.channel_set_voltage(3, 'B', 1, 0, I_bs)
    dc.channel_set_voltage(3, 'A', 1, 0, fluxs)
    
    dv.new('Power in Band scan, 410mV, -282mV',['Frequency [V]','Power [dBm]'],['Transmission[dB]'])
    
    p = dv.packet()
    for key in PNA.keys():
        p.add_parameter(key, PNA.__getitem__(key))
    for key in SA.keys():
        if key in PNA.keys():
            pass
        else:
            p.add_parameter(key, SA.__getitem__(key))
    p.send()         
    
    #PowerPoints=np.array([0,0])
    i=0
    for Freqs in Freq:

        PNA['CWfrequency']=Freqs     
        dataP=ps.powerScan(sample,save=False)
            
        power=np.asarray(dataP[:,0])
        power.shape = (len(power),1)
        FreqPoints = np.asarray([Freqs]*len(power))
        FreqPoints.shape = (len(power),1)

        magP = np.asarray(dataP[:,1])
        magP.shape = (len(magP),1)
            
        newDataP = np.hstack((FreqPoints, power, magP))
        
        #Power1dB=power[magP.argmax()]
        #Powervalue=Power1dB[0]
        #PData=np.hstack((Freqs,Powervalue))
        #PowerPoints = np.vstack((PowerPoints,PData))
        i=i+1
        dv.add(newDataP)  
    
    #dv.new('Extract 1dB compression point',['Frequency [GHz]'],['Transmission[dB]'])    
    #PowerSat=np.delete(PowerPoints,0,axis=0)        
    #dv.add(PowerSat)
            
    
# For a fixed frequency, it records Transmission as a function of bias current and flux.    
def CW_VS_fluxandBias(sample):
    print 'loading parameters'
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device()
    samp, dev = util.loadDeviceType(sample, 'PNA')
    SA = sample['noiseMeas']
    FS = sample['fluxScan']
    PNA = sample['PNA']
    dv = cxn.data_vault
    
    start=-1.4
    stop=0.4
    step=0.01
    
    I_b= np.arange(0.250,0.700,0.01)
    flux = np.arange(start,stop,step)
    
    
    PNA['number of points']=21
    SA['numberOfPoints']=21
    CWfreq=PNA['CWfrequency']
    
    SA['span']=1
    SA['freq']=PNA['CWfrequency']*1000+(SA['span']/2)
    
    dv.new('PNA scan CW scan at %f GHz' %CWfreq,['Current Bias [V]','Flux Bias[V]'],['Transmission [dB]','Noise[dBm]'])
    
    p = dv.packet()
    for key in PNA.keys():
        p.add_parameter(key, PNA.__getitem__(key))
    for key in SA.keys():
        if key in PNA.keys():
            pass
        else:
            p.add_parameter(key, SA.__getitem__(key))
    p.send()         
    
    
    for I_bs in I_b:
        
        dc.channel_set_voltage(3, 'B', 1, 0, I_bs)

        FluxPoints=np.arange(start,stop,step)
        FluxPoints.shape = (len(FluxPoints),1) 
            
        Ib_Points = np.asarray([I_bs]*len(flux))
        Ib_Points.shape = (len(FluxPoints),1)
            
        Transmission = np.asarray([0]*len(flux))
        Transmission.shape = (len(FluxPoints),1)

        Noisen = np.asarray([0]*len(flux))
        Noisen.shape = (len(FluxPoints),1)
            
        i=0    
        for fluxs in flux:            
                    
            dc.channel_set_voltage(3, 'A', 1, 0, fluxs)
            
            dataP=ps.powerScan(sample,save=False)
            Transmission[i] = dataP[0,1]
            
            try:
                datan = ns.takeNoiseScan(sample, save = False)
            except:
                print 'initial noise scan failed'
                print 'trying again'
                time.sleep(5)
                try:
                    datan = ns.takeNoiseScan(sample, save = False)
                except:
                    print 'something may be wrong'
                    datan = np.vstack((freqs,[0]*SA['numberOfPoints'])).T
            
            Noisen[i]=datan[0,1]
            
            i=i+1

        newData = np.hstack((Ib_Points, FluxPoints, Transmission))            
        newData = np.hstack((newData,Noisen))
        dv.add(newData)                
        
        