import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import frequencyScan as fs
import powerScan as ps
import noiseScan as ns

def measure(sample):

    cxn = sample._cxn
    samp, dev = util.loadDeviceType(sample, 'fluxScan')
    fluxScan = dev[0]
    points = fluxScan['points']
    hittite = cxn.hittite_t2100_server
    hittite.select_device(fluxScan['Hittite'])
    hittite.output(False)
    hittite.output(True)
    hittite.output(False)
    dac = fluxScan['dacBoard']
    chan = fluxScan['fastbiasDacChannel']
    PNA = sample['PNA']
    SA = sample['noiseMeas']
    pumpMethod = fluxScan['HittitePumpMethod']
    

    for point in points:
        bias = float(point[0])/1000
        freq = float(point[1])
        pumpPow = float(point[2])
        
        hittite.frequency(freq*1e9)
        hittite.amplitude(pumpPow)
        
        sb.set_fb(dac, cxn, bias*V, chan)
        
        SA['ParampPumpFreq'] = freq
        SA['ParampPumpPower'] = 'off'
        SA['ParampPumpMethod'] = pumpMethod
        SA['ParampBiasVoltage'] = bias
        SA['freq'] = freq*1000
        ns.takeNoiseScan(sample)
        hittite.output(True)
        time.sleep(20)
        SA['ParampPumpPower'] = pumpPow
        ns.takeNoiseScan(sample)
        hittite.output(False)
        
        PNA['startFreq'] = freq-0.05
        PNA['stopFreq'] = freq+0.05
        PNA['ParampPumpFreq'] = freq
        PNA['ParampPumpPower'] = 'off'
        PNA['ParampBiasVoltage'] = bias
        PNA['power'] = -85
        fs.freqScan(sample)
        
        hittite.output(True)
        PNA['pumpPower'] = pumpPow
        fs.freqScan(sample)
        
        PNA['startPow'] = -90
        PNA['stopPow'] = -50
        PNA['CWfrequency' ] = freq+0.003
        ps.powerScan(sample)
        
        hittite.output(False)
        
        print 'done with point'
        
        
        
        
def measureSideband(sample):

    cxn = sample._cxn
    samp, dev = util.loadDeviceType(sample, 'fluxScan')
    fluxScan = dev[0]
    points = fluxScan['points']
    hittite = cxn.hittite_t2100_server
    hittite.select_device(fluxScan['Hittite'])
    hittite.output(False)
    hittite.output(True)
    hittite.output(False)
    hittite.select_device(fluxScan['Hittite2'])
    hittite.output(False)
    hittite.output(True)
    hittite.output(False)
    dac = fluxScan['dacBoard']
    chan = fluxScan['fastbiasDacChannel']
    PNA = sample['PNA']
    SA = sample['noiseMeas']
    pumpMethod = fluxScan['HittitePumpMethod']
    detune = fluxScan['SidebandDetuning']
    

    for point in points:
        bias = float(point[0])/1000
        freq = float(point[1])
        pumpPow = point[2]
        pow1,pow2 = float(pumpPow.split(',')[0]), float(pumpPow.split(',')[1])
        
        hittite.select_device(fluxScan['Hittite'])
        hittite.frequency((freq+detune)*1e9)
        hittite.amplitude(pow1)
        
        hittite.select_device(fluxScan['Hittite2'])
        hittite.frequency((freq-detune)*1e9)
        hittite.amplitude(pow2)
        
        sb.set_fb(dac, cxn, bias*V, chan)
        
        SA['ParampPumpFreq'] = freq
        SA['ParampPumpPower'] = 'off'
        SA['ParampPumpMethod'] = pumpMethod
        SA['ParampBiasVoltage'] = bias
        SA['freq'] = freq*1000
        ns.takeNoiseScan(sample)
        hittite.select_device(fluxScan['Hittite'])
        hittite.output(True)
        hittite.select_device(fluxScan['Hittite2'])
        hittite.output(True)
        time.sleep(20)
        SA['ParampPumpPower'] = pumpPow
        ns.takeNoiseScan(sample)
        hittite.select_device(fluxScan['Hittite'])
        hittite.output(False)
        hittite.select_device(fluxScan['Hittite2'])
        hittite.output(False)
        
        PNA['startFreq'] = freq-0.05
        PNA['stopFreq'] = freq+0.05
        PNA['ParampPumpFreq'] = freq
        PNA['ParampPumpPower'] = 'off'
        PNA['ParampPumpMethod'] = pumpMethod
        PNA['ParampBiasVoltage'] = bias
        PNA['power'] = -85
        fs.freqScan(sample)
        
        hittite.select_device(fluxScan['Hittite'])
        hittite.output(True)
        hittite.select_device(fluxScan['Hittite2'])
        hittite.output(True)
        PNA['ParampPumpPower'] = pumpPow
        fs.freqScan(sample)
        
        PNA['startPow'] = -90
        PNA['stopPow'] = -50
        PNA['CWfrequency' ] = freq+0.003
        ps.powerScan(sample)
        
        hittite.select_device(fluxScan['Hittite'])
        hittite.output(False)
        hittite.select_device(fluxScan['Hittite2'])
        hittite.output(False)
        
        print 'done with point'
     
        
        
        
        
                