import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import frequencyScan as fs
import powerScan as ps
import noiseScan as ns
import time

def measure(sample):
    print 'loading parameters'
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device(0)
    samp, dev = util.loadDeviceType(sample, 'fluxScan')
    fluxScan = dev[0]
    points = fluxScan['points']
    pumpAddr = fluxScan['Anritsu']
    coolDown = fluxScan['waitTime']
    anritsu = cxn.anritsu_server
    anritsu.select_device(pumpAddr)
    anritsu.output(False)
    dac = fluxScan['dacBoard']
    chan = fluxScan['fastbiasDacChannel']
    PNA = sample['PNA']
    Noise = sample['noiseMeas']
    freq = float(points[0][1])
    print 'ok'
    print 'setting up PNA'
    PNA['startFreq'] = freq-0.5
    PNA['stopFreq'] = freq+0.5
    PNA['ParampPumpFreq'] = freq
    PNA['power'] = -60
    PNA['startPow'] = -70
    PNA['stopPow'] = -30
    PNA['number of points'] = 1001
    PNA['CWfrequency' ] = freq+0.001
    PNA['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    PNA['ParampPumpMethod'] = fluxScan['PumpMethod']
    PNA['bandwidth'] = 500
    print 'setting up Spectrum Analyzer'
    Noise['ParampPumpFreq'] = freq
    Noise['ParampPumpMethod'] = fluxScan['PumpMethod']
    Noise['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    Noise['freq'] = freq*1000
    Noise['span'] = 1000
    Noise['numberOfPoints']=1001
    
    print 'setting up Anritsu'
    if PNA['ParampPumpMethod'] == 'input w':
        anritsu.frequency((freq)*GHz)
    if PNA['ParampPumpMethod'] == '2w':
        anritsu.frequency((2*freq)*GHz)
    anritsu.output(True)
    print 'settling...'
    time.sleep(coolDown)

    print 'taking data'
    #counter = 0
    for point in points:
        try:
            # if counter > 5:
                # anritsu.output(False)
                # time.sleep(300)
                # counter = 0
            anritsu.output(True)
            bias = float(point[0])
            freq = float(point[1])
            pumpPow = float(point[2])
            PNA['ParampPumpPower'] = pumpPow
            PNA['ParampBiasVoltage'] = bias
            Noise['ParampPumpPower'] = pumpPow
            Noise['ParampBiasVoltage'] = bias
            setFlux(bias, fluxScan, sb, dc)
            anritsu.amplitude(pumpPow)
            ns.takeNoiseScan(sample)
            fs.freqScan(sample)
            ps.powerScan(sample)
            print 'done with point'
        #counter += 1
        except:
            try:
                # if counter > 5:
                # anritsu.output(False)
                # time.sleep(300)
                # counter = 0
                anritsu.output(True)
                bias = float(point[0])
                freq = float(point[1])
                pumpPow = float(point[2])
                PNA['ParampPumpPower'] = pumpPow
                PNA['ParampBiasVoltage'] = bias
                Noise['ParampPumpPower'] = pumpPow
                Noise['ParampBiasVoltage'] = bias
                setFlux(bias, fluxScan, sb, dc)
                anritsu.amplitude(pumpPow)
                ns.takeNoiseScan(sample)
                fs.freqScan(sample)
                ps.powerScan(sample)
                print 'done with point'
            except:
                print error
    
    setFlux(0, fluxScan, sb, dc)
        
        
def tuneMeasure(sample):
    print 'loading parameters'
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device(0)
    samp, dev = util.loadDeviceType(sample, 'fluxScan')
    fluxScan = dev[0]
    points = fluxScan['points']
    pumpAddr = fluxScan['Anritsu']
    coolDown = fluxScan['waitTime']
    anritsu = cxn.anritsu_server
    anritsu.select_device(pumpAddr)
    anritsu.output(False)
    dac = fluxScan['dacBoard']
    chan = fluxScan['fastbiasDacChannel']
    PNA = sample['PNA']
    Noise = sample['noiseMeas']
    
    freq = float(points[0][1])
    print 'setting up PNA'
    PNA['power'] = -20
    PNA['startPow'] = -30
    PNA['stopPow'] = 10
    PNA['number of points'] = 1001
    PNA['CWfrequency' ] = freq+0.001
    PNA['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    PNA['ParampPumpMethod'] = fluxScan['PumpMethod']
    PNA['bandwidth'] = 100
    print 'setting up Spectrum Analyzer'
    Noise['ParampPumpFreq'] = freq
    Noise['ParampPumpMethod'] = fluxScan['PumpMethod']
    Noise['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    Noise['span'] = 1000
    Noise['numberOfPoints']=1001
    print 'setting up Anritsu'
    if PNA['ParampPumpMethod'] == 'input w':
        anritsu.frequency((freq)*GHz)
    if PNA['ParampPumpMethod'] == '2w':
        anritsu.frequency((2*freq)*GHz)
    time.sleep(coolDown)
    
    print 'taking data'
    for point in points:
        # try:
        bias = float(point[0])
        freq = float(point[1])
        if PNA['ParampPumpMethod'] == 'input w':
            anritsu.frequency((freq)*GHz)
            PNA['ParampPumpFreq'] = freq
            Noise['ParampPumpFreq'] = freq
        if PNA['ParampPumpMethod'] == '2w':
            anritsu.frequency((2*freq)*GHz)
            PNA['ParampPumpFreq'] = 2*freq
            Noise['ParampPumpFreq'] = 2*freq
        pumpPow = float(point[2])
        PNA['ParampPumpPower'] = pumpPow
        PNA['ParampBiasVoltage'] = bias
        PNA['name'] = 'PNA Scan at %f GHz' %freq
        PNA['startFreq'] = freq-0.5
        PNA['stopFreq'] = freq+0.5
        PNA['CWfrequency'] = freq+0.001
        Noise['ParampPumpPower'] = pumpPow
        Noise['ParampBiasVoltage'] = bias
        Noise['name'] = 'Paramp Noise Scan at %f GHz' %freq
        Noise['freq'] = freq*1000
        setFlux(bias, fluxScan, sb, dc)
        anritsu.amplitude(pumpPow)
        anritsu.output(False)
        ns.takeNoiseScan(sample)
        fs.freqScan(sample)
        anritsu.output(True)
        time.sleep(2)
        ns.takeNoiseScan(sample)
        fs.freqScan(sample)
        ps.powerScan(sample)
        print 'done with point'
        # except:
            # print 'data point failed'
            
    setFlux(bias, fluxScan, sb, dc)       
        
        
def setFlux(bias,fluxScan,sb,dc):
    if fluxScan['dacBoard']=='DC rack':
        dc.channel_set_voltage(3, 'A', 1, 0, bias)
    else:
        sb.set_fb(fluxScan['dacBoard'], cxn, bias*V, fluxScan['fastbiasDacChannel'])
        
End = True


        
                