import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import frequencyScan as fs
import powerScan as ps
import noiseScan as ns

def measure(sample):
    print 'loading parameters'
    cxn = sample._cxn
    samp, dev = util.loadDeviceType(sample, 'fluxScan')
    fluxScan = dev[0]
    points = fluxScan['points']
    pumpAddr = fluxScan['Hittite']
    coolDown = fluxScan['waitTime']
    hittite = cxn.hittite_t2100_server
    hittite.select_device(pumpAddr)
    hittite.output(False)
    dac = fluxScan['dacBoard']
    chan = fluxScan['fastbiasDacChannel']
    PNA = sample['PNA']
    Noise = sample['noiseMeas']
    
    freq = float(points[0][1])
    print 'setting up PNA'
    PNA['startFreq'] = freq-0.05
    PNA['stopFreq'] = freq+0.05
    PNA['ParampPumpFreq'] = freq
    PNA['power'] = -70
    PNA['startPow'] = -80
    PNA['stopPow'] = -40
    PNA['CWfrequency' ] = freq+0.0005
    PNA['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    PNA['ParampPumpMethod'] = fluxScan['PumpMethod']
    print 'setting up Spectrum Analyzer'
    Noise['ParampPumpFreq'] = freq
    Noise['ParampPumpMethod'] = fluxScan['PumpMethod']
    Noise['ParampPumpRTAttenuation'] = fluxScan['PumpRTAtten']
    Noise['freq'] = freq*1000
    Noise['span'] = 100
    print 'setting up Hittite'
    if PNA['ParampPumpMethod'] == 'input w':
        hittite.frequency((freq)*GHz)
    if PNA['ParampPumpMethod'] == '2w':
        hittite.frequency((2*freq)*GHz)
    hittite.output(True)
    print 'settling...'
    time.sleep(coolDown)
    
    print 'taking data'
    for point in points:
        try:
            bias = float(point[0])
            freq = float(point[1])
            pumpPow = float(point[2])
            
            sb.set_fb(dac, cxn, bias*V, chan)
            hittite.amplitude(pumpPow)
            PNA['ParampPumpPower'] = pumpPow
            PNA['ParampBiasVoltage'] = bias
            fs.freqScan(sample)
            Noise['ParampPumpPower'] = pumpPow
            Noise['ParampBiasVoltage'] = bias
            ns.takeNoiseScan(sample)
            ps.powerScan(sample)
            print 'done with point'
        except:
            print 'data point failed'
    
    sb.set_fb(dac, cxn, 0*V, chan)
        
        
        
        
        
        
        
        
        
        
        
                