import numpy as np
import matplotlib.pyplot as plt
import pyle.envelopes as env
import pyle.dataking.DevJim.CatchRelease.gateCompiler as gc

TIME = 2500
STEP = 0.1


def runQubits(server, devices, stats, probs=None, dataFormat=None,
              debug=False):
    """A generic sequence for running multiple devices with the Qubit Sequencer.
    qubits
    These parameters control the sequence and how it should be run:
    
    server - labrad server to which we send the sequence (Qubit Sequencer).
    devices - list of dict-like objects (typically loaded from the registry)
        containing the configuration parameters of all devices involved in
        the sequence.
    stats - the number of repetitions of the experiment to be run.
    
    
    These parameters control how the data should be processed before it is
    returned to the caller (note that this processing is done by the Qubit
    Sequencer server for speed; see the server documentation for more info): 
    
    dataFormat - string description of how data should be returned
    probs - a list of which probabilities should be returned.  For example,
        when only measuring a single qubit this should be [1] since we only
        care about the |1>-state probability.
    
    DEPRICATED DATA FORMAT FLAGS
    raw - a flag indicating whether to return raw switching data (converted
        to microseconds) or processed probabilities (the default). Replaced
        by 'raw_microseconds' for SQUID readout and 'iq' for resonator
        readout.
    separate - a flag indicating whether to return separate probabilities
        for each qubit (N possibilities), or combined probabilities of the
        multi-qubit states (2**N posibilities; the default). Replaced by
        'probs_separate'.
    switches - Replaced by 'raw_switches'
    """
    plotTraces(devices)
#    p = server.packet()
#    makeSequence(p, devices)
#    p.build_sequence()
#    return sendPacket(p, debug)


def plotTraces(devices):
    plt.figure()
    times = np.linspace(0,TIME,TIME/STEP)
    for device in devices:
        if isinstance(device, gc.Qubit):
            channels = ['xy','z','rr']
        elif isinstance(device, gc.Resonator):
            channels = ['xy']
        elif isinstance(device, gc.TunableCoupler):
            channels = ['z']
        else:
            channels = []
        for channel in channels:
            trace = device.get(channel, env.NOTHING)
            plt.plot(times, trace(times).real, '.', label=device.__name__+' '+channel)
    plt.legend(loc='upper right')
    plt.xlabel('Time [ns]')
    plt.ylabel('Traces')
    plt.grid('on')