#Standard packages
import numpy as np
from math import atan2
import scipy
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import pyle

import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking import singleQubit as sq
from pyle.dataking import multipleQubit as mq
import pyle.plotting.dstools as ds
import pyle.plotting.plothelper as ph
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
from pyle.datavault import DataVaultWrapper
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.analysis import iq as iq
from pyle.analysis import readout as readout
from pyle.dataking.optimization import optimization as popt

from pyle.dataking.benchmarking import randomizedBenchmarking as rb

#import pyle.dataking.DevDaniel.Qubit.readout as readout

from pyle import signalprocessing as signalprocessing
import pyle.quantum as quantum
import time
import copy
from matplotlib.mlab import find
from scipy.interpolate import UnivariateSpline
import itertools
from pyle.dataking import zfuncs as zfuncs

class NestedDict(dict):
    def __missing__(self, key):
        self[key] = NestedDict()
        return self[key]
    
def reshapeRepCodeProbs(prob, qubits):
    probsAll=[]
    #get measure qubit data
    for idx,q in enumerate(qubits):
        if q['_surf']=='measure':
            probsMeasure=prob[idx,:,1]
            probsAll = np.hstack((probsAll,probsMeasure))
    #get data qubit data
    for idx,q in enumerate(qubits):
        if q.get('_surf',None) != 'measure':
            probsAll = np.hstack((probsAll,prob[idx,-1,1] ))
            
    return probsAll

def buildRepetitionCode(devs,inputStates,paramDic,cycles):
    #inputs:
    #-devs
    #-inputStates: [0,1,0...]
    #-paramDic: control sq phases, CZ params, -detunes & phases
    #-cycles (last cycle measures all)
    if inputStates is None:
        inputStates=[0]*len(devs)    
    alg = gc.Algorithm(agents=devs)    
    #build qubit dict: qubitDict[qubitName]=qubit
    qubitDict={}
    for qubit in alg.qubits:
        qubitDict[qubit.__name__]=qubit    
    #figure out the CZ controls and CZ control-target pairs   
    controlTargetList=[] #list of all control-target pairs
    controlList=[] #unique list of controls
    for qubit in alg.qubits:
        aczKeys = [x for x in qubit.keys() if 'aczAmp' in x] #aczAmp -> CZ actor
        for key in aczKeys:
            aczAmp = qubit[key]
            targetName = key.strip('aczAmp')        
            controlTargetList += [[ qubit.__name__ , targetName ]]
            if qubit.__name__ not in controlList:
                controlList+= [qubit.__name__]
    #set paramDic parameters. Syntax: paramDic['q0']['q1aczPhase']=0.1
    paramDicToQubits(alg.qubits, paramDic)
    
                #print key, params[key]
                #print qubitDict[qubit.__name__][key]
                #h46uy                
    #initialize qubits
    for st, q in zip(inputStates, alg.qubits):
        alg[gates.MoveToState([q], 0, st)]                     
    #X stabilizers
    for n in range(cycles):

        #old sequence
        for controlQubitName in controlList:
            alg[gates.PiHalfPulse([qubitDict[controlQubitName]],phase=3*np.pi/2.)] #-Y/2            
        for controlQubitName,targetQubitName in controlTargetList:
            alg[gates.adiabaticCZ([qubitDict[controlQubitName],qubitDict[targetQubitName]])]         
        for controlQubitName in controlList:
            alg[gates.PiHalfPulse([qubitDict[controlQubitName]],phase=np.pi/2.)] #Y/2                
            
        #measure qubits
        if n+1 != cycles:
            alg[gates.Measure( [ q for q in alg.qubits if q.get('_surf',None)=='measure' ] , name = n)] #measure qubits only  
    #end: measure all
    alg[gates.Measure(alg.qubits, name='Final')]
    alg.compile(correctXtalkZ=True) 
    return alg

def buildRepetitionCodeIS(alg,inputStates,cycles,instructionSet):
    #inputs:
    #-devs
    #-inputStates: [0,1,0...]
    #-paramDic: control sq phases, CZ params, -detunes & phases
    #-cycles (last cycle measures all)
    
    qubitDict = {}
    for qubit in alg.qubits:
    
        qubitDict[qubit.__name__] = qubit
    if inputStates is None:
        inputStates=[0]*len(alg.qubits)      
    #figure out the CZ controls and CZ control-target pairs   
    controlTargetList=[] #list of all control-target pairs
    controlList=[] #unique list of controls
    for qubit in alg.qubits:
        aczKeys = [x for x in qubit.keys() if 'aczAmp' in x] #aczAmp -> CZ actor
        for key in aczKeys:
            aczAmp = qubit[key]
            targetName = key.strip('aczAmp')        
            controlTargetList += [[ qubit.__name__ , targetName ]]
            if qubit.__name__ not in controlList:
                controlList+= [qubit.__name__]

    #initialize qubits
    for st, q in zip(inputStates, alg.qubits):
        alg[gates.MoveToState([q], 0, st)]                     
    #X stabilizers
    for n in range(cycles):        
        #new sequence
        length=np.max( [np.alen(instructionSet[key]) for key in instructionSet.keys()] )
        for step in range(length):
            alg[gates.Sync(alg.qubits)]
            for qubitName in instructionSet.keys():
                instruction=instructionSet[qubitName][step][0]
                if instruction=='-Y/2':
                    alg[gates.PiHalfPulse([qubitDict[qubitName]],phase=3*np.pi/2.)] #-Y/2  
                elif instruction=='Y/2':
                    alg[gates.PiHalfPulse([qubitDict[qubitName]],phase=np.pi/2.)] #Y/2
                elif instruction=='echo':
                    alg[gates.PiPulse([qubitDict[qubitName]],phase=0.)] #X                      
                    alg[gates.PiPulse([qubitDict[qubitName]],phase=np.pi)] #-X                     
                elif 'CZ' in instruction:
                    targetQubitName=instruction.strip('CZ')
                    alg[gates.adiabaticCZ([qubitDict[qubitName],qubitDict[targetQubitName]])]
                elif 'I':
                    pass
                else:                    
                    raise Exception('not implemented: '+instruction)            
        #measure qubits
        if n+1 != cycles:
            alg[gates.Measure( [ q for q in alg.qubits if q.get('_surf',None)=='measure' ] , name = n)] #measure qubits only  
    #end: measure all
    alg[gates.Measure(alg.qubits, name='Final')]  
    
def repetitionCodeIS(measure):
    #every 'cycle' consists of 1-2 single qubit gates, or 1 CZ gate.
    #we list only the CZ controls/measure, the rest follows naturally
    #I=idle, CZ# is CZ with qubit #. Other gates are named normally
    instructionSetFull={}
    instructionSetFull['q1']=[['-Y/2'] , ['CZq0']  , ['CZq2'] , ['Y/2']]
    instructionSetFull['q3']=[['-Y/2'] , ['CZq4']  , ['echo']   , ['CZq2'] , ['Y/2']]
    instructionSetFull['q5']=[['I']    , ['-Y/2'] , ['CZq4'] , ['CZq6'] , ['Y/2']]
    instructionSetFull['q7']=[['-Y/2'] , ['CZq6']  , ['CZq8'] , ['Y/2']]
    instructionSet={}
    for meas in measure:
        if meas in instructionSetFull.keys():
            instructionSet[meas]=instructionSetFull[meas]
    #make same length
    length=np.max( [np.alen(instructionSet[key]) for key in instructionSet.keys()] )
    for key in instructionSet.keys():
        diffL= length-np.alen(instructionSet[key])
        instructionSet[key] += [['I']]*diffL            
    return instructionSet
    
def testBuildRepetitionCode(Sample,measure,cycles=2):
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    paramDic=NestedDict()
    #paramDic['q0']['q1aczDetune']=4*GHz
    alg=buildRepetitionCode(devs,[0]*len(devs),paramDic,cycles)
    ph.pulses(alg)
    
def testBuildRepetitionCodeIS(Sample,measure,cycles=2):
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    alg = gc.Algorithm(agents=devs)
    buildRepetitionCodeIS(alg,[0]*len(measure),cycles,repetitionCodeIS(measure))
    alg.compile(correctXtalkZ = True)
    ph.pulses(alg, spacingByQubit=False, spacing = -1)    
    
def fixUnits(Sample, measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']):
    
    def stripUnits(thing):
        if isinstance(thing,Value):
            thing = thing.value
        return thing
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    for q,Q in zip(qubits,Qubits):
        rates = q['settlingRates']
        rates = [stripUnits(r) for r in rates]
        Q['settlingRates'] = [r*GHz for r in rates]
        Q['adc demod phase'] = stripUnits(Q['adc demod phase']) * rad

def setKey(Sample, measure = None, key = None, value = None):

    if measure=='all':
        measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']
    if measure=='data':
        measure = ['q0', 'q2', 'q4', 'q6', 'q8']
    if measure=='measure':
        measure = ['q1','q3','q5', 'q7']
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for Q in Qubits:
        Q[key] = value
        
def removeDeprecatedAczKeys(Sample, measure=None):

    """
    keys that should be in control (Measure) qubit:
        aczPadding
    keys that should be in target/parasitic (data) qubit:
        aczDetunePadding
        aczDetuneWidth
    keys to be deleted in data:
        qnqnAczDetunePadding
        qnqnAczDetuneWidth
        aczPadding
    """

    if measure is None:
        measure = Sample['config']
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits, Qubits):
        for key in q.keys():
            if key == 'aczPadding' and q['_surf'] == 'data':
                del Q[key]
                continue
            if 'aczDetunePadding' in key and key[0]=='q':
                del Q[key]
                continue
            if 'aczDetuneWidth' in key and key[0]=='q':
                del Q[key]
                continue
                
def portReadoutKeys(Sample, measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    for Q in Qubits:
        Q['readoutDemodLen'] = Q['readoutLen']
        Q['readoutDemodFrequency'] = Q['readoutFrequency']
        
        Q['readoutRingupLength'] = Q['preEmphLen']
        Q['readoutRingupFactor'] = Q['preEmphFactor']
        
        Q['readoutPulseLen'] = Q['readoutLen']
        Q['readoutPulsePower'] = Q['readoutPower']
        Q['readoutPulseWidth'] = Q['readoutWidth']
        Q['readoutPulseFrequency'] = Q['readoutFrequency']
        
        Q['readoutRingdownLength'] = Q['ringdownLen']
        
        del Q['readoutLen']
        del Q['readoutFrequency']
        del Q['readoutWidth']
        del Q['readoutPower']
        del Q['readoutIntegrateIntoRingdown']
            
def aczKeys(Sample, measure, key = 'acz'):
    '''
    key = 'acz' or 'detune'
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if key == 'acz':
    
        qTName = qubits[1].__name__
        qCName = qubits[0].__name__
         
        Qubits[0][qTName+'aczTime'] = 40*ns
        Qubits[0][qTName+'aczW'] = 1.5*ns
        Qubits[0][qTName+'aczAmp'] = -0.5*GHz
        Qubits[0][qTName+'aczLambda'] = [1.0, 0.05, 0.6, -0.2, 0.05]
        Qubits[0][qTName+'aczPhase'] = 0.0
        Qubits[1][qCName+'aczPhase'] = 0.0
        
def fixAczLambdaKey(Sample, measure):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    for q,Q in zip(qubits,Qubits):
        aczKeys = [x for x in q.keys() if x.endswith('aczLambda')]
        for key in aczKeys:
            oldLambda = q[key]
            newLambda = oldLambda[1:]
            Q[key] = newLambda
        
def aczKeyDefaults(Sample, measure):
    # if has aczDetune, make sure had aczDetunePadding, aczDetuneWidth, aczDetunePhase
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits,Qubits):
        aczKeys = [x for x in q.keys() if x.endswith('aczDetune')]
        for key in aczKeys:
            qName = key.strip('aczDetune')
            aczDetunePadding = q.get(qName+'aczDetunePadding',None)
            if aczDetunePadding != 1.5*ns:
                Q[qName+'aczDetunePadding'] = 1.5*ns
            aczDetuneWidth = q.get(qName+'aczDetuneWidth',None)
            if aczDetuneWidth != 1.5*ns:
                Q[qName + 'aczDetuneWidth'] = 1.5*ns
            aczDetunePhase = q.get(qName + 'aczDetunePhase',None)
            if aczDetunePhase is None:
                Q[qName + 'aczDetunePhase'] = 0.0
                
    for q,Q in zip(qubits,Qubits):
        aczKeys = [x for x in q.keys() if x.endswith('aczW')]
        for key in aczKeys:
            qName = key.strip('aczW')
            aczW = q.get(qName+'aczW',None)
            if aczW != 1.5*ns:
                Q[qName+'aczW'] = 1.5*ns
            aczPadding = q.get('aczPadding')
            if aczPadding != 1.5*ns:
                Q['aczPadding'] = 1.5*ns
            
def makeaSwapKey(Sample, measure, targetName = 'rr'):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if targetName:
        qTName = targetName
        qCName = qubits.__name__
    else:
        qTName = qubits[1].__name__
        qCName = qubits[0].__name__
    
    if isinstance(qubits, list):
        Qubit = Qubits[0]
    else:
        Qubit = Qubits
    
    Qubit[qTName+'aSwapTime'] = 40*ns
    Qubit[qTName+'aSwapW'] = 0.0*ns
    Qubit[qTName+'aSwapAmp'] = 0.5*GHz
    Qubit[qTName+'aSwapLambda'] = [1.0, 0.1, np.pi-0.1, -0.2, 0.0]

def allLevels(Sample, measure = ['q0','q1','q2','q3','q4']):
    
    ph.levels(Sample, measure)
    
    #assume odd qubits are measure qubits
    measureQubits = measure[1::2]
    dataQubits = measure[::2]
    
    pairs = []
    for k,mq in enumerate(measureQubits):
        pairs.append([mq, dataQubits[k]])
        pairs.append([mq, dataQubits[k+1]])
        
    for pair in pairs:  
        ph.aczLevels(Sample, measure, pair)

def dataset_to_comp(Sample, dvwData, removeOffset=True):
    
    if not isinstance(dvwData,list):
        dvwData = [dvwData]
    
    measure = dvwData[0].parameters['measure'][0]
    
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    zEnvs = []
    for data in dvwData:
    
        curve = fitting.minimaCuts(data, 'columns', ptsToFit=5, invert=True)
        indices = np.argsort(curve[:,0])
        acStarkShift = curve[:,1] - q['f10']['GHz']
        acStarkShift = acStarkShift[indices]
        
        #Convert frequency to z amplitude
        params = q['zFuncAna']
        M = params[0].value
        fmax = params[1].value
        offset = params[2].value
        fc = params[3].value
        def fitFunc(x, M, fmax, offset, fc):
            return fmax * np.sqrt(np.abs(np.cos(np.pi*M*np.abs(x-offset)))) +\
                   fc * (np.sqrt(np.abs(np.cos(np.pi*M*np.abs(x-offset)))) -1.0)
        zControl = np.linspace(-0.3, 0.3, 1000)
        dfControl = fitFunc(zControl, M, fmax, offset, fc) - q['f10']['GHz']
        
        interp_freq2z = scipy.interpolate.interp1d(dfControl[::-1], zControl[::-1], kind='linear')
        z = interp_freq2z(acStarkShift)
        if removeOffset:
            z += -z[0]
        zEnv = env.GRAPE_CZ(0, curve[indices,0], -z)
        zEnvs.append(zEnv)
        
    return sum(zEnvs)

def stark_compensation(Sample, measure, name='AC Stark Z Compensated',
                       df=0.250*GHz, npts = 30, delays=None, zCompensation=None, zEnv=None,
                       piPulse=False, excitationLen=None, excitationAmp=0.1,
                       stats=900, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    
    f10_GHz = q['f10']['GHz']
    freqs = np.linspace(f10_GHz-df['GHz']/2.,f10_GHz+df['GHz']/2.,npts)*GHz
    if delays is None:
        rl = (excitationLen or q['readoutLen'])['ns'] 
        maxDelay = rl+750.0
        delays = st.r[-30:maxDelay:10, ns]
    else:
        rl = (excitationLen or q['readoutLen'])['ns']
        maxDelay = rl+750.0
    
    axes=[(zCompensation, 'z compensation'), (delays, 'Delay'), (freqs, 'Frequency')]
    deps = [('Probability', 'P1', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)
    def func(server, currZ, delay, freq):
        alg = gc.Algorithm(agents=devs)
        for qubit in alg.agents:
            if 'fc' in qubit:
                qubit['fc'] = q['f10'] - 225*MHz
        if piPulse:
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.PiPulse([alg.q0], freq=freq, state=None)]
        alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        if excitationLen is None:
            alg.q0.measure(delay = -maxDelay-delay)
        else:
            df = alg.q0['readoutFrequency'] - alg.q0['readoutDevice']['carrierFrequency']
            alg.q0['rr'] += env.mix(env.flattop(-maxDelay-delay, excitationLen, 5.0, amp=excitationAmp), df)
        if zEnv is not None:
            start, end = alg.q0['z'].start, alg.q0['z'].end
            alg.q0['z'] += env.shift(zEnv, -delay)
            alg.q0['z'].start, alg.q0['z'].end = start, end
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray)
        returnValue([np.mean(prob)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    return data
    
def stark_compensation_on_demand(Sample, measure, name='AC Stark Z Compensation on demand',
                       zrange = 0.2, npts = 30, delays=None,
                       stats=900, save=True):
    
    zrange = np.linspace(-zrange/2., zrange/2., npts)
                       
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)

    if delays is None:
        rl = q['readoutLen']['ns']
        maxDelay = rl+750.0
        delays = st.r[-30:maxDelay:10, ns]
    else:
        rl = q['readoutLen']['ns']
        maxDelay = rl+750.0
    
    def sweep():
        for d in delays:
            for z in zrange:
                yield d,z
                
    sweepData = {
        'cumZ':np.zeros((len(delays))),
        'dIdx':0,
        'zIdx':0,
        'currProbs':0*zrange,
        }
        
    # construct the current z
    def genCurrZ(idx, z):
        cumZ = np.copy(sweepData['cumZ'])
        cumZ[idx] += z
        
        if idx>0:
            if idx<len(delays)-1:
                cumZ[idx+1] = 2*z + cumZ[idx-1]
                cumZ[idx+2] = 3*z + cumZ[idx-1]
        else:
            cumZ[idx+1] += 2*z
        return cumZ
    
    axes=[('Delay','ns'),('Z', 'a.u.')]
    deps = [('Probability', 'P1', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)
    
    def func(server, args):
        currDelay,currZ = args
        zIdx = sweepData['zIdx']
        dIdx = sweepData['dIdx']
        print 'currDelay',str(currDelay),'currZ',str(currZ),'zIdx',str(zIdx),'dIdx',str(dIdx)
        currZdata = genCurrZ(dIdx, currZ)
        zEnv = env.wigglyZ(0, np.array(list(delays)), np.array(currZdata))
        
        alg = gc.Algorithm(agents=devs)
        #alg[gates.Wait([alg.q0], alg.q0['piLen']*-1/2.)]
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        alg.q0.measure(delay = -maxDelay-currDelay)
        start, end = alg.q0['z'].start, alg.q0['z'].end
        alg.q0['z'] += env.shift(zEnv, -currDelay)
        alg.q0['z'].start, alg.q0['z'].end = start, end
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray))  
        
        sweepData['currProbs'][zIdx] = prob
        
        if zIdx+1 != npts: #Not done with this z
            sweepData['zIdx']+=1
        else: #Update fit
            bestZ = getMaxGauss(zrange, sweepData['currProbs'])
            print 'best Z was',str(bestZ)
            sweepData['cumZ'] = genCurrZ(dIdx,bestZ)
            sweepData['dIdx'] += 1
            sweepData['zIdx'] = 0
        
        returnValue([currDelay, currZ, prob])
    data = sweeps.run(func, sweep(), dataset=dataset, save=save, noisy=False)
    return data, env.wigglyZ(0, np.array(list(delays)), np.array(sweepData['cumZ']))
    
def rebuild_compensation_on_demand(Sample, dvwData):
    
    data = dvwData
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    delays = data[::dataSize, 0]
    zrange = data[:dataSize,1]
    probs = np.reshape(data[:,2], (numColumns, dataSize))

    cumZ = np.zeros((len(delays)))
        
    # construct the current z
    def genCurrZ(idx, z, cumZ):
        cumZ[idx] += z
        
        if idx>0:
            if idx<len(delays)-1:
                cumZ[idx+1] = 2*z + cumZ[idx-1]
                cumZ[idx+2] = 3*z + cumZ[idx-1]
        else:
            cumZ[idx+1] += 2*z
        return cumZ
    
    for delay, prob in zip(delays, probs):
        bestZ = getMaxGauss(zrange, prob)
        idx = np.argwhere(delays==delay)[0][0]
        cumZ = genCurrZ(idx, bestZ, cumZ)
        
    return env.wigglyZ(0, np.array(list(delays)), np.array(cumZ))
     
def dephasingVsTime(Sample, measure, echo = False, name = 'Dephasing vs. Time',
            stats=600, freq = None, delay = 500*ns, iteration = st.r[0:500:1],
            fringeFreq = 1*MHz,
            riseTime = 2*ns,
            detuneAmp=None, iterations = 1,
            reset=True, plot=True, update=True,
            noisy=True, save=True, collect=False,detuneLen=None):
        
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
        
    axes = [(iteration,'Iteration'), (delay,'delay time')]  
    deps = [('Phase Envelope', '', '')]
    kw = {'stats': stats, 'riseTime':riseTime}
    tomoPhases = {'+X': 0.0, '+Y':0.25, '-X': -0.5, '-Y':-0.25}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currIter, currDelay):

        reqs = []
        
        for tomoPhase in tomoPhases:
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0],currDelay)]
            alg[gates.PiHalfPulse([alg.q0], phase = 2*np.pi*(fringeFreq['GHz']*currDelay['ns'] + tomoPhases[tomoPhase]))]
            alg[gates.Measure([alg.q0])]
            alg.compile()
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))

        results = yield FutureList(reqs)
        data = []
            
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(prob)
                        
        phaseAmplitude = np.sqrt((data[0]-data[2])**2 + (data[1]-data[3])**2)
        
        returnValue([phaseAmplitude])    
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)

def repetitionCodeVsParam(Sample, measure, qubitAndParamNames, sweepRange, stats=900, cycles = 5, update=False,
                state=None, name = "scan RC vs param", save=True, plot=False, collect=False):
    #scans the Measure probs, and final probs as a function of qubits and parameters
    #syntax:  (s,['q0','q1','q2'],[['q1','q0aczPhase']],linspace(-0.2,0.2,11))
    #-> scans versus q0aczPhase
    #sweepRange: additive
    #units: must be passed in correctly
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    if state is None:
        state=[0]*len(qubits)

    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps(nelderMead=False, sweepRange=sweepRange)
    
    #deps: P1 q1 1, P1 q1 2, ... P1 q1 n-1, P1 q3.... n-1 P1 q0
    deps=[]
    for cycle in range(cycles):
        for q in qubits:
            if q.get('_surf',None) == 'measure':
                deps += [('P1 %s  cycle: %s (measure)' % (q.__name__,cycle) , '', '')]
    for q in qubits:
        if q.get('_surf',None) == 'data':
            deps+= [("P1 %s (data)" % q.__name__ ,'','')]
    
    kw = {'stats': stats,'state':state,'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, *args):
        currParamDict = param.argsToCurrParamDict(args)
        alg = gc.Algorithm(agents=devs)
        param.updateQubits(alg.qubits, currParamDict)
        buildRepetitionCodeIS(alg,state,cycles,repetitionCodeIS(measure))
        alg.compile(correctXtalkZ = True)
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRaw')
        prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
        probsAll = reshapeRepCodeProbs(prob, alg.qubits)
        returnValue(probsAll)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1)     
        
    if collect:
        return data
            
def repetitionCodeNM(Sample, measure, qubitAndParamNames, probsIdeal, state, xtol = 0.001, ftol = 0.001,
           nonzdelt = 0.02, zdelt = 0.02,
           averages = 1,
           name = 'Repitition Code NM',cycles=5,
           maxfun=None, k = 10, interleaved = False, 
           stats=900, correctXtalkZ=True,
           collect=False, save = True, noisy = True,
           settling=False, pulseShape=False):
    """
    qubitAndParamNames: keys to perform NM on. syntax: 
    #for parameters which are not lists themselves
    qubitAndParamNames=[ ['q0','piLen'] , ['q1','q0aczPhase'] , ...]
    
    #for parameters which are lists themselves: the 3rd entry indicates which indices to NM on. For aczLambda, settlingRates etc.
    qubitAndParamNames=[ [ 'q1','q0aczLambda',[1,2,3,4] ] , ...]
    
    ['q0','q1','q2'] example:
    state=[1,0,0]
    probsIdeal=[1,0,1,0,1,1,0]
    qubitAndParamNames=[ ['q1','q0aczPhase'],['q1','q0aczAmp'],['q0','q1aczPhase'],['q1','q2aczPhase'],['q1','q2aczAmp'],['q2','q1aczPhase']]
    qubitAndParamNames=[ ['q1','q0aczPhase'],['q1','q0aczAmp'],['q0','q1aczPhase'],['q1','q2aczPhase'],['q1','q2aczAmp'],['q2','q1aczPhase'],['q1',aczPadding']]
    
    #also on CZ parameters:
    state=[1,0,0]
    probsIdeal=[1,0,1,0,1,1,0]    
    qubitAndParamNames=[ ['q1','q0aczPhase'],['q1','q0aczAmp'],['q1','q0aczLambda',[1,2,3,4] ],['q1','q2aczLambda',[1,2,3,4] ],['q0','q1aczPhase'],['q1','q2aczPhase'],['q1','q2aczAmp'],['q2','q1aczPhase']]    
       
    
    """
    print "BUG: qubitAndParamNames is not saved to datavault"
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps()
    name += ' cycles:'+str(cycles)
    
    def makeInputsAddDepsNM(qubitAndParamNames, qubits, deps):
        inputs=[]
        for item in qubitAndParamNames:
            if np.alen(item)==2:
                qname=item[0]
                paramName=item[1]
                inputs+=[devs[qname][paramName]]
                deps  +=[ ( qname + ' '+str(paramName) , '' , '' ) ]
            elif np.alen(item)==3:
                qname=item[0]
                paramName=item[1]
                idxList=item[2]
                for idx in idxList:
                    inputs+=[devs[qname][paramName][idx]]
                    deps  +=[ ( qname + ' '+str(paramName)+ ' idx:'+str(idx) , '' , '' ) ]               
            else:
                raise Exception('see syntax')
        return inputs

    kw={'stats':stats, 'qubitAndParamNames':qubitAndParamNames, 'correctXtalkZ':correctXtalkZ,'probsIdeal':probsIdeal,'state':state,'k':k,'cycles':cycles}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)    
    def func(server, args):                  
        currParamDict = param.argsToCurrParamDict(args)
        alg = gc.Algorithm(agents=devs)
        param.updateQubits(alg.qubits, currParamDict)
        buildRepetitionCodeIS(alg,state,cycles,repetitionCodeIS(measure))
        alg.compile(correctXtalkZ = True)
        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRaw')
        prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
        probsAll = reshapeRepCodeProbs(prob, alg.qubits)
        print probsAll
        error = np.sum( (np.array(probsAll) - np.array(probsIdeal))**2 ) / (1.0*np.alen(probsIdeal)) #quadratic distance from ideal, normalized
        returnValue([error])
        
    funcWrapper = popt.makeFuncWrapper(averages, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
        
    return output


def makeInputStates(numQubits):
    states=[]
    for t in range(2**numQubits):
        state=[]
        txt=bin(t).split('0b')[1]
        txt=txt[::-1]
        txt += '0'* (numQubits-len(txt))
        for subtxt in txt:
            state += [  int( subtxt ) ]
        states += [ state ]
    return states
    
def runRepetitionCode(Sample, measure, stats=600, cycles = 2,
                state=[0,0,0],
                name = "run Repetition Code", save=True, plot=False, collect=False):

    if not isinstance(cycles, list):
        cycles = [cycles]
                
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    name += ' state:'+str(state) + ' cycles:'+str(cycles)
    
    axes = [(cycles,'Cycles')]
    deps = readout.depsForDV()
    
    kw = {'stats': stats, 'cycles':cycles,'state':state}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        paramDic=NestedDict()
        alg=buildRepetitionCodeIS(devs,state,paramDic,currCycle,repetitionCodeIS(measure))
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRaw')

        prob = readout.IQToProbs(data.asarray, alg.qubits, measure)

        dataForDV = readout.reshapeForDV(data.asarray)
        returnValue(dataForDV)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=0)      
        
    if collect:
        return readout.reshapeFromDV(data)
    
def cycle3(Sample, measure, stats=900, cycles = 5,
                phaseC0 = None, phaseC1 = None, state=[0,0,0],
                name = "Cycle 3", save=True, plot=False, collect=False):

    if not isinstance(cycles, list):
        cycles = [cycles]
                
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    origPhaseC0 = qubits[1][qubits[0].__name__+'aczPhase']
    origPhaseC1 = qubits[1][qubits[2].__name__+'aczPhase']
    
    maxDataLength = len(measure)*stats*np.max(cycles)*2
    
    axes = [(cycles,'Cycles')]
    deps = readout.depsForDV()
    
    kw = {'stats': stats, 'cycles':cycles}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        
        for st, q in zip(state, alg.qubits):
            alg[gates.MoveToState([q], 0, st)]
        for x in range(currCycle):
            alg[gates.PiHalfPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q0])]
            alg[gates.adiabaticCZ([alg.q1,alg.q2])]
            alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]

            if x +1 != currCycle:
                alg[gates.Measure([alg.q1], name = x)]
        alg[gates.Measure([alg.q0,alg.q1,alg.q2], name='Final')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRaw')
        dataForDV = readout.reshapeForDV(data.asarray)
        returnValue(dataForDV)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=0)      
        
    if collect:
        return readout.reshapeFromDV(data)
            
def cycle3Probs(Sample, measure, stats=900, cycles = 5, update=False,
                phaseC0 = None, phaseC1 = None, state=[0,0,0],
                name = "Cycle 3 probs", save=True, plot=False, collect=False):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    origPhaseC0 = qubits[1][qubits[0].__name__+'aczPhase']
    origPhaseC1 = qubits[1][qubits[2].__name__+'aczPhase']
    
    axes = [(phaseC0,'phaseC0'),
            (phaseC1,'phaseC1')]
    deps = []
    deps = [("Probability Measure |1>","Demod %d" % cycle,'') for cycle in range(cycles)]
    deps+= [("Probability Data0 |1>",'',''),("Probability Data1 |1>",'','')]
    
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currPhaseC0, currPhaseC1):
        alg = gc.Algorithm(agents=devs)
        if currPhaseC0:
            alg.q1[qubits[0].__name__+'aczPhase'] = origPhaseC0 + currPhaseC0
        if currPhaseC1:
            alg.q1[qubits[2].__name__+'aczPhase'] = origPhaseC1 + currPhaseC1
        
        for st, q in zip(state, alg.qubits):
            alg[gates.MoveToState([q], 0, st)]
        for x in range(cycles):
            alg[gates.PiHalfPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q0])]
            alg[gates.adiabaticCZ([alg.q1,alg.q2])]
            alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]

            if x +1 != cycles:
                alg[gates.Measure([alg.q1], name = x)]
        alg[gates.Measure([alg.q0,alg.q1,alg.q2], name='Final')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRaw')
        prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
        probsMeasure = prob[1,:,1]
        probData0 = prob[0,-1,1]
        probData1 = prob[2,-1,1]
        probs = np.hstack((probsMeasure, probData0, probData1))
        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)      
    
    if update:
        probs = data[:,1:-2] # first row is phase, last two rows are data qubits
        avgProbs = np.mean(probs, axis=1)
        bestPhase = getMaxGauss( data[:,0] , avgProbs, fitToMax=False)    
        if np.any(phaseC0):
            key = qubits[0].__name__+'aczPhase'
            value = origPhaseC0 + bestPhase
        elif np.any(phaseC1):
            key = qubits[2].__name__+'aczPhase'
            value = origPhaseC1 + bestPhase
        Qubits[1][key] = value
        
    if collect:
        return data
        
def cycle5(Sample, measure, stats=10, cycles = [5], name = "The flippin' repetition code (5)", save=True, plot=False, echo=False, state=[0,0,0,0,0]):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if isinstance(cycles,list) is False:
        cycles = [cycles]
    
    axes = [(cycles,'Cycles')]
    deps = []
    for q in qubits:
        deps.append(('%s Is' % q.__name__, '', '')) # for a given cycle, we are going to concatenate all of the stats
        deps.append(('%s Qs' % q.__name__, '', ''))
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        alg[gates.Measure(alg.qubits, sync='end')]
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        for x in range(currCycle):
            alg[gates.PiHalfPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q0])]
            if echo:
                alg[gates.PiPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q2])]
            alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]
            alg[gates.Sync([alg.q1,alg.q3])]
            
            alg[gates.PiHalfPulse([alg.q3])]
            alg[gates.adiabaticCZ([alg.q3,alg.q4])]
            if echo:
                alg[gates.PiPulse([alg.q3])]
            alg[gates.adiabaticCZ([alg.q3,alg.q2])]
            alg[gates.PiHalfPulse([alg.q3], phase = -np.pi)]
            
            if x +1 != currCycle:
                alg[gates.Measure([alg.q1,alg.q3], sync='start')]
        alg[gates.Measure([alg.q1,alg.q3], sync='start')]
        alg[gates.Sync(alg.qubits)]
        alg[gates.Measure([alg.q0,alg.q2,alg.q4], sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRawLocalDemod', localDemod=True)
        data_for_dv = iq.format_something_useful_for_data_vault(data)
        returnValue(data_for_dv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1)
    
def reset21(Sample, measure, name = 'Reset',
                      amplitude = np.linspace(-1.2*np.pi, 1.2*np.pi, 51),
                      freq = None, riseTime = 1*ns, resettime=100*ns,
                      averages = 1, 
                      stats=900, correctXtalkZ=False,
                      save=True, collect=False, noisy=True,state=2):
    ''' Make sure your 2-state is tuned up in amplitude and frequency. '''
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if freq is None:
        freq = qubit['readoutFrequency'] + (qubit['f10']-qubit['f21'])
    
    axes = [(freq,'Reset Freqency'),(resettime, 'Reset Time'),(amplitude, 'Pi12 Angle')]
    deps = [('Probability |1> -> Ae', '', ''),('Probability |1> -> Ag', '', '')]
    
    kw = {'stats': stats, 'averages':averages}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currFreq, currThermTime, currAngle):

        #ref: 1 state
        alg = gc.Algorithm(agents=devs)
        alg[gates.MoveToState([alg.q0], 0, 1)]    
        
        if currFreq:
            amp = fToAmp(alg.q0, currFreq)
            alg[gates.Detune([alg.q0], length=currThermTime, w=riseTime, amp=amp)]
        alg[gates.RotPulse([alg.q0], currAngle, state=1)]
        alg[gates.Measure([alg.q0])]
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probAe = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        #exp: 2 state
        alg = gc.Algorithm(agents=devs)
        alg[gates.MoveToState([alg.q0], 0, 2)]         
        if currFreq:
            amp = fToAmp(alg.q0, currFreq)
            alg[gates.Detune([alg.q0], length=currThermTime, w=riseTime, amp=amp)]
        alg[gates.RotPulse([alg.q0], currAngle, state=1)]
        alg[gates.Measure([alg.q0])]
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probAg = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        returnValue([probAe, probAg])        

    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def timingLagIQ(Sample, measure, delay=st.r[-10:10:1,ns],
              stats=600, update=True,
              save=True, name='Timing Lag IQ', plot=False, noisy=True):
              
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if np.alen(qubit.get('xyPadding',None))>1:
        if q.get('xyPadding',None)[0] != q.get('xyPadding',None)[1]:
            raise Exception("You have asymmetric XY padding, you're really about to screw yourself over")
            
    origAdcTimingLag = qubit['readoutDevice']['adcTimingLag']
    origtimingLagRRUwave = qubit['readoutDevice']['timingLagRRUwave']
    
    axes = [(delay, 'Time')]
    deps = [('Clicks', 'I for %s' %qubit.__name__, ''), ('Clicks', 'Q for %s'%qubit.__name__, '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currRRlag):
    
        alg = gc.Algorithm(agents=devs)
        alg.q0['readoutDevice']['timingLagRRUwave'] = currRRlag
        alg[gates.Measure([alg.q0])]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        I,Q = np.squeeze(readout.averageIQSingleDemod(data)).transpose()    
        returnValue([I,Q])
        
    result = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
def waitLagIQ(Sample, measure, delay=st.r[-10:10:1,ns],
              stats=600, update=True, before=True,
              save=True, name='Timing Lag IQ', plot=False, noisy=True):
              
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if np.alen(qubit.get('xyPadding',None))>1:
        if q.get('xyPadding',None)[0] != q.get('xyPadding',None)[1]:
            raise Exception("You have asymmetric XY padding, you're really about to screw yourself over")
            
    origAdcTimingLag = qubit['readoutDevice']['adcTimingLag']
    origtimingLagRRUwave = qubit['readoutDevice']['timingLagRRUwave']
    
    axes = [(delay, 'Time')]
    deps = [('Clicks', 'I for %s' %qubit.__name__, ''), ('Clicks', 'Q for %s'%qubit.__name__, '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currDelay):
    
        alg = gc.Algorithm(agents=devs)
        if before:
            alg[gates.Wait([alg.q0],currDelay)]
            alg[gates.Measure([alg.q0])]
        else:
            alg[gates.Measure([alg.q0])]
            alg[gates.Wait([alg.q0],currDelay)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        I,Q = np.squeeze(readout.averageIQSingleDemod(data)).transpose()    
        returnValue([I,Q])
        
    result = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
def detuneTroubleShootAg2(Sample, measure, name = 'Detune Trouble Shoot Agilent',
            thetas=np.linspace(-np.pi/4,7*np.pi/4,100), 
            stats=1200, iterations=6,
            update=False, padding=0*ns, 
            noisy=True, save=False, collect=False,plot=True):
        
    """
    """
    cxn=Sample._cxn
    scope = cxn.agilent_dso91304a_oscilloscope()
    scope.select_device()     
    
    scope.reset()
    print 'scope reset'  
    time.sleep(2) 
    
    SCOPECHANNEL=1
    TRIGGERCHANNEL=2
    
    p = scope.packet().\
    channelOnOff(SCOPECHANNEL,'ON').\
    channelOnOff(TRIGGERCHANNEL,'ON').\
    scale(SCOPECHANNEL,0.02).\
    scale(TRIGGERCHANNEL,0.5).\
    position(SCOPECHANNEL,0.0).\
    position(TRIGGERCHANNEL,0.0).\
    horiz_scale(150.0e-9).\
    horiz_position(0.0).\
    trigger_sweep('TRIG').\
    trigger_mode('EDGE').\
    trigger_edge_slope('POS').\
    trigger_at(TRIGGERCHANNEL,0.5).\
    averagemode(1).\
    numavg(512).\
    gpib_write('ACQUIRE:COMPLETE:STATE ON').\
    gpib_write('ACQUIRE:COMPLETE 100')
    #yield p.send()  
    p.send()
    
    #trigger_at(TRIGGERCHANNEL,0.5).\
    print 'scope packet sent'
    time.sleep(2)     
    #numavg(128).\
    #acquisition_mode('AVE')   
    
    #RUN SCRIPT
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(iterations, 'Detune Iteration')]
    deps = [('Ramsey Amplitude', '', ''),
            ('Ramsey Phase', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, iteration):
    
        reqs = []
        
        for theta in thetas:
        
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiHalfPulse([alg.q0])]
            
            alg[gates.Wait([alg.q0], padding)]
            
            for k in range(iteration):
                alg[gates.Detune([alg.q0])]
                if iteration>1 and k<iteration-1:
                    alg[gates.Wait([alg.q0],50*ns)]
                    
            alg[gates.Wait([alg.q0], padding)]
                
            alg[gates.PiHalfPulse([alg.q0], phase = theta)]
            alg[gates.Measure([alg.q0])]
            alg.compile()    
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        data = []
        
        results = yield FutureList(reqs)
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(prob)
        
        amp, phase = fitting.fitRamsey(thetas, data)
        
        returnValue([amp, phase])

    data = sweeps.grid(func, axes, dataset=False and dataset, noisy=False)
    
    
    #GET TRACE
    [timeAxis,y] = scope.get_trace(SCOPECHANNEL) #start and stop in ns
    
    if save:
        dv = cxn.data_vault
        dv.cd(sample._dir)
        savename = 'detuneTroubleShoot'
        dv.new(savename, ['time [ns]'],['Voltage'])
        all = [[timeAxis[i],y[i]] for i in range(len(timeAxis))]
        dv.add(all)
        print 'saved'      
    
    
    if plot:
        plt.figure()
        plt.plot(timeAxis,y)
        plt.show()
        
    if collect:
        return data 
        
def forceRingdownQST(Sample, measure, name='Forced Resonator Ringdown QST',
            amp1 = None, amp2 = None, phase1 = None, phase2 = None, 
            ringdown = True,
            delay = 0*ns, tomoDelay = 100*ns,
            stats=900, correctXtalkZ=True,
            save=True, collect=False, noisy=True):
    '''
    We want to try and tune up a ringdown pulse after readout to reduce the photon decay time.
    The tricky part is, we have to have pulses that ring down BOTH |0> and |1> simultaneously.
    We check how well we are doing by preparing a |0>, readout, then quantify how well a pi pulse works,
    and also preparing a |1> and doing the same thing. The Pi at the end of each of these experiments 
    will only work well if the reasonator is properly rung down.
    '''
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]   
    
    if amp1 is None:
        amp1, _ = qubit['ringdownPulseAmps']
    if amp2 is None:
        _, amp2 = qubit['ringdownPulseAmps']
    if phase1 is None:
        phase1, _ = qubit['ringdownPulsePhases']
    if phase2 is None:
        _, phase2 = qubit['ringdownPulsePhases']
    
    axes = [(amp1, 'Amplitude 1'), (amp2, 'Amplitude 2'), (phase1, 'Phase 1'),(phase2, 'Phase 2'),(delay,'Delay')]
    deps = [('Phase |0>', 'rad', ''), ('Phase |1>', 'rad', ''),('Amp 0','',''),('Amp 1','','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currAmp1, currAmp2, currPhase1, currPhase2, currDelay):
        
        reqs = []
            
        for op in opList:
        
            alg = gc.Algorithm(agents=devs)
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
            alg[gates.Wait([alg.q0], currDelay)]
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0], tomoDelay)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Readout([alg.q0])] # should end up in |1> state
            alg.compile()

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        for op in opList:
            
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
            alg[gates.Wait([alg.q0], currDelay)]
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0], tomoDelay)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Readout([alg.q0])] # should end up in |1> state
            alg.compile()
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))

        dataZero = []
        dataOne = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            dataZero.append(prob)
        for r in results[len(opList):]:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            dataOne.append(prob)
                
        [rho, residual] = tomo.qst_evan(np.array(dataZero), tomo.octomo_ops, stats, 1)        
        phaseZero = ps.getDensityPhase(rho)[0]
        ampZero = np.abs(rho[1,0])
        
        [rho, residual] = tomo.qst_evan(np.array(dataOne), tomo.octomo_ops, stats, 1)        
        phaseOne = ps.getDensityPhase(rho)[0] 
        ampOne = np.abs(rho[1,0])
                
        returnValue([phaseZero, phaseOne, ampZero, ampOne])
            
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def ringdownMixerTable(q, integrationTime = None, ampI = 127, ampQ = 127, ADC_MIX_TIMESTEP = 2.0):
    readoutPulseLen = q['readoutPulseLen'] + np.sum(q['readoutRingdownPulseLens'])*ns
    carrierFrequency = q['readoutDevice']['carrierFrequency']
    readoutDemodFrequency = q['readoutDemodFrequency']-carrierFrequency
    totalLen = readoutPulseLen + integrationTime
    entries = int(np.floor(totalLen / (ADC_MIX_TIMESTEP*ns)))
    mix_wfm = np.exp(-2*np.pi*1j*np.arange(entries) * ADC_MIX_TIMESTEP * readoutDemodFrequency)
    mixTable = np.array(np.column_stack((np.real(mix_wfm*ampI), np.imag(mix_wfm*ampQ))), dtype=int)
    maskIdx = readoutPulseLen/(ADC_MIX_TIMESTEP*ns)+1
    mixTable[:maskIdx,:] = 0 # don't demodulate during pulse
    return mixTable
        
def forceRingdownNM(Sample, measure, integrationTime = 100*ns,
            paramNames = ['readoutRingdownPulseAmps',
                          'readoutRingdownPulsePhases'],
            name='Forced Resonator Ringdown',
            averages=5,
            nonzdelt = 2, zdelt = 2,
            xtol = 0.001, ftol = 0.001,
            maxfun = None,
            stats=1500, correctXtalkZ=True,
            save=True, collect=False, noisy=True):
    '''
    '''
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    qubitAndParamNames = popt.addQubitNamesToParamNames(paramNames, qubit)
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps(nelderMead=True)
    kw = {'stats': stats, 'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw) 
    
    def func(server, args):
        currParamDict = param.argsToCurrParamDict(args)
        
        alg = gc.Algorithm(agents = devs)
        param.updateQubits(alg.qubits, currParamDict)
        currRingdownMixerTable = ringdownMixerTable(alg.q0, integrationTime)
        alg.q0['readoutDemodMixerTable'] = currRingdownMixerTable
        alg.q0['readoutDemodUseMixerTable'] = True
        alg.q0['ringdown'] = True
        alg.q0['readoutDemodLen'] = alg.q0['readoutPulseLen'] + np.sum(alg.q0['readoutRingdownPulseLens'])*ns + integrationTime
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Measure([alg.q0])]
        alg.compile() 
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        I,Q = readout.averageIQSingleDemod(data.asarray)
        returnValue([np.sqrt(I**2 + Q**2)])
            
    funcWrapper = popt.makeFuncWrapper(averages, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
        
    return output

        
def forceRingdownViaPiNM(Sample, measure, wait = -100*ns, iterations=1, 
               nonzdelt = 0.05, zdelt = 0.01,
               xtol = 0.001, ftol = 0.001,
               name='Forced Ringdown NM',
               averages = 10,
               stats=2100,
               save=True, collect=False, noisy=True):
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    amp1, amp2 = qubit['ringdownPulseAmps']
    phase1, phase2 = qubit['ringdownPulsePhases']

    
    axes = [('Function Evaluation','')]
    deps = [('Error','',''),
            ('Amplitude 1','dBm',''),
            ('Amplitude 2','dBm',''),
            ('Phase 1','Rad',''),
            ('Phase 2','Rad','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    
    def func(server, args):
        
        currAmp1, currAmp2, currPhase1, currPhase2, currAvg = args

        alg = gc.Algorithm(agents=devs)
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], wait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0], ringdown=False)] # should end up in |1> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probNoPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], wait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0], ringdown=False)] # should end up in |0> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        error = ((1-probNoPi)**2 + probPi**2)/2.
        print error
        returnValue([error])
            
    def funcWrapper(args):

        args=list(args)
        #currAmp1, currAmp2, currPhase1, currPhase2 = args
    
        def seqGen():
            for currAvg in range(averages):
                #yield currAmp1, currAmp2, currPhase1, currPhase2, currAvg
                yield args + [currAvg]
                    
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample, 'dummy dataset', axes, deps,measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save=False, noisy=False)
        
        avgError = np.mean(data)
        
        print 'avg error: %s, std: %s' % (avgError, np.std(data)/np.sqrt(averages))
        
        garbageCollect.collect()
        
        return avgError
            
    inputs = [amp1,amp2,phase1,phase2]
    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt)
    
    return output

    
    
def aczLeakageVsLength(Sample, measure, lengths = st.r[30:50:1, ns]):
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)

    for length in lengths:
        Qubits[0][qubits[1].__name__+'aczTime'] = length
        mq.aczPhaseTuner(s,[measure[0],measure[1]], amplitude=None)
        mq.aczShapeNM(s,['q1','q0'], maxfun = 50)
        mq.updateAczShapeNM(s, dvw[-1])
        mq.aczPhaseTuner(s,[measure[0],measure[1]], amplitude=None)
        mq.aczErrorFilter(s,['q1','q0'])
        
def hollisterBet(Sample, measure,
                 freq = None, state=1,
                 herald = False, plot=False,
                 name='Measurement (Hollister) Bet',
                 stats=900,
                 save=True, collect=False, noisy=True):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(freq,'Frequency')]
    deps = readout.genProbDeps(qubits, measure)
    
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currFreq):
        alg = gc.Algorithm(agents=devs)
        currAmp = zfuncs.FrequencyToAmp(alg.q0)(currFreq)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Sync([alg.q0,alg.q1])]
        alg[gates.Detune([alg.q1], length = alg.q0['readoutPulseLen'] + alg.q0['readoutRingdownLength'], amp = currAmp)]
        alg[gates.Readout([alg.q0], name = 'q0')]
        alg[gates.Sync([alg.q0,alg.q1])]
        alg[gates.Readout([alg.q1], name = 'q1')]
        alg.compile(correctXtalkZ=True)
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure)
        returnValue(np.hstack((probs[0][0],probs[1][1])))
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)     

def aczCheckRamsey(Sample, measure, name='Adiabatic CZ Check Ramsey',
             thetas=25, waitTime = None,
             iterations=1, herald=False,
             amp = None, length = None, extraWait = 0.0*ns,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True):
    '''
    measure = [control, target]
    '''
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]   
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)

    if waitTime is None:
        waitTime = controlQubit['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if herald:
            alg[gates.Measure(alg.qubits, name='Herald')]
        alg[gates.PiHalfPulse([controlQubit])]
        alg[gates.adiabaticCZ([controlQubit, targetQubit], amp=amp, length=length)]
        alg[gates.Wait([controlQubit], extraWait)]
        alg[gates.PiHalfPulse([controlQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)

        probs = readout.IQToProbs(data.asarray, alg.qubits, measure, herald=herald)[0][0][1]

        returnValue([probs])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[0][targetQubit.__name__+'aczPhase'] -= deltaPhi/iterations
    
    if collect:
        return data