#Standard packages
import numpy as np
from math import atan2
import scipy
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import pyle
import time

import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking import singleQubit as sq
from pyle.dataking import multipleQubit as mq
from pyle.dataking import automateDaily as ad
import pyle.plotting.dstools as ds
import pyle.plotting.plothelper as ph
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
from pyle.datavault import DataVaultWrapper
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.analysis import iq as iq

from pyle.dataking.benchmarking import randomizedBenchmarking as rb

#import pyle.dataking.DevDaniel.Qubit.readout as readout

from pyle import signalprocessing as signalprocessing
import pyle.quantum as quantum
import time
import copy
from matplotlib.mlab import find
from scipy.interpolate import UnivariateSpline
import itertools
from pyle.dataking import zfuncs as zfuncs

def telegraph(dt = 1*ns, length = 1e3*ns, gammaUp = 10*MHz, gammaDown = 10*MHz, amplitude = 1., aboutZero = True):
    """
    Generates random telegraph noise, switching between two states.
    We simulate this by initializing a random state, then generating discreen timesteps, where at
    each timestep there is a change to flip to the other state depending on the gamma in effect.
    Outputs a waveform that switches between 0 and amplitude.
    (False -> 0, True -> amplitude)
    """
    
    t = np.arange(0, length+dt, dt)
    
    initialState = np.random.rand(1)[0] < gammaDown/(gammaDown+gammaUp) # tbis might be backwards
    state = [initialState] 
    
    randoms = np.random.rand(len(t)-1)
    scalingDown = (dt * gammaDown).inBaseUnits().value
    scalingUp = (dt * gammaUp).inBaseUnits().value
    
    for ti,currRand in zip(t[1:],randoms): # first value was set by initialization
        currState = np.copy(state[-1])
        
        if currState: # if state is TRUE (amplitude) use gammaDown
            scalingProb = scalingDown
        else: # state is FALSE (0) use gammaUp
            scalingProb = scalingUp
        
        switch = currRand < scalingProb
        if switch:
            state.append(np.invert(currState))
        else:
            state.append(currState)
            
    state = np.array(state) * amplitude
    if aboutZero:
        # subtract off average value 
        avgAmp = float(amplitude) * gammaUp / (gammaUp + gammaDown)
        state -= avgAmp
            
    return np.vstack((t, state)).transpose()
    
def plotTelegraph(data):

    fourier = np.fft.fft(data[:,1])
    fftfreqs = np.fft.fftfreq(data[:,1].size,data[1,0]-data[0,0])

    plt.figure()
    plt.subplot(2,1,1)
    plt.plot(data[:,0], data[:,1])
    plt.subplot(2,1,2)
    plt.plot(fftfreqs, np.abs(fourier))
    
def ramseyTelegraph(Sample, measure, name='Ramsey Telegraph', riseTime = 1.5*ns, z=0.0,
           delay=st.r[0:4000:50,ns], phase=0.0, fringeFreq = 1*MHz,
           gammaUp = 10*MHz, gammaDown = 10*MHz, amplitude = 10*MHz, dt = 1*ns, 
           tomo=True, dualBlock=False, state=1,
           stats=150, randomize=False, averages = 10,
           save = True, collect=False, noisy=False,
           plot=True, update=True):
           
    """Ramsey sequence on one qubit. Can be single phase or 4-axis tomo, and
    can have randomized time axis and/or averaging over the time axis multiple
    times
    
    PARAMETERS
    sample: object defining qubits to measure, loaded from registry
    measure - scalar: number of qubit to measure. Only one qubit allowed.
    delay - iterable: time axis
    fringeFreq - value [Mhz]: Desired frequency of Ramsey fringes
    stats - scalar: number of times a point will be measured per iteration over
            the time axis. That the actual number of times a point will be
            measured is stats*averages
    name - string: Name of dataset.
    save - bool: Whether or not to save data to the datavault
    noisy - bool: Whether or not to print out probabilities while the scan runs
    collect - bool: Whether or not to return data to the local scope.
    randomize - bool: Whether or not to randomize the time axis.
    averages - scalar: Number of times to iterate over the time axis.
    tomo - bool: Set True if you want to measure all four tomo axes, False if
                 you only want the X axis (normal Ramsey fringes).
    """
    
    name += ' Amplitude: '+str(amplitude)
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    #Randomize time axis if you want
    if randomize:
        delay = st.shuffle(delay)
    
    #Generator that produces time delays. Iterates over the list of delays as many times as specified by averages.
    delay_gen = st.averagedScan(delay,averages,noisy=noisy)
    
    axes = [(delay_gen(), 'Delay'),(phase, 'Phase')]
    #If you want XY state tomography then we use all four pi/2 pulse phases
    if tomo:
        deps = [('Probability', '+X', ''), ('Probability', '+Y', ''),
                ('Probability', '-X', ''), ('Probability', '-Y', ''),
                ('Envelope','','')]
        tomoPhases = {'+X': 0.0, '+Y':0.25, '-X': -0.5, '-Y':-0.25}
    else:
        deps = [('Probability','','')]
        tomoPhases={'+X':0.0}
    
    kw = {'averages': averages, 'stats': stats, 'fringeFrequency': fringeFreq, 'z': z,'riseTime':riseTime}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, delay, phase):
        reqs = []
        for tomoPhase in tomoPhases:
            
            # generate RTS
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiHalfPulse([alg.q0], phase = 0.0, state=state)]
            alg[gates.Wait([alg.q0], delay)]
            alg[gates.PiHalfPulse([alg.q0], phase = 2*np.pi*(fringeFreq['GHz']*delay['ns'] + tomoPhases[tomoPhase]), state=state)]
            alg[gates.Measure([alg.q0])]
            alg.compile()
            
            RSTamplitude = zfuncs.freqDiffToAmp(alg.q0,amplitude)
            RTS = telegraph(dt = dt, length = delay+2*alg.q0['piLen'], gammaUp = gammaUp, gammaDown = gammaDown, amplitude = RSTamplitude)
            alg.q0['z']+=env.wigglyZ(0, RTS[:,0], RTS[:,1])
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        results = yield FutureList(reqs)
        data = []
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(prob)
        if tomo:
            envelope = np.sqrt((data[0]-data[2])**2 + (data[1]-data[3])**2)
            data = data + [envelope]
        returnValue(data)
    #Run the scan and save data
    data = sweeps.grid(func, axes, dataset = save and dataset, collect=collect, noisy=noisy)         
    if collect:
        return data

def basic_scans(s, m):
    idles = st.r[5:500:5, ns]
    ramseytimes = list(idles) + list(st.r[550:10000:50,ns])
    sq.T1(s, m, stats=2100, delay=st.r[50:100e3:1e3,ns])
    # quick ramsey
    sq.ramsey(s, m, delay=list(ramseytimes)[::2])

def orbit_w_ramsey(s, m, ramsey_end):
    idles = st.r[5:500:5, ns]
    ramseytimes = list(idles) + list(st.r[550:ramsey_end:50,ns])
    # orbit
    t = time.time()
    sq.randomizedBenchmarking(s, m, k=20)
    ORBITIdles(s, m, k=20, idles=idles)
    t2 = time.time()
    print "Duration of ORBIT scans: %.0f seconds" % (t2 - t)
    # long ramsey--while this is running, compare the ORBIT with the quick ramsey
    sq.ramsey(s, m, delay=ramseytimes, averages=5)
    t3 = time.time()
    print "Duration of Ramsey scan: %.0f seconds" % (t3-t2)
    
def figure1(s,m,ramsey_end, ref = True):
    # sq.freqtuner(s,m)
    # sq.randomizedBenchmarking(s, m, k=20)
    idles = list(st.r[5:100:5, ns]) + list(st.r[120:500:20, ns])
    idles_se = list(st.r[30:100:5, ns]) + list(st.r[120:500:20, ns])
    # sq.randomizedBenchmarkingIdles(s,m,k=10,steps=10, idlewaits = idles, doRef=False)
    # sq.freqtuner(s,m)
    # sq.randomizedBenchmarkingIdles(s,m,k=10,steps=10, idlewaits = idles_se, interleaved=[['SE']], doRef=False)
    idles = st.r[5:500:5, ns]
    ramseytimes = list(idles) + list(st.r[550:ramsey_end:50,ns])
    sq.ramsey(s, m, delay=ramseytimes, averages=5)
    sq.spinEcho(s, m, delay=ramseytimes, averages=5)
    sq.freqtuner(s,m)
    ORBITIdles(s, m, k=20, idles=idles)
    ORBITIdles(s, m, k=20, idles=idles_se, interleaved=[['SE']])

def ORBITIdles(Sample, measure, idles = st.r[5:100:10, ns],
                           name = 'ORBIT Idles',
                           m = None, k = 30, stats=900,
                           interleaved = [['IW']],
                           seqReductionFactor = 3.,
                           maxtime = 11000,
                           plot=True, collect=False,
                           save = True, noisy = True,randomize=False,
                           pipesize=50):
    '''
    '''
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)

    RBClass = rb.RandomizedBenchMarkingClifford(1)
    if isinstance(measure, str):
        RBClass.setGateLengths([qubit])
    else:
        RBClass.setGateLengths([qubit[0]])
    
    def lengthsToMs(idles):
        ms = []
        for idle in idles:
            qubit['identityWaitLen'] = idle
            RBClass.setGateLengths([qubit])
            if idle>300*ns:
                mstep = 1
            else:
                mstep = 5
            iterationsmax=rb.getlength(RBClass,maxtime,interleaved=interleaved,mstep=mstep)
            print "idle: %s, iterationsmax: %s" % (idle, iterationsmax)
            ms.append(np.floor(iterationsmax/seqReductionFactor))
        return ms
        
    if m is None:
        ms = lengthsToMs(idles)
        
    if isinstance(measure, list):
        RBClass.setGateLengths([qubit[1]])
        otherM=rb.getlength(RBClass,maxtime)

    def getSequence(m):
        sequenceData = RBClass.randGen([m], interleaved=interleaved,finish=True)
        return sequenceData

    if interleaved:
        name += ' interleaved: ' + str(interleaved)
                
        
    axes = [(range(k),'sequence'),(idles,'Idle Length')]
    
    deps = [('Sequence fidelity','',''),('m','','')]
    
    kw={'stats':stats, 'interleaved':interleaved, 'k':k,'axismode':'m',
        'dataFormat':'2Dswapped', 'm': m}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, currK, currIdle):
        if noisy:
            print 'K: ',str(currK)
            
        if m is None:
            idx = list(idles).index(currIdle)
            currM = ms[idx]
        else:
            currM = m
            
        fullGateList = getSequence(currM)
        alg = gc.Algorithm(agents=devs)
        alg.q0['identityWaitLen'] = currIdle
        alg[gates.RandomizedBenchmarkingSingleQubitClifford([alg.q0],fullGateList)]
        if isinstance(measure, list):
            otherGateList = RBClass.randGen([otherM], interleaved=None,finish=True)
            alg[gates.RandomizedBenchmarkingSingleQubitClifford([alg.q1],otherGateList)]
        alg[gates.Measure([alg.q0])]            
        alg.compile()
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = 1-np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        returnValue([prob,currM])
    
    data = sweeps.grid(func, axes, save=save, dataset=dataset ,pipesize=pipesize)
    
    if collect:
        return data
    
# now for multiprocessing
from multiprocessing import Process  
def _orbit_multi(conf, measure, **kwargs):
    # run the conf file to establish LabRAD connection and get sample dict.
    with open(conf) as f:
        exec(f)
    sq.freqtuner(s, measure)
    ORBITIdles(s, measure, **kwargs)
    
def ORBITMultiprocessing(Sample, measure, dvw, k=30,m=30,):
    ''' dvw is necessary because we look at all the starred datasets for the qubit params. '''
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    confDict = {'Peter': 'U:/Peter/labrad/dev/pyle/pyle/dataking/confPeter.py'}
    randomizedBenchmarking(Sample, measure, k = k, stats=stats)
    
    stars = dvw.keys(['star'])
    print "Found %s starred datasets" % len(stars)
    for star in stars:
        util.dvwsetToReg(s, dvw[star], [measure])
        try:
            # use another process to work around memory leaks
            try:
                conf = confDict[Sample._dir[1]]
            except KeyError:
                raise ValueError("conf file for user %s not in confDict; add to sq.randomizedBenchmarkingIdles." % s._dir[1])
            p = Process(target=_orbit_multi, args=(conf, measure),
                    kwargs = {'k':k, 'stats':stats, 'm':m, 'idles':st.r[5:100:10, ns]}
                )
            p.start()
            p.join()
        except Exception as e:
            #yes, I know, it's a jim, but it will keep running, and we'll get all the data
            print "Jim: %s" % e

def _idles_multi(conf, measure, **kwargs):
    with open(conf) as f:
        exec(f)
    sq.freqtuner(s, measure)
    sq.randomizedBenchmarkingIdles(s, measure, multiprocess=False, **kwargs)
            
def idles_vs_frequency(Sample, measure, dvw, k=30, stats=900, idlewaits=[2.5, 5, 10, 15, 20, 25, 30, 35, 40, 50, 75, 100]):
    kwargs = {
        'k': k,
        'stats': stats,
        'idlewaits': idlewaits,
    }
    confDict = {'Peter': 'U:/Peter/labrad/dev/pyle/pyle/dataking/confPeter.py'}
    stars = dvw.keys(['star'])
    print "Found %s starred datasets" % len(stars)
    for star in stars:
        util.dvwsetToReg(Sample, dvw[star], [measure])
        print "f10 = %s" % Sample[measure]['f10']
        try:
            # use another process to work around memory leaks
            try:
                conf = confDict[Sample._dir[1]]
            except KeyError:
                raise ValueError("conf file for user %s not in confDict; add to sq.randomizedBenchmarkingIdles." % Sample._dir[1])
            p = Process(target=_idles_multi, args=(conf, measure), kwargs=kwargs)
            p.start()
            p.join()
        except Exception as e:
            #yes, I know, it's a jim, but it will keep running, and we'll get all the data
            print "Jim: %s" % e
            
def figure3and4(s, measure = ['q2','q1'], zz=False):

    maxtime=11000
    
    sample, devs, qubit, Qubit = gc.loadQubits(s, measure, write_access=True)

    #lengths = [12*ns, 15*ns, 20*ns, 25*ns, 30*ns, 40*ns, 50*ns, 60*ns, 70*ns, 100*ns]
    #lengths = [40*ns, 50*ns, 60*ns, 70*ns, 100*ns]
    lengths = [100*ns]
    
    if zz:
        mq.zzXtalk(s,measure,delay=st.r[10:1000:5, ns])
    
    for length in lengths:
        for Q in Qubit:
            Q['piLen'] = length
            Q['piHalfLen'] = length
            Q['piLenZ'] = 2*length
            Q['identityWaitLen'] = 2*length
        for meas in measure:
            sq.rabihigh(s,meas, IQ=False, update=True, manual=False, stats=300, amps = np.linspace(0,1,101))
            sq.rabihigh(s,meas, IQ=False, update=True, manual=False, mode='piHalfPulse', stats=300, amps = np.linspace(0,1,101))
            sq.pituner(s, meas, state=1, diff=0.8, numPulses=3, mode='piPulse', stats=300)
            sq.pituner(s, meas, state=1, diff=0.8, numPulses=1, mode='piHalfPulse', stats=300)
            if length>50*ns:
                drag=False
            else:
                drag=True
            ad.calibrate_single_qubit_xy_gates(s,[meas],drag=drag)
        if length>50*ns:
            amp = sq.rabihighZQST(s,measure[0],stats=600,amps=np.linspace(0,0.05, 21))
            sq.orbitXYZ(s,measure[0], interleaved=[['Zpi']], piAmpZ = np.linspace(amp*0.95, amp*1.05, 21),m=50)
        else:
            amp = sq.rabihighZQST(s,measure[0],stats=600,amps=np.linspace(0,0.01, 21))
            sq.orbitXYZ(s,measure[0], interleaved=[['Zpi']], piAmpZ = np.linspace(amp*0.95, amp*1.05, 21),m=10)
        # EVERYTHING TUNED UP
        sq.randomizedBenchmarking(s,measure[0], maxtime=maxtime,readoutAngles=True, k=30) # Q2 REF
        mq.randomizedBenchmarkingSim(s,measure,measure, maxtime=maxtime, k=20) # Q2 Q1 SIMULT FIG 3
        #sq.randomizedBenchmarking(s,measure[0], maxtime=13000, interleaved=[['X']],readoutAngles=True) # interleaved X FIG 4
        sq.randomizedBenchmarking(s,measure[0], maxtime=maxtime, interleaved=[['IW']],readoutAngles=True, k=20) # interleaved I FIG 4
        sq.randomizedBenchmarking(s,measure[0], maxtime=maxtime, interleaved=[['Zpi']],readoutAngles=True, k=20) # interleave Z FIG 4
        sq.randomizedBenchmarking(s,measure[0], maxtime=maxtime, interleaved=[['X','X']],readoutAngles=True, k=20) # interleave Y,X FIG 4
        sq.randomizedBenchmarking(s,measure[0], maxtime=maxtime, interleaved=[['Y','X']],readoutAngles=True, k=20) # interleaved X,X FIG 4
        
def fix_my_error(s):
    
    # RETAKE ALL 100 ns
    # X RETAKE XX and YX for ALL TIMES
    
    dvw = DataVaultWrapper(s)
    #datasets = [33, 59, 85, 111, 137, 164, 190, 216, 242]
    datasets = [111, 137, 164, 190, 216, 242]
    maxtime = 11000
    
    for dataset in datasets:
        util.dvwsetToReg(s,dvw[dataset],['q2'])
        sq.freqtuner(s,'q2')
        sq.randomizedBenchmarking(s,'q2', maxtime=maxtime, interleaved=[['X','X']],readoutAngles=True, k=20) # interleave Y,X FIG 4
        sq.randomizedBenchmarking(s,'q2', maxtime=maxtime, interleaved=[['Y','X']],readoutAngles=True, k=20) # interleaved X,X FIG 4
        
def analyze3and4(s, A=0.48, B=0.47, plot=True, noisy=True):
    dvw = DataVaultWrapper(['', 'Julian', 'Qubit', '9Xmon', '140611'], cxn=s._cxn)
    # ['', 'Julian', 'Qubit', '9Xmon', '140611']
    # Q2 ref fidelities
    datasets = [33, 59, 85, 111, 137, 164, 190, 216, 242, 342]
    ref_fids = []
    ref_fids_err=[]
    refs = []
    lengths = []
    for dataNum in datasets:
        data = dvw[dataNum]
        r, params = rb.reference(data, A, B, plot=plot, noisy=noisy)
        ref_fids.append(r/1.875)
        ref_fids_err.append(params['rerr']/1.875)        
        refs.append(data)
        lengths.append(data.parameters[data.parameters['measure'][0]]['piLen'])
    # Q2 simult fidelities
    sim_fids = []
    sim_fids_err = []
    datasets = [34, 60, 86, 112, 138, 165, 191, 217, 243, 343]
    for dataNum in datasets:
        data = dvw[dataNum]
        all_rs, all_rs_errs, all_params = rb.simultaneous(data, A, B, plot=plot, noisy=noisy)
        sim_fids.append(all_rs[0]/1.875)
        sim_fids_err.append(params['rerr']/1.875)
    # interleaved IW fidelities
    IW_fids = []
    IW_fids_err = []    
    datasets = [35, 61, 87, 113, 139, 166, 192, 218, 244, 344]
    for dataNum,ref in zip(datasets,refs):
        data = dvw[dataNum]
        r, rc, _, params_inter = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        IW_fids.append(rc)
        IW_fids_err.append(params_inter['rcerr'])
    # interleaved Zpi fidelities
    Z_fids = []
    Z_fids_err=[]
    datasets = [36, 62, 88, 114, 140, 167, 193, 219, 245, 345]
    for dataNum,ref in zip(datasets,refs):
        data = dvw[dataNum]
        r, rc, _, params_inter = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        Z_fids.append(rc)
        Z_fids_err.append(params_inter['rcerr'])        
    # interleaved XX fidelities
    datasets = [297, 300, 303, 307, 310, 313, 316, 319, 322, 346]
    XX_fids = []
    XX_fids_err = []
    for dataNum,ref in zip(datasets,refs):
        data = dvw[dataNum]
        r, rc, _, params_inter = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        XX_fids.append(rc)
        XX_fids_err.append(params_inter['rcerr'])          
    # interleaved YX fidelities
    datasets = [298, 301, 305, 308, 311, 314, 317, 320, 323, 347]
    YX_fids = []
    YX_fids_err = []
    for dataNum,ref in zip(datasets,refs):
        data = dvw[dataNum]
        r, rc, _, params_inter = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        YX_fids.append(rc)   
        YX_fids_err.append(params_inter['rcerr'])          

    return lengths, ref_fids, sim_fids, IW_fids, Z_fids, XX_fids, YX_fids,ref_fids_err,sim_fids_err,IW_fids_err,Z_fids_err,XX_fids_err,YX_fids_err
    
    
def plot3and4(s):
    
    lengths, ref_fids, sim_fids, IW_fids, Z_fids, XX_fids, YX_fids,ref_fids_err,sim_fids_err,IW_fids_err,Z_fids_err,XX_fids_err,YX_fids_err = analyze3and4(s, A=0.48, B=0.47, plot=False, noisy=False)
    
    plt.figure()
    plt.plot(lengths, ref_fids, '-ob', label = 'reference')
    plt.plot(lengths, sim_fids, '-or', label = 'simultaneous')
    plt.plot(lengths, np.array(sim_fids)-np.array(ref_fids), '-ok', label = 'added simult error')
    plt.legend()
    plt.title('Fig 3')
    
    lengths = [currLen.value*2 for currLen in lengths]
    plt.figure()
    #plt.plot(lengths, IW_fids, '-o', label = 'I')
    #plt.plot(lengths, Z_fids, '-o', label = 'Z')
    #plt.plot(lengths, XX_fids, '-o', label = 'XX')
    #plt.plot(lengths, YX_fids, '-o', label = 'YX')
    plt.errorbar(lengths, IW_fids,yerr=IW_fids_err, fmt='-o', label = 'I')
    plt.errorbar(lengths, Z_fids, yerr=Z_fids_err,  fmt='-o', label = 'Z')
    plt.errorbar(lengths, XX_fids,yerr=XX_fids_err, fmt='-o', label = 'XX')
    plt.errorbar(lengths, YX_fids,yerr=YX_fids_err, fmt='-o', label = 'YX')    
    plt.legend()
    plt.title('Fig 4')
    # writetofile('d:\\ucsb\\python\\IW.txt',lengths,IW_fids,IW_fids_err)
    # writetofile('d:\\ucsb\\python\\Z.txt',lengths,Z_fids,Z_fids_err)
    # writetofile('d:\\ucsb\\python\\XX.txt',lengths,XX_fids,XX_fids_err)    
    # writetofile('d:\\ucsb\\python\\YX.txt',lengths,YX_fids,YX_fids_err)
    
def writetofile(fname,x,y,yerr): 
    if fname is not None:
        outfile = open(fname, 'w')
        print >> outfile, ''
        print >> outfile, 'length fid fiderr'
        print >> outfile, ''
        print >> outfile, ''
        for j in range(len(x)):
            towrite = ', '.join(map(str, [x[j],y[j],yerr[j]]))
            print >> outfile,  towrite
        outfile.close()  
        
def analyze1(s, A=None, B=None, plot=False, noisy=False):

    dvw = DataVaultWrapper(s)
    # ref = dvw[23]
    # inter_idles = range(24,64)
    # inter_ses = range(72,107)    
    ref = dvw[135]
    inter_idles = range(24,64)
    inter_ses = range(72,107)
    idle_lens = []
    idle_fids = []
    se_lens = []
    se_fids = []
    for inter_idle in inter_idles:
        data = dvw[inter_idle]
        r, rc, _, _ = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        idle_lens.append(data.parameters[data.parameters['measure'][0]]['identityWaitLen'])
        idle_fids.append(rc)
    for inter_se in inter_ses:
        data = dvw[inter_se]
        r, rc, _, _ = rb.interleaved(ref, data, A, B, plot=plot, noisy=noisy)
        se_lens.append(data.parameters[data.parameters['measure'][0]]['identityWaitLen'])
        se_fids.append(rc)
        
    return idle_lens, idle_fids, se_lens, se_fids
        
def plot1(s):
        
    idle_lens, idle_fids, se_fids = analyze1(s, A=0.495, B=0.47)
    
    plt.figure()
    plt.plot(idle_lens, idle_fids, 'o', label='I')
    plt.plot(idle_lens, se_fids, 'o', label='SE')
    plt.xlabel('Length [ns]')
    plt.ylabel('Error')
    plt.legend()
    
def omegazz(s):
    
    dvw = DataVaultWrapper(['', 'Julian', 'Qubit', '9Xmon', '140610fig34'], cxn=s._cxn)
    data = dvw[1]
    
    delay = data[:,0]
    phase = data[:,1]
    
    p = np.polyfit(delay, phase, 1) #p[0] = rad/ns
    omegazz = p[0]/(2*np.pi) # in GHz
    
def fitFig3(s):
    
    lengths, ref_fids, sim_fids, IW_fids, Z_fids, XX_fids, YX_fids,ref_fids_err,sim_fids_err,IW_fids_err,Z_fids_err,XX_fids_err,YX_fids_err = analyze3and4(s, A=0.48, B=0.47, plot=False, noisy=False)
    lengths = np.array(lengths)
    added_error = np.array(sim_fids)-np.array(ref_fids)
    added_error_errs = [np.sqrt(serr**2 + rerr**2) for serr,rerr in zip(sim_fids_err, ref_fids_err)]
    
    plt.figure()
    plt.plot(lengths, ref_fids, '-ob', label = 'reference')
    plt.plot(lengths, sim_fids, '-or', label = 'simultaneous')
    plt.plot(lengths, added_error, '-ok', label = 'added simult error')
    plt.legend()
    plt.title('Fig 3')
    
    dvw = DataVaultWrapper(['', 'Julian', 'Qubit', '9Xmon', '140610fig34'], cxn=s._cxn)
    data = dvw[1]
    
    delay = data[:,0]
    phase = data[:,1]
    
    p = np.polyfit(delay, phase, 1) #p[0] = rad/ns
    omegazz = p[0]/(2*np.pi) # in GHz
    plt.figure()
    plt.plot(delay, phase, 'ob')
    plt.plot(delay, np.polyval(p, delay), '-r', lw=2)
    plt.title('Omegazz: %1.2f [MHz]' % (omegazz*1e3))
    
    fitFunc = lambda tg, coef, offs: coef * (tg*omegazz)**2+offs
    
    p1,_ = curve_fit(fitFunc, lengths, added_error)
    plt.figure()
    lengths_fit = np.linspace(0, 105, 101)
    plt.plot(lengths, added_error, 'ob')
    plt.plot(lengths_fit, fitFunc(lengths_fit, *p1), '-k')
    
    data_ref = np.vstack((lengths, ref_fids, ref_fids_err)).transpose()
    data_sim = np.vstack((lengths, sim_fids, sim_fids_err)).transpose()
    data_added = np.vstack((lengths, added_error, added_error_errs)).transpose()
    data_omegazz = np.vstack((delay, phase)).transpose()
    data_omegazz_fit = np.vstack((delay, np.polyval(p, delay))).transpose()
    data_added_fit = np.vstack((lengths_fit, fitFunc(lengths_fit, *p1))).transpose()
    
    # np.savetxt('fig3_data_ref.csv', data_ref, delimiter = ',')
    # np.savetxt('fig3_data_sim.csv', data_sim, delimiter = ',')
    # np.savetxt('fig3_data_added.csv', data_added, delimiter = ',')
    # np.savetxt('fig3_data_added_fit.csv', data_added_fit, delimiter = ',')
    # np.savetxt('fig3_data_omegazz.csv', data_omegazz, delimiter = ',')
    # np.savetxt('fig3_data_omegazz_fit.csv', data_omegazz_fit, delimiter = ',')

    
    print p1
    
    
    