"""
Experiment Runner GUI
Using PyQt and LabRAD (synchronous method)
"""
import sys, os, random
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from ListArranger import ListArranger

#import matplotlib
#from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
#from matplotlib.backends.backend_qt4agg import NavigationToolbar2QTAgg as NavigationToolbar
#from matplotlib.figure import Figure

import labrad, labrad.types as T

class AppForm(QMainWindow):
    def __init__(self, parent=None, initialFile = None):
        QMainWindow.__init__(self, parent)
        self.setWindowTitle('Experiment Runner')
        self.cxn = labrad.connect('localhost')

        self.create_menu()
        self.create_main_frame()
        self.create_status_bar()
        self.refresh_server_list()
    
    def create_main_frame(self):
        ''' create a generic widget as the main frame.
        use nested HBoxLayouts and VBoxLayouts to hold the widgets. '''
        mainVBox = QVBoxLayout()
        hbox = QHBoxLayout()
        hbox2 = QHBoxLayout()
        # the labrad host display
        label = QLabel("<b>LabRAD Host</b>")
        self.labrad_host_edit = QLineEdit(self.cxn.host)
        reconnect = QPushButton("Reconnect")
        reconnect.clicked.connect(self.reconnect)
        vbox = QVBoxLayout()
        vbox.addWidget(label)
        hbox2.addWidget(self.labrad_host_edit)
        hbox2.addWidget(reconnect)
        vbox.addLayout(hbox2)

        # server list
        label = QLabel("<b>Servers:</b>")
        self.server_list = QListWidget()
        self.server_list.currentTextChanged.connect(self.server_list_changed)
        vbox.addWidget(label)
        vbox.addWidget(self.server_list)
        hbox.addLayout(vbox)
        
        # next column: settings
        vbox = QVBoxLayout()
        label = QLabel("<b>Server functions:</b>")
        self.setting_list = QListWidget()
        self.setting_list.currentTextChanged.connect(self.setting_list_changed)
        vbox.addWidget(label)
        vbox.addWidget(self.setting_list)
        hbox.addLayout(vbox)
        
        ## next column: about the setting/parameters ##
        # label, list and controls
        label = QLabel("<b>Parameters:</b>")
        self.parameter_list = QListWidget()
        listarr = ListArranger(None, self.parameter_list, self.statusBar())
        # add button and edit box
        add_param = QPushButton("Add Parameter")
        self.parameter_edit = QLineEdit()
        add_param.clicked.connect(self.add_parameter)
        help = QPushButton("?")
        help.setToolTip("List of arguments to pass to the function.\n  Values are eval'd before passing, so put strings in quotes!\n  " + 
                "Use \"VAL\" to indicate where a numeric value should be used in a sweep function.\n  " + 
                "Otherwise, it is automatically added to the end (if you use this as a sweep function).")
        help.setMaximumWidth(help.fontMetrics().boundingRect(help.text()).width() + 15)
        test_button = QPushButton("Test Current Selection")
        test_button.clicked.connect(self.test_selection)
        test_button.setToolTip("Run current function and display results.\nNote: this actually runs the function with given parameters. So be careful!")
        # do the layout
        vbox = QVBoxLayout()
        hbox2 = QHBoxLayout(); hbox2.addWidget(help); hbox2.addWidget(label)
        vbox.addLayout(hbox2)
        vbox.addWidget(self.parameter_list)
        vbox.addWidget(listarr)
        hbox2 = QHBoxLayout(); hbox2.addWidget(self.parameter_edit); hbox2.addWidget(add_param)
        vbox.addLayout(hbox2)
        vbox.addWidget(test_button)
        hbox.addLayout(vbox)
        
        mainVBox.addLayout(hbox)
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        mainVBox.addWidget(line)
        
        
        ### finished upper row. now do lower. ###
        hbox = QHBoxLayout()
        # first, config function stuff
        label = QLabel("<b>Config functions:</b>")
        self.config_func_list = QListWidget()
        listarr = ListArranger(None, self.config_func_list,self.statusBar())
        self.add_config_func_button = QPushButton("Add as Config Func")
        self.add_config_func_button.clicked.connect(self.add_config_func)
        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addWidget(self.config_func_list)
        vbox.addWidget(listarr)
        vbox.addWidget(self.add_config_func_button)
        hbox.addLayout(vbox)
        
        # next, sweep/scan funcs
        label = QLabel("<b>Sweep functions:</b>")
        self.sweep_func_list = QListWidget()
        add_sweep_func = QPushButton("Add as Sweep Func")
        add_sweep_func.clicked.connect(self.add_sweep_func)
        listarr = ListArranger(None, self.sweep_func_list,self.statusBar())
        l1 = QLabel("Start")
        self.start_edit = QLineEdit()
        l2 = QLabel("Stop")
        self.stop_edit = QLineEdit()
        l3 = QLabel("Step")
        self.step_edit = QLineEdit()
        self.set_as_get_checkbox = QCheckBox("Use this as get function")
        self.set_as_get_checkbox.setToolTip("To get the value of this parameter, use the same function, without the argument.")
        self.simple_compare_checkbox = QCheckBox("Use simple comparison")
        self.simple_compare_checkbox.setToolTip("Use a simple numerical comparison to determine if we've reached the set value.\n(i.e. no additional check for system stability, etc.)")
        self.simple_compare_edit = QLineEdit('0.01')
        self.simple_compare_edit.setToolTip("Fraction to use for simple numerical comparison")
        
        hbox2 = QHBoxLayout()
        hbox2.addWidget(l1); hbox2.addWidget(self.start_edit)
        hbox2.addWidget(l2); hbox2.addWidget(self.stop_edit)
        hbox2.addWidget(l3); hbox2.addWidget(self.step_edit)
        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addWidget(self.sweep_func_list)
        vbox.addWidget(listarr)
        vbox.addLayout(hbox2)
        hbox2 = QHBoxLayout(); hbox2.addWidget(self.set_as_get_checkbox); hbox2.addWidget(self.simple_compare_checkbox); hbox2.addWidget(self.simple_compare_edit)
        vbox.addLayout(hbox2)
        vbox.addWidget(add_sweep_func)
        hbox.addLayout(vbox)
        
        # finally, measurement funcs
        label = QLabel("<b>Measure functions:</b>")
        self.measure_func_list = QListWidget()
        listarr = ListArranger(None, self.sweep_func_list,self.statusBar())
        self.add_measure_func_button = QPushButton("Add as Measure Func")
        self.add_measure_func_button.clicked.connect(self.add_measure_func)
        vbox = QVBoxLayout()
        vbox.addWidget(label)
        vbox.addWidget(self.measure_func_list)
        vbox.addWidget(listarr)
        vbox.addWidget(self.add_measure_func_button)
        hbox.addLayout(vbox)
        
        mainVBox.addLayout(hbox)
        
        # wrap it up in a widget
        self.main_frame = QWidget()
        self.main_frame.setLayout(mainVBox)
        self.setCentralWidget(self.main_frame)
        self.add_mode = None    # mode is either None, 'get_func', 'at_func' -- latter two for when we need to add a specific func
    
    def reconnect(self):
        self.cxn = labrad.connect(str(self.labrad_host_edit.text()))
        self.refresh_server_list()
    
    def refresh_server_list(self):
        self.server_list.clear()
        for s in self.cxn.servers:
            self.server_list.addItem(s)
            
    def server_list_changed(self, text):
        ''' when we select a new server, list the settings. '''
        self.setting_list.clear()
        if not text:
            return
        for s in self.cxn[str(text)].settings:
            i = QListWidgetItem(str(s))
            i.setToolTip(str(self.cxn[str(text)][str(s)]))
            self.setting_list.addItem(i)
            
    def setting_list_changed(self, text):
        ''' when we choose a new setting, update the stats about it. '''
        pass
        
    def add_parameter(self):
        p = str(self.parameter_edit.text())
        if p:
            self.parameter_list.addItem(p)
        else:
            self.statusBar().showMessage("No blank parameters allowed.", 3000)
            
    def test_selection(self):
        ''' call the currently configured function and print results '''
        func = self.create_func_wrapper()
        res = str(func())
        QMessageBox.information(self, func.string, res)
            
    def add_config_func(self):
        '''' callback. creates a function wrapper and sticks it in the list item. '''
        func = self.create_func_wrapper()
        item = QListWidgetItem(func.string)
        item.func = func
        self.config_func_list.addItem(item)
        
    def add_sweep_func(self):
        ''' adding a sweep function is more complicated. need to have a set function, get function, test equality function... '''
        if not self.start_edit.text() or not self.stop_edit.text():
            self.statusBar().showMessage("Must have a start and stop value for a sweep.")
            return
        start, stop = float(str(self.start_edit.text())), float(str(self.stop_edit.text()))
        if self.step_edit.text():
            step = float(str(self.step_edit.text()))
        else:
            step = 0

        if self.add_mode == None:
            func = self.create_func_wrapper(addVal=True)
            item = QListWidgetItem(func.string)
            item.func = func
            item.sweep = [start, stop, step]
            self.sweep_func_list.addItem(item)
            # get the get function
            if self.simple_compare_checkbox.isChecked():
                item.compare_val = float(self.simple_compare_edit.text())
            else:
                item.compare_val = None
                self.add_mode = 'at_func'
            if self.set_as_get_checkbox.isChecked():
                item.get_func = self.create_func_wrapper(ignoreVal=True)
            else:
                self.add_mode = 'get_func'

        elif self.add_mode == 'get_func':
            item = self.sweep_func_list.item(self.sweep_func_list.count()-1)
            item.get_func = self.create_func_wrapper()
            if item.compare_val == None:
                self.add_mode = 'at_func'
            else:
                self.add_mode = None
        elif self.add_mode == 'at_func':
            item = self.sweep_func_list.item(self.sweep_func_list.count()-1)
            item.compare_func = self.create_func_wrapper(addVal=True)
            self.add_mode = None
        
        if self.add_mode == 'at_func':
            self.add_config_func_button.setEnabled(False)
            self.add_measure_func_button.setEnabled(False)
            self.statusBar().showMessage('Next function must be the function to determine when the sweep is AT a given value.')
        elif self.add_mode == 'get_func':
            self.add_config_func_button.setEnabled(False)
            self.add_measure_func_button.setEnabled(False)
            self.statusBar().showMessage('Next function must be the function to GET what the sweep value is.')
        else:
            self.add_config_func_button.setEnabled(True)
            self.add_measure_func_button.setEnabled(True)
            self.statusBar().showMessage('Thank you for the function.', 3000)
        
    def add_measure_func(self):
        func = self.create_func_wrapper()
        item = QListWidgetItem(func.string)
        item.func = func
        self.measure_func_list.addItem(item)
        
    def create_func_wrapper(self, ignoreVal = False, addVal = False):
        ''' create a function wrapper out of the currently selected options.
        if ignoreVal = True, remove a parameter that starts with VAL (if found).
        otherwise, replace VAL with the argument to the function wrapper every time it's called.
        (e.g. for setting a value to be used in a sweep or scan.)
        (ignoreVal is used when you have the same function for setting a sweep param and measuring where it is.)
        if addVal, then add a 'VAL' entry if we don't have one already'''
        srv = str(self.server_list.currentItem().text())
        setting = str(self.setting_list.currentItem().text())
        plist = []
        valIndex = None
        for i in xrange(self.parameter_list.count()):
            p = str(self.parameter_list.item(i).text())
            if p.upper().startswith('VAL'):
                valIndex = i
            else:
                p = self.handle_argument(p)
            plist.append(p)
        if addVal and not valIndex:
            plist.append('VAL')
            valIndex = len(plist)-1
        if ignoreVal and valIndex:
            plist.pop(valIndex)
            def wrapper():
                return self.cxn[srv][setting](*plist)
        else:
            def wrapper(v):
                plist[valIndex] = v
                return self.cxn[srv][setting](*plist)
        wrapper.string = '%s.%s(%s)' % (srv, setting, ','.join([str(p) for p in plist]))
        return wrapper
        
    def handle_argument(self, arg):
        ''' takes a string argument, and attempts to process it inot a python type '''
        return eval(arg)
    
    def create_status_bar(self):
        self.statusBar().showMessage("Create your experiment!")
        
    def create_menu(self):        
        self.file_menu = self.menuBar().addMenu("&File")
        
        quit_action = self.create_action("&Quit", slot=self.close, 
            shortcut="Ctrl+Q", tip="Close the application")
        
        self.add_actions(self.file_menu, 
            (None, quit_action))

    def add_actions(self, target, actions):
        for action in actions:
            if action is None:
                target.addSeparator()
            else:
                target.addAction(action)

    def create_action(  self, text, slot=None, shortcut=None, 
                        icon=None, tip=None, checkable=False, 
                        signal="triggered()"):
        action = QAction(text, self)
        if icon is not None:
            action.setIcon(QIcon(":/%s.png" % icon))
        if shortcut is not None:
            action.setShortcut(shortcut)
        if tip is not None:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot is not None:
            self.connect(action, SIGNAL(signal), slot)
        if checkable:
            action.setCheckable(True)
        return action
        
    def closeEvent(self, event):
        self.cxn.disconnect()
        event.accept()


def main():
    initialFile = None
    for arg in sys.argv[1:]:
        if not arg.startswith('-'):
            initialFile = arg
    app = QApplication(sys.argv)
    form = AppForm(initialFile=initialFile)
    form.show()
    app.exec_()


if __name__ == "__main__":
    main()
