#Standard packages
import numpy as np
from math import atan2
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import scipy
import itertools
import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
from pyle import signalprocessing as signalprocessing
import pyle.envelopes as env
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import measurement
from pyle.dataking import squid
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.dataking import singleQubit as sq
from pyle.util import sweeptools as st
from pyle.dataking import utilMultilevels as ml
from pyle.plotting import dstools as ds
from pyle.fitting import fitting
from pyle.util import structures
from pyle.datavault import DataVaultWrapper
from pyle import tomo as tomo
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray

from pyle.sim import pyleSim as ps
from pyle.math import tensor

from pyle import gateCompiler as gc
from pyle import gates

from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss

from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.dataking.benchmarking import randomizedBenchmarking as rb

from pyle.dataking.DevDaniel.Qubit import readout
from pyle.dataking import multipleQubit as mq

from pyle import quantum
# from classCZ import adiabaticCZtuning
import copy
import time
from zfuncs import *



def calLGBI(Sample, measure, name='aCZ ancilla tuner',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.15:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]  
    if newSpline:
        a1Spline = mq.arbPhaseTunerQST(Sample, measure, 0,1,amplitude=st.r[0.25:0.82:0.01],stats=1500,retSpline=True,plot=False)
        Sample['q0']['aczPhaseToAmp'] = (dvw._dir,int(dvw[-1].name.split('-')[0]))
        a1CorrSpline = mq.arbPhaseFitterQST(Sample, measure, 0, 1, amps = st.r[0.25:0.82:0.01], iterations = 3)
        Sample['q0']['aczAmpToPhase'] = (dvw._dir,int(dvw[-1].name.split('-')[0]))
        a2Spline  = mq.arbPhaseTunerQST(Sample, measure, 2,3,amplitude=st.r[0.25:0.76:0.01],stats=1500,retSpline=True,plot=False)
        Sample['q3']['aczPhaseToAmp'] = (dvw._dir,int(dvw[-1].name.split('-')[0]))
        a2CorrSpline = mq.arbPhaseFitterQST(Sample, [2,3], 2, 3, amps = st.r[0.25:0.76:0.01], iterations = 3)
        Sample['q3']['aczAmpToPhase'] = (dvw._dir,int(dvw[-1].name.split('-')[0]))
    
    else:
        path = Sample['q0']['aczPhaseToAmp'][0]
        amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
        relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
        amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
        absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
        #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
        phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
        a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
        a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
        if checkSpline:
            plt.plot(relPhases,amps1)
            plt.plot(phase, a1Spline(phase))
        
        path = Sample['q3']['aczPhaseToAmp'][0]
        amps1 = dvw[int(Sample['q3']['aczPhaseToAmp'][1])][:,0]
        relPhases = dvw[int(Sample['q3']['aczPhaseToAmp'][1])][:,1]
        amps2 = dvw[int(Sample['q3']['aczAmpToPhase'][1])][:,0]
        absPhases = np.unwrap(dvw[int(Sample['q3']['aczAmpToPhase'][1])][:,1])/3
        #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
        phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
        a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
        a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
        if checkSpline:
            plt.plot(relPhases,amps1)
            plt.plot(phase, a2Spline(phase))
        

            
def LGBI(Sample, measure, name='LBGI measurement',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.15:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, ancON = True,manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)


    path = Sample['q0']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        plt.plot(relPhases,amps1)
        plt.plot(phase, a1Spline(phase))
    
    path = Sample['q3']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q3']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q3']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q3']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q3']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        plt.plot(relPhases,amps1)
        plt.plot(phase, a2Spline(phase))
        
        
    anc1 = 0
    
    bell2 = 2
    bell1 = 1
     
    anc2 = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qBellT, QBellT = qubits[bell1], Qubits[bell1]
    qBellC, QBellC = qubits[bell2], Qubits[bell2]
    # q0=qubits[0]
    qa1, Qa1=qubits[anc1],Qubits[anc1]
    qa2, Qa2=qubits[anc2],Qubits[anc2]

    qBellT['readout'] = True
    qBellC['readout'] = True
    qa1['readout'] = True 
    qa2['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    bell2Qubit = devs[bell2]
    bell1Qubit = devs[bell1]
    anc1Qubit = devs[anc1]
    anc2Qubit = devs[anc2]
        
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(phase, 'Ancilla Cphase phase shift'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Prob |q0>', '', ''),('Prob |q1>', '', ''),('Prob |q2>','',''),('Prob |q3>','',''),('Coralator', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase, currLen, it,waitTime):
        
        reqs = []
        

        alg = gc.Algorithm(agents=devs)
        if bellON:
            #Next 4 lines make a Bell state
            alg[gates.PiHalfPulse([ bell1Qubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ bell2Qubit ])] #X_pi/2                \
            alg[gates.adiabaticCZ( devs, [bell2Qubit , bell1Qubit] )]   #   ||   This is apparently a CNOT
            alg[gates.PiHalfPulse([ bell2Qubit ],phase=np.pi )] #-X_pi/2   /
        
        if ancON:
            alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]
            alg[gates.PiQuarterPulseCorr([anc1Qubit],phase=np.pi )]
            alg[gates.PiQuarterPulseCorr([anc2Qubit],phase=np.pi )]
            alg[gates.PiThreeEigthsPulseCorr([bell2Qubit],phase=-np.pi/2 )]
            
            alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]

            for k in range(it):
                alg[gates.PiHalfPulse([ anc1Qubit],phase=np.pi/2)] #X_pi/2
                alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase=a1CorrSpline(a1Spline(currPhase)), amp=a1Spline(currPhase),length=currLen)]
                alg[gates.PiHalfPulse([ anc1Qubit],phase=-np.pi/2)]
                alg[gates.PiHalfPulse([ anc2Qubit],phase=np.pi/2)]
                alg[gates.adiabaticCPhase(devs,[bell2Qubit, anc2Qubit],corrPhase=a2CorrSpline(a2Spline(currPhase)), amp=a2Spline(currPhase),length=currLen)]
                alg[gates.PiHalfPulse([ anc2Qubit],phase=-np.pi/2)]
                # if piPulse:
                    # alg[gates.PiPulse([targetQubit])]                
                    # alg[gates.PiPulse([devs[0]])]     
                    # alg[gates.PiHalfPulse([devs[3]])]                    
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
             
            alg[gates.PiQuarterPulseCorr([anc1Qubit])]
            alg[gates.PiQuarterPulseCorr([anc2Qubit])]
            alg[gates.PiQuarterPulseCorr([anc1Qubit],phase=-np.pi/2 )]
            alg[gates.PiQuarterPulseCorr([anc2Qubit],phase=-np.pi/2 )]
            alg[gates.PiQuarterPulseCorr([bell1Qubit],phase=-np.pi/2 )]
            alg[gates.PiQuarterPulseCorr([bell2Qubit],phase=np.pi/2 )]
            
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        #ps.plotPulses(devs)

        
        data = yield runQubits(server, devs, stats, dataFormat='iqRaw')

            # Getting results for loaded (with Bell state) case for all 4 qubits
        pa1 = mq.mqReadoutProbabilities(data, devs, measure, correlated = False)[anc1]
        pa2 = mq.mqReadoutProbabilities(data, devs, measure, correlated = False)[anc2]
        pb1 = mq.mqReadoutProbabilities(data, devs, measure, correlated = False)[bell1]
        pb2 = mq.mqReadoutProbabilities(data, devs, measure, correlated = False)[bell2]


        
        def correlator(a1,a2,b1,b2):
            C = a1*a2+a1*b2+b1*a2-b1*b2
            return C
        
        returnValue([pa1,pb1,pb2,pa2,correlator(pa1,pa2,pb1,pb2)])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    return data

def LGBICalVis(Sample, measure, name='LBGI measurement',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.15:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, ancON = True,manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)


    path = Sample['q0']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a1Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a1CorrSpline(amps2))
    
    path = Sample['q2']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a2Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a2CorrSpline(amps2))
        
        
    anc1 = 0
    
    bell2 = 2
    bell1 = 1
     
    anc2 = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qBellT, QBellT = qubits[bell1], Qubits[bell1]
    qBellC, QBellC = qubits[bell2], Qubits[bell2]
    # q0=qubits[0]
    qa1, Qa1=qubits[anc1],Qubits[anc1]
    qa2, Qa2=qubits[anc2],Qubits[anc2]

    qBellT['readout'] = True
    qBellC['readout'] = True
    qa1['readout'] = True 
    qa2['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    bell2Qubit = devs[bell2]
    bell1Qubit = devs[bell1]
    anc1Qubit = devs[anc1]
    anc2Qubit = devs[anc2]
        
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(phase, 'Ancilla Cphase phase shift'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Prob |q0> unloaded', '', ''),('Prob |q0> loaded', '', ''),('Prob |q3> unloaded','',''),('Prob |q3> loaded','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase, currLen, it,waitTime):
        reqs=[]
        currAmp1 = a1Spline(currPhase)
        corrPhase1 = a1CorrSpline(currAmp1)
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiHalfPulse([anc1Qubit],phase=np.pi/2.0)] #X_pi/2
        alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase1, amp=currAmp1)]
        alg[gates.ChangePhase([anc1Qubit],xyPhase = -currPhase/2.0)]
        alg[gates.ChangePhase([bell1Qubit],xyPhase = -currPhase/2.0)]
        alg[gates.PiHalfPulse([anc1Qubit],phase=-np.pi/2.0)]
        
        currAmp2 = a2Spline(currPhase)
        corrPhase2 = a2CorrSpline(currAmp2)
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi/2)]
        alg[gates.adiabaticCPhase(devs,[bell2Qubit, anc2Qubit],corrPhase = -corrPhase2, amp=currAmp2)]
        alg[gates.ChangePhase([anc2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.ChangePhase([bell2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.PiHalfPulse([anc2Qubit],phase=-np.pi/2)]

        alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
         
        alg[gates.PiHalfPulse([anc1Qubit],phase=0)]
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi)]
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        dataUnLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        pU1 = mq.mqReadoutProbabilities(dataUnLoaded, devs, measure, correlated = False)[anc1]
        pU2 = mq.mqReadoutProbabilities(dataUnLoaded, devs, measure, correlated = False)[anc2]
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([bell1Qubit])]
        alg[gates.PiPulse([bell2Qubit])]
        # alg[gates.PiHalfPulse([anc1Qubit],phase=0)] #X_pi/2
        alg[gates.PiHalfPulse([anc1Qubit], phase = np.pi/2.0)] #X_pi/2
        alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase1, amp=currAmp1)]
        alg[gates.ChangePhase([anc1Qubit],xyPhase = -currPhase/2.0)]
        alg[gates.ChangePhase([bell1Qubit],xyPhase = -currPhase/2.0)]
        alg[gates.PiHalfPulse([anc1Qubit],phase=-np.pi/2.0)]
        
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi/2)]
        alg[gates.adiabaticCPhase(devs,[bell2Qubit, anc2Qubit],corrPhase = -corrPhase2, amp=currAmp2)]
        alg[gates.ChangePhase([anc2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.ChangePhase([bell2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.PiHalfPulse([ anc2Qubit],phase=-np.pi/2)]
        
        alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
         
        alg[gates.PiHalfPulse([anc1Qubit],phase=0)]
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi)]
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        dataLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')

            # Getting results for loaded (with Bell state) case for all 4 qubits
        pL1 = mq.mqReadoutProbabilities(dataLoaded, devs, measure, correlated = False)[anc1]
        pL2 = mq.mqReadoutProbabilities(dataLoaded, devs, measure, correlated = False)[anc2]

        returnValue([pU1,pL1,pU2,pL2])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
        
def LGBICal01(Sample, measure, name='LBGI measurement',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.3:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, ancON = True,manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)


    path = Sample['q0']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a1Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a1CorrSpline(amps2))
    
    path = Sample['q2']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a2Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a2CorrSpline(amps2))
        
        
    anc1 = 0
    
    bell2 = 2
    bell1 = 1
     
    anc2 = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qBellT, QBellT = qubits[bell1], Qubits[bell1]
    qBellC, QBellC = qubits[bell2], Qubits[bell2]
    # q0=qubits[0]
    qa1, Qa1=qubits[anc1],Qubits[anc1]
    qa2, Qa2=qubits[anc2],Qubits[anc2]

    qBellT['readout'] = True
    qBellC['readout'] = True
    qa1['readout'] = True 
    qa2['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    bell2Qubit = devs[bell2]
    bell1Qubit = devs[bell1]
    anc1Qubit = devs[anc1]
    anc2Qubit = devs[anc2]
        
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(phase, 'Ancilla Cphase phase shift'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Prob |q0> unloaded', '', ''),('Prob |q0> loaded', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase, currLen, it,waitTime):
        currAmp = a1Spline(currPhase)
        corrPhase = a1CorrSpline(currAmp)
        reqs = []
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiHalfPulse([anc1Qubit],phase=np.pi/2.0)] #X_pi/2
        print currPhase,a1Spline(currPhase),a1CorrSpline(a1Spline(currPhase))
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        alg[gates.PiHalfPulse([anc1Qubit],phase=-np.pi/2.0)]
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        dataUnLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        pU1 = mq.mqReadoutProbabilities(dataUnLoaded, devs, measure, correlated = False)[anc1]
        
        # reqs = []
        # alg = gc.Algorithm(agents=devs)
        # alg[gates.PiHalfPulse([bell1Qubit],phase=-np.pi/2.0)] #X_pi/2
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit], corrPhase=a1CorrSpline(a1Spline(currPhase)), amp=a1Spline(currPhase),length=currLen)]
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        # alg[gates.PiHalfPulse([bell1Qubit],phase=np.pi/2.0)]
        # alg[gates.Measure(devs)]
        # alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        # dataUnLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        # pU1 = mq.mqReadoutProbabilities(dataUnLoaded, devs, measure, correlated = False)[bell1]
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([bell1Qubit])]
        alg[gates.PiHalfPulse([anc1Qubit],phase=np.pi/2.0)] #X_pi/2
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        alg[gates.PiHalfPulse([anc1Qubit],phase=-np.pi/2.0)]
        alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        dataLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        pL1 = mq.mqReadoutProbabilities(dataLoaded, devs, measure, correlated = False)[anc1]
        
        # alg = gc.Algorithm(agents=devs)
        # alg[gates.PiPulse([anc1Qubit])]
        # alg[gates.PiHalfPulse([bell1Qubit],phase=-np.pi/2.0)] #X_pi/2
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase=a1CorrSpline(a1Spline(currPhase)), amp=a1Spline(currPhase),length=currLen)]
        # alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
        # alg[gates.PiHalfPulse([bell1Qubit],phase=np.pi/2.0)]
        # alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
        # alg[gates.Measure(devs)]
        # alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        # reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        # dataLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        # pL1 = mq.mqReadoutProbabilities(dataLoaded, devs, measure, correlated = False)[bell1]

        returnValue([pU1,pL1])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def LGBICalDetunePhase(Sample, measure, name='Ancilla Phase Calibration',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.3:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, ancON = True,manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)

    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]
    
    path = Sample['q0']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a1Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a1CorrSpline(amps2))
    
    path = Sample['q2']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a2Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a2CorrSpline(amps2))
        
        
    anc1 = 0
    
    bell2 = 2
    bell1 = 1
     
    anc2 = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qBellT, QBellT = qubits[bell1], Qubits[bell1]
    qBellC, QBellC = qubits[bell2], Qubits[bell2]
    # q0=qubits[0]
    qa1, Qa1=qubits[anc1],Qubits[anc1]
    qa2, Qa2=qubits[anc2],Qubits[anc2]

    qBellT['readout'] = True
    qBellC['readout'] = True
    qa1['readout'] = True 
    qa2['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    bell2Qubit = devs[bell2]
    bell1Qubit = devs[bell1]
    anc1Qubit = devs[anc1]
    anc2Qubit = devs[anc2]
        
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(phase, 'Ancilla Cphase phase shift'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Ancilla Phase', '', ''),('Ancilla x/2 Probability ', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase, currLen, currIt,waitTime):
        
        reqs = []
        #amp = a1Spline(currPhase)
        currAmp = a1Spline(currPhase)
        corrPhase = a1CorrSpline(currAmp)
        for op in opList:
            # Control Ramsey
            alg = gc.Algorithm()
            alg[gates.PiHalfPulse([anc1Qubit])]
            alg[gates.PiPulse([bell1Qubit])]
            
            for k in range(currIt):
                alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase = -corrPhase, amp=currAmp)]
                if piPulse:
                    alg[gates.PiPulse([anc1Qubit],phase=phaseC)]                
                if currIt>1 and k<currIt-1:
                    alg[gates.Wait([anc1Qubit], waitTime)] #time between CZ's
                    
            #alg[gates.Wait([controlQubit,targetQubit], waitTime)] #time after last CZ.
            alg[gates.Tomography([anc1Qubit], op)]
            alg[gates.Measure(devs)]
            alg.compile(correctXtalkZ=correctXtalkZ,beSmart=True)

            reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
         
        # Target Ramsey
        alg = gc.Algorithm()
        alg[gates.PiHalfPulse([anc1Qubit], phase = np.pi/2.0)]
        alg[gates.PiPulse([bell1Qubit])]
        
        for k in range(currIt):
            alg[gates.adiabaticCPhase(devs,[anc1Qubit, bell1Qubit],corrPhase=-corrPhase, amp=currAmp)]
            if piPulse:
                alg[gates.PiPulse([bell1Qubit],phase=phaseT)]                
            if currIt>1 and k<currIt-1:
                alg[gates.Wait([anc1Qubit], waitTime)] #time between CZ's
                
        #alg[gates.Wait([controlQubit,targetQubit], waitTime)] #time after last CZ.            
        alg[gates.PiHalfPulse([anc1Qubit], phase = -np.pi/2.0)]
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=True)
    
        dataProb = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        
        dataControl = []
        dataTarget = []
        msorted=list( np.sort(measure) )
        results = yield FutureList(reqs)
        for r in results:
            Is,Qs = r[msorted.index(anc1)]
            prob = np.mean(sq.readoutProbabilities(anc1Qubit,np.array(Is)+1.0j*np.array(Qs)))
            dataControl.append(prob)
        pL1 = mq.mqReadoutProbabilities(dataProb, devs, measure, correlated = False)[anc1]
        
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl = ps.getDensityPhase(rho)[0] -np.pi/2.    
   
        
        returnValue([phaseControl, pL1])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    
def LGBICal23(Sample, measure, name='LBGI measurement',
             length=None, amplitude = st.r[0.5:0.8:0.01],phase=st.r[0.15:np.pi:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, ancON = True,manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False,newSpline=True, checkSpline = False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    def fitFunc(x,a,b,c):
        #fit func to calculate phase vs. amplitude for arbitraty phase
        return a*(np.exp(b*x))+c*x
    dvw = DataVaultWrapper(Sample)


    path = Sample['q0']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q0']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q0']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a1Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a1CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a1Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a1CorrSpline(amps2))
    
    path = Sample['q2']['aczPhaseToAmp'][0]
    amps1 = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,0]
    relPhases = dvw[int(Sample['q2']['aczPhaseToAmp'][1])][:,1]
    amps2 = dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,0]
    absPhases = np.unwrap(dvw[int(Sample['q2']['aczAmpToPhase'][1])][:,1])/3
    #using just a spline was malfunctioning thats why we use a spline to invert a curve fit.
    phaseFit, error = curve_fit(fitFunc,amps1,relPhases,[0.01,5,0.1])
    a2Spline = scipy.interpolate.UnivariateSpline(fitFunc(amps1,phaseFit[0],phaseFit[1],phaseFit[2]), amps1, s=0)
    a2CorrSpline = scipy.interpolate.UnivariateSpline(amps2, absPhases, s=0.5)
    if checkSpline:
        # plt.plot(relPhases,amps1)
        # plt.plot(phase, a2Spline(phase))
        plt.plot(amps2,absPhases)
        plt.plot(amps2, a2CorrSpline(amps2))
        
        
    anc1 = 0
    
    bell2 = 2
    bell1 = 1
     
    anc2 = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qBellT, QBellT = qubits[bell1], Qubits[bell1]
    qBellC, QBellC = qubits[bell2], Qubits[bell2]
    # q0=qubits[0]
    qa1, Qa1=qubits[anc1],Qubits[anc1]
    qa2, Qa2=qubits[anc2],Qubits[anc2]

    qBellT['readout'] = True
    qBellC['readout'] = True
    qa1['readout'] = True 
    qa2['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    bell2Qubit = devs[bell2]
    bell1Qubit = devs[bell1]
    anc1Qubit = devs[anc1]
    anc2Qubit = devs[anc2]
        
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(phase, 'Ancilla Cphase phase shift'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Prob |q0> unloaded', '', ''),('Prob |q0> loaded', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase, currLen, it,waitTime):
         
        reqs = []
        #Unloaded, no pi/2 pulse
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi/2)]
        alg[gates.adiabaticCPhase(devs,[bell2Qubit, anc2Qubit],corrPhase=a2CorrSpline(a2Spline(currPhase)), amp=a2Spline(currPhase),length=currLen)]
        alg[gates.ChangePhase([anc2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.ChangePhase([bell2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.PiHalfPulse([anc2Qubit],phase=-np.pi/2)]

        

        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        dataUnLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        pU1 = mq.mqReadoutProbabilities(dataUnLoaded, devs, measure, correlated = False)[anc2]
        
        alg = gc.Algorithm(agents=devs)
        #Loaded, has pi/2 pulse
        alg[gates.PiPulse([bell2Qubit])]
        alg[gates.PiHalfPulse([anc2Qubit],phase=np.pi/2)]
        alg[gates.adiabaticCPhase(devs,[bell2Qubit, anc2Qubit],corrPhase=a2CorrSpline(a2Spline(currPhase)), amp=a2Spline(currPhase),length=currLen)]
        alg[gates.ChangePhase([anc2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.ChangePhase([bell2Qubit],xyPhase = currPhase/2.0)]
        alg[gates.PiHalfPulse([ anc2Qubit],phase=-np.pi/2)]
        
        alg[gates.Sync([anc1Qubit,anc2Qubit,bell1Qubit,bell2Qubit])]        
        
        alg[gates.Measure(devs)]
        alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        dataLoaded = yield runQubits(server, devs, stats, dataFormat='iqRaw')

            # Getting results for loaded (with Bell state) case for all 4 qubits
        pL1 = mq.mqReadoutProbabilities(dataLoaded, devs, measure, correlated = False)[anc2]

        returnValue([pU1,pL1])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    return data    
    
def aczPhaseVaryer(Sample, measure, control, target, name='aCZ ancilla tuner',
             length=None, amplitude = st.r[0.5:0.8:0.01],
             iterations = 1, waitTime = 0.0,bellON=True, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True,piPulse=False,beSmart=False):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]    
    
    bellControl = 2
    bellTarget = 1
    
    anc1Control = 1
    anc1Target  = 0
    
    anc2Control = 2
    anc2Target = 3

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = qubits[target], Qubits[target]
    # q0=qubits[0]
    q2=qubits[2]
    q3=qubits[3]
    qC, QC = qubits[control], Qubits[control]
    q['readout'] = True
    qC['readout'] = True
    q2['readout'] = True 
    q3['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    controlQubit = devs[control]
    targetQubit = devs[target]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.98, origAmplitude*1.02, 21)
    
    axes = [(amplitude, 'Adiabatic Cphase Amplitude'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Relative Phase |q0>', '', ''),('Relative Phase |q3>', '', ''),('Probability |q1>','',''),('Probability |q2>','',''),('Phase Coherence unloaded', '', ''),('Phase Coherence loaded', '', '')]
    # deps = [('Relative Phase |q0>', '', ''),('Phase Coherence unloaded', '', ''),('Phase Coherence loaded', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp, currLen, it,waitTime):
        
        reqs = []
        
        for op in opList:
            # unloaded control
            alg = gc.Algorithm(agents=devs)
            # alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.PiHalfPulse([devs[0]])]
            alg[gates.PiHalfPulse([devs[3]])]
            
            for k in range(it):
                alg[gates.adiabaticCZ(devs,[controlQubit, targetQubit],amp=currAmp,length=currLen)]
                alg[gates.adiabaticCZ(devs,[devs[2], devs[3]],amp=currAmp,length=currLen)]
                if piPulse:
                    # alg[gates.PiPulse([targetQubit])]                
                    alg[gates.PiPulse([devs[0]])]                
                    alg[gates.PiPulse([devs[3]])]                
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
                    alg[gates.Wait([devs[3]], waitTime)]
            
            alg[gates.Tomography([devs[3]], op)]
            alg[gates.Tomography([controlQubit], op)]
            # alg[gates.Tomography([devs[0],devs[1]], op)]
            # alg[gates.Tomography([controlQubit,devs[3]], op)]
            alg[gates.Measure(devs)]
            alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)
        
            reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
         
        for op in opList:
            # loaded control
            alg = gc.Algorithm(agents=devs)
            
            if bellON:
                #Next 4 lines make a Bell state
                alg[gates.PiHalfPulse([ devs[1] ] )] #X_pi/2
                alg[gates.PiHalfPulse([ devs[2] ] )] #X_pi/2                \
                alg[gates.adiabaticCZ( devs , [ devs[2] , devs[1]] )]   #   ||   This is apparently a CNOT
                alg[gates.PiHalfPulse([ devs[2] ],phase=np.pi )] #-X_pi/2   /
            if not bellON:
            # alg[gates.PiPulse([controlQubit])]
                alg[gates.PiPulse([devs[1]])]
                alg[gates.PiPulse([devs[2]])]
            # alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.PiHalfPulse([devs[0]])]
                
            alg[gates.PiHalfPulse([devs[3]])]
            
            for k in range(it):
                alg[gates.adiabaticCZ(devs, [controlQubit, targetQubit],amp=currAmp,length=currLen)]
                alg[gates.adiabaticCZ(devs,[devs[2], devs[3]],amp=currAmp,length=currLen)]
                if piPulse:
                    # alg[gates.PiPulse([targetQubit])]                
                    alg[gates.PiPulse([devs[0]])]     
                    alg[gates.PiHalfPulse([devs[3]])]                    
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
                    
                    
            alg[gates.Tomography([devs[3]], op)]
            alg[gates.Tomography([controlQubit], op)]
            # alg[gates.Tomography([devs[0],devs[1]], op)]
            # alg[gates.Tomography([controlQubit,devs[3]], op)]
            # alg[gates.Tomography([targetQubit,devs[3]], op)]
            alg[gates.Measure(devs)]
            alg.compile(correctXtalkZ=correctXtalkZ,beSmart=beSmart)

            reqs.append(runQubits(server, devs, stats, dataFormat='iqRaw'))
        
        dataUnLoaded1,dataUnLoaded2 = [],[]
        dataLoaded1, dataLoaded2 = [],[]
        dataProb1,dataProb2 = [],[]
        data = yield runQubits(server, devs, stats, dataFormat='iqRaw')
        results = yield FutureList(reqs)
        # print 'results',results
        for r in results[:len(opList)]:
            Is,Qs = r[0]
            # prob = np.mean(sq.readoutProbabilities(q,np.array(Is)+1.0j*np.array(Qs)))
            probAnc1 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[0]
            
            
            probAnc2 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[3]
            
            dataUnLoaded1.append(probAnc1)
            dataUnLoaded2.append(probAnc2)
        for r in results[len(opList):]:
            Is,Qs = r[0]
            # prob = np.mean(sq.readoutProbabilities(q,np.array(Is)+1.0j*np.array(Qs)))
            probAnc1 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[0]
            probAnc2 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[3]
            probQ1 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[1]
            probQ2 = mq.mqReadoutProbabilities(r, devs, measure, correlated = False)[2]
            dataLoaded1.append(probAnc1)
            dataLoaded2.append(probAnc2)
            dataProb1.append(probQ1)
            dataProb2.append(probQ2)
            
        # [rho1, residual1] = tomo.qst_evan(np.array(dataUnLoaded1), tomo.octomo_ops, stats, 2)        
        [rho1, residual1] = tomo.qst_evan(np.array(dataUnLoaded1), tomo.octomo_ops, stats, 1)        
        # [rho2, residual2] = tomo.qst_evan(np.array(dataUnLoaded2), tomo.octomo_ops, stats, 2)        
        [rho2, residual2] = tomo.qst_evan(np.array(dataUnLoaded2), tomo.octomo_ops, stats, 1)        
        phaseUnLoaded1 = ps.getDensityPhase(rho1)[0]   
        phaseUnLoaded2 = ps.getDensityPhase(rho2)[0]   
        ampUnLoaded1 = abs(rho1[0,1])
        ampUnLoaded2 = abs(rho2[0,1])
        
        [rho1, residual1] = tomo.qst_evan(np.array(dataLoaded1), tomo.octomo_ops, stats, 1)        
        [rho2, residual2] = tomo.qst_evan(np.array(dataLoaded2), tomo.octomo_ops, stats, 1)        
        phaseLoaded1 = ps.getDensityPhase(rho1)[0]   
        phaseLoaded2 = ps.getDensityPhase(rho2)[0]   
        ampLoaded1 = abs(rho1[0,1])
        ampLoaded2 = abs(rho2[0,1])
        bellProb1, bellProb2 = np.mean(dataProb1),np.mean(dataProb2)
        deltaPhi1 = phaseUnLoaded1 - phaseLoaded1
        deltaPhi2 = phaseUnLoaded2 - phaseLoaded2
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi1),mapToBranch(deltaPhi2),bellProb1,bellProb2, ampUnLoaded1, ampLoaded1])
        # print [mapToBranch(deltaPhi1),mapToBranch(deltaPhi2), ampUnLoaded1, ampLoaded1]
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update and manual:
        amps = data[:,0]
        traces = [{'x': amps, 'y': data[:,1], 'args': ('-r',)},
                  {'x': amps, 'y': [np.pi]*len(amps), 'args': ('-k',)}]
        params = [{'name': 'Adiabatic Cphase amp', 'val': float(controlQubit[targetQubit.__name__+'aczAmp']), 
                   'range':(min(amps), max(amps)), 'axis': 'x', 'color': 'b'}]
        result = adjust.adjust(params, traces)
        if result:
            QC[targetQubit.__name__+'aczAmp'] = result['Adiabatic Cphase amp']
        return result['Adiabatic Cphase amp']
                
    elif update and not manual:
        amps = data[:,0]
        phases = np.unwrap(data[:,1])
        
        spline = scipy.interpolate.UnivariateSpline(amps, phases, s=0.5)
        bestAmp = scipy.optimize.fsolve(lambda x: spline(x)-np.pi, controlQubit[targetQubit.__name__+'aczAmp'])[0]
        QC[targetQubit.__name__+'aczAmp'] = bestAmp
        return bestAmp
        if not manual and iterations>1:
            raise Exception("Branch cut/unwrap issues make it hard to determine which pi you are aiming for")
        
        if plot:
            plt.plot(amps, phases)
            plt.plot(amps, spline(amps))
            