#Standard packages
import numpy as np
from math import atan2
import scipy
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import pyle

import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
import pyle.dataking.utilMultilevels as ml
from pyle.dataking import singleQubit as sq
from pyle.dataking import multipleQubit as mq
import pyle.plotting.dstools as ds
import pyle.plotting.plothelper as ph
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
from pyle.datavault import DataVaultWrapper
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.analysis import iq as iq
from pyle.analysis import readout as readout

from pyle.dataking.benchmarking import randomizedBenchmarking as rb

#import pyle.dataking.DevDaniel.Qubit.readout as readout

from pyle import signalprocessing as signalprocessing
import pyle.quantum as quantum
import time
import copy
from matplotlib.mlab import find
from scipy.interpolate import UnivariateSpline
import itertools
from pyle.dataking.zfuncs import *


def createBellState(alg,controlQubit, targetQubit, bellState = 'Psi+'):
    '''
    Creates a Bell State so we don't have to type this in all the time
    '''      
    if bellState=='Psi+':
        # 01 + 10
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2    
    elif bellState=='Psi-':
        # 01 - 10
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ targetQubit ],phase=np.pi)] #-X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
      
    elif bellState=='Phi+':
        # 00 + 11
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ controlQubit ],phase=np.pi)] #-X_pi/2                    
    elif bellState=='Phi-':
        # 00 - 11
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ targetQubit ], phase=np.pi)] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ controlQubit ],phase=np.pi)] #-X_pi/2
    else:
        raise Exception('not implemented')
        
def correlationTester(Sample, measure,name='Corrlation Tester',phase=st.r[0.15:np.pi:0.01],  stats=900,
                      correctXtalkZ=True, save=True, collect=False, noisy=True):            
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(phase, 'Phase')]
    deps=[]
    for state_bin_repr in range(2**len(measure)): 
    #This thing displays the correlated probabilities in the correct order
        probstr='|'+np.binary_repr(state_bin_repr,len(measure))+'>'
        deps+=[('Probability', probstr, '')]
    # deps = [('Prob |>', '', ''),('','',''),('','',''),('','','')]
    # deps = [('Prob |q0>', '', ''),('Prob |q1>', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit,phase=currPhase)
        alg[gates.Measure(alg.qubits)]
        alg.compile()
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure,correlated = True)
        # probs = mq.mqReadoutProbabilities(data, alg.qubits, measure,correlated = False)
        # print probs
        
        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)     
    
def correlationTesterArbRotation(Sample, measure,name='Corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        
        createBellState(alg,controlQubit, targetQubit)
        if equator:
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        pairRot(alg,currAmp,rotationQubit=controlQubit)
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probC = pairCorr(alg,measure,data)
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

def ancillaCorrelationTesterArbRotation(Sample, measure,name='ancilla corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
            
    '''
    Tests the degree of correlation vs. some angle rotation between bell states using the weak ancilla measurement.
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        ancilla1 = alg.q0
        controlQubit = alg.q1
        targetQubit = alg.q2
        ancilla2 = alg.q3
        
        createBellState(alg,controlQubit, targetQubit)
        pairRot(alg,currAmp,rotationQubit=targetQubit)
        alg[gates.Sync([ancilla1,controlQubit,targetQubit,ancilla2])]
        
        alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlQubit , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        alg[gates.Sync([ancilla1,controlQubit,targetQubit,ancilla2])]
        
        alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetQubit] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        
        # alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
        # alg[gates.PiHalfPulse([ targetQubit ]) ]
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probC = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])
        # probC = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

def correlationTesterHeralding(Sample, measure,name='Corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        ml.setMultiKeys(alg.q0,1)
        ml.setMultiLevels(alg.q0, 'piAmp', currAmp, 1)
        
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit,phase = np.pi)
        if equator:
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiPulse([controlQubit])]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        prob00 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[0]
        prob01 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[1]
        prob10 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[2]
        prob11 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[3]
        
        probC = (prob00+prob11-prob10-prob01)
         
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)  

    
    
def measureBellStateProb(Sample, measure,name='Bell state pop.',length=None, trial = st.r[1:1000:1],
                         stats=900, correctXtalkZ=True, update=False, save=True, collect=False, noisy=True):
    '''
    Makes a Bell state using the new qubit code. Measures the probability as a function of trial
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    
    
    axes = [(trial, 'Trial')]
    deps = [('Prob |q0>', '', ''),('Prob |q1>', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currTrial):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit)
        alg[gates.Measure(alg.qubits)]
        alg.compile()
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure)
        returnValue(probs)


    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def bellViolationArbRotation(Sample, measure,name='Bell violation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, rotation = 'control',
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True, delay = 0*ns, herald = False):            
    
    '''
    attempts to recreate the bell violation by comparing the correlator constructed using 4 distinct measurements vs. some angle rotation.
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('Bell term 1','',''),('Bell term 2','',''),('Bell term 3','',''),('Bell term 4','',''),('Bell Corelator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        
        reqs = []
        
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        
        if herald:
            alg[gates.Measure(alg.qubits, name = 'Herald')]
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            alg[gates.AmpPulse([ controlQubit ],amp = currAmp)]
        else:
            alg[gates.AmpPulse([ targetQubit ], amp = currAmp)]
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits, name = 'Measure')]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM1 = pairCorr(alg,measure,data)
        
        
        ##### Detector Settings 2
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        if herald:
            alg[gates.Measure(alg.qubits, name = 'Herald')]
            alg[gates.Sync(alg.qubits)]
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            alg[gates.AmpPulse([ controlQubit ],amp = currAmp)]
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        else:
            alg[gates.AmpPulse([ targetQubit ], amp = currAmp)]
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits, name = 'Measure')]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM2 = pairCorr(alg,measure,data)
        
        
        ###### Detector Settings 3
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        if herald:
            alg[gates.Measure(alg.qubits, name = 'Herald')]
            alg[gates.Sync(alg.qubits)]
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            alg[gates.AmpPulse([ controlQubit ],amp = currAmp)]
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            alg[gates.AmpPulse([ targetQubit ], amp = currAmp)]
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits, name = 'Measure')]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM3 = pairCorr(alg,measure,data)
        
        ##### Detector Settings 4
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        if herald:
            alg[gates.Measure(alg.qubits, name = 'Herald')]
            alg[gates.Sync(alg.qubits)]
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            alg[gates.AmpPulse([ controlQubit ],amp = currAmp)]
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            alg[gates.AmpPulse([ targetQubit ], amp = currAmp)]
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits, name = 'Measure')]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        results = yield FutureList(reqs)
        probM1 = pairCorr(alg,measure,results[0], herald = herald)
        probM2 = pairCorr(alg,measure,results[1], herald = herald)
        probM3 = pairCorr(alg,measure,results[2], herald = herald)
        probM4 = pairCorr(alg,measure,results[3], herald = herald)
        
        returnValue([probM1,probM2,probM3,probM4,(probM2+probM3+probM4-probM1)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

def bellViolationNMOptimization(Sample, measure,name='NM optimization for maximum bell violation',
                                 averages = 20, stats=900, initialRotation = 0.161, rotation = 'control',
                                 nonzdelt = 0.1, zdelt = 0.00025, xtol = 0.001, ftol = 0.001,
                                 maxfun=None, correctXtalkZ=True, save=True, collect=False, noisy=True, delay = 0*ns):            
    
    '''
    attempts to optimize the bell violation by using Nelder Mead optimization of the cz parameters
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    controlAmp = controlQubit[targetQubit.__name__+'aczAmp']
    controlPhase = controlQubit[targetQubit.__name__+'aczPhase']
    targetPhase = targetQubit[controlQubit.__name__+'aczPhase']
   
    axes = [('Function Evaltuation','')]
    deps=[('Bell Corelator','',''),('control Amp','GHz','GHz'),('control Phase','',''),('target Phase','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ, 'initialRotation':initialRotation}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, args):
        
        reqs = []
        currAmp = initialRotation
        currControlAmp, currControlPhase, currTargetPhase, currSeq = args
        
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM1 = pairCorr(alg,measure,data)
        
        
        ##### Detector Settings 2
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM2 = pairCorr(alg,measure,data)
        
        
        ###### Detector Settings 3
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probM3 = pairCorr(alg,measure,data)
        
        ##### Detector Settings 4
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))    
        # data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        results = yield FutureList(reqs)
        probM1 = pairCorr(alg,measure,results[0])
        probM2 = pairCorr(alg,measure,results[1])
        probM3 = pairCorr(alg,measure,results[2])
        probM4 = pairCorr(alg,measure,results[3])
        
        returnValue([-(probM2+probM3+probM4-probM1)])
    def funcWrapper(args):
        args =list(args)
        currControlAmp = args[0]
        currControlPhase = args[1]
        currTargetPhase = args[2]
        
        def seqGen():
            for currSeq in range(averages):
                yield currControlAmp, currControlPhase, currTargetPhase, currSeq
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample,'dummy dataset', axes, deps, measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save = False, noisy = noisy)
        avgViolation = np.mean(data)
        return avgViolation
    
    inputs = []
    inputs.append(controlAmp)
    inputs.append(controlPhase)
    inputs.append(targetPhase)
    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt, maxfun=maxfun)
    
    return output



def LGBIalg(alg, qubits, currAng, initialRotation):

    arbRot0 = alg.q0['arbRotateCal'].value
    arbRot1 = alg.q1['arbRotateCal'].value
    arbRot2 = alg.q2['arbRotateCal'].value
    arbRot3 = alg.q3['arbRotateCal'].value

    ancilla1,controlBell,targetBell,ancilla2 = qubits
    createBellState(alg, controlBell, targetBell)
    alg[gates.AmpPulse([ targetBell ], amp = initialRotation)]
    
    alg[gates.PiPulse([ controlBell ])]
    alg[gates.PiPulse([ targetBell ], phase = np.pi)]
    
    alg[gates.AmpPulse([ ancilla1 ], amp = (currAng/(arbRot0)))]
    alg[gates.AmpPulse([ ancilla2 ], amp = (currAng/(arbRot3)))]
    alg[gates.adiabaticCZ([controlBell , ancilla1], detuneOthers = False)]
    alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
    alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
    alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]

    alg[gates.PiHalfPulse([ controlBell ], phase = np.pi)]
    alg[gates.PiHalfPulse([ targetBell ])]
    
def LGBIViolation(Sample, measure, name='LGBI violation with arbitrary rotaions',
                                 angles=st.r[(np.pi/2):0:0.01], initialRotation = 0.155,  stats=900, recal = False,
                                 correctXtalkZ=True, save=True, collect=False, averages = None,
                                 noisy=True):            
    
    '''
    Full LGBI violation algorithm including weak measurement.
    '''
    
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if recal:
        for device in measure:
            calArbRotation(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    arbRot0 = sample['q0']['arbRotateCal'].value
    arbRot1 = sample['q1']['arbRotateCal'].value
    arbRot2 = sample['q2']['arbRotateCal'].value
    arbRot3 = sample['q3']['arbRotateCal'].value
    if averages:
         angle_gen = st.averagedScan(angles, averages, noisy = noisy)
         axes = [(angle_gen(), 'Ancilla rotation')]
    else:
        axes = [(angles, 'Ancilla rotation')]
    deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ, 'averages':averages}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAng):
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        # targetBell = alg.q2
        # controlBell = alg.q1
        # ancilla1 = alg.q0
        # ancilla2 = alg.q3
        
        LGBIalg(alg,alg.qubits,currAng,initialRotation)
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        # ph.pulses(alg)
        # raise Exception("blah")
        # ph.pulses(alg)
        # raise exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        # probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 + a1b2 - b1a2 - b1b2)]
        # def correlator(alg,measure,data):
            # correction = 1.0/np.sin(currAng)
            # probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = False)
            # a1 = (probs[0]*-2+1)*correction
            # a2 = (probs[3]*-2+1)*correction
            # b1 = (probs[1]*-2+1)
            # b2 = (probs[2]*-2+1)
            # return [a1*a2,a1*b2,b1*a2,b1*b2,(-a1*a2 - a1*b2 + b1*a2 - b1*b2)]
            
        returnValue(correlator(alg,measure,data))
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)    

def LGBIViolationHerald(Sample, measure, name='LGBI violation with arbitrary rotaions and heralding',
                                 angles=st.r[(np.pi/2):0:0.01],  stats=200, recal = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
    
    '''
    Full LGBI violation algorithm including weak measurement and heralding.
    '''
    
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if recal:
        for device in measure:
            calArbRotation(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    arbRot0 = sample['q0']['arbRotateCal'].value
    arbRot1 = sample['q1']['arbRotateCal'].value
    arbRot2 = sample['q2']['arbRotateCal'].value
    arbRot3 = sample['q3']['arbRotateCal'].value
   
    axes = [(angles, 'Ancilla rotation')]
    deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAng):
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3
        piAmp = targetBell['piAmp']
        
        #Herald readout
        alg[gates.Measure(alg.qubits, sync='start')]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]

        #initialize Bell State and rotate
        createBellState(alg, controlBell, targetBell)
        pairRot(alg,0.155,targetBell)
        
        # weak measurement with ancilla 1
        pairRot(alg,(currAng/(arbRot0)),ancilla1)
        # alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlBell , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        alg[gates.PiPulse([ controlBell ])]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        alg[gates.Wait([ controlBell ], 100*ns)]
        
        # weak measurement with ancilla 2
        pairRot(alg,(currAng/(arbRot3)),ancilla2)
        # alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # final rotations
        alg[gates.PiPulse([ controlBell ])]
        alg[gates.PiHalfPulse([ controlBell ])]
        alg[gates.PiHalfPulse([ targetBell ], phase = np.pi)]
        
        alg[gates.Measure(alg.qubits, sync='start')]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        # ph.pulses(alg)
        # raise Exception("blah")
        runList = []
        if stats> 100:
            runs = stats/100
            finalstats = stats%100
            for run in range(runs):
                print run
                rv = yield runQubits(server, alg.agents, 100, dataFormat='iqRaw', measure = measure, localDemod = True)
                runList.append(rv)
            if finalstats:
                rv = yield runQubits(server, alg.agents, finalstats, dataFormat='iqRaw', measure = measure, localDemod = True)
                runList.append(rv)
            runData = iq.concatenate_stats(runList)
        else:
            runData = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure, localDemod = True)
        data = sq.do_post_select(runData, alg)
        # probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 - a1b2 + b1a2 - b1b2)]
        
        returnValue(correlator(alg,measure,data))
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize = 1)    
    return data
    
def LGBIReadoutOptimizer(Sample, measure, name='LGBI as a function of readout len',
                                 readoutLen=st.r[400*ns:800*ns:20*ns],  stats=900, recal = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):
    """
    Runs LGBI experiment at a single ancilla coupling strength to measure terms as a function of readout time
    """
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    # name = 'LGBI as a function of readout len, angle: ', ang
    if recal:
        for device in measure:
            calArbRotation(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    
    #This arbRot thing is hacked in, can we please fix this?
    arbRot0 = sample['q0']['arbRotateCal']
    arbRot1 = sample['q1']['arbRotateCal']
    arbRot2 = sample['q2']['arbRotateCal']
    arbRot3 = sample['q3']['arbRotateCal']
    
   
    axes = [(readoutLen, 'Readout Length')]
    deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currLen):
        currAng = 0.01
        # Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        allQubits = [alg.q0,alg.q1,alg.q2,alg.q3]
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3
        
        #Changes Readout length of all Qubits
        for qubit in allQubits:
            ml.setMultiKeys(qubit,1)
            ml.setMultiLevels(qubit, 'readoutLen', currLen, 1)
        #initialize Bell State and rotate
        # LGBIalg(alg,controlBell,targetBell,ancilla1,ancilla2,currAng)
        
        createBellState(alg, controlBell, targetBell)
        pairRot(alg,0.155,targetBell)
        
        # weak measurement with ancilla 1
        pairRot(alg,(currAng/(2*arbRot0)),ancilla1)
        # alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlBell , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        # correct phase for ancilla 2 from first measurement
        alg[gates.ChangePhase([ targetBell ], xyPhase = targetBell['q1q0aczPhase'])]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # weak measurement with ancilla 2
        pairRot(alg,(currAng/(2*arbRot3)),ancilla2)
        # alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # final rotations
        alg[gates.PiHalfPulse([ controlBell ])]
        alg[gates.PiHalfPulse([ targetBell ], phase = np.pi)]
        
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        # ph.pulses(alg)
        # raise exception("blah")
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 - a1b2 + b1a2 - b1b2)]

        returnValue(correlator(alg,measure,data))
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)    
 
def LGBIRotationOptimizer(Sample, measure, name='LGBI as a function of initial bell rotation',
                                 amps = st.r[0:1:0.01],  stats=900, recal = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):
    """
    Runs LGBI experiment at a single ancilla coupling strength to measure terms as a function of the initial bell pair rotation/correlation angle.
    """
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    # name = 'LGBI as a function of readout len, angle: ', ang
    if recal:
        for device in measure:
            calArbRotation(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    
    #This arbRot thing is hacked in, can we please fix this?
    arbRot0 = sample['q0']['arbRotateCal']
    arbRot1 = sample['q1']['arbRotateCal']
    arbRot2 = sample['q2']['arbRotateCal']
    arbRot3 = sample['q3']['arbRotateCal']
    
   
    axes = [(amps, 'rotation amplitude')]
    # deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    deps=[('bell1-bell2 term','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        currAng = 0.5
        # Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        allQubits = [alg.q0,alg.q1,alg.q2,alg.q3]
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3

        #initialize Bell State and rotate
        # LGBIalg(alg,controlBell,targetBell,ancilla1,ancilla2,currAng)
        
        createBellState(alg, controlBell, targetBell)
        pairRot(alg,currAmp,targetBell)
        
        # weak measurement with ancilla 1
        pairRot(alg,(currAng/(2*arbRot0)),ancilla1)
        # alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlBell , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        # correct phase for ancilla 2 from first measurement
        alg[gates.ChangePhase([ targetBell ], xyPhase = targetBell['q1q0aczPhase'])]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # weak measurement with ancilla 2
        pairRot(alg,(currAng/(2*arbRot3)),ancilla2)
        # alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # final rotations
        alg[gates.PiHalfPulse([ controlBell ])]
        alg[gates.PiHalfPulse([ targetBell ], phase = np.pi)]
        
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        # ph.pulses(alg)
        # raise exception("blah")
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 - a1b2 + b1a2 - b1b2)]

        # returnValue(correlator(alg,measure,data))
        returnValue([correlator(alg,measure,data)[3]])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)    
 
def LGBIWithQST(Sample, measure, name='LGBI with bell state QST',  stats=900, recal = False,
                                 averages = 1, correctXtalkZ=True, save=True, collect=False,
                                 bellState='Psi+', waitTime=0.0*ns, noisy=True):
    """
    Runs LGBI experiment while taking qst data on the initial bell state at various points to diagnose the potential problems with the cz's
    """
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    # name = 'LGBI as a function of readout len, angle: ', ang
    if recal:
        for device in measure:
            calArbRotation(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    arbRot0 = sample['q0']['arbRotateCal']
    arbRot1 = sample['q1']['arbRotateCal']
    arbRot2 = sample['q2']['arbRotateCal']
    arbRot3 = sample['q3']['arbRotateCal']
    
   
    ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, len(qubits[1:3])))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps = [('Probability', '|00>', ''),('Probability', '|01>', ''),('Probability', '|10>', ''),('Probability', '|11>', '')]
    
    kw = {'stats': stats, 'bellState':bellState, 'waitTime':waitTime ,'correctXtalkZ':correctXtalkZ}    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    def func(server, i):
        
        postOp = ops[i]
        
        if noisy:
            print postOp
            
        alg = gc.Algorithm(agents = devs)
        allQubits = [alg.q0,alg.q1,alg.q2,alg.q3]
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3
        
        createBellState(alg, controlBell, targetBell, bellState = bellState)
        # pairRot(alg,0.163,targetBell)
        
        # weak measurement with ancilla 1
        # pairRot(alg,(currAng/(2*arbRot0)),ancilla1)
        # alg[gates.PiHalfPulse([ ancilla1 ])]
        # alg[gates.adiabaticCZ([controlBell , ancilla1] )]
        # alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        # alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # weak measurement with ancilla 2
        # pairRot(alg,(currAng/(2*arbRot3)),ancilla2)
        # alg[gates.PiHalfPulse([ ancilla2 ])]
        # alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
        # alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        # alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # final
        # alg[gates.PiHalfPulse([ controlBell ])]
        # alg[gates.PiHalfPulse([ targetBell ], phase = np.pi)]
        alg[gates.PiPulse([ controlBell ])]
        alg[gates.PiPulse([ targetBell ], phase = np.pi)]
        
        alg[gates.Tomography([controlBell,targetBell], postOp)]
        alg[gates.Sync([controlBell,targetBell])]
        # alg[gates.Tomography([ancilla1,ancilla2], postOp)]
        # alg[gates.Sync([ancilla1,ancilla2])]
        alg[gates.Wait(alg.qubits, 100*ns)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise Exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data[1:3], alg.qubits[1:3], measure[1:3], correlated = True)
        # probs = mq.mqReadoutProbabilities(data[::3], alg.qubits[::3], measure[::3], correlated = True)
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    rho = tomo.qst(data[:,1:], 'octomo2')    
    ps.plotDensity(rho)
    
    if collect:
        return data 

def LGBIViolationNMOptimization(Sample, measure,name='NM optimization for maximum LGBI violation',
                                 averages = 30, stats=900, initialRotation = 0.155, measAngle =0.2,
                                 nonzdelt = 0.1, zdelt = 0.00025, xtol = 0.001, ftol = 0.001,
                                 maxfun=None, correctXtalkZ=True, save=True, collect=False, noisy=True, delay = 0*ns):            
    
    '''
    attempts to optimize the LGBI violation by using Nelder Mead optimization of the cz parameters
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    arbRot0 = sample['q0']['arbRotateCal'].value
    arbRot1 = sample['q1']['arbRotateCal'].value
    arbRot2 = sample['q2']['arbRotateCal'].value
    arbRot3 = sample['q3']['arbRotateCal'].value
    
    q0q1Phase = qubits[0]['q1aczPhase']
    q1q0Phase = qubits[1]['q0aczPhase']
    q1q2Phase = qubits[1]['q2aczPhase']
    # q1q0q2phase = qubits[2]['q1q0aczDetunePhase']
    q2q1phase = qubits[2]['q1aczPhase']
    q2q3phase = qubits[2]['q3aczPhase']
    q3q2phase = qubits[3]['q2aczPhase']
    #ACZ amplitudes
    q0aczAmp = qubits[1]['q0aczAmp']
    q1q2aczAmp = qubits[1]['q2aczAmp']
    q3q2aczAmp = qubits[3]['q2aczAmp']
   
    axes = [('Function Evaltuation','')]
    deps=[('LGBI Corelator','',''),('q0q1Phase','',''),('q1q0Phase','',''),('q1q2Phase','',''),#('q1q0q2phase','',''),
          ('q2q1phase','',''),('q2q3phase','',''),('q3q2phase','',''),('q0aczAmp','',''),('q1q2aczAmp','',''),('q3q2aczAmp','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ, 'initialRotation':initialRotation}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, args):
    
        currAng = measAngle
        alg = gc.Algorithm(agents=devs)
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3
        
        ancilla1['q1aczPhase'] = args[0]
        controlBell['q0aczPhase'] = args[1]
        controlBell['q2aczPhase'] = args[2]
        # targetBell['q1q0aczDetunePhase'] = args[3]
        targetBell['q1aczPhase'] = args[3]
        targetBell['q3aczPhase'] = args[4]
        ancilla2['q2aczPhase'] = args[5]
        controlBell['q0aczAmp'] = args[6]
        controlBell['q2aczAmp'] = args[7]
        ancilla2['q2aczAmp'] = args[8]
        
        LGBIalg(alg,alg.qubits,currAng,initialRotation)
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        # ph.pulses(alg)
        # raise Exception("blah")

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 + a1b2 - b1a2 - b1b2)]

        returnValue([-correlator(alg,measure,data)[4]])
        
    def funcWrapper(args):
        args =list(args)
        # currq0q1Phase = args[0]
        # currq1q0Phase = args[1]
        # currq1q2Phase = args[2]
        # currq1q0q2phase = args[3]
        # currq2q1phase = args[3]
        # currq2q3phase = args[4]
        # currq3q2phase = args[5]
        
        def seqGen():
            for currSeq in range(averages):
                #yield currq0q1Phase, currq1q0Phase, currq1q2Phase, currq2q1phase, currq2q3phase, currq3q2phase, currSeq
                yield args + [currSeq]
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample,'dummy dataset', axes, deps, measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save = False, noisy = noisy)
        avgViolation = np.mean(data)
        std = np.std(data)
        print "Standard deviation: %s" % std
        print "Average Value: %s" % avgViolation
        return avgViolation, std
    
    inputs = []
    inputs.append(q0q1Phase)
    inputs.append(q1q0Phase)
    inputs.append(q1q2Phase)
    # inputs.append(q1q0q2phase)
    inputs.append(q2q1phase)
    inputs.append(q2q3phase)
    inputs.append(q3q2phase)
    inputs.append(q0aczAmp)
    inputs.append(q1q2aczAmp)
    inputs.append(q3q2aczAmp)

    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt, maxfun=maxfun)
    
    return output

def updateLGBIParams(Sample,measure,dataSet,iteration = 'min'):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    data = dvw[dataSet]
    if iteration == None:
        optimalParams = list(data[-1,2:])
    elif iteration == 'min':
        bestIdx = np.argmin(data[:,1])
        optimalParams = list(data[bestIdx,2:])
        print 'best IDX is: %s' % (bestIdx+1)
    else:
        optimalParams = list(data[iteration-1,2:])
    
    Qubits[0]['q1aczPhase'] = optimalParams[0]
    Qubits[1]['q0aczPhase'] = optimalParams[1]
    Qubits[1]['q2aczPhase'] = optimalParams[2]
    # Qubits[2]['q1q0aczDetunePhase'] = optimalParams[3]
    Qubits[2]['q1aczPhase'] = optimalParams[3]
    Qubits[2]['q3aczPhase'] = optimalParams[4]
    Qubits[3]['q2aczPhase'] = optimalParams[5]
    #ACZ amplitudes
    Qubits[1]['q0aczAmp'] = optimalParams[6]
    Qubits[1]['q2aczAmp'] = optimalParams[7]
    Qubits[3]['q2aczAmp'] = optimalParams[8]


    
def weakMeasureTest(Sample, measure, name='weak measurement test with ancilla',
                                 angles=st.r[(np.pi/2):0:0.01],  stats=900, recal = False, pulse = None,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
    
    '''
    Tests the weak measurement protocol by comparing a pi pulse on the measured qubit with no pi pulse
    on the measured qubit vs. initial rotation of the measuring qubit.
    '''
    
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if recal:
        for device in meausre:
            calArbPhase(Sample, measure[0],2*devs[measure[0]]['piAmp'].value, dvw)
    
    arbRot0 = rot2AmpFunc(dvw,sample[measure[0]])[1]
    arbRot1 = rot2AmpFunc(dvw,sample[measure[1]])[1]
    
   
    axes = [(angles, 'Ancilla rotation')]
    deps=[('probaility no pi pulse ','',''),('probability pi pulse','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAng):
        ##### Detector Settings 1
        
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        #initialize Bell State
        if pulse == 'target':
            pairRot(alg,(currAng/(2*arbRot1)),targetQubit)
            # alg[gates.PiHalfPulse([ targetQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ targetQubit ], phase = np.pi)]
        else:
            pairRot(alg,currAng/(2*arbRot0),controlQubit)
            # alg[gates.PiHalfPulse([ controlQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
            
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)

        dataUnloaded = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probsUL = mq.mqReadoutProbabilities(dataUnloaded, alg.qubits, measure, correlated = False)
        
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        #initialize Bell State
        if pulse == 'target':
            alg[gates.PiPulse([ controlQubit ])]
            pairRot(alg,currAng/(2*arbRot1),targetQubit)
            # alg[gates.PiHalfPulse([ targetQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ targetQubit ], phase = np.pi)]
        else:
            alg[gates.PiPulse([ targetQubit ])]
            pairRot(alg,currAng/(2*arbRot0),controlQubit)
            # alg[gates.PiHalfPulse([ controlQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
            
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)

        dataLoaded = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probsL = mq.mqReadoutProbabilities(dataLoaded, alg.qubits, measure, correlated = False)
        
        if pulse == 'target':
            pUL,PL = probsUL[1],probsL[1]
        else:
            pUL,PL = probsUL[0],probsL[0]
            
        returnValue([pUL,PL])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
        
def aczPRamsey(Sample, measure, name='Adiabatic CZ Parasitic Ramsey ',
             thetas=25, waitTime = None,
             iterations=1,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True,ampfactor=1.0,extraWait=0.0):
    '''
    measure = [control, target, parasitic]
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
       
    if update and not qubits[2].get(qubits[0].__name__+qubits[1].__name__+'aczDetune'):
        raise Exception("You should not have to correct phase for a qubit that is not moving -> Z crosstalk correction will fix this")
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)

    if waitTime is None:
        waitTime = qubits[0]['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Control no Pi',''),('Probability |1>','Control Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        parasiticQubit = alg.q2
        alg[gates.PiHalfPulse([parasiticQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.PiHalfPulse([parasiticQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
  
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        parasiticQubit = alg.q2
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiHalfPulse([parasiticQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]    
        alg[gates.PiPulse([controlQubit])]
        alg[gates.Sync(alg.qubits)]        
        alg[gates.PiHalfPulse([parasiticQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        Is0, Qs0 = data0[2]
        probNoPi = np.mean(sq.readoutProbabilities(parasiticQubit,np.array(Is0)+1.0j*np.array(Qs0)))
        
        data1 = results[1]
        Is1, Qs1 = data1[2]
        probPi = np.mean(sq.readoutProbabilities(parasiticQubit,np.array(Is1)+1.0j*np.array(Qs1)))
 
        returnValue([probNoPi, probPi])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[2][qubits[0].__name__+qubits[1].__name__+'aczDetunePhase'] -= deltaPhi/iterations
    
    if collect:
        return data
        
def aczPRamseyArb(Sample, measure, ramseyQ, piQ, name='Adiabatic CZ Parasitic Ramsey Arb',
             thetas=25, waitTime = None,
             iterations=1,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True,ampfactor=1.0,extraWait=0.0):
    '''
    measure = [control, target, parasitic]
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
       
    if update and not qubits[2].get(qubits[0].__name__+qubits[1].__name__+'aczDetune'):
        raise Exception("You should not have to correct phase for a qubit that is not moving -> Z crosstalk correction will fix this")
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)

    if waitTime is None:
        waitTime = qubits[0]['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Control no Pi',''),('Probability |1>','Control Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        ramseyQubit = alg.qubits[measure.index(ramseyQ)]
        piQubit = alg.qubits[measure.index(piQ)]
        alg[gates.PiHalfPulse([ramseyQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.PiHalfPulse([ramseyQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
  
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        ramseyQubit = alg.qubits[measure.index(ramseyQ)]
        piQubit = alg.qubits[measure.index(piQ)]
        alg[gates.PiPulse([piQubit])]
        alg[gates.PiHalfPulse([ramseyQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]    
        alg[gates.Sync(alg.qubits)]        
        alg[gates.PiHalfPulse([ramseyQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        Is0, Qs0 = data0[measure.index(ramseyQ)]
        probNoPi = np.mean(sq.readoutProbabilities(ramseyQubit,np.array(Is0)+1.0j*np.array(Qs0)))
        
        data1 = results[1]
        Is1, Qs1 = data1[measure.index(ramseyQ)]
        probPi = np.mean(sq.readoutProbabilities(ramseyQubit,np.array(Is1)+1.0j*np.array(Qs1)))
 
        returnValue([probNoPi, probPi])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[2][qubits[0].__name__+qubits[1].__name__+'aczDetunePhase'] -= deltaPhi/iterations
    
    if collect:
        return data

def parasiticQST(Sample, measure, name='qst on parasitic qubit during acz pulse',  stats=900,
                                 averages = 1, correctXtalkZ=True, save=True, collect=False,
                                 bellState='Psi+', waitTime=0.0*ns, noisy=True, ampfactor=1.0,extraWait=0.0):
    """
    Runs QST on two qubits which should not be entangled at all to asses their degree of entanglement
    """
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

   
    ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, len(qubits[1:3])))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps = [('Probability', '|00>', ''),('Probability', '|01>', ''),('Probability', '|10>', ''),('Probability', '|11>', '')]
    
    kw = {'stats': stats, 'bellState':bellState, 'waitTime':waitTime ,'correctXtalkZ':correctXtalkZ}    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    def func(server, i):
        
        postOp = ops[i]
        
        if noisy:
            print postOp
            
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        parasiticQubit = alg.q2
        alg[gates.PiPulse([controlQubit])]
        # alg[gates.PiHalfPulse([parasiticQubit])]
        alg[gates.Sync(alg.qubits)]
        alg[gates.adiabaticCZ([controlQubit, targetQubit])]
        alg[gates.Wait([controlQubit], extraWait)]    
        # alg[gates.PiPulse([controlQubit])]
        alg[gates.Sync(alg.qubits)]        
        # alg[gates.PiHalfPulse([parasiticQubit], phase = 0)]
        alg[gates.Tomography([controlQubit,parasiticQubit], postOp)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise Exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data[::2], alg.qubits[::2], measure[::2], correlated = True)
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    rho = tomo.qst(data[:,1:], 'octomo2')    
    ps.plotDensity(rho)
    
    if collect:
        return data

def aczPhaseMatrix(Sample, measure, aczpair, name='ACZ Phase Matrix',
             iterations = 1, waitTime = 0.0, stats=900,
             correctXtalkZ=True, save=True, collect=False, noisy=True):
             
    """
    Runs the adiabatic cphase pulse and looks at the phase accumulated on all qubits in measure.  This includes qubits not involved in the acz phase detuning pulse at all
    """
    
    name += 'from  ' + aczpair[0] + aczpair[1] + 'aczPulse'
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    
    axes = [(range(len(measure)), 'measure index'), (iterations, 'number of iterations')]
    deps = [('Accumulated Phase Control Unloaded','',''),('Accumulated Phase Control Loaded','','')]

    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    def func(server, currInd, it):
        
        reqs = []
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.qubits[measure.index(aczpair[0])]
            targetQubit = alg.qubits[measure.index(aczpair[1])]
            phaseQubit = alg.qubits[currInd]
            alg[gates.PiHalfPulse([phaseQubit])]
            alg[gates.Sync(alg.qubits)]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit])]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Sync(alg.qubits)]
            alg[gates.Tomography([phaseQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.qubits[measure.index(aczpair[0])]
            targetQubit = alg.qubits[measure.index(aczpair[1])]
            phaseQubit = alg.qubits[currInd]
            alg[gates.PiPulse([controlQubit])]
            alg[gates.PiHalfPulse([phaseQubit])]
            alg[gates.Sync(alg.qubits)]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit])]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Sync(alg.qubits)]
            alg[gates.Tomography([phaseQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        phaseDataUnloaded = []
        phaseDataLoaded = []
        for r in results[:len(opList)]:
            I,Q = r[currInd]
            prob = np.mean(sq.readoutProbabilities(alg.qubits[currInd],np.array(I)+1.0j*np.array(Q)))
            phaseDataUnloaded.append(prob)
        for r in results[len(opList):]:
            I,Q = r[currInd]
            prob = np.mean(sq.readoutProbabilities(alg.qubits[currInd],np.array(I)+1.0j*np.array(Q)))
            phaseDataLoaded.append(prob)

        [rho, residual] = tomo.qst_evan(np.array(phaseDataUnloaded), tomo.octomo_ops, stats, 1)        
        accumulatedPhase = ps.getDensityPhase(rho)[0]
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(accumulatedPhase)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    return data


    
def pairCorr(alg,measure,data, herald = False, *kwargs):
    '''
    measures the correlation among pairs so we don't have to 
    write it over and over again

    returns correlated probability
    '''
    prob = np.squeeze(readout.IQToProbs(data.asarray, alg.qubits, measure, herald = herald, correlated = True, noisy = True))
    return (prob[0]+prob[3]-prob[2]-prob[1])

def arbPairCorr(alg,measure,data,terms,*kwargs):
    '''
    measures the correlation among an arbitrary set of pairs so we don't have to 
    write it over and over again

    returns correlated probability
    '''
    prob = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
    corr = 0
    for value in terms[0]:
        corr += prob[value]
    for value in terms[1]:
        corr -= prob[value]
    return (corr)
        
def calArbRotation(Sample, measure, maxAmp, dvw):         
    sq.rabihigh(Sample,measure,stats = 3000, amps = np.linspace(0,maxAmp,50), IQ = False, update = False)
    dvw = dvw.cd(Sample._dir)
    xdata = dvw[-1][:,0]
    ydata = dvw[-1][:,1]
    angleToAmp = rot2AmpFunc(xdata,ydata)[1]
    Sample[measure]['arbRotateCal'] = (2*angleToAmp)
            
def sinFitFunc(x,a,b,c):
    return a*(np.sin(b*x))**2 + c
      
def rot2AmpFunc(xdata, ydata):
    fitParams, errors = curve_fit(sinFitFunc, xdata, ydata, p0=[0.7,np.pi,0.1])
    return fitParams
    
def aczAmpTuner(Sample, measure, name='ACZ Phase Tuner',
             length=None, amplitude = st.r[0.5:0.8:0.01, GHz],
             iterations = 1, waitTime = 0.0, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=True,
             save=True, collect=False, noisy=True):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.97, origAmplitude*1.03, 21)
    
    axes = [(amplitude, 'Adiabatic Cphase Amplitude'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('|1> state probability unloaded', '', ''),('|1> state probability loaded', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    def func(server, currAmp, currLen, it,waitTime):
        
        reqs = []
        
        # unloaded control
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.PiHalfPulse([targetQubit])]
        
        for k in range(it):
            alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
            if it>1 and k<it-1:
                alg[gates.Wait([controlQubit], waitTime)]
        
        alg[gates.PiHalfPulse([targetQubit], phase = np.pi)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
    
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        # loaded control
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiHalfPulse([targetQubit])]
        
        for k in range(it):
            alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
            if it>1 and k<it-1:
                alg[gates.Wait([controlQubit], waitTime)]
                
        alg[gates.PiHalfPulse([targetQubit],phase = np.pi)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)

        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        Is0, Qs0 = data0[1]
        probNoPi = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is0)+1.0j*np.array(Qs0)))
        
        data1 = results[1]
        Is1, Qs1 = data1[1]
        probPi = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is1)+1.0j*np.array(Qs1)))

        returnValue([probNoPi, probPi])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update and manual:
        amps = data[:,0]
        traces = [{'x': amps, 'y': data[:,1], 'args': ('-r',)},
                  {'x': amps, 'y': [np.pi]*len(amps), 'args': ('-k',)}]
        params = [{'name': 'Adiabatic Cphase amp', 'val': float(controlQubit[targetQubit.__name__+'aczAmp']), 
                   'range':(min(amps), max(amps)), 'axis': 'x', 'color': 'b'}]
        result = adjust.adjust(params, traces)
        if result:
            Qubits[0][targetQubit.__name__+'aczAmp'] = result['Adiabatic Cphase amp']*GHz
        return result['Adiabatic Cphase amp']
                
    elif update and not manual:
        amps = data[:,0]
        phases = np.unwrap(data[:,1])
        
        spline = scipy.interpolate.UnivariateSpline(amps, phases, s=0.5)
        
        bestAmp = scipy.optimize.fsolve(lambda x: spline(x)-np.pi, controlQubit[targetQubit.__name__+'aczAmp'])[0]
        Qubits[0][targetQubit.__name__+'aczAmp'] = bestAmp*GHz
        
        if not manual and iterations>1:
            raise Exception("Branch cut/unwrap issues make it hard to determine which pi you are aiming for")
        
        if plot:
            plt.figure()
            plt.plot(amps, phases)
            plt.plot(amps, spline(amps))

def switchHeraldOnOff(Sample, measure, herald = 'off'):
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    if herald == 'on':
        for meas in measure:
            Qubits[measure.index(meas)]['ringdownLen'] = 3500*ns
    if herald == 'off':      
        for meas in measure:
            Qubits[measure.index(meas)]['ringdownLen'] = 600*ns
    
def heraldingCorrelationTester(Sample, measure,name='Improved correlation with heralding',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.Measure(alg.qubits, sync='start')]
        alg[gates.Sync(alg.qubits)]
        createBellState(alg,controlQubit, targetQubit)
        if equator:
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        pairRot(alg,currAmp,rotationQubit=controlQubit)
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        runList = []
        if stats> 100:
            runs = stats/100
            finalstats = stats%100
            for run in range(runs):
                print run
                rv = yield runQubits(server, alg.agents, 100, dataFormat='iqRaw', measure = measure, localDemod = True)
                runList.append(rv)
            if finalstats:
                rv = yield runQubits(server, alg.agents, finalstats, dataFormat='iqRaw', measure = measure, localDemod = True)
                runList.append(rv)
            runData = iq.concatenate_stats(runList)
        else:
            runData = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure, localDemod = True)
        data = sq.do_post_select(runData, alg)
        
        probC = pairCorr(alg,measure,data)
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False, pipesize = 1)    
    
def heraldingStatePrepTest(Sample, measure, stats = 300, correctXtalkZ = True):

    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    probs0 = []
    probs1 = []
    
    alg = gc.Algorithm(agents=devs)
    alg[gates.Measure(alg.qubits, sync='end')]
    alg[gates.Sync(alg.qubits)]
    alg[gates.Measure(alg.qubits, sync='start')]
    alg.compile(correctXtalkZ=correctXtalkZ)
    
    runList = []
    if stats> 100:
        runs = stats/100
        finalstats = stats%100
        for run in range(runs):
            print run
            rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, 100, dataFormat='iqRaw', measure = measure, localDemod = True)
            runList.append(rv.wait())
        if finalstats:
            rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, finalstats, dataFormat='iqRaw', measure = measure, localDemod = True)
            runList.append(rv.wait())
        runData = iq.concatenate_stats(runList)
    else:
        runData = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats, dataFormat='iqRaw', measure = measure, localDemod = True)
        runData = runData.wait()
    data = sq.do_post_select(runData, alg)
    for i in range(len(measure)):
        Is, Qs = data[i]
        probs0.append(np.mean(sq.readoutProbabilities(alg.qubits[i],np.array(Is)+1.0j*np.array(Qs))))
    
    alg = gc.Algorithm(agents=devs)
    alg[gates.Measure(alg.qubits, sync='end')]
    alg[gates.Sync(alg.qubits)]
    alg[gates.PiPulse([alg.q0])]
    alg[gates.PiPulse([alg.q1])]
    alg[gates.PiPulse([alg.q2])]
    alg[gates.PiPulse([alg.q3])]
    alg[gates.Sync(alg.qubits)]
    alg[gates.Measure(alg.qubits, sync='start')]
    alg.compile(correctXtalkZ=correctXtalkZ)
    
    runList = []
    if stats> 100:
        runs = stats/100
        finalstats = stats%100
        for run in range(runs):
            print run
            rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, 100, dataFormat='iqRaw', measure = measure, localDemod = True)
            runList.append(rv.wait())
        if finalstats:
            rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, finalstats, dataFormat='iqRaw', measure = measure, localDemod = True)
            runList.append(rv.wait())
        runData = iq.concatenate_stats(runList)
    else:
        runData = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats, dataFormat='iqRaw', measure = measure, localDemod = True)
        runData = runData.wait()
    data = sq.do_post_select(runData, alg)
    for i in range(len(measure)):
        Is, Qs = data[i]
        probs1.append(np.mean(sq.readoutProbabilities(alg.qubits[i],np.array(Is)+1.0j*np.array(Qs))))
    
    return probs0,probs1

def sameReadout(Sample, measure):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    readoutParams = []
    for q,Q in zip(qubits,Qubits):
        len = q['readoutLen']
        ringdown = q['ringdownLen']
        ringdownIntegration = q['readoutIntegrateIntoRingdown']
        currParams = [len, ringdown, ringdownIntegration]
        readoutParams.append(currParams)
        Q['readoutBak'] = currParams
    
    maxLen, maxRingdown, maxRingdownIntegration = np.max(np.array(readoutParams),axis=0)
    for Q in Qubits:
        Q['readoutLen'] = maxLen*ns
        Q['ringdownLen'] = maxRingdown*ns
        Q['readoutIntegrateIntoRingdown'] = maxRingdownIntegration*ns
    
    for meas in measure:
        sq.calculateReadoutAxis(Sample, meas)
        
def restoreReadout(Sample, measure):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits,Qubits):
        len, ringdown, ringdownIntegration = q['readoutBak']
        Q['readoutLen'] = len
        Q['ringdownLen'] = ringdown
        Q['readoutIntegrateIntoRingdown'] = ringdownIntegration
        
    for meas in measure:
        sq.calculateReadoutAxis(Sample, meas)
        
def setKey(Sample, measure = None, key = None, value = None):

    if measure=='all':
        measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']
    if measure=='data':
        measure = ['q0', 'q2', 'q4', 'q6', 'q8']
    if measure=='measure':
        measure = ['q1','q3','q5', 'q7']
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for Q in Qubits:
        Q[key] = value
        
def killKey(Sample, key, measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for Q in Qubits:
        try: 
            del Q[key] 
        except:
            print 'missing key'
            
    '''
    n.killKey(s,'q0aczAmp')
    n.killKey(s,'q1aczAmp')
    n.killKey(s,'q2aczAmp')
    n.killKey(s,'q3aczAmp')
    n.killKey(s,'q4aczAmp')

    n.killKey(s,'q0aczTime')
    n.killKey(s,'q1aczTime')
    n.killKey(s,'q2aczTime')
    n.killKey(s,'q3aczTime')
    n.killKey(s,'q4aczTime')

    n.killKey(s,'q0aczPhase')
    n.killKey(s,'q1aczPhase')
    n.killKey(s,'q2aczPhase')
    n.killKey(s,'q3aczPhase')
    n.killKey(s,'q4aczPhase')

    n.killKey(s,'q0aczLambda')
    n.killKey(s,'q1aczLambda')
    n.killKey(s,'q2aczLambda')
    n.killKey(s,'q3aczLambda')
    n.killKey(s,'q4aczLambda')

    n.killKey(s,'q0aczW')
    n.killKey(s,'q1aczW')
    n.killKey(s,'q2aczW')
    n.killKey(s,'q3aczW')
    n.killKey(s,'q4aczW')
    '''
    
#end         