from datetime import datetime
import itertools

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import leastsq, fsolve
import time
import random
import pyle.gateCompiler as gc
import labrad

from labrad.units import Unit
ns, us, GHz, MHz = [Unit(s) for s in ('ns', 'us', 'GHz', 'MHz')]
#from scripts import GHz_DAC_bringup

from pyle.dataking import multipleQubit as mq
from pyle.dataking import singleQubit as sq
from pyle.dataking import util as util
from pyle.util import sweeptools as st

from pyle.plotting import dstools as ds
from pyle import tomo

'''
def smsNotify(cxn, msg, username):
    return cxn.telecomm_server.send_sms('automate daily', msg, username)

def getBoardGroup(cxn, Sample):
    """ Get the board group used by the experiment associated to sample"""
    fpgas = cxn.ghz_fpgas
    boardGroups = fpgas.list_board_groups()
    def getAnyBoard():
        for dev in Sample.values():
            try:
                #Look in channels to see if we can find any FPGA board
                return dict(dev['channels'])['uwave'][1][0]
            except (KeyError,TypeError):
                #Don't do anything, just try the next one
                pass
    board = getAnyBoard()
    if board is None:
        return board
    for bg in boardGroups:
        if board in [name[1] for name in fpgas.list_devices(bg)]:
            return bg
    return None        
    
def getReadoutType(Sample, device):
    pass
    
def bringupBoards(cxn, boardGroup):
    ok = True
    resultWords = {True:'ok',False:'failed'}
    fpgas = cxn.ghz_fpgas
    try:
        successDict = GHz_DAC_bringup.bringupBoardGroup(fpgas, boardGroup)
        for board, successes in successDict.items():
            for item,success in successes.items():
                if not success:
                    print 'board %s %s failed'%(board,item)
                    ok = False
    except Exception:
        ok = False
    return ok
    
def bringup_dcBias(Sample, pause=True):
    pass 
'''

def set_paramp(Sample):
    sample, qubits = util.loadQubits(Sample, write_access=False)
    cxn = Sample._cxn
    readoutConfig = sample[qubits[0]['readoutConfig']]
    
    ar = cxn.anritsu_server
    ar.select_device(readoutConfig['pumpGPIB'])
    ar.frequency(readoutConfig['pumpFrequency'])
    ar.amplitude(readoutConfig['pumpPower'])
    ar.output(True)

def bringup_single_qubits(Sample, measure, spectroscopy=False, spectroscopy2=True, rabi=True, fine=True):

    for meas in measure:
        if spectroscopy:
            sq.spectroscopy(Sample, meas, freq = spectroscopy, update=True)
        
        if spectroscopy2:
            sq.spectroscopy2State(Sample, meas, freq = None, update=True)
            
        if rabi:
            sq.rabihigh(Sample, meas, mode='piPulse',update=True, stats=300)
            sample, devs, qubit, Qubit = gc.loadQubits(Sample, meas, write_access=True)
            Qubit['piHalfAmp'] = qubit['piAmp']/2.
            #sq.rabihigh(Sample, meas, mode='piHalfPulse',update=True)
            
        sq.calculateReadoutCenters(Sample, meas)
            
        if fine:
            sq.freqtuner(Sample, meas, state=1, delay = st.r[0:200:3, ns], fringeFreq = 50*MHz, stats=300)
            
            sq.pituner(Sample, meas, state=1, diff=0.8, numPulses=3, mode='piPulse', stats=300)
            sq.pituner(Sample, meas, state=1, diff=0.8, numPulses=1, mode='piHalfPulse', stats=300)
            
def bringup_boards(Sample):
    
    fp = Sample._cxn.ghz_fpgas
    
    for dev in fp.list_devices():
        if 'DAC' in dev[1]:
            fp.select_device(dev[0])
            fp.dac_bringup()
    
def calibrate_readout(Sample, measure, freq=True):

    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)

    for meas in measure:
        
        idx = sample['config'].index(meas)
        readoutFreq = qubits[idx]['readoutPulseFrequency']
        
        if freq:
            sq.lollipop(Sample, meas, freq = st.r[readoutFreq-5*MHz:readoutFreq+5*MHz:1e-4, GHz], 
                     update=True, plot=False, dataFormat='iq')
        sq.calculateReadoutCenters(Sample, meas)
 
def calibrate_readout2state(Sample, measure, freq=True,states=[0,1,2]):

    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)

    for meas in measure:
        
        idx = sample['config'].index(meas)
        readoutFreq = qubits[idx]['readoutPulseFrequency']
        
        if freq:
            sq.lollipop(Sample, meas,df=8*MHz,states=states,update=True, plot=False,save=True, dataFormat='iq')
        sq.calculateReadoutCenters(Sample,meas,states=states)                
 
def characterize_coherence(Sample, measure, T1=st.lr[50:75e3:100, ns], T2=st.r[0:5500:50, ns], fringeFreq=2*MHz, twoState=False):

    for meas in measure:
        if T1:
            sq.T1(Sample, meas, delay = T1)
            if twoState:
                sq.T1(Sample, meas, delay = T1, state=2)
            
        if T2:
            sq.ramsey(Sample, meas, update=False, plot=False, delay = T2, fringeFreq = fringeFreq)
            sq.spinEcho(Sample, meas, update=False, plot=False, delay = T2, fringeFreq = fringeFreq)
            if twoState:
                sq.ramsey(Sample, meas, update=False, plot=False, delay = T2, fringeFreq = fringeFreq, state=2)
                sq.spinEcho(Sample, meas, update=False, plot=False, delay = T2, fringeFreq = fringeFreq, state=2)
    
    
def calibrate_single_qubit_xy_gates(Sample, measure, manual=False,
                          freq=True,
                          coarse=False, drag=True, fine=True,
                          cal2state=False, orbitFreq=False):
                              
    for meas in measure:
        sample, devs, qubit, Qubit = gc.loadQubits(Sample, meas, write_access=True)

        if freq:
            sq.freqtuner(Sample, meas, state=1)
            
        if cal2state:
            sq.freqtuner(Sample, meas, state=2)
            sq.pituner(Sample, meas, state=2)
            
        if coarse:
            sq.pituner(Sample, meas, state=1, diff=0.8, numPulses=3, mode='piPulse')
            sq.pituner(Sample, meas, state=1, diff=0.8, numPulses=1, mode='piHalfPulse')
            if drag:
                result = sq.dragScanAlpha(Sample, meas, pingPongs = 3, collect=True, manual = manual)
            
        if fine:
            sq.pituner(Sample, meas, state=1, diff=0.75, numPulses=5, mode='piPulse')
            sq.pituner(Sample, meas, state=1, diff=0.75, numPulses=5, mode='piHalfPulse')
            if drag:
                #sq.orbitXYZ(Sample, meas, wwFactor = np.linspace(0.2, 1.8, 21), update=True)
                sq.orbitXYZ(Sample, meas, alpha = np.linspace(-0.4, 0.4, 21), update=True)
        if orbitFreq:
            #sq.orbitXYZ(Sample, meas, interleaved = [['IW']], m=5, freq = np.linspace(-0.0003,0.0003, 21)*GHz, update=True, identityWaitLen = 1000*ns, stats=900)  
            sq.orbitXYZ(Sample, meas, interleaved = [['IW']], m=5, freq = np.linspace(-0.0003,0.0003, 21)*GHz, update=True, identityWaitLen = 1000*ns, stats=900)          
        
def calibrate_cz(Sample, pairs, amplitude=True, controlPhase=True, targetPhase=False, shape=False):
    
    for pair in pairs:
        if amplitude:
            mq.aczPhaseTuner(Sample, pair, amplitude=None)
        if controlPhase:
            mq.aczCRamsey(Sample, pair, update=True)
        if targetPhase:
            mq.aczIRamsey(Sample, pair, update=True)
            
def calibrate_cz_phases(Sample, measure):
    
    sample, devs, qubits = gc.loadQubits(Sample, measure, write_access=False)
    
    qubitDict = {}
    for qubit in qubits:
        qubitDict[qubit.__name__] = qubit
    
    for qubit in qubits:
        aczKeys = [x for x in qubit.keys() if 'aczAmp' in x]
        for key in aczKeys:
            parasiticQubit=None
            # find if there is parasitic qubit, only works if there's ONE parasitic qubit
            for otherQubit in qubits:
                aczDetuneKeys = [x for x in otherQubit.keys() if qubit.__name__+key.strip('aczAmp')+'aczDetune' in x]
                if len(aczDetuneKeys)>0:
                    parasiticQubit = otherQubit # if there is more than one parasitic qubit then this will overwrite the last one
            # have control, target, parasitic at this point
            # control = qubit
            targetQubitName = key.strip('aczAmp')
            targetQubit = qubitDict[targetQubitName]
            
            if parasiticQubit:
                mq.aczPhaseMatrix(Sample, [qubit.__name__, targetQubit.__name__, parasiticQubit.__name__], update=True)
            else:
                mq.aczPhaseMatrix(Sample, [qubit.__name__, targetQubit.__name__], update=True)
            
def batch_run(Sample, measure, scanName, sargs = ''):
    for meas in measure:
        scan = scanName + '(Sample,'+"'"+str(meas)+"'"+','+sargs+')'
        eval(scan)
    