#Standard packages
import numpy as np
from math import atan2
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import scipy
import itertools
import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
from pyle import signalprocessing as signalprocessing
import pyle.envelopes as env
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import utilMultilevels as ml
import pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.dataking import singleQubit as sq
from pyle.dataking import zfuncs
from pyle.util import sweeptools as st
from pyle.plotting import dstools as ds
from pyle.fitting import fitting
from pyle.util import structures
from pyle.datavault import DataVaultWrapper
from pyle import tomo as tomo
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray

from pyle.sim import pyleSim as ps
from pyle.analysis import readout as readout
from pyle.dataking.optimization import optimization as popt
from pyle.math import tensor

from pyle import gateCompiler as gc
from pyle import gates

from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss

from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.dataking.benchmarking import randomizedBenchmarking as rb

from pyle import quantum
import copy
import time
    
def dualQubitTest(Sample, measure, stats=900):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    alg = gc.Algorithm(agents=devs)
    alg[gates.Measure([alg.q0,alg.q1])]
    alg.compile()
    rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats=stats, dataFormat='iqRaw', debug=True)
    return rv.wait()
 
def MQTest(Sample, measure, piOnThese, stats=900, herald=False):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    alg = gc.Algorithm(agents=devs)
    if herald:
        alg[gates.Measure(alg.qubits, name='Herald')]
    if piOnThese:
        for pi in piOnThese:
            alg[gates.PiPulse([ alg.qubits[measure.index(pi)] ])]
    alg[gates.Measure(alg.qubits)]
    alg.compile()
    rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats=stats, dataFormat='iqRaw', measure = measure, debug=False)
    data= rv.wait()
    data = data.asarray
    print data.shape
    probs = readout.IQToProbs(data, alg.qubits, measure, correlated = True)
    print np.size(probs)
    return data,probs
    
def measurementXtalk(Sample, measure, name='Measurement X Talk',
             stats=3000, correlated=False, update=False, plot=True,
             save=True, collect=False, noisy=True):
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    statesList = [qubit['readoutCenterStates'] for qubit in qubits]
    
    experiments = list(itertools.product(*statesList))
    exptIndex = range(len(experiments))
        
    axes = [(exptIndex,'Experiment Index')]
    deps = [('Mean I','',''),('Mean Q','',''),('Expected state','',''),('IQ distance from expected','','')]*len(measure)
    # deps = [("IQ distance from expected readout center",'','')]*len(measure)
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currIdx):
        
        alg = gc.Algorithm(agents = devs)
        currExpt = experiments[currIdx]
        for qidx,algq in zip(currExpt, alg.qubits):
            if qidx:
                alg[gates.PiPulse([algq])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        data = data.asarray
        # data = readout.averageIQSingleDemod(data.asarray) # (channels, mean I or Q)
        if len(data.shape)==1:
            data = [data]
        toReturn = []
        for channel, q, state in zip(data, alg.qubits, currExpt):
            readoutCenters = q['readoutCenters']
            centerZ, _, _ = readout.IQToReadoutCenter(channel)
            exptI, exptQ = centerZ.real, centerZ.imag
            if state == 1:
                expectedI, expectedQ = readoutCenters[1].real, readoutCenters[1].imag
            elif state == 0:
                expectedI, expectedQ = readoutCenters[0].real, readoutCenters[0].imag
            distance = np.sqrt((expectedI-exptI)**2 + (expectedQ-exptQ)**2)
            toReturn.extend([exptI, exptQ, state, distance])
        returnValue(toReturn)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    measure = data.parameters['measure']
    qubits = [data.parameters[meas] for meas in measure]
    if plot:
        plt.figure()
        cols='krgbmc'
    for idx, qubit in enumerate(qubits):
        Is0 = []
        Is1 = []
        Qs0 = []
        Qs1 = []
    
        Is = data[:,idx*4+1]
        Qs = data[:, idx*4+2]
        states = data[:, idx*4+3]
        for I,Q,state in zip(Is, Qs, states):
            if state==0:
                Is0.append(I)
                Qs0.append(Q)
            elif state==1:
                Is1.append(I)
                Qs1.append(Q)
        if plot:
            plt.plot(Is0, Qs0, label=measure[idx] + '|0>', marker='o',c=cols[idx], markersize = 5, alpha=0.3,lw=0)
            plt.plot(Is1, Qs1, label=measure[idx] + '|1>', marker='o',c=cols[idx], markersize = 5, alpha=0.3,lw=0)
            readoutCenters = qubit['readoutCenters']
            Z0 = readoutCenters[0]
            Z1 = readoutCenters[1]
            plt.plot(Z0.real, Z0.imag, label=measure[idx] + '|0> registry', marker = 's', markersize = 10, markeredgecolor = cols[idx], mfc='None',markeredgewidth=2)
            plt.plot(Z1.real, Z1.imag, label=measure[idx] + '|1> registry', marker = 's', markersize = 10, markeredgecolor = cols[idx], mfc='None',markeredgewidth=2)
        if update:
            Qubit = Qubits[idx]
            Qubit['readoutCenters'] = [np.mean(Is0) + 1.0j*np.mean(Qs0), np.mean(Is1) + 1.0j*np.mean(Qs1)]
            Qubit['readoutCenterStates'] = [0,1]
    if plot:
        plt.legend()
        
    if collect:
        return data
        
def readoutNM(Sample, measure, 
               paramNames = ['readoutPulsePower', 'readoutRingupFactor', 'readoutRingupLength',
                             'readoutPulseFrequency', 'readoutDemodFrequency'],
               nonzdelt = 0.2, zdelt = 0.2,
               xtol = 0.001, ftol = 0.001,
               name='Readout NM MQ',
               averages = 10,
               maxfun=None,
               update = False, plot = False,
               stats=2100,
               save=True, collect=False, noisy=True):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    qubitAndParamNames = sum([popt.addQubitNamesToParamNames(paramNames, qubit) for qubit in qubits])
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps(nelderMead=True)
    kw = {'stats': stats, 'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw) 
    
    experiments = np.array(list(itertools.product([0,1], repeat = len(qubits))))
    
    def func(server, args):
        currParamDict = param.argsToCurrParamDict(args)
        
        reqs = []
        
        for expt in experiments:
        
            alg = gc.Algorithm(agents = devs)
            param.updateQubits(alg.qubits, currParamDict)
            for k,qubit in enumerate(alg.qubits):
                if expt[k]:
                    alg[gates.PiPulse([qubit])]
            alg[gates.Measure([alg.q0])]
            alg.compile()
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        results = yield FutureList(reqs)
        
        errors = []
        for k,qubit in enumerate(alg.qubits):
            # find out which experiments this qubit was |0> and |1>
            zeroIdx = np.argwhere(experiments[:,0]==0).flatten()
            oneIdx = np.argwhere(experiments[:,0]==1).flatten()
            zeroData = np.contatenate([results[zidx].asarray for zidx in zeroIdx], axis = 1) # add along stat axis
            oneData = np.contatenate([results[oidx].asarray for oidx in oneIdx], axis = 1)
            fids, probs, centers, _ = readout.IQToReadoutFidelity([zeroData,oneData], states=[0,1])
            errors.append(1-np.mean(fids))
            
        returnValue([np.sum(errors)])
            
    funcWrapper = popt.makeFuncWrapper(averages, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
        
    return output
        
def measurementDephasing(Sample, measure, stats=600,
                      power = np.linspace(-60, -25, 51)*dBm, readoutWidth = None,
                      noisy=True,
                      save=True, collect=False, name = 'Measurement Dephasing'):
    
    sample, devs, qubits = gc.loadQubits(Sample, measure, write_access=False)
    
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)   
    
    axes=[(power, 'Readout Power'),(readoutWidth,'Readout Width')]
    deps = [('Phase Coherence no Measure','',''), ('Phase Coherence Measure','','')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps ,measure=measure, kw=kw)
    def func(server, currPower, currReadoutWidth):
    
        reqs = []
        
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            alg.q1['readoutPulsePower'] = -200*dBm
            alg.q1['ringdownLen'] = 0*ns
            if currReadoutWidth:
                alg.q1['readoutPulseWidth'] = currReadoutWidth
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Sync(alg.qubits)]
            alg[gates.Readout([alg.q1],name='dephase')]
            alg[gates.Sync(alg.qubits)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Measure([alg.q0])]
            alg.compile()
                        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            if currPower:
                alg.q1['readoutPulsePower'] = currPower
            if currReadoutWidth:
                alg.q1['readoutPulseWidth'] = currReadoutWidth
            alg.q1['ringdownLen'] = 0*ns
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Sync(alg.qubits)]
            alg[gates.Readout([alg.q1],name='dephase')]
            alg[gates.Sync(alg.qubits)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Measure([alg.q0])]
            alg.compile()

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        dataControl = []
        dataMeasured = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataControl.append(prob[0][1][1])
            
        for r in results[len(opList):]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataMeasured.append(prob[0][1][1])
            
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl = ps.getDensityPhase(rho)[0]   
        ampControl = abs(rho[0,1])
        
        [rho, residual] = tomo.qst_evan(np.array(dataMeasured), tomo.octomo_ops, stats, 1)        
        phaseMeasured = ps.getDensityPhase(rho)[0]   
        ampMeasured = abs(rho[0,1])

        deltaPhi = phaseMeasured-phaseControl
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([ampControl, ampMeasured])
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=noisy)
    return data
    
def rabihigh(Sample, measure, name='RabiHigh MQ',
             amps = None,
             stats=600, state=1,
             save=True, collect=False, noisy=True):
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    if amps is None:
        amps = np.linspace(0, 2.0, 50)
    
    axes = [(amps,'Pi pulse amp')]
    deps = readout.genProbDeps(qubits, measure)
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, amp):
        
        alg = gc.Algorithm(agents = devs)
        for qi in alg.qubits:
            ml.setMultiLevels(qi, 'piAmp', amp, state)
            alg[gates.PiPulse([qi])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure)
        returnValue(np.squeeze(probs).flatten())
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def rabilong(Sample, measure, length = st.r[10:200:1, ns],
             name='Rabi Long',
             stats = 1500,
             herald = False,
             save=True, collect=False, noisy=True):
    """Power rabi oscillations to find optimal pi pulse amplitude"""
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(length,'Pi pulse length')]
    deps = readout.genProbDeps(qubits, measure)
    kw = {'stats': stats}

    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currLen):
        
        alg = gc.Algorithm(agents = devs)
        if herald:
            alg[gates.Measure(alg.qubits, name = 'Herald')]
        for qi in alg.qubits:
            qi['piLen'] = currLen
            alg[gates.PiPulse([qi])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure, herald=herald)
        returnValue(np.squeeze(probs).flatten())
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

    if collect:
        return data
        
def rabihigh2QCorrelated(Sample, measure, name='RabiHigh MQ Correlated',
             amps = None,
             stats=600,
             save=True, collect=False, noisy=True):
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    if amps is None:
        amps = np.linspace(0, 2.0, 50)
    
    axes = [(amps,'Pi pulse amp')]
    deps = [('Probability |%s>'%k,'','') for k in [00,01,10,11]]
    kw = {'stats': stats}
    
    q0piAmp = qubits[0]['piAmp']
    q1piAmp = qubits[1]['piAmp']
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, amp):
        
        alg = gc.Algorithm(agents = devs)
        alg.q0['piAmp'] = q0piAmp * amp
        alg.q1['piAmp'] = q1piAmp * amp
        alg[gates.PiPulse([alg.q0])]
        alg[gates.PiPulse([alg.q1])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, alg.qubits, measure, correlated=True)
        returnValue(probs)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data

def rabiXtalk(Sample, measure, name='RabiHigh X Talk', freq = None,
             lengths = np.logspace(np.log10(10), np.log10(750), 40)*ns,
             driveAmp = 0.5,
             stats=600, plot=True,
             save=True, collect=False, noisy=True):
             
    '''
    '''
    
    if len(measure)<2:
        raise Exception("You need to put 2+ qubits into measure, e.g. [0,1]")
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    experiments = list(itertools.product(range(len(measure)), range(len(measure))))
    
    axes = [(lengths,'Pi Pulse Length'), (freq, 'Frequency')]
    deps = [('Probability |1> %s'%qubits[qmeas].__name__,'Drive %s'%qubits[qdrive].__name__,'Rabi Height') for qmeas,qdrive in experiments]
    kw = {'stats': stats, 'driveAmp':driveAmp}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currLen, currFreq):
    
        reqs = []
        probs = []
            
        for qmeas, qdrive in experiments:
            alg = gc.Algorithm(agents = devs)
            #print "qmeas: %s, qdrive: %s" % (qmeas, qdrive)
            
            if currFreq:
                # zAmp = fToAmp(alg.qubits[qmeas], currFreq)
                zAmp = zfuncs.FrequencyToAmp(alg.qubits[qmeas])(currFreq)
                alg.qubits[qdrive]['f10'] = currFreq
                alg.qubits[qdrive]['f21'] = currFreq-200*MHz
                for q in alg.qubits:
                    q['fc'] = currFreq + 0.150*GHz
            else:
                alg.qubits[qdrive]['f10'] = qubits[qmeas]['f10']
                alg.qubits[qdrive]['f21'] = qubits[qmeas]['f10'] - 200*MHz
                
            for q in alg.qubits:
                q['piLen'] = currLen
            if currFreq:
                alg[gates.RabiXTalk([alg.qubits[qmeas], alg.qubits[qdrive]], driveAmp, z=zAmp)]
            else:
                alg[gates.RabiXTalk([alg.qubits[qmeas], alg.qubits[qdrive]], driveAmp, z=0)]
            alg[gates.Measure(alg.qubits)]
            alg.compile()
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        probs = []
        
        for r,expt in zip(results,experiments):
            qmeas,qdrive = expt
            # Is, Qs = r[qmeas] #THIS GOES WRONG IF qmeas is not given in the order of measure
            # prob = np.mean(sq.readoutProbabilities(alg.qubits[qmeas],np.array(Is)+1.0j*np.array(Qs)))
            # probs.append(prob)
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            probs.append(prob[qmeas][0][1])

        returnValue(probs)
            
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if plot:
        ph.isolationMatrix(data)

    if collect:
        return isolationMatrixNorm
        
def piPulseXtalk(Sample, measure, name='piPulse X Talk', lambda2=np.linspace(-1.0,1.0,51),
             save=True, collect=False, noisy=True,stats=900):
             
    '''
    '''
    
    if len(measure)!=2:
        raise Exception("You need to put 2 qubits into measure, e.g. ['q0','q1']")
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    qdrive=measure[0]
    qmeas=measure[1]
    
    name += ' ' + measure[0] + ' on ' + measure[1]
    
    axes = [(lambda2,'lambda2')]
    deps = [('Probability |1> drive','','') , ('Probability |1> meas','','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currLambda2):   

        alg = gc.Algorithm(agents = devs)
        alg.q0['piLambdas'] = [currLambda2*ns]
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        prob = mqReadoutProbabilities(data, alg.qubits, measure) # if holding, readout only works for stationary qubit
        print prob
        returnValue([prob])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data

def piPulseXtalk2(Sample, measure, name='piPulse X Talk', lambda2=np.linspace(-1.0,1.0,11),lambda3=np.linspace(-1.0,1.0,11),
             save=True, collect=False, noisy=True,stats=900):
             
    '''
    '''
    
    if len(measure)!=3:
        raise Exception("You need to put 3 qubits into measure, e.g. ['q0','q1']")
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    qdrive=measure[0]
    qmeas=measure[1]
    qmeas2=measure[2]    
    
    name += ' ' + measure[0] + ' on ' + measure[1] + ' and ' + measure[2]   
    
    axes = [(lambda2,'lambda2'),(lambda3,'lambda3')]
    #deps = [('Probability |1> drive','','') , ('Probability |1> meas','','')]
    deps = [('parasitic prob','','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currLambda2,currLambda3):   

        alg = gc.Algorithm(agents = devs)
        alg.q0['piLambdas'] = [currLambda2*ns,currLambda3*ns]
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        prob = mqReadoutProbabilities(data, alg.qubits, measure) # if holding, readout only works for stationary qubit
        print prob
        err = prob[1]+prob[2]
        returnValue([err])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def spectroscopy2tone(Sample, measure, otherQubitsOnCarrier = None, name='Spectroscopy 2 tone',
                      zrange=np.arange(-1.0,1.0,0.02), freq = st.r[5.5:6.0:5e-3, GHz],
                      stats=450, sb_freq=-100.0*MHz,
                      save=True, collect=False, noisy=True,
                      update=True,correctXtalkZ=True):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(zrange, 'z pulse height'),(freq, 'qubit frequency')]
    # deps = [('Probability |%s>'%k,'','') for k in [00,01,10,11]]
    deps = [('Probability |1>','',''),('Probability |1>','','')]
    kw = {'stats': stats,'correctXtalkZ':correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)   
   
    def func(server, currZamp, currFreq):
        alg = gc.Algorithm(agents=devs)
        for q in alg.qubits:
            q['fc'] = currFreq - sb_freq
        alg[gates.Spectroscopy([alg.q0], df=sb_freq, z=currZamp, zlonger=100*ns)]
        alg[gates.Spectroscopy([alg.q1], df=sb_freq, z=0, zlonger=100*ns)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        # probs = mqReadoutProbabilities(data, alg.qubits, measure, correlated=True)
        probs = mqReadoutProbabilities(data, alg.qubits, measure, correlated=False)
        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)        
        
def spectroscopyZState(Sample, measure, otherQubitsOnCarrier = None, name='Spectroscopy Z State',
                      zrange = 'coarse', fspan=40.0*MHz,
                      state = 0,
                      fspansteps=31,
                      stats=450, sb_freq=-100.0*MHz,
                      save=True, collect=False, noisy=True,
                      update=True,correctXtalkZ=False):
    """stuff
    Known issues:
    The pipeline makes it so that the computation of the fit function lags
    the actual data taking by one row. This is subtle, ask Peter or Dan.
    """
    
    if zrange == 'coarse':
        halfRange = np.logspace(np.log10(0.001), np.log10(1.0), 20)
        zrange = np.hstack((halfRange, -halfRange))
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    zrange = zrange[np.argsort(abs(zrange))] #reorder zrange to start at 0, and then move up and down
    
    fsweep=np.linspace(-1.0,1.0,fspansteps)*fspan['MHz']
    fsweep = fsweep[np.argsort(abs(fsweep))]*MHz
    fStep = 1.0*MHz/fspansteps
    freqPoints=len(fsweep)
    
    axes = [('Frequency','GHz'),('Z','')]
    deps = [("Probabiliy |1>",'','')]   
    kw = {'stats': stats,'correctXtalkZ':correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)   
    
    def qubitFreqModel(v, A, B, C):
        """Approximate qubit frequency as a function of bias
        
        This ignores the offset from Ec so do not use this to fit data
        """
        return A*np.cos(B*v + C)**0.5
    
    sweepData={
        'freqIdx': 0,
        'fluxIdx': 0,
        'freq': 0*fsweep,
        'prob': 0*fsweep,
        'maxima': 0*zrange,
        'freqFunc': np.array([0]),
        'parameters': (qubit[0]['f10']['GHz'], 1.0, 0.0),
        'linearFit': None
        }
    
    def sweep():
        for z in zrange:
            if sweepData['fluxIdx'] in [0,1,2] or sweepData['parameters'] is None:
                freqCenter_GHz = qubit[0]['f10']['GHz']
            else:
                freqCenter_GHz = qubitFreqModel(z, *sweepData['parameters'])
            freqCenter_GHz = st.nearest(freqCenter_GHz, fStep['GHz'])
            for df in fsweep:
                yield z,(df+(freqCenter_GHz*GHz))['GHz']*GHz #Don't ask

    def func(server, args):
        print('fluxIdx: %d'%sweepData['fluxIdx'])
        zAmplitude, frequency = args
        print frequency, zAmplitude
        #Set carrier and do spectroscopy
        alg = gc.Algorithm(agents=devs)
        alg.q0['fc']=frequency - sb_freq
        alg[gates.MoveToState([alg.q1], 0, state)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.Spectroscopy([alg.q0], df=sb_freq, z=zAmplitude, zlonger=100*ns)]
        if otherQubitsOnCarrier:
            for otherQubitName in otherQubitsOnCarrier: # null out other lines w same carrier
                alg.agents[otherQubitName]['fc'] = frequency - sb_freq
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        idx = sweepData['freqIdx']
        sweepData['freq'][idx] = frequency
        sweepData['prob'][idx] = prob      
        
        if idx+1 != freqPoints: #Not done with this flux
            sweepData['freqIdx'] += 1
        else: #Update fit
            #Store max frequency
            fluxIdx = sweepData['fluxIdx']
            maxFreqIdx = np.argmax(sweepData['prob'])
            maxFreq = sweepData['freq'][maxFreqIdx]
            sweepData['maxima'][fluxIdx] = maxFreq
            #update running fit
            fluxToFit = zrange[:sweepData['fluxIdx']+1]
            freqToFit_GHz = sweepData['maxima'][:sweepData['fluxIdx']+1]
            if fluxIdx < 4:
                pass
            else:
                try:
                    guess = list(sweepData['parameters'])
                    print "fluxToFit: %s"%fluxToFit
                    print "freqToFit_GHz: %s"%freqToFit_GHz
                    print "guess: ",guess
                    pOpt,pCov = scipy.optimize.curve_fit(qubitFreqModel, fluxToFit, freqToFit_GHz, guess, maxfev=10000)
                    sweepData['parameters'] = pOpt
                except:
                    print("WARNING: fitting did not converge on this pass")
            sweepData['fluxIdx'] += 1
            sweepData['freqIdx'] = 0
            if fluxIdx > 8:
                print("Switching to 40MHz freq span")
                fsweep=np.linspace(-1.0,1.0,fspansteps)*40
        returnValue([frequency, zAmplitude, prob])
    data = sweeps.run(func, sweep(), dataset=dataset, save=save, noisy=False)     
        
def swaptuner(Sample, measure, name = 'Swap Tuner',idlingName=None,
              swapAmp = None, swapTime = None, swapW = None,
              iterations = 2, state=1,
              ampBound = 0.075, ampPoints = 41, timeBound = 6*ns, timePoints = 41,
              save=True, update=True, stats=600, noisy=True):
    '''
    measure = [MovingQubit, IdlingQubit]
    '''   
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    moving = qubits[0]
    if idlingName is None:
        idling = qubits[1]
        idlingName = idling.__name__
    
    if swapAmp is None:
        swapAmp = moving[idlingName+'SwapAmp']
    if swapTime is None:
        swapTime = moving[idlingName+'SwapTime']
    if swapW is None:
        swapW = moving[idlingName+'SwapW']
        
    for iteration in range(iterations):
        dataStats=stats*(2**iteration)
        dSwapAmp = ampBound/(2.0**iteration)
        dSwapTime = (timeBound/(2.0**iteration))['ns']
        data = sq.swapSpectroscopy(Sample,
                                swapLen=swapTime, riseTime = swapW,
                                swapAmp=st.PQlinspace(swapAmp-dSwapAmp,swapAmp+dSwapAmp,ampPoints, None),
                                measure=moving.__name__, state=state, collect=True, save=save, freqScan=False,
                                stats=dataStats, noisy=noisy, name=name)
                                
        swapAmp = getMaxGauss( np.array(data[:,0]) , np.array(data[:,1]), fitToMax=False)
        data = sq.swapSpectroscopy(Sample,
                                swapLen=st.PQlinspace(swapTime-dSwapTime,swapTime+dSwapTime,timePoints, ns),
                                swapAmp=swapAmp, riseTime = swapW,
                                measure=moving.__name__, state=state, collect=True, save=save, freqScan=False,
                                stats=dataStats, noisy=noisy, name=name)
        swapTime = getMaxGauss( np.array(data[:,0]) , np.array(data[:,1]), fitToMax=False)        
        swapTime = swapTime*ns
        
    if update:
        Qubits[0][idlingName+'SwapAmp'] = swapAmp
        Qubits[0][idlingName+'SwapTime'] = swapTime
        Qubits[0][idlingName+'SwapW'] = swapW            
            
def testdelaySwap(Sample, measure, freq, length,w=1*ns,name='Qubit-Qubit delay, direct swap',
             startTime=st.r[-30:30:0.5,ns],
             waitLen=5*ns,
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True, update=False):
    #raise Exception("Julian make sure the minus sign conventions are correct")
    '''
    measure = [master, slave]
    
    here, inject is the master, and retrieve is the slave. This means, that if you have a 5 qubit chain, and can only define 
    delays with respect to their nearest neighbour, you need to 
    -define a master, say q2
    -measure q2,q1 & q1 q0
    -q0 timinglag = q2-q1 timinglag + q1-q0 timinglag
    '''
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    master = qubits[0]
    slave = qubits[1]

    # amp0=sq.fToAmp(master,freq)
    # amp1=sq.fToAmp(slave,freq)
    
    amp0=zfuncs.FrequencyToAmp(master)(freq)
    amp1=zfuncs.FrequencyToAmp(slave)(freq)
    
    #freqAmp = zfuncs.FrequencyToAmp(master)(freq)

    axes = [(startTime, 'Control swap start time')]
    deps = [('Probability |1>','Inject',''),('Probability |1>','Retrieve','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currTime):
        
        if noisy:
            print currTime
        
        alg = gc.Algorithm(agents = devs)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.testdelaySwap([alg.q0, alg.q1], [length,w,amp0,amp1],startTime=currTime)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure)
        prob=[probs[0][0][1],probs[1][0][1]]
        returnValue(prob)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    xfound = getMaxGauss(data[:,0],data[:,2],fitToMax=True)
    print xfound
    if update:
        Qubits[1]['timingLagWrtMaster'] -= xfound*ns
    if collect:
        return data
    else:
        return xfound

def testdelayAllSwap(Sample, master, pairs, freqs, lengths, w=1*ns, name='Qubit-Qubit delay, direct swap',
             waitLen=5*ns,startTime=st.r[-30:30:0.5,ns],
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True, update=True):
    '''
    pairs must be in order, starting from the master down:
    master = 2
    pairs = [ [2,1] , [1,0] ]
    freqs = [ 4.8*GHz, 4.8*GHz]
    lengths = [10*ns, 10*ns]
    master = 2
    pairs = [ [2,3] , [3,4] ]    
    freqs = [ 4.8*GHz, 4.8*GHz]    
    RB
    
    mq.testdelayAllSwap(s,'q2',[['q2','q1'],['q1','q0'],['q2','q3'],['q3','q4']],[5.1*GHz]*4, [10*ns]*4)
    '''
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    
    indexDict = {}
    for qubit in qubits:
        indexDict[qubit.__name__] = Sample['config'].index(qubit.__name__)
    
    for idx, pair,freq,length in zip(range(len(pairs)),pairs,freqs,lengths):   
        print pair, freq
        xfound = testdelaySwap(Sample, [pair[0], pair[1]], freq, length,w=w,name='Qubit-Qubit delay, direct swap',startTime=startTime,correctXtalkZ=True, collect=False, update=False,stats=stats)       
        if pair[0] == master:
            #update the slave immediately
            qubits[ indexDict[pair[1]] ] ['timingLagWrtMaster'] -= xfound*ns
            if update:
                Qubits[ indexDict[pair[1]] ] ['timingLagWrtMaster'] -= xfound*ns 
        else:
            masterfound=False
            #walk up the chain until you get the master
            totaldelay = xfound*ns  
            for idx2 in np.arange( idx-1, -1, -1 ): #goes from idx-1...0
                print idx2
                #totaldelay += -qubits[ pairs[idx2][1] ]['timingLagWrtMaster'] #not here, because this qubit should already be corrected for
                if pairs[idx2][0]==master:
                    masterfound=True
                    qubits[ indexDict[pair[1]] ] ['timingLagWrtMaster'] -= totaldelay    
                    if update:
                        Qubits[ indexDict[pair[1]] ] ['timingLagWrtMaster'] -= totaldelay
                    break
            if masterfound is False:
                raise Exception('no master')
        
def measZXtalk(Sample, measure, name = 'Meas Z Xtalk',
              amps = np.linspace(-0.2,0.2,11), length = 200*ns, 
              selfFactor = 0.05,
              plot=True, correctXtalkZ = False, collect = False,
              stats=600, noisy=True, save=True):
              
    '''
    measure = [target, xtalk]
    '''
              
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]
    
    if correctXtalkZ:
        name += ' corrected'
    if measure[0] == measure[1]:
        ampsPython = selfFactor*amps
        measure = [measure[0]]
    else:
        ampsPython = amps
              
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    if len(measure)>1:
        qt = qubits[0]
        qx = qubits[1]
    else:
        qt = qubits[0]
        qx = qubits[0]
    
    name += ':' + qt.__name__ + ' ' + qx.__name__
    
    axes = [(ampsPython, 'Z Pulse Amplitude')]
    deps = [('Qubit Phase', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    calZpaXtalkInvOrig = qt['calZpaXtalkInv']
    
    def func(server, currAmp): 
        
        reqs = []

        for op in opList:
            # Experiment Ramsey
            alg = gc.Algorithm(agents = devs)
            if len(measure)>1:
                qt = alg.q0
                qx = alg.q1
            else:
                qt = alg.q0
                qx = alg.q0
            alg[gates.PiHalfPulse([qt])]
            alg[gates.Sync([qt,qx])]
            alg[gates.Detune([qx], amp = currAmp, length = length, phase=False)]
            alg[gates.Sync([qt,qx])]
            alg[gates.Tomography([qt], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        data = []      
        
        results = yield FutureList(reqs)

        for r in results:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)[0][0][1]
            data.append(prob)
            
        [rho, residual] = tomo.qst_evan(np.array(data), tomo.octomo_ops, stats, 1)        
        phaseControl    = ps.getDensityPhase(rho)[0]        
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(phaseControl)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def measZXtalkAll(Sample, measure, amps = np.linspace(-0.2,0.2,11),
                  correctXtalkZ = False, length = 200*ns, selfFactor = 0.005, 
                  update=True, debugMode = False):
    
    '''
    Calibrates z crosstalk between all pairs of qubits (fills in the entire xtalk matrix)
    measure is a list. e.g. [0,1,2]
    
    Z Xtalk:
    
    V_actual = M * V_ideal
    phi = V_actual * factor_(V->phi)
    
    step 1: measure the phase on all qubits due to 1 qubit. place this in a column in a matrix. iterate over all control qubits to fill the matrix
    step 2: divide each row by its diagonal element. This also normalizes the qubit-dependent V->phi factor away.
    step 3 (optional - not implemented): supplant this matrix into calZpaXtalk matrix
    step 4: invert and save row-by-row in the calZpaXtalkInv key in each qubit.
    
    Note: the Xtalk correction assumes you placed the matrix in alphabetic order. Can be overruled by passing in the config, then it's in the order of the config.
    '''
    
    if update and correctXtalkZ:
        raise Exception("You're doing something dumb")
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    
    dim = len(Sample['config'])
    indexDict = {}
    for meas in measure:
        indexDict[meas] = Sample['config'].index(meas)
    
    data = []
    
    experiments = list(itertools.product(measure, measure)) #order: target, xtalk qubit
    for expt in experiments:
        stillWorking = True
        currAmps = amps
        while stillWorking:
            currData = measZXtalk(Sample, [expt[0], expt[1]], amps = amps, collect=True, correctXtalkZ=correctXtalkZ)
            p = np.polyfit(currData[:,0], np.unwrap(currData[:,1]), 1)
            data.append([expt, p[0]])
            if debugMode:
                while stillWorking:
                    currData = measZXtalk(Sample, [expt[0], expt[1]], amps=currAmps, length=length, selfFactor=selfFactor, collect=True, correctXtalkZ=correctXtalkZ)
                    yes_no = raw_input('Data ok? [y/n]: ')
                    if yes_no.capitalize() == 'Y':
                        p = np.polyfit(currData[:,0], np.unwrap(currData[:,1]), 1)
                        data.append([expt, p[0]])
                        stillWorking=False
                    elif yes_no.capitalize() == 'N':
                        trunc_amps = raw_input('[T]runcate or adjust [A]mps? [t/a]: ')
                        if trunc_amps.capitalize() == 'T':
                            # do adjust to find new dataset
                            traces = [{'x': currData[:,0], 'y': currData[:,1], 'args': ('-ok',)}]
                            params = [{'name': 'Z Amp lower val', 'val': min(currData[:,0]), 'range':(min(currData[:,0]), max(currData[:,0])), 'axis': 'x', 'color': 'b'},
                                      {'name': 'Z Amp upper val', 'val': max(currData[:,0]), 'range':(min(currData[:,0]), max(currData[:,0])), 'axis': 'x', 'color': 'r'}]
                            result = adjust.adjust(params, traces)
                            lower = result['Z Amp lower val']
                            upper = result['Z Amp upper val']
                            idx = (currData[:,0]>lower) & (currData[:,0]<upper)
                            p = np.polyfit(currData[idx,0], np.unwrap(currData[idx,1]), 1)
                            data.append([expt, p[0]])
                            stillWorking=False
                        elif trunc_amps.capitalize() == 'A':
                            # adjust currAmps
                            str_amps = raw_input('Enter new amps, e.g. np.linspace(-1,1,11): ')
                            currAmps = eval(str_amps)
                        else:
                            print "Please pick [t/a]"
                    else:
                        print "Please pick [y/n]"
            else:
                # currData = measZXtalk(Sample, [expt[0], expt[1]], amps=currAmps, length=length, selfFactor=selfFactor, collect=True, correctXtalkZ=correctXtalkZ)
                # p = np.polyfit(currData[:,0], np.unwrap(currData[:,1]), 1)
                # data.append([expt, p[0]])
                stillWorking=False
    
    xTalkMatrixUnnormalized = np.diag([1.]*dim)
    for dat in data:
        expt, slope = dat
        xTalkMatrixUnnormalized[indexDict[expt[0]],indexDict[expt[1]]] = slope
    
    xTalkMatrixNormalized = np.copy(xTalkMatrixUnnormalized)
    for row in range(dim):
        normalizationFactor = xTalkMatrixUnnormalized[row,row]
        xTalkMatrixNormalized[row,:] /= normalizationFactor
    
    invertedXTalkMatrix = np.linalg.inv(xTalkMatrixNormalized)
    if update:
        for k,Q in enumerate(Qubits):
            Q['calZpaXtalkInv'] = invertedXTalkMatrix[k,:]
    
    return data
            
def zzXtalk(Sample, measure, name='ZZ Xtalk',
            delay=st.r[10:100:10,ns],freq=None,
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True, plot=True):
             
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops] 
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    axes = [(delay, 'detune length'),(freq,'detune freq')]
    deps = [('Single Qubit Phase', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    
    def func(server, currDelay, currFreq):          
        
        reqs = []
        
        for op in opList:
            # no pi pulse
            alg = gc.Algorithm(agents = devs)
            targetQubit = alg.q0
            controlQubit = alg.q1
            if currFreq:
                freqAmp = zfuncs.FrequencyToAmp(controlQubit)(currFreq)
            else:
                freqAmp = 0.0
            alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.Sync([targetQubit, controlQubit])] 
            alg[gates.Detune([controlQubit], length=currDelay,amp=freqAmp)]  
            alg[gates.Sync([targetQubit, controlQubit])]             
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        for op in opList:
            # pi pulse
            alg = gc.Algorithm(agents = devs)
            targetQubit = alg.q0
            controlQubit = alg.q1
            if currFreq:
                # freqAmp = fToAmp(controlQubit, currFreq)
                freqAmp = zfuncs.FrequencyToAmp(controlQubit)(currFreq)
            else:
                freqAmp = 0.0
            alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.PiPulse([controlQubit])]
            alg[gates.Sync([targetQubit, controlQubit])]            
            alg[gates.Detune([controlQubit], length=currDelay,amp=freqAmp)]  
            alg[gates.Sync([targetQubit, controlQubit])]             
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        dataControl = []
        dataExperiment = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataControl.append(prob[0][0][1])
            
        for r in results[len(opList):]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataExperiment.append(prob[0][0][1])
            
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl    = ps.getDensityPhase(rho)[0]        
        
        [rho, residual] = tomo.qst_evan(np.array(dataExperiment), tomo.octomo_ops, stats, 1)        
        phaseExperiment = ps.getDensityPhase(rho)[0] 
            
        deltaPhi = phaseExperiment-phaseControl
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)   
    
    if plot:
        lengths, freqs, phases = ds.arrange2Dscan(data)
        fitted = []
        for freq,phase in zip(freqs[::-1],phases):
            p = np.polyfit(lengths, np.unwrap(phase), 1)
            omegazz = p[0]/(2*np.pi)
            fitted.append([freq, omegazz])
        fitted = np.array(fitted)
        plt.plot(fitted[:,0], fitted[:,1],'ob')
        plt.xlabel('Frequency [GHz]')
        plt.ylabel('Omega zz [GHz]')
    
def aSwap(Sample, measure, name='aSWAP', hold=True, 
             length = st.r[20:50:1, ns], amplitude = st.r[0:-1.0:0.01, GHz],
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True):
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(amplitude, 'Swap Amplitude'), (length, 'Swap Length')]
    deps = [('Retrieve Probability |1>', '', '')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currAmp, currLen):
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        if hold:
            alg[gates.adiabaticSWAP([alg.q0,alg.q1], length = currLen, deltaf1 = currAmp, holdTime = alg.q1['readoutLen']+100*ns, holdTimeTiming = False)]
        else:
            alg[gates.adiabaticSWAP([alg.q0,alg.q1], length = currLen, deltaf1 = currAmp)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=True)

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        prob = mqReadoutProbabilities(data, alg.qubits, measure)[1] # if holding, readout only works for stationary qubit
        returnValue([prob])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def aSwapNM(Sample, measure, hold=True, nonzdelt = 0.05, zdelt = 0.01,
               xtol = 0.001, ftol = 0.001,
               name='a SWAP Shape Nelder Mead',
               averages = 10,
               iterations = 1,
               maxfun=None,
               update = False, plot = False,
               stats=2100, correctXtalkZ=True,
               save=True, collect=False, noisy=True):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    amp = controlQubit[targetQubit.__name__+'aSwapAmp']
    thi = controlQubit[targetQubit.__name__+'aSwapLambda'][1]
    thf = controlQubit[targetQubit.__name__+'aSwapLambda'][2]
    lam2 = controlQubit[targetQubit.__name__+'aSwapLambda'][3]
    lam3 = controlQubit[targetQubit.__name__+'aSwapLambda'][4]
    w = controlQubit[targetQubit.__name__+'aSwapW']
    
    axes = [('Function Evaluation','')]
    deps = [('1-Population |1>','',''),('1-Population |1> STD','',''),
            ('Amplitude','GHz',''),
            ('Theta Initial','',''),
            ('Theta Final','',''),
            ('Lambda 2','',''),
            ('Lambda 3','',''),
            ('W','','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    
    def func(server, args):
        
        currAmp, currThi, currThf, currLam2, currLam3, currW, currAvg = args

        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        controlQubit[targetQubit.__name__+'aSwapAmp'] = currAmp
        controlQubit[targetQubit.__name__+'aSwapLambda'][1] = currThi
        controlQubit[targetQubit.__name__+'aSwapLambda'][2] = currThf
        controlQubit[targetQubit.__name__+'aSwapLambda'][3] = currLam2
        controlQubit[targetQubit.__name__+'aSwapLambda'][4] = currLam3
        controlQubit[targetQubit.__name__+'aSwapW'] = currW
        alg[gates.PiPulse([controlQubit])]
        if hold:
            alg[gates.adiabaticSWAP([alg.q0,alg.q1], deltaf1 = currAmp, holdTime = alg.q1['readoutLen']+100*ns, holdTimeTiming = False)]
        else:
            alg[gates.adiabaticSWAP([alg.q0,alg.q1], deltaf1 = currAmp)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)    
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        prob = mqReadoutProbabilities(data, alg.qubits, measure)[1]
        print 1-prob
        returnValue([1-prob])
            
    def funcWrapper(args):

        args=list(args)
        
        #currAmp, currThi, currThf, currLam2, currLam3, currW = args
    
        def seqGen():
            for currAvg in range(averages):
                #yield currAmp, currThi, currThf, currLam2, currLam3, currW, currAvg
                yield args + [currAvg]
                    
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample, 'dummy dataset', axes, deps,measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save=False, noisy=False)
        
        avgError = np.mean(data)
        std = np.std(data)/np.sqrt(averages)
        print 'avg error: %s, std: %s' % (avgError, std)
        
        garbageCollect.collect()
        
        return avgError, std
            
    inputs = [amp, thi, thf, lam2, lam3, w]
    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt, maxfun=maxfun)
    
    return output
    
def aczCRamsey(Sample, measure, name='Adiabatic CZ Control Ramsey',
             thetas=25, waitTime = None,
             iterations=1, herald=False,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True,ampfactor=1.0,extraWait=0.0):
    '''
    measure = [control, target]
    '''
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]   
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)

    if waitTime is None:
        waitTime = controlQubit['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Control no Pi',''),('Probability |1>','Control Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if herald:
            alg[gates.Measure(alg.qubits, name='Herald')]
        alg[gates.PiHalfPulse([controlQubit])]

        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]
        alg[gates.PiHalfPulse([controlQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if herald:
            alg[gates.Measure(alg.qubits, name='Herald')]
        alg[gates.PiPulse([targetQubit])]
        alg[gates.PiHalfPulse([controlQubit])]

        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]                
        alg[gates.PiHalfPulse([controlQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        probs0 = readout.IQToProbs(data0.asarray, alg.qubits, measure, herald=herald)[0][0][1]
        
        data1 = results[1]
        probs1 = readout.IQToProbs(data1.asarray, alg.qubits, measure, herald=herald)[0][0][1]
        
        returnValue([probs0, probs1])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[0][targetQubit.__name__+'aczPhase'] -= deltaPhi/iterations
    
    if collect:
        return data

def aczTRamsey(Sample, measure, name='Adiabatic CZ Target Ramsey',
             thetas=25, waitTime = None,
             iterations=1, herald=True,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True):
             
    '''
    measure = [control, target]
    '''
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]  
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Target no Pi',''),('Probability |1>','Target Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, theta):
        
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if herald:
            alg[gates.Measure(alg.qubits, name='Herald')]
        alg[gates.PiHalfPulse([targetQubit])]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        
        alg[gates.PiHalfPulse([targetQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if herald:
            alg[gates.Measure(alg.qubits, name='Herald')]
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiHalfPulse([targetQubit])]

        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        
        alg[gates.PiHalfPulse([targetQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        probs0 = readout.IQToProbs(data0.asarray, alg.qubits, measure, herald=herald)[1][0][1]
        
        data1 = results[1]
        probs1 = readout.IQToProbs(data1.asarray, alg.qubits, measure, herald=herald)[1][0][1]
        
        returnValue([probs0, probs1])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[1][controlQubit.__name__+'aczPhase'] -= deltaPhi/iterations
    
    if collect:
        return data
      
def aczPRamsey(Sample, measure, name='Adiabatic CZ Parasitic Ramsey ',
             thetas=25, waitTime = None,
             iterations=1,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True,ampfactor=1.0,extraWait=0.0):
    '''
    measure = [control, target, parasitic]
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
       
    if update and not qubits[2].get(qubits[0].__name__+qubits[1].__name__+'aczDetune'):
        raise Exception("You should not have to correct phase for a qubit that is not moving -> Z crosstalk correction will fix this")
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)

    if waitTime is None:
        waitTime = qubits[0]['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Control no Pi',''),('Probability |1>','Control Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        parasiticQubit = alg.q2
        alg[gates.PiHalfPulse([parasiticQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.PiHalfPulse([parasiticQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
  
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        parasiticQubit = alg.q2
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiHalfPulse([parasiticQubit])]
        alg[gates.Sync(alg.qubits)]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Wait([controlQubit], extraWait)]    
        alg[gates.Sync(alg.qubits)]        
        alg[gates.PiHalfPulse([parasiticQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        probs0 = readout.IQToProbs(data0.asarray, alg.qubits, measure)[2][0][1]
        
        data1 = results[1]
        probs1 = readout.IQToProbs(data1.asarray, alg.qubits, measure)[2][0][1]
 
        returnValue([probs0, probs1])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[2][qubits[0].__name__+qubits[1].__name__+'aczDetunePhase'] -= deltaPhi/iterations
    
    if collect:
        return data
        
def aczCSimultRamsey(Sample, measure, name='Adiabatic CZ Control Simultaneous Ramsey ',
             thetas=25, waitTime = None,
             iterations=1,
             stats=600, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True,ampfactor=1.0,extraWait=0.0):
    '''
    measure = [control, target, control2, target2]
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','Control0 no Pi',''),('Probability |1>','Control0 Pi',''),
            ('Probability |1>','Control1 no Pi',''),('Probability |1>','Control1 Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
 
    def func(server, theta):      
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit0 = alg.q0
        targetQubit0 = alg.q1
        controlQubit1 = alg.q2
        targetQubit1 = alg.q3
        alg[gates.PiHalfPulse([controlQubit0])]
        alg[gates.PiHalfPulse([controlQubit1])]
        alg[gates.Sync(alg.qubits)]
        alg[gates.adiabaticCZ([controlQubit0, targetQubit0], detuneOthers=False)]
        alg[gates.adiabaticCZ([controlQubit1, targetQubit1], detuneOthers=False)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.PiHalfPulse([controlQubit0], phase = theta)]
        alg[gates.PiHalfPulse([controlQubit1], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
  
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit0 = alg.q0
        targetQubit0 = alg.q1
        controlQubit1 = alg.q2
        targetQubit1 = alg.q3
        alg[gates.PiPulse([targetQubit0])]
        alg[gates.PiPulse([targetQubit1])]
        alg[gates.PiHalfPulse([controlQubit0])]
        alg[gates.PiHalfPulse([controlQubit1])]
        alg[gates.Sync(alg.qubits)]
        alg[gates.adiabaticCZ([controlQubit0, targetQubit0], detuneOthers=False)]
        alg[gates.adiabaticCZ([controlQubit1, targetQubit1], detuneOthers=False)]
        alg[gates.Sync(alg.qubits)]
        alg[gates.PiHalfPulse([controlQubit0], phase = theta)]
        alg[gates.PiHalfPulse([controlQubit1], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
  
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        Is, Qs = data0[0]
        probNoPi0 = np.mean(sq.readoutProbabilities(controlQubit0,np.array(Is)+1.0j*np.array(Qs)))
        Is, Qs = data0[2]
        probNoPi1 = np.mean(sq.readoutProbabilities(controlQubit1,np.array(Is)+1.0j*np.array(Qs)))
        
        data1 = results[1]
        Is, Qs = data1[0]
        probPi0 = np.mean(sq.readoutProbabilities(controlQubit0,np.array(Is)+1.0j*np.array(Qs)))
        Is, Qs = data1[2]
        probPi1 = np.mean(sq.readoutProbabilities(controlQubit1,np.array(Is)+1.0j*np.array(Qs)))
 
        returnValue([probNoPi0, probPi0, probNoPi1, probPi1])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data

def aczDynamicPhase(Sample, measure, name='ACZ Dynamic Phase',
             iterations = st.r[1:8:1], update=True,
             waitTime = 5*ns,
             stats=600, correctXtalkZ=True,
             save=True, collect=False, noisy=True, plot=True,piPulse=False,phaseC=0.0,phaseT=0.0):
             
    """
    measure = [control, target]
    Runs the Cphase operation repeatedly to accumulate single qubit phases. Then looks 
    at how the phase accumulates as a function of Cphase operation to be able to compensate
    single qubit XY phases.
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]
    
    #thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]  
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
    
    axes = [(iterations, 'Number of CZ operations')]
    deps = [('Control Ramsey Relative Phase', '', ''),
            ('Target Relative Phase', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ, 'waitTime':waitTime}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    
    def func(server, iteration):          
        
        reqs = []
        
        for op in opList:
            # Control Ramsey
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([controlQubit])]
            for k in range(iteration):
                alg[gates.adiabaticCZ([controlQubit, targetQubit])]
                if piPulse:
                    alg[gates.PiPulse([controlQubit],phase=phaseC)]                
                if iteration>1 and k<iteration-1:
                    alg[gates.Wait([controlQubit], waitTime)] #time between CZ's
                    
            alg[gates.Tomography([controlQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        for op in opList:
            # Target Ramsey
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(iteration):
                alg[gates.adiabaticCZ([controlQubit, targetQubit])]
                if piPulse:
                    alg[gates.PiPulse([targetQubit],phase=phaseT)]                
                if iteration>1 and k<iteration-1:
                    alg[gates.Wait([controlQubit], waitTime)] #time between CZ's
                               
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        dataControl = []
        dataTarget = []
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataControl.append(prob[0][0][1])
        for r in results[len(opList):]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataTarget.append(prob[1][0][1])
        
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl = ps.getDensityPhase(rho)[0] --np.pi/2.    
        
        [rho, residual] = tomo.qst_evan(np.array(dataTarget), tomo.octomo_ops, stats, 1)        
        phaseTarget = ps.getDensityPhase(rho)[0] --np.pi/2.    
        
        returnValue([phaseControl, phaseTarget])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        phaseControl = data[:,1]
        phaseTarget = data[:,2]
        
        phasePerGateControl = np.polyfit(iterations, np.unwrap(phaseControl), 1)[0]
        phasePerGateTarget = np.polyfit(iterations, np.unwrap(phaseTarget), 1)[0]
        
        Qubits[0][targetQubit.__name__+'aczPhase'] += phasePerGateControl
        Qubits[1][controlQubit.__name__+'aczPhase'] += phasePerGateTarget
        
    if plot:
        phaseControl = np.unwrap(data[:,1])
        phaseTarget = np.unwrap(data[:,2])
        plt.figure()
        plt.plot(iterations,phaseControl,'r',iterations,phaseTarget,'b')
        plt.show()

def aczPhaseTuner(Sample, measure, name='ACZ Phase Tuner',
             length=None, amplitude = 'guess',
             iterations = 1, waitTime = 0.0, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=True,
             save=True, collect=False, noisy=True):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.97, origAmplitude*1.03, 21)
        
    elif amplitude is 'guess':
        f21Control = controlQubit['f21']
        f10Target = targetQubit.get(controlQubit.__name__+'aczDetune', targetQubit['f10'])
        Delta = -f21Control+f10Target
        amplitude = np.linspace(Delta*0.8, Delta*1.05, 21)
    
    axes = [(amplitude, 'Adiabatic Cphase Amplitude'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Relative Phase', '', ''),('Phase Coherence unloaded', '', ''),('Phase Coherence loaded', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    def func(server, currAmp, currLen, it, waitTime):
        
        reqs = []
        
        for op in opList:
            # unloaded control
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        for op in opList:
            # loaded control
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiPulse([controlQubit])]
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
                    
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        dataUnloaded = []
        dataLoaded = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataUnloaded.append(prob[1][0][1])
            
        for r in results[len(opList):]:
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)
            dataLoaded.append(prob[1][0][1])
            
            
        [rho, residual] = tomo.qst_evan(np.array(dataUnloaded), tomo.octomo_ops, stats, 1)        
        phaseUnloaded = ps.getDensityPhase(rho)[0]   
        ampUnLoaded = abs(rho[0,1])
        
        [rho, residual] = tomo.qst_evan(np.array(dataLoaded), tomo.octomo_ops, stats, 1)        
        phaseLoaded = ps.getDensityPhase(rho)[0]   
        ampLoaded = abs(rho[0,1])

        deltaPhi = phaseLoaded-phaseUnloaded
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi), ampUnLoaded, ampLoaded])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update and manual:
        amps = data[:,0]
        traces = [{'x': amps, 'y': data[:,1], 'args': ('-r',)},
                  {'x': amps, 'y': [np.pi]*len(amps), 'args': ('-k',)}]
        params = [{'name': 'Adiabatic Cphase amp', 'val': float(controlQubit[targetQubit.__name__+'aczAmp']), 
                   'range':(min(amps), max(amps)), 'axis': 'x', 'color': 'b'}]
        result = adjust.adjust(params, traces)
        if result:
            Qubits[0][targetQubit.__name__+'aczAmp'] = result['Adiabatic Cphase amp']*GHz
        return result['Adiabatic Cphase amp']
                
    elif update and not manual:
        amps = data[:,0]
        phases = np.unwrap(data[:,1])
        
        spline = scipy.interpolate.UnivariateSpline(amps, phases, s=0.5)
        
        bestAmp = scipy.optimize.fsolve(lambda x: spline(x)-np.pi, controlQubit[targetQubit.__name__+'aczAmp'])[0]
        Qubits[0][targetQubit.__name__+'aczAmp'] = bestAmp*GHz
        print 'bestAmp:', bestAmp
        if not manual and iterations>1:
            raise Exception("Branch cut/unwrap issues make it hard to determine which pi you are aiming for")
        
        if plot:
            plt.figure()
            plt.plot(amps, phases)
            plt.plot(amps, spline(amps))
    return

def aczPhaseTuner2state(Sample, measure, name='ACZ Phase Tuner 2 state',
             length=None, amplitude = st.r[0.5:0.8:0.01, GHz],
             iterations = 1, waitTime = 0.0, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=True,
             save=True, collect=False, noisy=True):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.97, origAmplitude*1.03, 21)
    
    axes = [(amplitude, 'Adiabatic Cphase Amplitude'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Relative Phase', '', ''),('Relative Phase 2', '', ''),('Phase Coherence unloaded', '', ''),('Phase Coherence loaded', '', ''),('Phase Coherence loaded2', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    def func(server, currAmp, currLen, it,waitTime):
        
        reqs = []
        
        for op in opList:
            # unloaded control
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
         
        for op in opList:
            # loaded control
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiPulse([controlQubit])]
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
                    
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        for op in opList:
            # loaded control
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.MoveToState([controlQubit], 0, 2)]
            alg[gates.PiHalfPulse([targetQubit])]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
                    
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))            
        
        dataUnLoaded = []
        dataLoaded = []
        dataLoaded2 = []
        results = yield FutureList(reqs)
        n=len(opList)
        for r in results[:n]:
            Is,Qs = r[1]
            prob = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataUnLoaded.append(prob)
        for r in results[n:(2*n)]:
            Is,Qs = r[1]
            prob = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataLoaded.append(prob)
        for r in results[(2*n):]:
            Is,Qs = r[1]
            prob = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataLoaded2.append(prob)            
            
        [rho, residual] = tomo.qst_evan(np.array(dataUnLoaded), tomo.octomo_ops, stats, 1)        
        phaseUnLoaded = ps.getDensityPhase(rho)[0]   
        ampUnLoaded = abs(rho[0,1])
        
        [rho, residual] = tomo.qst_evan(np.array(dataLoaded), tomo.octomo_ops, stats, 1)        
        phaseLoaded = ps.getDensityPhase(rho)[0]   
        ampLoaded = abs(rho[0,1])
        
        [rho, residual] = tomo.qst_evan(np.array(dataLoaded2), tomo.octomo_ops, stats, 1)        
        phaseLoaded2 = ps.getDensityPhase(rho)[0]   
        ampLoaded2 = abs(rho[0,1])        

        deltaPhi = phaseUnLoaded - phaseLoaded
        deltaPhi2 = phaseUnLoaded - phaseLoaded2        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi), mapToBranch(deltaPhi2), ampUnLoaded, ampLoaded, ampLoaded2])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update and manual:
        amps = data[:,0]
        traces = [{'x': amps, 'y': data[:,1], 'args': ('-r',)},
                  {'x': amps, 'y': [np.pi]*len(amps), 'args': ('-k',)}]
        params = [{'name': 'Adiabatic Cphase amp', 'val': float(controlQubit[targetQubit.__name__+'aczAmp']), 
                   'range':(min(amps), max(amps)), 'axis': 'x', 'color': 'b'}]
        result = adjust.adjust(params, traces)
        if result:
            Qubits[0][targetQubit.__name__+'aczAmp'] = result['Adiabatic Cphase amp']*GHz
        return result['Adiabatic Cphase amp']
                
    elif update and not manual:
        amps = data[:,0]
        phases = np.unwrap(data[:,1])
        
        spline = scipy.interpolate.UnivariateSpline(amps, phases, s=0.5)
        
        bestAmp = scipy.optimize.fsolve(lambda x: spline(x)-np.pi, controlQubit[targetQubit.__name__+'aczAmp'])[0]
        Qubits[0][targetQubit.__name__+'aczAmp'] = bestAmp*GHz
        print 'bestAmp:', bestAmp
        if not manual and iterations>1:
            raise Exception("Branch cut/unwrap issues make it hard to determine which pi you are aiming for")
        
        if plot:
            plt.figure()
            plt.plot(amps, phases)
            plt.plot(amps, spline(amps))
    return
      
def aczPop2state(Sample, measure, name='ACZ Pop 2 state', state=1,
             length=None, amplitude = st.r[0.5:0.8:0.01, GHz],
             iterations = 1, waitTime = 0.0, manual=False, plot=True,
             stats=900, correctXtalkZ=True, update=True,
             save=True, collect=False, noisy=True):
             
    """
    Runs the adiabatic cphase and looks at the retrieve ramsey to tune in CZ gate
    """
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
        
    if amplitude is None:
        origAmplitude = controlQubit[targetQubit.__name__+'aczAmp']
        amplitude = np.linspace(origAmplitude*0.97, origAmplitude*1.03, 21)
    
    axes = [(amplitude, 'Adiabatic Cphase Amplitude'),(length, 'Adiabatic Cphase Length'), (iterations,'ACZ Iterations'),(waitTime,'waitTime')]
    deps = [('Target pop, control in |0>', '', ''),('Target pop, control in |1>', '', ''),('Target pop, control in |2>', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    def func(server, currAmp, currLen, it,waitTime):
        
        reqs = []
        

        # unloaded control
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.MoveToState([targetQubit], 0, state)]  
        
        for k in range(it):
            alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
            if it>1 and k<it-1:
                alg[gates.Wait([controlQubit], waitTime)]
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
     

        # loaded control
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.PiPulse([controlQubit])]
        alg[gates.MoveToState([targetQubit], 0, state)]  
        
        for k in range(it):
            alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
            if it>1 and k<it-1:
                alg[gates.Wait([controlQubit], waitTime)]
                

        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)

        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        

        # loaded control
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.MoveToState([controlQubit], 0, 2)]
        alg[gates.MoveToState([targetQubit], 0, state)]        
        
        for k in range(it):
            alg[gates.adiabaticCZ([controlQubit, targetQubit],amp=currAmp,length=currLen)]             
            if it>1 and k<it-1:
                alg[gates.Wait([controlQubit], waitTime)]
                

        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)

        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))            
        

        results = yield FutureList(reqs)

        Is,Qs = results[0][1]
        prob0 = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
        
        Is,Qs = results[1][1]
        prob1 = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
        
        Is,Qs = results[2][1]
        prob2 = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))        

        
        returnValue([prob0,prob1,prob2])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    return
   
def aczPhaseMatrix(Sample, measure, name='ACZ Phase Matrix',
             iterations = 1, waitTime = 0.0, stats=900,
             update=True,
             correctXtalkZ=True, save=True, collect=False, noisy=True):
             
    """
    measure = [control, target, parasitic], OR [control, target]
    Runs the adiabatic cphase pulse and looks at the phase accumulated on all qubits in measure.  This includes qubits not involved in the acz phase detuning pulse at all
    """
    
    print "[Control, Target, Parasitic]: %s" % measure
    
    aczpair = [measure[0], measure[1]]
    
    name += 'from  ' + aczpair[0] + aczpair[1] + 'aczPulse'
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = list(ops)    
   
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)

    
    axes = [(range(len(measure)), 'measure index'), (iterations, 'number of iterations')]
    deps = [('Accumulated Phase','','')]

    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw) 

    def mapToBranch(phi):
        while phi<=-np.pi/4. or phi>7/4.*np.pi:
            if phi<=-np.pi/4.:
                phi+=2*np.pi
            elif phi>7/4.*np.pi:
                phi-=2*np.pi
        return phi
    
    def func(server, currInd, it):
        
        reqs = []
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.qubits[measure.index(aczpair[0])]
            targetQubit = alg.qubits[measure.index(aczpair[1])]
            phaseQubit = alg.qubits[currInd]
            alg[gates.PiHalfPulse([phaseQubit])]
            alg[gates.Sync(alg.qubits)]
            
            for k in range(it):
                alg[gates.adiabaticCZ([controlQubit, targetQubit])]             
                if it>1 and k<it-1:
                    alg[gates.Wait([controlQubit], waitTime)]
            
            alg[gates.Sync(alg.qubits)]
            alg[gates.Tomography([phaseQubit], op)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)
        
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))

        
        results = yield FutureList(reqs)
        phaseData = []
        for r in results:
            #I,Q = r[currInd]
            prob = readout.IQToProbs(r.asarray, alg.qubits, measure)[currInd][0][1]
            phaseData.append(prob)

        [rho, residual] = tomo.qst_evan(np.array(phaseData), tomo.octomo_ops, stats, 1)        
        accumulatedPhase = ps.getDensityPhase(rho)[0]
        
        returnValue([mapToBranch(accumulatedPhase)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        controlPhase = data[0,1] + np.pi/2
        targetPhase = data[1,1] + np.pi/2
        origControlPhase = qubits[0][qubits[1].__name__+'aczPhase']
        origTargetPhase = qubits[1][qubits[0].__name__+'aczPhase']
        Qubits[0][qubits[1].__name__+'aczPhase'] = mapToBranch(origControlPhase + controlPhase)
        Qubits[1][qubits[0].__name__+'aczPhase'] = mapToBranch(origTargetPhase + targetPhase)
        
        if len(data.parameters['measure'])>2:
            parasiticPhase = data[2,1] + np.pi/2
            origParasiticPhase = qubits[2][qubits[0].__name__+qubits[1].__name__+'aczDetunePhase']
            Qubits[2][qubits[0].__name__+qubits[1].__name__+'aczDetunePhase'] = mapToBranch(origParasiticPhase + parasiticPhase)

    return data
    
def aczTailC(Sample, measure, name='ACZ Tail Control',
             delay = np.logspace(np.log10(1), np.log10(400), 50)*ns,
             pulseLength=None, amplitude = None, wiggly=True,
             plot=True, 
             iterations = 1,
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True):
             
    if iterations>1:
        name += ' iterations: %d' % iterations
             
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    axes = [(delay, 'Ramsey Delay')]
    deps = [('QST Relative Phase', '', '')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currDelay):
        
        reqs = []
        
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([controlQubit])]
            for k in range(iterations):
                alg[gates.adiabaticCZ([controlQubit, targetQubit], amp=0, wiggly=False)]   
                if iterations>1 and k<iterations-1:
                    alg[gates.PiPulse([controlQubit])]
            alg[gates.Wait([controlQubit], currDelay)]
            alg[gates.Tomography([controlQubit], op)]
            alg[gates.Measure(alg.qubits)]
            
            alg.compile(correctXtalkZ=correctXtalkZ)    

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([controlQubit])]
            for k in range(iterations):
                alg[gates.adiabaticCZ([controlQubit, targetQubit], wiggly=wiggly)]
                if iterations>1 and k<iterations-1:
                    alg[gates.PiPulse([controlQubit])]                   
            alg[gates.Wait([controlQubit], currDelay)]
            alg[gates.Tomography([controlQubit], op)]
            alg[gates.Measure(alg.qubits)]
            
            alg.compile(correctXtalkZ=correctXtalkZ)    

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        dataControl = []
        dataExperiment = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(controlQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataControl.append(prob)
        for r in results[len(opList):]:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(controlQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataExperiment.append(prob)
            
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl    = ps.getDensityPhase(rho)[0]        
        
        [rho, residual] = tomo.qst_evan(np.array(dataExperiment), tomo.octomo_ops, stats, 1)        
        phaseExperiment = ps.getDensityPhase(rho)[0] 
            
        deltaPhi = phaseControl-phaseExperiment
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi)])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
    
def aczTailT(Sample, measure, name='ACZ Tail Target',
             delay = np.logspace(np.log10(1), np.log10(400), 50)*ns,
             nullSwap=False,
             pulseLength=None, amplitude = None,
             plot=False, 
             stats=900, correctXtalkZ=True, update=False,
             save=True, collect=False, noisy=True):
             
    # generate tomo operations
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    axes = [(delay, 'Ramsey Delay')]
    deps = [('QST Relative Phase', '', '')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currDelay):
        
        reqs = []
        
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.adiabaticCZ(devs,[controlQubit, targetQubit])]           
            alg[gates.Wait([targetQubit], currDelay)]
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            
            alg.compile(correctXtalkZ=correctXtalkZ)   

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        for op in opList:
            alg = gc.Algorithm(agents = devs)
            controlQubit = alg.q0
            targetQubit = alg.q1
            alg[gates.PiHalfPulse([targetQubit])]
            alg[gates.adiabaticCZ(devs,[controlQubit, targetQubit])]           
            alg[gates.Wait([targetQubit], currDelay)]
            alg[gates.Tomography([targetQubit], op)]
            alg[gates.Measure(alg.qubits)]
            
            alg.compile(correctXtalkZ=correctXtalkZ)    

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
            
        dataControl = []
        dataExperiment = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            Is,Qs = r[1]
            prob = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataControl.append(prob)
        for r in results[len(opList):]:
            Is,Qs = r[1]
            prob = np.mean(sq.readoutProbabilities(targetQubit,np.array(Is)+1.0j*np.array(Qs)))
            dataExperiment.append(prob)
            
        [rho, residual] = tomo.qst_evan(np.array(dataControl), tomo.octomo_ops, stats, 1)        
        phaseControl    = ps.getDensityPhase(rho)[0]        
        
        [rho, residual] = tomo.qst_evan(np.array(dataExperiment), tomo.octomo_ops, stats, 1)        
        phaseExperiment = ps.getDensityPhase(rho)[0] 
            
        deltaPhi = phaseControl-phaseExperiment
        
        def mapToBranch(phi):
            while phi<=-np.pi/4. or phi>7/4.*np.pi:
                if phi<=-np.pi/4.:
                    phi+=2*np.pi
                elif phi>7/4.*np.pi:
                    phi-=2*np.pi
            return phi
        
        returnValue([mapToBranch(deltaPhi)])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    t=data[:,0]
    phase=data[:,1]
    #kick out stuff below 0
    idx=matplotlib.mlab.find(t>=0.0)
    t=t[idx]
    phase=phase[idx]
    phase=np.unwrap(phase)
    v, cov, fitFunc=fitting.fitCurve('exponential',t,phase,[phase[0]-phase[-1],50,phase[-1]])
    decayTime=v[1]        
    print v
    print decayTime
    if plot:
        plt.figure()
        plt.plot(t,phase,'ob', label = 'data')
        plt.plot(t,fitFunc(t, *v),'-r', label='fit')
        plt.xlabel('Ramsey Delay')
        plt.ylabel('Ramsey Phase')
        plt.show()
    
    if collect:
        return data
                
def aczErrorFilter(Sample, measure, name='Adiabatic CZ Error Filter',
             delay = st.r[10:20:0.1, ns],
             iterations = 1,
             measure2=True,
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True, averages = 1):
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    delay_gen = st.averagedScan(delay,averages,noisy=noisy)
    
    axes = [(delay_gen(), 'acz length')]
    if measure2:
        deps = [("Probabiliy |%s>" % k,'','') for k in range(3)]
    else:
        deps = [('Retrieve Probability |1>', '', '')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currLen):
        
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiPulse([targetQubit])]
        alg[gates.Sync([controlQubit, targetQubit])]
        
        for k in range(iterations):
        
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            alg[gates.Wait([controlQubit], currLen)]
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            alg[gates.Wait([controlQubit], currLen)]
        
        if not measure2:
            alg[gates.Sync([controlQubit, targetQubit])]
            alg[gates.PiPulse([controlQubit])] # map 1 state onto 0, so we can clearly see 2 state population
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)    

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        if measure2:
            Is,Qs = data[0]
            probs = sq.averageMultiProbs(alg.q0,sq.readoutProbabilityAngles(alg.q0, np.array(Is)+1.0j*np.array(Qs)))
        else:
            probs = mqReadoutProbabilities(data, alg.qubits, measure)

        returnValue(probs)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def aczSpectroscopy(Sample, measure, name='acz Spectroscopy',
             length = st.r[0:50:1, ns], amplitude = st.r[0:0.8:0.01],
             state = [1,0],
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True):
             
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    axes = [(amplitude, 'Swap Amplitude'), (length, 'Swap Length')]
    deps = [('Inject Probability |1>', '', ''),
            ('Retrieve Probability |1>', '', '')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currAmp, currLen):
        
        alg = gc.Algorithm(agents = devs)  
        controlQubit = alg.q0
        targetQubit = alg.q1
        if state[0] is 1:
            alg[gates.PiPulse([controlQubit])]
        if state[1] is 1:
            alg[gates.PiPulse([targetQubit])]
        alg[gates.adiabaticCZ([controlQubit, targetQubit], length = currLen, amp = currAmp)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)    

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, alg.qubits, measure)
        returnValue(probs)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if collect:
        return data
        
def aczFPop(Sample, measure, name='Adiabatic CZ F state Pop',
             length = None, amplitude = None,
             thi = None, thf = None, 
             lam2 = None, lam3 = None,
             w = None,
             iterations=1,
             measure2 = True,
             update = False, plot = False,
             stats=900, correctXtalkZ=True,
             save=True, collect=False, noisy=True):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    
    axes = [(length, 'acz length'), (amplitude, 'acz amplitude'),
            (thi, 'Theta Initial'),
            (thf, 'Theta Final'),
            (lam2, 'Lambda 2'),
            (lam3, 'Lambda 3'),
            (w, 'W'),
            (iterations,'iterations')]
    if measure2:
        deps = readout.genProbDeps(qubits, measure)
        print deps
    else:
        deps = [('Control Probability |1>', '', ''),('Target Probability |1>', '', '')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currLen, currAmp, currThi, currThf, currLam2, currLam3, currW, currIter):

        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        if currThi:
            controlQubit[targetQubit.__name__+'aczLambda'][0] = currThi
        if currThf:
            controlQubit[targetQubit.__name__+'aczLambda'][1] = currThf
        if currLam2:
            controlQubit[targetQubit.__name__+'aczLambda'][2] = currLam2
        if currLam3:
            controlQubit[targetQubit.__name__+'aczLambda'][3] = currLam3
        if currW:
            controlQubit[targetQubit.__name__+'aczW'] = currW
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiPulse([targetQubit])]
        for k in range(currIter):
            alg[gates.adiabaticCZ([controlQubit, targetQubit], length=currLen, amp=currAmp)]   
            if currIter>1 and k<currIter-1:
                alg[gates.Wait([controlQubit],controlQubit['piLen'])]
        if not measure2:
            alg[gates.PiPulse([controlQubit])] # map 1 state onto 0, so we can clearly see 2 state population
            alg[gates.PiPulse([targetQubit])] # map 1 state onto 0, so we can clearly see 2 state population
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)    
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        if measure2:
            prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
            #print prob[0][0]
            probs = [prob[0][0][2] , prob[1][0][1]] 
        else:
            prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
            probs = [prob[0][0][1]] + [prob[1][0][1]]

        returnValue(probs)
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    x = data[:,0]
    if measure2:
        y = data[:,1]
    else:
        y = data[:,1]

    spline = scipy.interpolate.UnivariateSpline(x, y, s=0.05)
    bestAmp = scipy.optimize.fmin(lambda x: spline(x), x[len(x)/2])[0]
    
    if update:
        if bestAmp > x[-1] or bestAmp<x[0]: # outside of bounds, do manual pick
        
            traces = [{'x': x, 'y': y, 'args': ('-ob',)}]
            params = [{'name': 'Pulse Shape Parameter', 'val': x[len(x)/2], 'range':(min(x), max(x)), 'axis': 'x', 'color': 'b'}]
            bestAmp = adjust.adjust(params, traces)['Pulse Shape Parameter']
        
    print 'best amp is: ',bestAmp
    
    if plot:
    
        plt.plot(x, y)
        plt.plot(x, spline(x))
        
        
    if update:
        if np.any(length):
            Qubits[0][targetQubit.__name__+'aczTime'] = bestAmp*ns
        if np.any(thi):
            aczLambda = controlQubit[targetQubit.__name__+'aczLambda']
            aczLambda[0] = bestAmp
            Qubits[0][targetQubit.__name__+'aczLambda'] = aczLambda
        if np.any(thf):
            aczLambda = controlQubit[targetQubit.__name__+'aczLambda']
            aczLambda[1] = bestAmp
            Qubits[0][targetQubit.__name__+'aczLambda'] = aczLambda
        if np.any(lam2):
            aczLambda = controlQubit[targetQubit.__name__+'aczLambda']
            aczLambda[2] = bestAmp
            Qubits[0][targetQubit.__name__+'aczLambda'] = aczLambda
        if np.any(lam3):
            aczLambda = controlQubit[targetQubit.__name__+'aczLambda']
            aczLambda[3] = bestAmp
            Qubits[0][targetQubit.__name__+'aczLambda'] = aczLambda
        if np.any(w):
            Qubits[0][targetQubit.__name__+'aczW'] = bestAmp*ns
        
        if plot:
            plt.plot(x, y)
            plt.plot(x, spline(x))
    
    if collect:
        return data
                
def aczShapeNM(Sample, measure, 
               qubitAndParamNames = None,
               nonzdelt = 0.02, zdelt = 0.02,
               xtol = 0.001, ftol = 0.001,
               name='Adiabatic CZ Shape Nelder Mead',
               averages = 10,
               iterations = 1,
               maxfun = None,
               measure2=True,
               update = False, plot = False,
               stats=2100, correctXtalkZ=True,
               save=True, collect=False, noisy=True):
    
    if qubitAndParamNames is None:
        qubitAndParamNames = [[measure[0], measure[1]+'aczLambda']]
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps()
    kw = {'stats': stats, 'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    
    def func(server, args):
        currParamDict = param.argsToCurrParamDict(args)

        alg = gc.Algorithm(agents = devs)
        param.updateQubits(alg.qubits, currParamDict)
        controlQubit = alg.q0
        targetQubit = alg.q1
        alg[gates.PiPulse([controlQubit])]
        alg[gates.PiPulse([targetQubit])]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]   
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit],controlQubit['piLen'])]
        if not measure2:
            alg[gates.PiPulse([controlQubit])] # map 1 state onto 0, so we can clearly see 2 state population
            alg[gates.PiPulse([targetQubit])] # map 1 state onto 0, so we can clearly see 2 state population
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)    
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        if measure2:
            prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
            probs = [prob[0][0][2]]
        else:
            prob = readout.IQToProbs(data.asarray, alg.qubits, measure)
            probs = [prob[0][0][1]] + [prob[1][0][1]]
        print probs
        returnValue(probs)
            
    funcWrapper = popt.makeFuncWrapper(averages, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
    
    return output
    
def randomizedBenchmarkingSim(Sample, measure, measureRB, ms = None, steps = 30, k = 20, maxtime = 6500, stats=900, plot=True, collect=False,
                           name = 'Sim RB', correlatedProbs = True,
                           save = True, noisy = True, 
                           correctXtalkZ=True, correctXtalkXY=False, interleaved=False):
    '''
    Simultaneous RB.
    syntax for a measurement of C x I = IC: randomizedBenchmarkingSimClifford2D(s,[0,1],[1])
    mq.randomizedBenchmarkingSim(s,['q0','q2'],['q0'])
    mq.randomizedBenchmarkingSim(s,['q0','q2'],['q2'])
    mq.randomizedBenchmarkingSim(s,['q0','q2'],['q0','q2'])    
    '''

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    RBClass = rb.RandomizedBenchMarkingClifford(1) # we are doing simultaneous single qubit Randomized Benchmarking
    RBClass.setGateLengths([qubits[0]])
        
    name += ' ' + str(measureRB)
        
    def getSequence(m):
        sequenceData = RBClass.randGen([m], interleaved=interleaved, finish=True)
        sramLen=15000 #in ns
        seqLength=RBClass.computeGateLength(sequenceData)
        return sequenceData
        
    if ms == None:
        iterationsmax=rb.getlength(RBClass,maxtime,interleaved=interleaved)
        print 'max iterations %d' % iterationsmax
        ms = st.ilr[1:iterationsmax:steps]._range()
    
    iter = np.zeros(len(ms),dtype=int)
    for tel in range(len(iter)):
        iter[tel] = ms[tel]
    ms=iter
        
    axesname='m - number of Cliffords'
    if interleaved:
        name += ' interleaved: ' + str(interleaved)
        axesname='m - number of sets of Clifford+interleaved'
    axes = [(range(k),'sequence'),(ms, axesname)]
    if correlatedProbs:
        deps=[]
        for state_bin_repr in range(2**np.alen(measure)):
            probstr='|'+np.binary_repr(state_bin_repr,np.alen(measure))+'>'
            deps+=[('Probability', probstr, '')]
    else:
        deps = [('Sequence Fidelity ' + q.__name__,'','') for q in qubits]

    kw={'stats':stats, 'interleaved':interleaved, 'correctXtalkZ':correctXtalkZ, 'measureRB':measureRB,
    'axismode':'m', 'dataFormat': 'stacked2Dswapped','k':k, 'correlatedProbs':correlatedProbs}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, ki, m):                  
        alg = gc.Algorithm(agents = devs)
        for mRB in measureRB:            
            fullGateList = getSequence(m)          
            alg[gates.RandomizedBenchmarkingSingleQubitClifford( [ alg.agents[mRB] ] ,fullGateList)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ) 
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data,alg.qubits,measure,correlated = correlatedProbs)
        if correlatedProbs is False:
            probs = 1-np.array(probs)
        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset = save and dataset, pipesize=50)
            
    if collect:
        return data

def orbitSimNM(Sample, measure, qubitAndParamNames = None,
                           m = 50, k = 20, stats=900, collect=False,
                           nonzdelt = 0.2, zdelt = 0.2,
                           xtol = 0.001, ftol = 0.001,
                           maxfun=None,
                           name = 'ORBIT Sim NM', save = True, noisy = True, 
                           correctXtalkZ=True, interleaved=False):
                           
    if qubitAndParamNames == None:
        qubitAndParamNames = []
        for meas in measure:
            qubitAndParamNames.append([meas,'piAmp'])
            qubitAndParamNames.append([meas,'piHalfAmp'])
            qubitAndParamNames.append([meas,'alphaDRAG'])
            qubitAndParamNames.append([meas,'piLambdas'])

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps()
    kw={'stats':stats, 'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)    
    
    RBClass = rb.RandomizedBenchMarkingClifford(1) # we are doing simultaneous single qubit Randomized Benchmarking
    RBClass.setGateLengths([qubits[0]])
    
    def func(server, args):         
        currParamDict = param.argsToCurrParamDict(args)
        alg = gc.Algorithm(agents = devs)
        currParamDict = param.argsToCurrParamDict(args)
        for qubit in alg.qubits:
            sequenceData = RBClass.randGen([m], interleaved=interleaved, finish=True)
            alg[gates.RandomizedBenchmarkingSingleQubitClifford( [qubit], sequenceData)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        prob = readout.IQToProbs(data.asarray,alg.qubits,measure,correlated=True) 
        returnValue([1-np.squeeze(prob)[0]])
        
    funcWrapper = popt.makeFuncWrapper(k, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
        
    return output
                  
def orbitCZ(Sample, measure, 
                             phaseC = np.linspace(-0.3,0.3,31), phaseT = None, aczAmp = None, linearFF = 0.0,
                             settlingAmps = None, settlingRates = None, 
                             cf10 = None,
                             resetSettling=True,
                             wigglyAmps = None,
                             aczPadding = None,
                             xyPadding = None,
                             name = 'RB ACZ Optimization',
                             iterations = 20, seq = 10, interleaved = False, 
                             stats=900, beSmart=True, correctXtalkZ=True,
                             plot=False, collect=False, save = True, noisy = True, syncEachClifford=True,
                             update=True):
    
    # measure must be passed in the right order such that the adiabatic CZ is passed in the right order
    # measure[0] is adiabaticCZ control
    # measure[1] is adiabaticCZ target
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    for qi in q:
        qi['readout']=True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]
    
    origAczCPhase = qubits[0][qubits[1].__name__+'aczPhase']
    origAczTPhase = qubits[1][qubits[0].__name__+'aczPhase']
    origAczAmp = qubits[0][qubits[1].__name__+'aczAmp']
    origSettlingAmps = qubits[0]['settlingAmplitudes']
    origSettlingRates = qubits[0]['settlingRates']
    origCf10 = qubits[0]['f10']
    
    RBClass = rb.RandomizedBenchMarkingClifford(len(measure))    
    RBClass.setGateLengths(qubits)            
    
    
    axesname='m - number of Cliffords'
    if interleaved:
        name += ' interleaved: ' + str(interleaved)
        axesname='m - number of sets of Clifford+interleaved'
    axes = [(aczAmp,'ACZ Amp'), (settlingAmps,'settling Amp'),(settlingRates,'settling Rate'), (wigglyAmps, 'wiggly Amps'),(iterations, axesname),(phaseC,'ACZ C Phase'),(phaseT,'ACZ T Phase'),(cf10,'control f10'),(aczPadding,'aczPadding'),(xyPadding,'xyPadding')]
    deps = [('Sequence Fidelity', '', '')]
    
    sequenceData = [RBClass.randGen([iterations], interleaved=interleaved) for k in range(seq)] # contains all of the gates for all of the sequences for the system    

    #For labrad, the kw parameters need to be rectangular lists. So, we add IGN (for ignore) to make all elements the same length
    seqDataForDV=[]
    setToLen=11 #longest sequence is 11 gates: 3 from C1, 6 from swap, 2 from S1
    for seqs in sequenceData:
        #per sequence
        seqnew=[]
        for cliff in seqs:
            #per clifford
            cliffnew=[]
            for qseq in cliff:
                #per qubit
                cliffnew.append( qseq+['IGN']*(setToLen-len(qseq)) )
            seqnew.append(cliffnew)
        seqDataForDV.append(seqnew)

    kw={'stats':stats, 'interleaved':interleaved, 'seq':seq, 'correctXtalkZ':correctXtalkZ,'axismode':'m', 'syncEachClifford': syncEachClifford, 'sequences': seqDataForDV ,'dataFormat': 'stacked'}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, currAmp, currSetAmp, currSetRate, currWigglyAmp, it, currCPhase, currTPhase, currCf10, currAczPadding, currxyPadding):        
    
        if currSetRate is not None:
            if resetSettling:
                qubits[0]['settlingRates'] = [rate*GHz for rate in currSetRate]
            else:
                qubits[0]['settlingRates'] = origSettlingRates + [currSetRate]
        if currSetAmp is not None:
            if resetSettling:
                qubits[0]['settlingAmplitudes'] = [currSetAmp]
            else:
                qubits[0]['settlingAmplitudes'] = origSettlingAmps + [currSetAmp]
        if currWigglyAmp is not None:
            qubits[0]['aczWigglyAmp'] = currWigglyAmp
        if currCPhase is not None:
            qubits[0][qubits[1].__name__+'aczPhase'] = origAczCPhase + currCPhase
        if currTPhase is not None:
            qubits[1][qubits[0].__name__+'aczPhase'] = origAczTPhase + currTPhase
        if currAmp is not None:
            qubits[0][qubits[1].__name__+'aczAmp'] = origAczAmp + currAmp
            if linearFF is not None:
                qubits[0][qubits[1].__name__+'aczPhase'] = origAczCPhase + currAmp*linearFF + currCPhase
                
        if currCf10:
            qubits[0]['f10'] = origCf10 + currCf10
        if currAczPadding:
            qubits[0]['aczPadding'] = currAczPadding
        if currxyPadding:
            qubits[0]['xyPadding'] = currxyPadding
            qubits[1]['xyPadding'] = currxyPadding
        #print qubits[0]['settlingRates'],qubits[0]['settlingAmplitudes']
        
    
        reqs = []        
        print 'interation %d' % it        
        for sequenceNum in range(seq):            
            print 'sequence number %d' % sequenceNum
            if interleaved:
                #c1 - c - c2 - c - c(m+1), m=2
                idx=it*2
            else:
                idx=it
            currGateList = sequenceData[sequenceNum][:idx]
            fullGateList = RBClass.finishGateList(currGateList)            
            alg = gc.Algorithm(agents = devs)
            alg[gates.RandomizedBenchmarkingMultiQubitClifford(qubits,fullGateList,syncEachClifford=syncEachClifford)]
            alg[gates.Measure(alg.qubits)]            
            alg.compile(beSmart=beSmart, correctXtalkZ=correctXtalkZ)            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)) 
            del currGateList
            del fullGateList            
            del alg
        #time.sleep(1) #to keep the memory from overflowing...
        results = yield FutureList(reqs)
        data = []
        for r in results:
            probs = mqReadoutProbabilities(r,devs,list(np.sort(measure)),correlated = True)
            if len(measure) == 1:
                data.append(1-probs[0])# to get fidelity of |00> state in the stupid special case you're doing MQ RB with 1 qubit
            else:
                data.append(probs[0]) # to get fidelity of |00> state
        returnValue([np.mean(data)])     
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    if update:
        amp = getMaxGauss( np.array(data[:,0]) , np.array(data[:,1]), fitToMax=True)    
        if phaseC is not None:
            Q[0][qubits[1].__name__+'aczPhase']+=amp
        elif phaseT is not None:
            Q[1][qubits[0].__name__+'aczPhase']+=amp
        elif aczAmp is not None:
            Q[0][qubits[1].__name__+'aczAmp']+=amp
            
    if plot:
        data=np.array(data)
        dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
        dataLen = len(data)
        numColumns = dataLen/dataSize
        newDataLen = numColumns*dataSize
        data = data[:newDataLen, :] # we take off any unfinished rows
        corrections = data[::dataSize, 0]
        phases = data[:dataSize,1]
        fidelities = np.reshape(data[:,2], (numColumns, dataSize))
        
        for f,c in zip(fidelities, corrections):
            plt.plot(phases,f, label=str(c))
            
        plt.legend()
        plt.xlabel('correction')
        plt.ylabel('sequence fidelity')
    
    if collect:
        return data            
        
def orbitXYXtalk(Sample, measure, m = 25, k = 10,
                             correctXtalkXY = False,
                             freq = None, piLen = None, wwFreq1=None,
                             name = 'RB XY Xtalk',
                             interleaved = False, 
                             stats=900, correctXtalkZ=True,
                             plot=False, collect=False, save = True, noisy = True,
                             update=True):

    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    name += ' ' + qubits[0].__name__ + ' degraded by ' + qubits[1].__name__
    
    origF10 = qubits[1]['f10']
    origF21 = qubits[1]['f21']
    nonLin = origF10-origF21
    
    origPiLen = qubits[1]['piLen']
    origPiHalfLen = qubits[1]['piHalfLen']
    origPiAmp = qubits[1]['piAmp']
    origPiHalfAmp = qubits[1]['piHalfAmp']
    
    if freq is None:
        fc = qubits[0]['fc']
        upperBound = fc + 0.34
        lowerBound = fc - 0.34
        freq = np.linspace(lowerBound, upperBound, 51)*GHz
        inds = np.abs(freq-fc)>0.01*GHz # remove where sideband mixing is < 10 MHz
        freq = freq[inds]
    
    RBClass = rb.RandomizedBenchMarkingClifford(1)    
    RBClass.setGateLengths(qubits)            
    
    axesname='m - number of Cliffords'
    if interleaved:
        name += ' interleaved: ' + str(interleaved)
        axesname='m - number of sets of Clifford+interleaved'
    
    def getSequence(m):
        sequenceData = RBClass.randGen([m], interleaved=interleaved,finish=True)
        return sequenceData 
    
    axes = [(freq,'X Talk Frequency'),(piLen,'Pi Pulse Length'),(wwFreq1,'ww Freq 1')]
    deps = [('Sequence Fidelity', '', '')]
        
    kw={'stats':stats, 'interleaved':interleaved, 'correctXtalkZ':correctXtalkZ,'axismode':'m','dataFormat': 'stacked'}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, currFreq, currPiLen, currWWfreq1):        
        reqs = []               
        for ki in range(k):    
            print 'k =',str(ki)
            alg = gc.Algorithm(agents = devs)
            if currFreq:
                alg.q1['f10'] = currFreq
                alg.q1['f21'] = currFreq - nonLin
                if ki==0:
                    print currFreq
            if currPiLen:
                alg.q1['piLen'] = currPiLen
                alg.q1['piAmp'] = origPiAmp * origPiLen/currPiLen
                alg.q1['piHalfLen'] = currPiLen
                alg.q1['piHalfAmp'] = origPiHalfAmp * origPiHalfLen/currPiLen
                if ki==0:
                    print currPiLen
            if currWWfreq1:
                alg.q1['ww'][1] = currWWfreq1
                if ki==0:
                    print currWWfreq1
            for q in alg.qubits:
                fullGateList = getSequence(m)
                alg[gates.RandomizedBenchmarkingSingleQubitClifford([q],fullGateList)]
            alg[gates.Measure(alg.qubits)]
            alg.compile(correctXtalkZ=correctXtalkZ)          
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)) 
        results = yield FutureList(reqs)
        data = []
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(1-prob) # to get fidelity
        returnValue([np.mean(data)])     
    data = sweeps.grid(func, axes, dataset = save and dataset)

    if collect:
        return data 

def orbitXYXtalkAll(Sample, measure):
            
    experiments = list(itertools.permutations(measure,2))
    listOfDvwData = [orbitXYXtalk(Sample, expt, collect=True) for expt in experiments]
    ph.orbitXYXtalk(listOfDvwData)

def calRBXYXtalk(Sample, measure, xAmp = None, xPhase = 0.0,
                             freq = None, piLen = None,
                             name = 'Cal RB XY Xtalk',
                             iterations = 25, seq = 10, interleaved = False, 
                             stats=900, beSmart=True, correctXtalkZ=True,
                             plot=False, collect=False, save = True, noisy = True,
                             update=True):

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    q[0]['readout']=True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]

    config = Sample['config']
    qt = qubits[0]
    qx = qubits[1]
    targetqInd = config.index(qt.__name__)
    xtalkqInd = config.index(qx.__name__)
    calXYXtalkInvOrig = qt['calXYXtalkInv']
    
    name += ' ' + qubits[0].__name__ + ' degraded by ' + qubits[1].__name__
    
    origF10 = qubits[1]['f10']
    origF21 = qubits[1]['f21']
    nonLin = origF10-origF21
    
    origPiLen = qubits[1]['piLen']
    origPiHalfLen = qubits[1]['piHalfLen']
    origPiAmp = qubits[1]['piAmp']
    origPiHalfAmp = qubits[1]['piHalfAmp']
    
    RBClass = rb.RandomizedBenchMarkingClifford(1)    
    RBClass.setGateLengths(qubits)            
    
    axesname='m - number of Cliffords'
    if interleaved:
        name += ' interleaved: ' + str(interleaved)
        axesname='m - number of sets of Clifford+interleaved'
    axes = [(xAmp, 'X Talk Amp'),(xPhase,'X Talk Phase'),(freq,'X Talk Frequency'),(piLen,'Pi Pulse Length')]
    deps = [('Sequence Fidelity', '', '')]
    
    sequenceData = [[RBClass.randGen([iterations], interleaved=interleaved) for k in range(seq)] for q in qubits] # contains all of the gates for all of the sequences for the system    

    kw={'stats':stats, 'interleaved':interleaved, 'seq':seq, 'correctXtalkZ':correctXtalkZ,'axismode':'m','dataFormat': 'stacked'}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, currAmp, currPhase, currFreq, currPiLen):        
        
        if currFreq:
        
            qubits[1]['f10'] = currFreq
            qubits[1]['f21'] = currFreq - nonLin
        
            print currFreq
        
        if currPiLen:
            
            qubits[1]['piLen'] = currPiLen
            qubits[1]['piAmp'] = origPiAmp * origPiLen/currPiLen
            qubits[1]['piHalfLen'] = currPiLen
            qubits[1]['piHalfAmp'] = origPiHalfAmp * origPiHalfLen/currPiLen
            
            print currPiLen
            
        calXYXtalkInv = calXYXtalkInvOrig
        calXYXtalkInv[xtalkqInd] = [currAmp, currPhase]
        qt['calXYXtalkInv'] = calXYXtalkInv
        
        reqs = []        
        print 'interation %d' % iterations        
        for sequenceNum in range(seq):            
            print 'sequence number %d' % sequenceNum
            if interleaved:
                #c1 - c - c2 - c - c(m+1), m=2
                idx=iterations*2
            else:
                idx=iterations
                
            alg = gc.Algorithm(agents = devs)
            for k,q in enumerate(qubits):
                currGateList = sequenceData[k][sequenceNum][:idx]
                fullGateList = RBClass.finishGateList(currGateList)
                alg[gates.RandomizedBenchmarkingSingleQubitClifford([q],fullGateList)]
            alg[gates.Measure(alg.qubits)]            
            alg.compile(beSmart=beSmart, correctXtalkZ=correctXtalkZ, correctXtalkXY=True)            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)) 
            del currGateList
            del fullGateList            
            del alg
        results = yield FutureList(reqs)
        data = []
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(qubits[0],np.array(Is)+1.0j*np.array(Qs)))
            data.append(1-prob) # to get fidelity
        returnValue([np.mean(data)])     
    data = sweeps.grid(func, axes, dataset = save and dataset)

    if collect:
        return data           
     
def orbitMQNM(Sample, measure, 
           qubitAndParamNames = None,
           xtol = 0.001, ftol = 0.001,
           nonzdelt = 0.02, zdelt = 0.02,
           name = 'RB MQ NM',
           maxfun=None,
           settling=False, 
           shape=False,
           m = 20, k = 20, interleaved = False, 
           stats=900, correctXtalkZ=True,
           collect=False, save = True, noisy = True, syncEachClifford=True):

    if qubitAndParamNames is None:
        qubitAndParamNames = [[measure[0], measure[1]+'aczPhase'],
                              [measure[0], measure[1]+'aczAmp'],
                              [measure[1], measure[0]+'aczPhase']]
        if shape:
            qubitAndParamNames.append([measure[0], measure[1]+'aczLambda'])
        if settling:
            qubitAndParamNames.append([measure[0],'settlingRates'])
            qubitAndParamNames.append([measure[0],'settlingAmplitudes'])
            qubitAndParamNames.append([measure[1],'settlingRates'])
            qubitAndParamNames.append([measure[1],'settlingAmplitudes'])
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    param = popt.Parameters(qubitAndParamNames, devs, measure)
    axes, deps, inputs = param.makeInputsAxesDeps()
    kw = {'stats': stats, 'qubitAndParamNames':qubitAndParamNames}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  
    
    RBClass = rb.RandomizedBenchMarkingClifford(len(measure))    
    RBClass.setGateLengths(qubits)            
    
    def func(server, args):        
        currParamDict = param.argsToCurrParamDict(args)
        
        alg = gc.Algorithm(agents = devs)
        param.updateQubits(alg.qubits, currParamDict)
        fullGateList = RBClass.randGen([m], interleaved=interleaved,finish=True)
        alg[gates.RandomizedBenchmarkingMultiQubitClifford([alg.q0,alg.q1],fullGateList,syncEachClifford=syncEachClifford)]
        alg[gates.Measure(alg.qubits)]            
        alg.compile(correctXtalkZ=correctXtalkZ)     
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure, correlated = True)
        if noisy:
            print probs
        returnValue([1-np.squeeze(probs)[0]]) #error is 1-P00
        
    funcWrapper = popt.makeFuncWrapper(k, func, axes, deps, measure, sample)
    output = sweeps.fmin(funcWrapper, inputs, 
                         dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, 
                         zdelt = zdelt, maxfun=maxfun)
    
    return output
    
def randomizedBenchmarkingMQ(Sample, measure, name = 'MQ RB',
                             ms = None, steps=30, maxtime=6250, k = 20, interleaved = False, 
                             stats=900, correctXtalkZ=True,
                             plot=True, collect=False, save = True, noisy = True,syncEachClifford=True):
    '''
    Clifford randomized benchmarking, generates fully random sequences for each sub-experiment
    syntax: interleaved=[['CZ'],['CZ']]
    measure must be passed in the right order such that the adiabatic CZ is passed in the right order
    measure[0] is adiabaticCZ control
    measure[1] is adiabaticCZ target
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    RBClass = rb.RandomizedBenchMarkingClifford(len(measure))    
    RBClass.setGateLengths(qubits)            
    
    if ms == None:
        iterationsmax=rb.getlength(RBClass,maxtime,interleaved=interleaved,mstart=1)
        ms = st.ilr[1:iterationsmax:steps]._range()     

    def getSequence(m):
        sequenceData = RBClass.randGen([m], interleaved=interleaved,finish=True)
        return sequenceData
        
    axesname='m - number of Cliffords'
    if interleaved:
        name += ' interleaved: ' + str(interleaved)
        axesname='m - number of sets of Clifford+interleaved'
    axes = [(range(k),'sequence'),(ms, axesname)]
    deps = readout.genProbDeps(qubits, measure, correlated=True)   

    kw={'stats':stats, 'interleaved':interleaved, 'k':k, 'correctXtalkZ':correctXtalkZ,'axismode':'m', 'syncEachClifford': syncEachClifford, 'dataFormat': 'stacked2Dswapped'}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw) 
    
    def func(server, currK, currM):             
        alg = gc.Algorithm(agents = devs)
        fullGateList = getSequence(currM) 
        alg[gates.RandomizedBenchmarkingMultiQubitClifford([alg.q0,alg.q1],fullGateList,syncEachClifford=syncEachClifford)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ) 
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = readout.IQToProbs(data.asarray, alg.qubits, measure, correlated = True)
        returnValue(probs)        
    data = sweeps.grid(func, axes, dataset = save and dataset,pipesize=50)
            
    if collect:
        return data     

def qptPrep(Sample, measure, averages = 1, stats = 2100,
            save=True, collect=False, noisy=True,waitTime=0.0*ns,correctXtalkZ=True,tomokey='octomo',reset=False):
            
    name = 'QPT - %s qubit prep'%np.alen(measure)       
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    for qi in q:
        qi['readout']=True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]
            
    tomonames = eval('tomo.'+tomokey+'_names')               
    ops = list(tomo.gen_qpt_tomo_ops(tomonames, len(qubits)))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps=[]
    for state_bin_repr in range(2**len(measure)):
        probstr='|'+np.binary_repr(state_bin_repr,len(measure))+'>'
        deps+=[('Probability', probstr, '')]

    kw = {'stats': stats, 'waitTime':waitTime,'tomokey':tomokey,'reset':reset}         
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    
    def func(server, i):
        
        preOp, postOp = ops[i]
        
        if noisy:
            print preOp, postOp
            
        alg = gc.Algorithm(agents = devs)
        alg[gates.Tomography(qubits, preOp)]
        alg[gates.Sync(qubits)]       
        alg[gates.Wait(qubits,waitTime)]
        alg[gates.Sync(qubits)]
        alg[gates.Tomography(qubits, postOp)]
        alg[gates.Measure(alg.qubits)]            
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)      
        probs = mqReadoutProbabilities(data, devs, measure, correlated = True)
        
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
            
    if collect:
        return data       

def qptAcz(Sample, measure, control, target, name = 'QPT - Adiabatic CZ',
            averages = 1, stats = 2100,
            save=True, collect=False, noisy=True,correctXtalkZ=True):
    """Be aware of the order of things here, it needs to match with qptPrep. RB."""     
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    for qi in q:
        qi['readout'] = True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]    
    targetQubit = devs[target]
    controlQubit = devs[control]

    ops = list(tomo.gen_qpt_tomo_ops(tomo.octomo_names, 2))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps = [('Probability', '|00>', ''),('Probability', '|01>', ''),('Probability', '|10>', ''),('Probability', '|11>', '')]
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure)
    
    def func(server, i):
        
        preOp, postOp = ops[i]
        
        if noisy:
            print preOp, postOp
            
        alg = gc.Algorithm(agents = devs)
        alg[gates.Tomography(qubits, preOp)]
        alg[gates.Sync(devs)]           
        alg[gates.adiabaticCZ([controlQubit, targetQubit])]
        alg[gates.Sync(devs)]           
        alg[gates.Tomography(qubits, postOp)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, devs, measure, correlated = True)
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
            
    if collect:
        return data        

def qstPrep(Sample, measure,averages = 1, stats = 3600,
            save=True, collect=False, noisy=True,waitTime=0.0*ns,tomokey='octomo',reset=False):
            
    name = 'QST - %s qubit prep'%np.alen(measure)         
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    for qi in q:
        qi['readout']=True
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    allQubits=[]
    for q in devs:
        if q['_type']=='transmon':
            allQubits.append(q)    
    qubits = [devs[meas] for meas in measure]
    
    #ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, len(qubits)))
    tomonames = eval('tomo.'+tomokey+'_names')        
    ops = list(tomo.gen_qst_tomo_ops(tomonames, len(qubits)))    
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps=[]
    for state_bin_repr in range(2**len(measure)):
        probstr='|'+np.binary_repr(state_bin_repr,len(measure))+'>'
        deps+=[('Probability', probstr, '')]
        
    kw = {'stats': stats, 'waitTime':waitTime,'tomokey':tomokey,'reset':reset}       
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
      
    
    def func(server, i):
        
        postOp = ops[i]
        
        if noisy:
            print postOp
            
        alg = gc.Algorithm(agents = devs)
        if reset:
            alg[gates.Reset(allQubits)]
            alg[gates.Wait(allQubits,250.0*ns)]
            alg[gates.Sync(allQubits)]        
        alg[gates.Sync(qubits)]       
        alg[gates.Wait(qubits,waitTime)]
        alg[gates.Sync(qubits)]
        alg[gates.Tomography(qubits, postOp)]
        alg[gates.Sync(qubits)]
        alg[gates.Measure(alg.qubits)]            
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)

        if len(measure)==1:     
            Is,Qs = data[0]
            prob = np.mean(sq.readoutProbabilities(qubits[0],np.array(Is)+1.0j*np.array(Qs)))
            probs = [1-prob,prob]
        if len(measure)>=2:
            probs = mqReadoutProbabilities(data, devs, measure, correlated = True)            
        #sq.readoutProbabilities(q,np.array(Is)+1.0j*np.array(Qs)) 
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    if len(measure)<=2:
        if len(measure)==1:
            rho = tomo.qst(data[:,1:], tomokey) 
            ps.plotDensity(rho)            
        if len(measure)==2 and tomokey=='octomo':    
            rho = tomo.qst(data[:,1:], 'octomo2')    
            ps.plotDensity(rho)
    
    if collect:
        return data       

def qstBellState(Sample, measure, control,target, name = 'QST - Bell state',
            averages = 1, stats = 2100,
            save=True, collect=False, noisy=True,waitTime=0.0*ns,bellState='Psi+',correctXtalkZ=True):
    """Generates and does qst on the 4 bell states. RB"""  
    name += ' - '
    name += bellState
    
    # sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    # q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure] 
    # for qi in q:
        # qi['readout']=True
    # devs = gc.registryWrapper2GateCompiler(sample, qubits)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    qubits = [devs[meas] for meas in measure]
            
    ops = list(tomo.gen_qst_tomo_ops(tomo.octomo_names, len(qubits)))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    axes = [(index_gen(), 'Tomo Operation')]
    deps = [('Probability', '|00>', ''),('Probability', '|01>', ''),('Probability', '|10>', ''),('Probability', '|11>', '')]
    
    kw = {'stats': stats, 'bellState':bellState, 'waitTime':waitTime ,'correctXtalkZ':correctXtalkZ}    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure,kw=kw)
    q1=qubits[control]
    q2=qubits[target]
    def func(server, i):
        
        postOp = ops[i]
        
        if noisy:
            print postOp
            
        alg = gc.Algorithm(agents = devs)
        q1 = alg.q0
        q2 = alg.q1
        if bellState=='Psi+':
            # 01 + 10
            alg[gates.PiHalfPulse([q1])] #X_pi/2
            alg[gates.PiHalfPulse([q2])] #X_pi/2
            alg[gates.adiabaticCZ([q1,q2])]
            alg[gates.PiHalfPulse([q1])] #X_pi/2
        elif bellState=='Psi-':
            # 01 - 10
            alg[gates.PiHalfPulse([q1])] #X_pi/2
            alg[gates.PiHalfPulse([q2],phase=np.pi)] #-X_pi/2
            alg[gates.adiabaticCZ([q1,q2])]
            alg[gates.PiHalfPulse([q1])] #X_pi/2            
        elif bellState=='Phi+':
            # 00 + 11
            alg[gates.PiHalfPulse([q1])] #X_pi/2
            alg[gates.PiHalfPulse([q2])] #X_pi/2
            alg[gates.adiabaticCZ([q1,q2])]
            alg[gates.PiHalfPulse([q1],phase=np.pi)] #-X_pi/2                
        elif bellState=='Phi-':
            # 00 - 11
            alg[gates.PiHalfPulse([q1])] #X_pi/2
            alg[gates.PiHalfPulse([q2],phase=np.pi)] #-X_pi/2
            alg[gates.adiabaticCZ([q1,q2])]
            alg[gates.PiHalfPulse([q1],phase=np.pi)] #-X_pi/2
        else:
            raise Exception('not implemented')
        alg[gates.Sync([q1,q2])]       
        alg[gates.Wait([q1,q2],waitTime)]
        alg[gates.Sync([q1,q2])]
        alg[gates.Tomography([q1,q2], postOp)]
        alg[gates.Sync([q1,q2])]
        alg[gates.Measure(alg.qubits)]            
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        returnValue(probs)
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    rho = tomo.qst(data[:,1:], 'octomo2')    
    ps.plotDensity(rho)
    
    if collect:
        return data

def qstGHZState(Sample, measure, measureAct, CTpairs,CNOTpairs, name = 'QST - GHZ state',
            averages = 1, stats = 3600, useTomo=True, analyze = True, echo=False,
            thetas = np.linspace(-np.pi,np.pi,51),
            save=True, collect=False, noisy=True,waitTime=0.0*ns,correctXtalkZ=True,tomokey='octomo',reset=False):
    """Generates and does qst on the N GHZ state. 
    syntax: 
    GHZ:
    measure=[0,1,2,3,4] #these get control
    measureAct=[0,1,2,3,4] #these get actually measured
    CTpairs= [ [0,1] , [2,1], [2,3], [4,3] ] # cz control-target pairs. the control does the trajectory.
    CNOTpairs = [ [0,1], [1,2], [2,3] ,[3,4]  ] # cnot control-target pairs. +-Y/2 pulses are placed such to generate the gate.
    
    build a 2 qubit bell state:
    measure=[0,1,2,3,4] #these get control
    measureAct=[0,1] #these get actually measured
    CTpairs= [ [0,1] ] # cz control-target pairs. the control does the trajectory.
    CNOTpairs = [ [0,1]  ] # cnot control-target pairs. +-Y/2 pulses are placed such to generate the gate.
    
    RB"""  
    GHZstate = len(CTpairs)+1
    if len(CTpairs) is not len(CNOTpairs):
        raise Exception('dissimilar pair length')
    name += ' - N='+str(GHZstate)

    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure] 
    
    #only these are measured
    qm = [qubits[meas] for meas in measureAct]   
    for qi in qm:
        qi['readout']=True
        
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]
            
    tomonames = eval('tomo.'+tomokey+'_names')        
    ops = list(tomo.gen_qst_tomo_ops(tomonames, len(measureAct)))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)

    axes = [(index_gen(), 'Tomo Operation')]
    deps=[]
    for state_bin_repr in range(2**len(measureAct)):
        probstr='|'+np.binary_repr(state_bin_repr,len(measureAct))+'>'
        deps+=[('Probability', probstr, '')]
        
    kw = {'stats': stats, 'GHZstate':GHZstate, 'waitTime':waitTime ,'correctXtalkZ':correctXtalkZ,'reset':reset,'tomokey':tomokey}    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measureAct,kw=kw)

    def func(server, i):
        
        postOp = ops[i]
        
        if noisy:
            print postOp
            
        alg = gc.Algorithm(agents = devs)
        if reset:
            alg[gates.Reset(devs)]
            alg[gates.Wait(devs,50.0*ns)]
            alg[gates.Sync(devs)]        
        # GHZ GENERATION
        alg[gates.PiHalfPulse([ qubits[measureAct[0]] ],phase=np.pi/2.)] #Y_pi/2        
        for CTpair,CNOTpair in zip(CTpairs,CNOTpairs):            
            alg[gates.PiHalfPulse([ qubits[ CNOTpair[1] ] ] ,phase=-np.pi/2.)] #-Y_pi/2 on CNOT target
            alg[gates.adiabaticCZ( devs , [ qubits[CTpair[0]] , qubits[CTpair[1]] ]) ] #CHANGE ME INTO A CZ WITH DETUNING OTHERS
            alg[gates.PiHalfPulse([ qubits[ CNOTpair[1] ] ] ,phase=np.pi/2.)] #Y_pi/2 on CNOT target
        # TOMOGRAPHY
        if echo:
            alg[gates.NPiPulses([qubits[measure[0]]],10)]
            alg[gates.NPiPulses([qubits[measure[1]]],14)]
            alg[gates.NPiPulses([qubits[measure[2]]],8)]
            alg[gates.NPiPulses([qubits[measure[3]]],2)]
        for qi in measureAct:
            alg[gates.ChangePhase([qubits[qi]], qubits[qi].get(str(len(measureAct))+'GHZphase',0.0 ) )]            
        alg[gates.Tomography( [qubits[m] for m in measureAct ], postOp)]
        alg[gates.Measure(alg.qubits)]            
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, devs, measureAct, correlated = True)
        returnValue(probs)

    
    data = sweeps.grid(func, axes, dataset = save and dataset)
            
    if collect:
        return data    
    
def qstGHZStateSE(Sample, measure, measureAct, CTpairs,CNOTpairs, name = 'QST - GHZ state',
            averages = 1, stats = 3600, useTomo=True, analyze = True, echo=False,
            thetas = np.linspace(-np.pi,np.pi,51),
            save=True, collect=False, noisy=True,waitTime=0.0*ns,correctXtalkZ=True,pos=None,fulltime=None,tomokey='octomo',reset=False):
    """Generates and does qst on the N GHZ state. 
    syntax: 
    GHZ:
    measure=[0,1,2,3,4] #these get control
    measureAct=[0,1,2,3,4] #these get actually measured
    CTpairs= [ [0,1] , [2,1], [2,3], [4,3] ] # cz control-target pairs. the control does the trajectory.
    CNOTpairs = [ [0,1], [1,2], [2,3] ,[3,4]  ] # cnot control-target pairs. +-Y/2 pulses are placed such to generate the gate.
    
    build a 2 qubit bell state:
    measure=[0,1,2,3,4] #these get control
    measureAct=[0,1] #these get actually measured
    CTpairs= [ [0,1] ] # cz control-target pairs. the control does the trajectory.
    CNOTpairs = [ [0,1]  ] # cnot control-target pairs. +-Y/2 pulses are placed such to generate the gate.
    
    RB"""  
    GHZstate = len(CTpairs)+1
    if len(CTpairs) is not len(CNOTpairs):
        raise Exception('dissimilar pair length')
    name += ' - N='+str(GHZstate)
    if useTomo is False:
        name += ' PARITY'
    else:
        name += ' SE'
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    q, Q = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure] 
    
    #only these are measured
    qm = [qubits[meas] for meas in measureAct]   
    for qi in qm:
        qi['readout']=True
        
    devs = gc.registryWrapper2GateCompiler(sample, qubits)
    qubits = [devs[meas] for meas in measure]
    tomonames = eval('tomo.'+tomokey+'_names')        
    ops = list(tomo.gen_qst_tomo_ops(tomonames, len(measureAct)))
    index = range(len(ops))
    
    index_gen = st.averagedScan(index,averages,noisy=noisy)
    
    if useTomo:
        axes = [(index_gen(), 'Tomo Operation')]
        deps=[]
        for state_bin_repr in range(2**len(measureAct)):
            probstr='|'+np.binary_repr(state_bin_repr,len(measureAct))+'>'
            deps+=[('Probability', probstr, '')]
    else:
        axes = [(thetas, 'Analyzer Angle')]
        deps = [('Parity','','')]
        
    # figure out stupid parity indices
    evenInd = []
    oddInd = []
    for stateInd in range(2**len(measureAct)):
        state = np.binary_repr(stateInd,len(measureAct))
        parity = state.count('1') & 0x1
        if parity:
            oddInd.append(stateInd)
        else:
            evenInd.append(stateInd)
        
    kw = {'stats': stats, 'GHZstate':GHZstate, 'waitTime':waitTime ,'correctXtalkZ':correctXtalkZ,'pos':pos, 'fulltime':fulltime,'tomokey':tomokey,'reset':reset}    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measureAct,kw=kw)

    def func(server, i):
        
        if useTomo:
            postOp = ops[i]
            
            if noisy:
                print postOp
            
        alg = gc.Algorithm(agents = devs)
        if reset:
            alg[gates.Reset(devs)]
            alg[gates.Wait(devs,250.0*ns)]
            alg[gates.Sync(devs)]         
        # GHZ GENERATION
        alg[gates.PiHalfPulse([ qubits[measureAct[0]] ],phase=np.pi/2.)] #Y_pi/2        
        for CTpair,CNOTpair in zip(CTpairs,CNOTpairs):            
            alg[gates.PiHalfPulse([ qubits[ CNOTpair[1] ] ] ,phase=-np.pi/2.)] #-Y_pi/2 on CNOT target
            alg[gates.adiabaticCZ( devs , [ qubits[CTpair[0]] , qubits[CTpair[1]] ]) ] #CHANGE ME INTO A CZ WITH DETUNING OTHERS
            alg[gates.PiHalfPulse([ qubits[ CNOTpair[1] ] ] ,phase=np.pi/2.)] #Y_pi/2 on CNOT target
        if pos and fulltime:
            for p,mAct in zip(pos,measureAct):
                if p:
                    q=qubits[mAct] 
                    time=1.*fulltime-p
                    wait=(time-q['piLen']*2)/4
                    if wait>5.:
                        alg[gates.Wait([q],wait)]
                        alg[gates.PiPulse([q])]
                        alg[gates.Wait([q],2.*wait)] 
                        alg[gates.PiPulse([q])]
                        alg[gates.Wait([q],wait)]            
        # TOMOGRAPHY
        if echo:
            alg[gates.NPiPulses([qubits[measure[0]]],10)]
            alg[gates.NPiPulses([qubits[measure[1]]],14)]
            alg[gates.NPiPulses([qubits[measure[2]]],8)]
            alg[gates.NPiPulses([qubits[measure[3]]],2)]
        for qi in measureAct:
            alg[gates.ChangePhase([qubits[qi]], qubits[qi].get(str(len(measureAct))+'GHZphase',0.0 ) )]
        if useTomo:
            alg[gates.Tomography( [qubits[m] for m in measureAct ], postOp)]
        else:
            for qi in measureAct:
                alg[gates.PiHalfPulse([qubits[qi]], phase=i)]
        alg[gates.Measure(alg.qubits)]            
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mqReadoutProbabilities(data, devs, measureAct, correlated = True)
        if useTomo:
            returnValue(probs)
        else:
            probs = np.array(probs)
            evenParity = sum(probs[evenInd])
            oddParity = sum(probs[oddInd])
            returnValue([evenParity-oddParity])
    
    data = sweeps.grid(func, axes, dataset = save and dataset)
    
    if analyze and not useTomo:
        t = data[:,0]
        p = data[:,1]
        
        p = p-np.mean(p)
        positive = np.mean(p[np.argwhere(p>0)])
        negative = np.mean(p[np.argwhere(p<0)])
        
        amplitude = positive-negative
        print amplitude
            
    if collect:
        return data    

def gateTester(Sample, measure, stats=900):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    alg = gc.Algorithm(agents=devs)
    alg[gates.Measure([alg.q1], name = 'Herald')]
    alg[gates.Wait([alg.q1], 10*ns)]
    alg[gates.PiPulse([alg.q1])]
    alg[gates.PiPulse([alg.q0])]
    alg[gates.PiPulse([alg.q2])]
    alg[gates.adiabaticCZ([alg.q1,alg.q0])]
    alg[gates.PiPulse([alg.q1])]
    alg[gates.PiPulse([alg.q0])]
    alg[gates.PiPulse([alg.q2])]
    alg[gates.Measure(alg.qubits, name = 'Final')]
    alg.compile()
    
    ph.pulses(alg,spacing=-2, timeStep = 0.1, spacingByQubit=False)
    
    # rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats=stats, dataFormat='iqRaw', measure = measure, debug = False)
    # data= rv.wait()
    # probs = mqReadoutProbabilities(data, alg.qubits, measure, correlated = False)
    # return data,probs
    return alg
    
def minPiLens(Sample, measure, minLen = 12*ns, maxAmp = 1.1):
    
    sample, qubits, Qubits = util.loadQubits(Sample, write_access=True)
    qubits, Qubits = [qubits[meas] for meas in measure], [Qubits[meas] for meas in measure]
    
    for q,Q in zip(qubits, Qubits):
        piLen = q['piLen']
        piAmp = q['piAmp']
        piHalfLen = q['piHalfLen']
        piHalfAmp = q['piHalfAmp']
        
        newPiLen = round(piAmp/maxAmp * piLen,1)*ns
        newPiAmp = piAmp * piLen/newPiLen
        if newPiLen < minLen:
            newPiLen = minLen
            newPiAmp = newPiAmp = piAmp * piLen/newPiLen
        
        newPiHalfLen = round(piHalfAmp/maxAmp * piHalfLen,1)*ns
        newPiAmp = piAmp * piLen/newPiLen
        if newPiHalfLen < minLen:
            newPiHalfLen = minLen
            newPiHalfAmp = piHalfAmp * piHalfLen/newPiHalfLen
        
        Q['piLen'] = newPiLen
        Q['piAmp'] = newPiAmp
        Q['piHalfLen'] = newPiHalfLen
        Q['piHalfAmp'] = newPiHalfAmp