# Author: Daniel Sank
# Created: 2011

# CHANGELOG
#

import sys
import time
import numpy as np
from msvcrt import getch, kbhit
import matplotlib.pyplot as plt

import labrad
from labrad.units import Unit, Value
s,Hz,kHz,MHz = (Unit(u) for u in ['s','Hz','kHz','MHz'])

import pyle.registry
import pyle.dataking.util as dataUtil
import pyle.util.sweeptools as st

######################
### Gain of SR445A ###
######################

def gain445A(cxn, channels):
    """Measure the gain of the SR445A vs frequency using GHz DAC as source
    
    This function has not been tested yet.
    """
    #<dac = load DAC object using fpgaTest stuff>
    rmsNoAmp = np.array([])
    rmsAmp = np.array([])
    #Set up scope
    scope = cxn[dac['oscilloscopeServer']]
    scope.select_device(dac['oscilloscopeId'])
    scope.reset()
    scope.termination(1, 50)
    scope.measure_type(1, 'RMS')
    scope.coupling(1, 'AC')
    print 'INSTRUCTIONS: Set up measurements on ch1. Set scope to RMS meaure'
    #Get amplitude vs frequency without the amplifier
    scope.scale(2, np.sqrt(2)*Value(260,'mV')*amp/3.0)
    for freq in freqs:
        dac['signalFrequency'] = [freq,freq]
        fpgaTest.dacSignal(sample, cxn.ghz_fpgas, reps=30, loop=True, getTimingData=False, trigger=None, dacs=[dac])
        scope.horiz_scale(Value(0.5/freq)) #5 cycles on screen
        time.sleep(3)
        rmsNoAmp = np.hstack((rmsNoAmp, scope.measure_value(1)))
    #Get amplitude vs frequency with amplifier
    print 'INSTRUCTIONS: Set up measurement on ch1 through amplifier'
    scope.scale(2, 5*channels*np.sqrt(2)*Value(260,'mV')*amp/3.0)
    for freq in freqs:
        dac['signalFrequency'] = [freq,freq]
        fpgaTest.dacSignal(sample, cxn.ghz_fpgas, reps=30, loop=True, getTimingData=False, trigger=None, dacs=[dac])
        scope.horiz_scale(Value(0.5/freq)) #5 cycles on screen
        time.sleep(3)
        rmsAmp = np.hstack((rmsAmp, scope.measure_value(1)))
    gain = rmsAmp/rmsNoAmp
    return np.vstack((freqs,gain)).T
    
##############################
### Fastbias low frequency ###
##############################

#This provides a useful scan over the range accessible by the SR770
           #span        start       ave,  overlap
RANGES = [(50*kHz,      50*kHz,    1000,   0.0),
          (50*kHz,      30*kHz,    1000,   0.0),
          (50*kHz,      5*kHz,     1000,   0.0),
          (10*kHz,      10*kHz,     100,   0.0),
          (10*kHz,      5*kHz,      100,   0.0),
          (10*kHz,      1*kHz,      100,   0.0),
          (1.1*kHz,     1100*Hz,     20,   0.0),
          (1.1*kHz,     500*Hz,      20,   0.0),
          (1.1*kHz,     100*Hz,      20,   0.0),
          (100*Hz,      100*Hz,      10,   0.0),
          (100*Hz,      50*Hz,       10,   0.0),
          (100*Hz,      10*Hz,       10,   0.0),
          (10*Hz,       10*Hz,        5,   0.0),
          (10*Hz,        5*Hz,        5,   0.0),
          (10*Hz,        1*Hz,        5,   0.0),
          #(1*Hz,        0.5*Hz,     20,    0.0),
          #(1*Hz,        0.1*Hz,     20,    0.0),
          #(0.2*Hz,      0.01*Hz,    20,     90.0),
          ]

def doDatasets(s, cxn, save, recipient=None, verbose=False):
    datas = []
    if recipient is None:
        recipient = s._dir[1]
    sample,devices = dataUtil.loadDevices(s)
    for key, dev in devices.items():
        if '_model' in dev and dev['_model'] in ('SR770', 'SR780'):
            analyzer = dev
            break
    else:
        raise ValueError("No spec analyzer found in devices.")
    dataDir = s._dir
    for frequencySpan,frequencyStart,averageNum,averageOverlap in RANGES:
        analyzer['frequencySpan']=frequencySpan
        analyzer['frequencyStart']=frequencyStart
        analyzer['averageNum']=averageNum
        analyzer['averageOverlap']=averageOverlap
        currData = getSpectralAmplitude(cxn, sample, devices, analyzer, dataDir, save=save)
        datas.append(currData)
        if verbose:
            try:
                cxn.telecomm_server.send_sms('noise test', 'done with span=%s' % frequencySpan, recipient)
            except:
                print 'telecomm failed'
        else:
            print 'done with span=%s' % frequencySpan
    try:
        cxn.telecomm_server.send_sms('noise test', 'measurement complete', recipient)
    except:
        print 'Text message failed'
    return np.vstack(datas)
    
def getSpectralAmplitude(cxn, sample, devices, analyzer, dataDir, save=True):
    analyzerServer = cxn.servers['Signal Analyzer ' + analyzer['_model']]
    #Set up the device and wait for settling
    setupSR770(analyzer, analyzerServer)
    #Acquire spectral density, will wait for averaging to complete
    t = time.time()
    print "Taking data... ",
    data = analyzerServer.power_spectral_amplitude_ASCII(0).asarray
    print "%.2f minutes for dataking." % ((time.time() - t)/60.0,)
    #Save to datavault
    indeps = [('Frequency','Hz')]
    deps = [('Spectral amplitude','','V/Hz^1/2')]
    if save is True:
        makeDataset(cxn, sample, devices, indeps, deps, data, dataDir, 
                    extraName = ' - %s' % analyzer['frequencySpan'])
    return data
    
def setupSR770(device, server):
    """Sets up the SR770 and updates the device object"""
    input = [
               ('inputRange','input_range', lambda x:x),
               ('inputCoupling','coupling', lambda x:x),
               ('inputGrounding','grounding', lambda x:x)]
    measure = [
               ('measureDisplay','display', lambda x:x[1]),
               ('measureType','measure', lambda x:x[1]),
               ('measureUnits','units', lambda x:x[1]),
               ('measureWindow','window', lambda x:x)]
    average = [
               ('average','average', lambda x:x),
               ('averageNum','num_averages', lambda x:x),
               ('averageOverlap','overlap', lambda x:x)]
    p = server.packet()
    p.select_device(device['_addr'])
    trace = device['trace']
    #Input setup
    for param, setting, _ in input:
        p[setting](device[param], key=param)
    #Measurement settings
    for param, setting, _ in measure:
        p[setting](trace, device[param], key=param)
    #Go to desired span and start frequency, and wait for settling
    if 'minimumSettle' in device:
        ms = device['minimumSettle']
    else:
        ms = 0*s
    p.freq_and_settle(device['frequencySpan'],device['frequencyStart'], ms, key='spanAndStart')
    
    #Averaging setup
    for param, setting, _ in average:
        p[setting](device[param], key=param)
    print "Settling... ",
    t = time.time()
    result = p.send()
    print " %.2f minutes to configure and settle at freq %s" % ((time.time()-t)/60.0, device['frequencySpan'])
    for param, setting, func in input+measure+average:
        device[param] = func(result[param])
    device['frequencySpan'] = result['spanAndStart'][0]
    device['frequencyStart'] = result['spanAndStart'][1]
    
def makeDataset(cxn, sample, devices, independents, dependents, data, dataDir, extraName = ""):
    name = sample['dut']['name'] + extraName
    parameters = parseSampleParameters(sample)
    if name.upper() in ['DAC', 'FPGA']:
        parameters.extend(getDACParameters(cxn))

    for device,parms in devices.items():
        for key,value in devices[device].items():
            parameters.append((device+'.'+key,value))
    dv = cxn.data_vault
    dv.cd('')
    dv.cd(dataDir, True)
    dv.new(name, independents, dependents)
    p = dv.packet()
    for parameter in parameters:
        p.add_parameter(parameter)
    p.send()
    dv.add(data)
    
def parseSampleParameters(sample): #this is a total hack
    parameters=[]
    for key,value in sample.items():
        if isinstance(value,dict):
            continue
        else:
            parameters.append((key,value))
    return parameters
    
def getDACParameters(cxn): # even more of a hack
    c = cxn.registry.context()
    p = cxn.registry.packet(context=c)
    p.cd(['', 'TestUser', 'fpgaTest', 'runningman DAC 21'], context=c)
    p.dir(context=c)
    a = p.send(context=c)
    p = cxn.registry.packet(context=c)
    for k in a.dir[1]:
        p.get(k)
    b = p.send()
    return [('dac.'+k, v) for k, v in zip(a.dir[1], b.get)]
    
###############################
### Fastbias high frequency ###
###############################

def fastbiasHighFreq(cxn):
    pass
    #<Set spectrum analyzer>
    #<10MHz to 300MHz>
    #<BW = 10KHz>
    #<Trace average mode with some number of averages>
    
    
    
#Measure DAC+Diff amp at low frequency with SR770 and SR560

#Measure DAC+Diff amp noise at high frequency with spectrum analyzer and SR445A

