import numpy as np
#import matplotlib.pyplot as plt
#import time

def interpol1d_cubic(x,h):
    """Fast cubic interpolator. Returns a function.
    Uses linear interpolation at the edges, and returns the values at the edges outside of the range. x needs to be equidistant. It can be monotonically rising or lowering. RB."""    
    def func(xdet):
        dx=x[1]-x[0]
        if type(xdet) is not list and type(xdet) is not np.ndarray:
            xdet=np.array(xdet)
        yout=np.linspace(0,0,len(xdet))
        xmin=min(x)
        xmax=max(x)
        for j,xdetelement in enumerate(xdet):
            #find idx            
            if xdetelement>=xmax:
                xdetelement=xmax
                if dx>0:
                    idx=len(x)-1
                else:
                    idx=0
                yout[j]=h[idx]
            elif xdetelement<=xmin:
                xdetelement=xmin
                if dx>0:
                    idx=0
                else:
                    idx=len(x)-1
                yout[j]=h[idx]
            else:
                xstart=x[0]
                xend=x[-1]
                idx = int(np.fix((xdetelement - xstart) / dx))
                if idx>=1 and (idx+2)<len(x):
                    #cubic interpol                
                    hp2=h[idx+2]
                    hp1=h[idx+1]
                    hp0=h[idx]
                    hm1=h[idx-1]     
                    d=hp0
                    c=(hp1-hm1)/2
                    b=(-hp2+4*hp1-5*hp0+2*hm1)/2
                    a=(hp2-3*hp1+3*hp0-hm1)/2
                    xi=(xdetelement-x[idx])/dx
                    yout[j]=a*xi**3 + b*xi**2 + c*xi+d
                else:
                    #linear interpol
                    if idx==(len(x)-1):
                        direction=-1
                    else:
                        direction=1
                    x1=x[idx]
                    x2=x[idx+direction]
                    h1=h[idx]
                    h2=h[idx+direction]
                    yout[j]=(h2-h1)/(x2-x1)*(xdetelement-x1)+h1
        return yout            
    return func
            
"""            
x=np.linspace(0,-2,40)
y=np.sin(6.23*x)*x**2
x2=np.linspace(-4,4,2001)
t=time.time()
y2=interpol1d_cubic(x,y)(x2)
print time.time()-t
plt.figure()
plt.plot(x,y,'k.',x2,y2) 
plt.show()   
"""