#!/usr/bin/python
'''

insterspersed_visualization.py

'''

import numpy as np

import numpy_states as nps
import dephasing_funcs_np as dfn

import pyle.plotting.qutip as q

def p(s):
    print "{: .8f} + {: .8f}i   {: .8f} + {: .8f}i".format(s[0].real, s[0].imag, s[1].real, s[1].imag)

pi = np.pi

def visualize_gate(gate, phi_g=1e-1, start_state=None, interspersed=10, verbose=False, show=True, start_state_2=None):
    x = float(interspersed)
    if start_state is None:
        state1 = nps.ground()
    else:
        state1 = start_state.copy()
    if start_state_2 is None:
        state2 = nps.ground()
    else:
        state2 = start_state_2.copy()
    noise = nps.rotz(phi_g/x)
    noise_full = nps.rotz(phi_g)

    inter_states1 = []
    inter_states2 = []

    for i in range(int(x)):
        state1 = np.dot(gate, state1)
        state1 = np.dot(noise, state1)
        inter_states1.append(state1.copy())
        state2 = np.dot(gate, state2)
        inter_states2.append(state2.copy())
        if verbose:
            print
            p(state1)
            p(state2)

    state2 = np.dot(noise_full, state2)
    inter_states2.append(state2.copy())
    if verbose:
        p(state2)

    b = q.prep_bloch()
    b.add_states([q.from_nps(x) for x in inter_states1])
    p1x, p1y, p1z = [x[0] for x in b.vectors], [x[1] for x in b.vectors], [x[2] for x in b.vectors]
    b.clear()
    b.add_states([q.from_nps(x) for x in inter_states2])
    p2x, p2y, p2z = [x[0] for x in b.vectors], [x[1] for x in b.vectors], [x[2] for x in b.vectors]
    b.clear()
    b.add_points([p1x, p1y, p1z])
    b.add_points([p2x, p2y, p2z])
    print "Distance 1->2: %s" % np.sqrt((p1x[-1]-p2x[-1])**2 + (p1y[-1]-p2y[-1])**2 + (p1z[-1]-p2z[-1])**2)
    print "Noise distance 1: %s" % np.sqrt((p1x[-1]-p2x[-2])**2 + (p1y[-1]-p2y[-2])**2 + (p1z[-1]-p2z[-2])**2)
    print "Noise distance 2: %s" % np.sqrt((p2x[-1]-p2x[-2])**2 + (p2y[-1]-p2y[-2])**2 + (p2z[-1]-p2z[-2])**2)
    if show:
        b.show()
    return b, state1, state2
    
def visualize_two_gates(gates, phi_g=1e-1, start_state=None, interspersed=10, gates_inv=None):
    print "First gate:"
    b1, state1, state2 = visualize_gate(gates[0], phi_g=phi_g, start_state=start_state, interspersed=interspersed, show=False)
    print "Second gate:"
    b2, state1, state2 = visualize_gate(gates[1], phi_g=phi_g, start_state=state1, start_state_2=state2, interspersed=interspersed, show=False)
    b2.add_points(b1.points[0])
    b2.add_points(b1.points[1])
    b2.point_color = ['b', 'r']
    if gates_inv:
        f1 = np.dot(gates_inv[1], np.dot(gates_inv[0], state1))
        f2 = np.dot(gates_inv[1], np.dot(gates_inv[0], state2))
        print "Error: "
        print "Blue: %s" % (nps.measure(f1))
        print "Red: %s" % (nps.measure(f2))
    b2.show()
    return b2
