import numpy as np
import numpy.linalg as LA

dtype = np.complex128
rtol, atol = 1e-10, 1e-15

pi = np.pi
I = np.array([[1,0],[0,1]], dtype=dtype)
X = np.array([[0,1],[1,0]], dtype=dtype)
Y = np.array([[0,-1j],[1j,0]], dtype=dtype)
Z = np.array([[1,0],[0,-1]], dtype=dtype)
def rotx(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*X
def roty(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Y
def rotz(x):
    return np.cos(x/2.0)*I - 1j*np.sin(x/2.0)*Z

def ground():
    return np.array([1,0], dtype=dtype)

def excited():
    return np.array([0,1], dtype=dtype)
    
def norm(state):
    return LA.norm(state)
    
def renorm(state):
    return state / norm(state)
    
def measure(state):
    return (np.conj(state[1]) * state[1]).real

def reduce(state):
    ''' factor out global phase factor to make the |0> coefficient
    real and non-negative.
    also, if the state is close to pure |0> or |1>, returns |0> or |1>.
    comparison made with np.allclose. '''
    if np.allclose(np.abs(state[0]), 0, rtol=rtol, atol=rtol):
        return excited()
    if np.allclose(np.abs(state[1]), 0, rtol=rtol, atol=rtol):
        return ground()
    return np.array([np.abs(state[0]), np.exp(-1j*np.angle(state[0])) * state[1]], dtype=dtype)

def equal(a, b):
    return np.allclose(np.dot(reduce(a.conj()), reduce(b)), 1, rtol=rtol, atol=atol)
    #return np.allclose(reduce(a), reduce(b))
    
def opposite(a, b):
    return np.allclose(np.dot(reduce(a.conj()), reduce(b)), 0, rtol=rtol, atol=atol)
    
def dotMany(a, b):
    for g in a:
        b = np.dot(g, b)
    return b
    
def angle_between(a, b):
    return np.arccos(np.dot(a.conj(), b))
    
equator = [renorm(ground() - 1j*excited()), renorm(ground() + excited()),
           renorm(ground() - excited()), renorm(ground() + 1j*excited()),]
octagon = equator + [ground(), excited()]
pigates = [rotx(pi), roty(pi), rotz(pi)]
pigates_inv = [rotx(-pi), roty(-pi), rotz(-pi)]
pi2gates = [rotx(pi/2), roty(pi/2), rotz(pi/2)]
pi2gates_inv = [rotx(-pi/2), roty(-pi/2), rotz(-pi/2)]

