# davidutil.py
# very common utilities for david's scripts
# includes error logging

# author    : david low
# date      : 120816
# tested    : 120816

import time
import os

LOCOFLOG = "./log/"
LOGEXTN  = ".log"
TOTALLOG = "completeLog"
ERRORLOG = "errorLog"
ERRORMSG = "ERR from "
TOTALMSG = "MSG from "
SPACING  = "    "

def safePyle():
    import sys
    import re
    syspath = sys.path
    hasPylePath = 0
    for entry in syspath:
        match = re.search(r"^N:\\dev\\pyle", entry)
        if (match != None):
            hasPylePath = 1
            break
    if (hasPylePath != 1):
        sys.path.append("N:\\dev\\pyle")
        print "Pyle Not found, adding Pyle"
    #print sys.path
    
def getTimes(): 
    ''' Get the current Time and returns [time since eppoc, human time]
        in 1 sec precision
    '''
    startTime = time.time()
    tmptime = time.localtime()
    startTimeH = ''.join(
                    [str(tmptime.tm_year % 100),
                     str(tmptime.tm_mon).zfill(2),
                     str(tmptime.tm_mday).zfill(2),
                     '_',
                     str(tmptime.tm_hour).zfill(2), ":",
                     str(tmptime.tm_min).zfill(2),":",
                     str(tmptime.tm_sec).zfill(2)])
    return [startTime, startTimeH]

def stopLog():
    ''' Stops the error logs, meaning it saves them in their own folder based on
        time
    '''
    time = getRoughTime()
    dir  = ''.join([LOCOFLOG,str(time),'/'])
    if not os.path.exists(dir):
        os.makedirs(dir)
    alllogs = os.listdir(LOCOFLOG)
    print str(alllogs)
    for file in alllogs:
        if file.find(LOGEXTN) == -1:
            continue
        else:
            new = ''.join([dir, file])
            print file
            print new
            os.rename(''.join([LOCOFLOG,file]),new)
        
def getRoughTime():
    ''' returns an approximate time in human readable form (1min precision)
    '''
    tmptime = time.localtime()
    return ''.join(
                [str(tmptime.tm_year % 100),
                 str(tmptime.tm_mon).zfill(2),
                 str(tmptime.tm_mday).zfill(2),
                 '_',
                 str(tmptime.tm_hour).zfill(2),
                 str(tmptime.tm_min).zfill(2)
                ])

def printErr(msg,lognames, source):
    ''' void printErr (string, array of string, string)
        simply prints to files with appropriate stuffs
        also prints Once to STDOUT
    '''
    msg = msg.replace("\n",''.join(["\n",SPACING]))
    time = getTimes()[1]
    for log in lognames:
        try:
            file = open(''.join(["./log/",str(log),'.log']),"a")
            file.write(''.join([
                                '\n\n', 
                                str(time), ': ', 
                                ERRORMSG, " ",
                                source, ':\n'
                               ])
                       )    
            file.write(''.join([SPACING,msg]))
            file.close()
        except Exception, e:
            print ''.join(["Cannot write to file: ", log, ". Error: ", str(e)])
    if ((TOTALLOG not in lognames) or 
        (ERRORLOG not in lognames)      
        ):
        printErr(msg,[TOTALLOG,ERRORLOG],source)
        print msg

def printLog(msg, lognames, source):
    ''' void printLog (string, array of String, string)
        prints to files with appropriate stuff
        also prints once to STDOUT
    '''
    msg = msg.replace("\n",''.join(["\n",SPACING]))
    time = getTimes()[1]

    for log in lognames:
        try:
            file = open(''.join(["./log/",str(log),'.log']),"a")
            file.write(''.join(['\n\n', 
                                str(time), ': ', 
                                TOTALMSG, " ",
                                source, ':\n'
                               ])
                       )    
            file.write(''.join([SPACING,msg]))
            file.close()
        except Exception, e:
            print ''.join(["Cannot write to file: ", log, ". Error: ", str(e)])
    if ( (TOTALLOG not in lognames) ):
        printLog(msg,[TOTALLOG],source)
        print msg

# TODO: ctrl+c handling


def freeze(cls):
    '''class freeze (class):
    Prevents someone external from setting new variables inside of the given 
    class.  Called as a descriptor
    '''
    _init = cls.__init__
    def init(self, *args, **kw):
        _init(self, *args, **kw)
        self.freeze = True
    cls.__init__ = init 

    def _setattr(self, attr, value):
        if getattr(self, "freeze", None) and \
           (attr=="freeze" or not hasattr(self, attr)
           ):
            raise AttributeError("Thou shall not set attributes!")
        super(cls, self).__setattr__(attr, value)
    cls.__setattr__ = _setattr

    return cls
       
       
@freeze
class Fds(object): 
    ''' 
        Fds: An object out of convenience.  
        It's sole purpose is to hold lots of variables that one can reference 
        by name.  In this way, we can pass all of the variables without a long 
        list / array of inputs as parameters.  It has the exact functionality
        of a c-struct.
        
        With the freeze descriptor, it is immutable in some sense.  One cannot
        add new variable names to it.  For instance, "self.newData = True" is 
        valid, while "self.newDeta = True" will give an error.  This helps 
        prevent typos in the code where one can update a variable using the 
        wrong name (ie mDataCaled vs mDataCal) but not realize the mistake. 
        This would cause very annoying run time errors that may not be obvious.
        
        If you want to add more variables to Fds, enter them inside Fds'
        __init__ method as shown below.  Adding them any other way should now
        work.
    '''
    def __init__(self, sample, funct, time):
        self.newData           = False
        self.isComplete        = True
        self.mCalInt           = None
        self.pCalInt           = None
        self.guessInv          = []
        self.guessReg          = []
        self.fData             = []
        self.mData             = []
        self.pData             = []
        self.mDataCaled        = []
        self.pDataCaled        = []
        self.mDataCaledInv     = []
        self.pDataCaledInv     = []
        self.mDataCaledRegdB   = []
        self.mDataCaledInvdB   = []
        self.covarInv          = []
        self.covarReg          = []
        self.mFitInvdB         = []
        self.pFitInv           = []
        self.mFitRegdB         = []
        self.pFitReg           = []
        self.pwr_fitParams_Inv = []
        self.pwr_fitParams_Reg = []
        self.fitdir            = [''.join(['Res',str(sample['resnum']).zfill(2)])]
        self.reAvg             = None
        self.imAvg             = None
        self.currentAveCtr     = 0
        self.numPhoInv         = 1e6
        self.QiErrorInv        = 1e10
        self.QiSigmaInv        = 1e10
        self.numPhoReg         = 1e6
        self.QiErrorReg        = 1e10
        self.QiSigmaReg        = 1e10
        self.func              = funct
        self.startTime         = time
        self.dataNum           = -1
    
