from numpy import *

def photonNumber(Qi, Qc, f0, power):
	x=f0
	y0=-89.61545
	A1=41.117
	t1=33205616613.59624
	A2=5.24706
	t2=989193583.18749
	A3=	41.33906
	t3=	33220921395.62012
	y=A1*exp(-x/t1) + A2*exp(-x/t2) + A3*exp(-x/t3) + y0
	y=y/2
	y=-y

	cableAtt = -1*abs(y)
	otherAtt = power
	PAC = cableAtt + otherAtt
	Qi = float(Qi)
	Qc = float(Qc)
	f0 = float(f0)
	PAC = float(PAC)
	Ql = (Qi**-1 + Qc**-1)**-1
	
	capacitance = 147e-12
	length = (3e8/sqrt(5.7)/f0/4)
	plankConstant = 6.63*1e-34
	
	powerInternal = (2/pi)*(Ql**2 / Qc)*10**(PAC/10)*.001		# Ramis Thesis eq 3.30
	voltageResonator = 2*(50*powerInternal)**.5						
	energyInternal = .5*capacitance*voltageResonator**2*length
	numberPhotons = energyInternal / (plankConstant * f0)
	
	return numberPhotons

def calibrate_data(mCalInt, pCalInt, fData, mData, pData):
	reData = [mData[i]*cos(pData[i]) for i in range(len(mData))]
	imData = [mData[i]*sin(pData[i]) for i in range(len(mData))]
	complex_data = [reData[i] + 1j*imData[i] for i in range(len(fData))]
	
	reFit = [mCalInt(x)*cos(pCalInt(x)) for x in fData]
	imFit = [mCalInt(x)*sin(pCalInt(x)) for x in fData]
	complex_fit = [reFit[i] + 1j*imFit[i] for i in range(len(fData))]
	
	complex_calibrated = [complex_data[i] / complex_fit[i] for i in range(len(fData))]
	
	mDataCal = [abs(x) for x in complex_calibrated]
	pDataCal = [angle(x) for x in complex_calibrated]
	
	return mDataCal, pDataCal
	
def invert_data(mData, pData):
	mDataInv = [x**-1 for x in mData]
	pDataInv = [x*-1 for x in pData]
	
	return mDataInv, pDataInv

def phase_correction(pData):
	changed = 0
	for i in range(len(pData))[:-1]:
		if abs(pData[i] - pData[i+1]) >= pi and pData[i] < pData[i+1]:
			pData[i+1] = pData[i+1] - 2*pi
			changed = 1
		if abs(pData[i] - pData[i+1]) >= pi and pData[i] > pData[i+1]:
			pData[i+1] = pData[i+1] + 2*pi
			changed = 1

	return list(pData), changed

def get_dataset(cxn, dvNum, dvDir):
    dv = cxn.data_vault
    dv.cd(dvDir)
    dv.open(dvNum)
    
    # Get Data In Chunks #
    data = []
    j=[]
    while j != [[]]:
        j = list(dv.get(500))
        data = data + j
    data = data[:-1]    
    # Get Data In Chunks #
    
    fData = [x[0]*1e9 for x in data]
    mData = [10**(x[1]/20.0) for x in data]
    pData = [x[2] for x in data]

    return fData, mData, pData
    
def ConvertMP(mData, pData):
    reData = [mData[i]*cos(pData[i]) for i in range(len(mData))]
    imData = [mData[i]*sin(pData[i]) for i in range(len(mData))]
    
    return reData, imData
 
def ConvertRI(realPart, imagPart):
    mData = [(realPart[i]**2 + imagPart[i]**2)**.5 for i in range(len(realPart))]
    pData = [arctan2(imagPart[i], realPart[i]) for i in range(len(realPart))]
    
    return mData, pData