'''

FluxNoiseAnalysis/live.py

This code is for analyzing RTO/calibration data as it comes in, for tune-up.
'''

def getRTOdatasets(s):
    if 'analysis' in s._dir:
        dv = s._cxn.data_vault
        ctx = dv.context()
        dv.cd(s['path'])
        d = dv.dir()[1]
    else:
        d = s._cxn.data_vault.dir()[1]
    named = [x for x in d if 'RTO Sweep' in x]
    times = [int(x.rpartition(':')[2][:-2]) for x in named]
    inds = [int(x.partition('-')[0]) for x in named]
    # only get contiguous datasets
    inds.reverse()
    keeps = []
    for i in inds:
        if not keeps:
            keeps.append(i)
        elif keeps[-1] - i == 1:
            keeps.append(i)
        else:
            break
    if not keeps:
        raise Exception("Could not find any RTO Sweep datasets")
    print "using these: %s" % keeps
    return keeps
    
def plotRTOtomo(s, datasets=None):
    if not datasets:
        datasets = getRTOdatasets(s)
    for ds in datasets:
        #plt.figure()
        d = getDataset(s._cxn.data_vault, ds, [])
        plotTomography(d)
        plt.title(d.parameters.holdTime)
        
def plotRTOtomo2(s, datasets=None):
    COLORS = ['red', 'green', 'blue', 'cyan', 'magenta', 'goldenrod', 'black', 'Brown', 'Grey', 'OliveDrab', 'DarkCyan', 'DarkMagenta', 'MediumSlateBlue']
    if not datasets:
        datasets = getRTOdatasets(s)
        datasets.reverse()
    plt.figure()
    plt.xlim([-1,1])
    plt.ylim([-1,1])
    i = 0
    for ds in datasets:
        d = getDataset(s._cxn.data_vault, ds, [])
        plt.plot(2*d.data[:,1]-1, 2*d.data[:,2]-1, '.', alpha=0.5, color=COLORS[i%len(COLORS)])
        cx, cy = (2*d.data[:,1]-1).mean(), (2*d.data[:,2]-1).mean()
        plt.text(cx, cy, str(d['parameters']['holdTime']), horizontalalignment='center', verticalalignment='center')
        i += 1
    
def plotRTOsweep(s, datasets=None, plot=True, plotUnwrap=True):
    ''' plot std dev vs hold time from RTO sweep datasets. '''
    if not datasets:
        datasets = getRTOdatasets(s)
    # now plot the datasets
    r = [checkPhases(s, ds=ds, returnPhases=True, returnTime=True, plot=plot) for ds in datasets]
    plt.figure()
    plt.plot([p[1] for p in r], [np.std(p[0]) for p in r], 'o')
    plt.xlabel('Hold Time [ns]')
    plt.ylabel(r'STD of $\Delta \theta$')
    if plotUnwrap:
        [checkUnwrap(s, x, title=p[1]) for x, p in zip(datasets, r)]
