from scipy.misc import factorial
from scipy.special import laguerre
from numpy import abs, ones, sqrt, exp, conjugate, arange, shape, reshape, trace, dot, asarray, size, newaxis, transpose, zeros, pi, real,imag, outer, inner, nonzero, identity, diag, min,  max, linspace, polyval, resize, sign, array
from pylab import iterable, imshow, xlabel, ylabel, show, draw, clf, colorbar, gca, axes
from numpy.linalg.linalg import lstsq
from numpy.linalg.linalg import svd
from myPlotUtils import mycolormap
import pdb


def complexgrid(r,n):
    alpha = linspace(-r,r,n,endpoint=True)
    return alpha[newaxis,:]+ 1.0j * alpha[:,newaxis]

def leastsqinv(m):
    """calculate the inverse of a (not necessarily square) matrix M in the
    least square sense via sigular value decomposition.  Returns M1,
    M2 such that M1*M = I and M*M2 = I. M1 and M2 are hermitian
    conjugates"""
    u,s,vh = svd(m,full_matrices=0)
    s = conjugate(1/s)
    u = dot(u * s[newaxis,:], vh)
    return conjugate(transpose(u)), u


def Hconjugate(m):
    """Hermitian conjugate. If m has more than 2 dimensions, transpose the
       last 2."""
    perm = arange(len(shape(m)))
    perm[-2:] = perm[-1], perm[-2]
    return transpose(conjugate(m),perm)

def packmatrix(N):
    """Helper function for Hpack and Hunpack"""
    d = arange(N)
    d = d[:,None] - d[None,:]
    return 1.0 * (d==0) + sqrt(2) * (d>0) + 1j*sqrt(2) * (d<0)

def Hpack(rho):
    """Transform a hermitian matrix into a real matrix"""
    s = shape(rho)
    l = size(rho)
    n = arange(s[-2])
    m = arange(s[-1])
    rho = reshape(rho,(l/s[-2]/s[-1],s[-2],s[-1]))
    if len(s) == 2:
        ns = (l,)
    else:
        ns = (l/s[-2]/s[-1],s[-2]*s[-1])
    return reshape((rho*conjugate(packmatrix(s[-1])[None,:,:])).real, ns)

def Hunpack(rho):
    """Transform a real matrix into a hermition matrix,
    undoes what Hpack does"""
    s = shape(rho)
    l = size(rho)
    N = int(round(sqrt(s[-1])))
    ns = s[:-1] + (N,N)
    rho = reshape(rho,(l/(N**2),N,N))
    n = arange(N)
    rho = rho * packmatrix(N)[None,:,:]
    return reshape(0.5*(rho + Hconjugate(rho)),ns)

    

def displacement(alpha,N=10,parity=False):
    """Calculate the displacement operator for amplitude alpha in
    the Fock basis |0>, ...., |N-1>.
    If parity=True, multiply with the parity operator.
    See writeup for algorithm."""
    if iterable(N):
         M = N[1]
         N = N[0]
    else:
        M = N
    alpha = asarray(alpha)
    salpha = shape(alpha)
    alpha = reshape(alpha,size(alpha))
    alphaS = -abs(alpha)**2
    n=arange(N)
    m=arange(M)
    n = n[:,None] + 0*m[None,:]
    m = m[None,:] + 0*n

    s = min([n,m],axis=0)
    d = abs(n-m)
    # keep the iteration real to speed up calculation
    iterator = (1.0/factorial(s)/factorial(d))[newaxis,:,:] \
        + 0.0 * alphaS[:,newaxis,newaxis]
    result = 1.0*iterator
    for x in arange(1,min(N,M)):
        iterator *= ((s+1-x) / float(x) / (x+d))[newaxis,:,:] \
            * alphaS[:,newaxis,newaxis] 
        result += iterator
    if not parity:
        result *= 1 - 2 * (m % 2)
    # complex values only from here on
    return reshape(result \
       * alpha[:,newaxis,newaxis]**(n-s)[newaxis,:,:] \
       * conjugate(alpha)[:,newaxis,newaxis]**(m-s)[newaxis,:,:] \
       * exp(0.5*alphaS)[:,newaxis,newaxis] \
       * (sqrt(factorial(n)*factorial(m)) * (1-2*(s%2)))[newaxis,:,:],
                   salpha + (N,M))


def wignermatrix(alpha, N=10):
    """calculate the Matrix Displacement(alpha)*Parity*Displacement(-alpha)
    = Displacement(2*alpha)*Parity (see write-up)"""
    return 2.0/pi*displacement(2.0*alpha, N=N, parity=True)


def Qkmatrix(alpha, K=10, N=10):
    """Calculate <m|D(alpha)|k><k|D(-alpha)|n>. k=0...K-1, n,m=0...N-1.
    The return value has the following shape:
    [shape(alpha), K, N, N]"""
    
    s = shape(alpha)
    alpha = reshape(alpha,size(alpha))
    M = displacement(-alpha,N=(K,N))
    return M[:,:,None,:] * conjugate(M)[:,:,:,None]
    
    
def wigner(rho,alpha,checkimag=True):

    sia = size(alpha)
    # pdb.set_trace()
    sir = size(rho)
    sizelim = 1<<20
    sha = shape(alpha)
    alpha = resize(alpha,sia)
    blocksize = sizelim / sir
    blocksize += (blocksize == 0)
    result = zeros(sia,dtype=complex)
    for n in arange((sia+blocksize-1)/blocksize):
        start = n*blocksize
        stop = min(start+blocksize,sia)
        result[start:stop]=trace(dot(wignermatrix(alpha[start:stop],
                                                  N=shape(rho)[0]),rho),
                                 axis1=-2,axis2=-1)
    if max(abs(imag(result)))>1e-12:
        print 'Wigner function has imaginary part. Is your density matrix hermitian?'
    else:
        result=real(result)
    result = reshape(result,sha)
    return result
                    
def wigner1(rho,alpha):
    alpha = asarray(alpha)
    s = shape(alpha)
    alpha = reshape(alpha,size(alpha))
    result = 0.0j * alpha
    N=shape(rho)[1]
    for n,a in enumerate(alpha):
        result[n] = trace(dot(rho,wignermatrix(a,N=N)))
    if max(abs(imag(result)))>1e-4:
        print 'Wigner function contains imaginary part. Is your density matrix hermitian?'
    else:
        result=real(result)
    return reshape(2.0/pi*result,s)


def wignerN(n,alpha):
    """Calculate the Wigner function of a n photon Fock state at points
    alpha."""
    alpha = abs(alpha)**2
    return 2.0/pi * (-1)**n * exp(-2*alpha) * polyval(laguerre(n),4*alpha)
        
def plotW(rho,r=2, points=101,interpolation='bicubic',cmap=mycolormap('wrkbw'),showBar=False,vmin=-2/pi,vmax=2/pi):
    """plot the Wigner function corresponding to the density matrix rho"""
    alpha = complexgrid(r,points)
    w = wigner(rho,alpha)
    r = r + 0.5*r/(points-1)
    ax = gca()
    imshow(w, cmap = cmap, interpolation = interpolation, aspect = 'equal', extent = (-r,r,-r,r), vmin=vmin, vmax=vmax, origin='lower')
    xlabel(r'Re($\alpha$)')
    ylabel(r'Im($\alpha$)')
    if showBar:
        ax=colorbar()
        ax.set_label(r'$W(\alpha)$')

def coherent(alpha,K=100):
    """return the first K elements in the Fock basis of a coherent state with
     aplitude alpha"""
    n=arange(K)
    return exp(-0.5*abs(alpha)**2)*alpha**n/sqrt(factorial(n))

def pure(psi):
    """Caluculate the density matrix of the pure state psi."""
    psi = 1.0*asarray(psi)
    cpsi = conjugate(psi)
    return outer(psi,cpsi)/inner(psi,cpsi)



def WtoRho(alpha, W, sigma=1.0, N=5, returnSigma=False, returnCovar=False):
    """Calculate the NxN density matrix corresponding to the Wigner
    function W, evaluated at coordinates alpha. The shapes of alpha
    and W have to match.  If returnSigma=True, return also the
    uncertainty in the density matrix (NxN matrix). If
    returnCovar=True return also the covariance of the density matrix
    (NxNxNxN matrix)."""
    if shape(alpha) != shape(W):
        print 'alpha and W must have the same shape'
        return
    sigma = sigma + W * 0.0
    s = size(alpha)
    W = reshape(W,s)
    sigma = reshape(sigma,s)
    alpha = reshape(alpha,s)
    linv,rinv = leastsqinv(Hpack(wignermatrix(alpha,N=N)))
    rho = Hunpack(dot(linv,W))
    if returnSigma or returnCovar:
        covar = dot(linv * (sigma**2)[newaxis,:], rinv)
        rho = [rho]
    if returnSigma:
        rho += [reshape(sqrt(diag(covar)),(N,N))]        
    if returnCovar:
        rho += [reshape(covar,(N,N,N,N))]
    return rho
    

    
def QktoRho(alpha, Qk, sigma=1.0, N=5, returnSigma=False, returnCovar=False):
    """Calculate the NxN density matrix corresponding to the photon number
    Probabilities Qk, measured after a displacements alpha. The shape
    of Qk has to be the same as alpha but with and added last
    dimension giving the photon number.  If returnSigma=True, return
    also the uncertainty in the density matrix (NxN matrix). If
    returnCovar=True return also the covariance of the density matrix
    (NxNxNxN matrix)."""
    
    sigma = sigma + Qk * 0.0
    s = size(alpha)
    K = shape(Qk)[-1]
    Qk = reshape(Qk,s*K)
    sigma = reshape(sigma,s*K)   
    alpha = reshape(alpha,s)
    linv,rinv = leastsqinv(Hpack(Qkmatrix(alpha,K=K,N=N)))
    rho = Hunpack(dot(linv,Qk))
    if returnSigma or returnCovar:
        covar = dot(linv * (sigma**2)[newaxis,:], rinv)
        rho = [rho]
    if returnSigma:
        rho += [reshape(sqrt(diag(covar)),(N,N))]
    if returnCovar:
        rho += [reshape(covar,(N,N,N,N))]       
    return rho

def jinjang(r,n):
    alpha = complexgrid(r,n)
    result = 1.0*sign(real(alpha))*(abs(alpha)<=r)
    outer = abs(alpha-0.5j*r)<0.5*r
    inner = abs(alpha-0.5j*r)<0.2*r
    result = result * (1-outer) + outer - 2*inner
    outer = abs(alpha+0.5j*r)<0.5*r
    inner = abs(alpha+0.5j*r)<0.2*r
    result = result * (1-outer) - outer + 2*inner
    return result
    

