import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K')]
import numpy as np
import time
import pyle.dataking.util as util

def tdrTrace(sample, TDR = None):

    cxn = sample._cxn
    if TDR:
        tdr = TDR
    else:
        samp, dev = util.loadDeviceType(sample, 'TDR')
        tdr = dev[0]
    start = tdr['timeStart']
    stop = tdr['timeStop']
    channel = tdr['channel']
    nameString = 'TDR Measurement ' +  tdr['device']
    address = tdr['TDRAddress']
    tek = cxn.tektronix_dsa_8300_sampling_scope
    tek.select_device(address)
    
        
    if tdr['units'] == 'ohms':
        independents = ['Time [ns]']
        dependents = ['Impedance [Ohms]']
    elif tdr['units'] == 'volts':
        independents = ['Time [ns]']
        dependents = ['Voltage [mV]']
    
    # get the parameters from the data vault
    
    
    if tdr['units'] == 'ohms':
        data = tek.get_trace_ohms(channel)
    elif tdr['units'] == 'volts':
        data = tek.get_trace(channel)
    
    dv = cxn.data_vault
    dv.cd(sample._dir, True)
    dv.new(nameString, independents, dependents)
    p = dv.packet()
    data = np.vstack(data)
    new = data.T
    p = dv.packet()
    for key in tdr.keys():
        p.add_parameter(key, tdr.__getitem__(key))
    p.send()
    dv.add(new)

