import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import frequencyScan as fs
import powerScan as ps
import freqFluxPhase as ffp
import freqPowerPhase as fpp
import frequencyScan as fs
from pyle.analysis import plotParampData as ps
import noiseScan as ns
from scipy.interpolate import UnivariateSpline as US

fluxMap = None

def calibrate(sample):
    #take a phase vs flux and frequency data set that maps out frequency points
    #for the entire frequency range of interest
    cxn = sample._cxn
    params = sample['PNA']     #load the pna sample object to make changes to the measurement parameters
    tune = sample['tuneUp']    #load the tuneUp sample object to keep track of our calibration data
    #these next few lines set the parameters of the measurement to the
    #right range for the initial flux scan then initiates the masurement
    params['MeasurementDivisions'] = 100 
    params['bandwidth'] = 2000 
    params['power'] = -50      
    params['startFreq'] = 3.5    
    params['stopFreq'] = 9    
    ffp.measure(sample, limits = [0.25,1.5])
    # the next few lines record which file number in the data vault
    # is storing the flux maping data for use later
    dv = sample._cxn.data_vault
    dv.cd(sample._dir)
    stuff = dv.dir()[1][-1]
    tune['FluxMapFile'] = int(stuff.split('-')[0])
    #this resets the fluxMap to none so it is regenerated
    #the next time autoTune is called
    global fluxMap
    fluxMap = None
    
def autoTune(sample, pump, freq, gain, save = False):
    #this function actually tunes up the amplifier to an arbitrary freq and gain.
    #based on the flux map data from the calibrate function.
    cxn = sample._cxn
    dv = cxn.data_vault
    dv.cd(sample._dir)
    params = sample['PNA']
    nparams = sample['noiseMeas']
    tune = sample['tuneUp']
    fmap = tune['FluxMapFile']
    tune['FluxSlope'] = 'neg'
    if fluxScan['dacBoard'] == 'DC rack':
        dc = cxn.dc_rack_server
        dc.select_device(0)
    corrFreq = fluxCorrection(freq)
    print corrFreq
    flux = 900
    print flux
    params['bandwidth'] = 100
    params['number of points'] = 501
    params['startFreq'] = freq-0.25
    params['stopFreq'] = freq+0.25
    params['power'] = -20
    # fs.freqScan(sample)
    # gainCal = dv.dir()[1][-1]
    # tune['GainCalFile'] = int(gainCal.split('-')[0])
    nparams['name'] = 'Paramp Noise Scan at %f GHz' %freq
    nparams['freq'] = freq*1000
    nparams['span'] = 500
    ns.takeNoiseScan(sample)
    noiseCal = dv.dir()[1][-1]
    tune['NoiseCalFile'] = int(noiseCal.split('-')[0])
    pump.output(True)
    if params['ParampPumpMethod'] == 'input w':
        bestGain, bestPow, bestFlux = zeroInGain(sample, pump, freq, gain, power, flux, savePoints = save)
    if params['ParampPumpMethod'] == '2w':
        bestGain, bestPow, bestFlux = zeroInGain2w(sample, pump, freq, gain, power, flux, savePoints = save)
    if abs(bestGain - gain) < 0.5 and not save:
        params['bandwidth'] = 100
        params['number of points'] = 1001
        params['startFreq'] = freq-0.05
        params['stopFreq'] = freq+0.05
        fs.freqScan(sample)
        gainFinal = dv.dir()[1][-1]
        tune['GainFile'] = int(gainFinal.split('-')[0])
        nparams['freq'] = freq*1000
        nparams['span'] = 100
        ns.takeNoiseScan(sample)
        pump.output(False)
        noiseFinal = dv.dir()[1][-1]
        tune['NoiseFile'] = int(noiseFinal.split('-')[0])
        ampNoise = ps.calcParampNoise(cxn, sample._dir, tune['GainFile'], tune['GainCalFile'], tune['NoiseFile'], tune['NoiseCalFile'])
        return bestGain, bestPow, bestFlux, ampNoise
    elif save:
        params['bandwidth'] = 100
        params['number of points'] = 1001
        params['startFreq'] = freq-0.5
        params['stopFreq'] = freq+0.5
        return 0,0,0,0
    else:
        print 'unable to achieve desired gain'
        return 0,0,0,0
    
def makeFluxMap(cxn, dir, file, freqRange = [4.5,8]):
    ps.getCal(cxn)
    dv = cxn.data_vault
    dv.cd(dir)
    dv.open(file)
    divs = dv.get_parameter('MeasurementDivisions')
    points = dv.get_parameter('number of points')
    start = freqRange[0]
    stop = freqRange[1]
    done = False
    data = dv.get(10000).asarray
    while not done:
        addData = dv.get(10000).asarray
        if len(addData) == 0 or len(addData) == 1:
            done = True
        else:
            data = np.vstack((data,addData))
    fluxes = np.reshape(data[:,0], (divs+1, points))
    freqs = np.reshape(data[:,1], (divs+1, points))
    phases = np.reshape(data[:,2], (divs+1, points))
    flpoints = np.array([])
    fpoints = np.array([])
    fslice = freqs[0,:]
    for i in range(len(fslice)):
        if not (fslice[i]<freqRange[0] or fslice[i]>freqRange[1]):
            phaseturn = phases.min()+(phases.max() - phases.min())/2
            results = np.where(abs(phases[:,i]-phaseturn)<30)[0]
            flresults = fluxes[:,i][results]
            closest = abs(flresults).min()
            flpoints = np.append(flpoints,closest)
            fpoints = np.append(fpoints,fslice[i])
    map = US(fpoints,flpoints,s=1)
    return map
           
def findPowerLevel(cxn, dir, file, freq):
    #find the correct power to operate the amplifer
    dv = cxn.data_vault
    dv.cd(dir)
    dv.open(file)
    divs = dv.get_parameter('MeasurementDivisions')
    points = dv.get_parameter('number of points')
    done = False
    data = dv.get(10000).asarray
    while not done:
        addData = dv.get(10000).asarray
        if len(addData) == 0 or len(addData)==1:
            done = True
        else:
            data = np.vstack((data,addData))
    powervalues = []
    try:
        detune = 0
        while not len(powervalues) and detune < 10:
            freq = freq + float(detune)/10
            divs = len(data[:,0])/points
            freqs = np.reshape(data[:,0], (divs,points))
            powers = np.reshape(data[:,1], (divs,points))
            phases = np.reshape(data[:,2], (divs,points))
            center =  np.where(abs(freqs[:,0]-freq)<0.01)[0][0]
            phaseZero = data[:,2].min()+((data[:,2].max()-data[:,2].min())/2)+20
            powerPlaces = np.where(abs(phases[center,:]-phaseZero)<1)[0]
            powervalues = powers[center,:][powerPlaces]
            detune += 1
        detune = 0
        while not len(powervalues) and detune > -10:
            divs = len(data[:,0])/points
            freq = freq + float(detune)/10
            freqs = np.reshape(data[:,0], (divs,points))
            powers = np.reshape(data[:,1], (divs,points))
            phases = np.reshape(data[:,2], (divs,points))
            center =  np.where(abs(freqs[:,0]-freq)<0.01)[0][0]
            phaseZero = data[:,2].min()+((data[:,2].max()-data[:,2].min())/2)+20
            powerPlaces = np.where(abs(phases[center,:]-phaseZero)<1)[0]
            powervalues = powers[center,:][powerPlaces]
            detune -= 1
        if not len(powervalues):
            pumpPower = 16
        else:
            pumpPower = powervalues.mean()+68+20 #30 change from 40 for flux pump - Josh
    except:
        pumpPower = 16
    return pumpPower
    
def findPowerLevel2w(cxn, dir, file, freq):
    #find the correct power to operate the amplifer
    dv = cxn.data_vault
    dv.cd(dir)
    dv.open(file)
    divs = dv.get_parameter('MeasurementDivisions')
    points = dv.get_parameter('number of points')
    done = False
    data = dv.get(10000).asarray
    while not done:
        addData = dv.get(10000).asarray
        if len(addData) == 0 or len(addData)==1:
            done = True
        else:
            data = np.vstack((data,addData))
    powervalues = []
    try:
        detune = 0
        while not len(powervalues) and detune < 10:
            freq = freq + float(detune)/10
            divs = len(data[:,0])/points
            freqs = np.reshape(data[:,0], (divs,points))
            powers = np.reshape(data[:,1], (divs,points))
            phases = np.reshape(data[:,2], (divs,points))
            center =  np.where(abs(freqs[:,0]-freq)<0.01)[0][0]
            phaseZero = data[:,2].min()+((data[:,2].max()-data[:,2].min())/2)+20
            powerPlaces = np.where(abs(phases[center,:]-phaseZero)<1)[0]
            powervalues = powers[center,:][powerPlaces]
            detune += 1
        detune = 0
        while not len(powervalues) and detune > -10:
            divs = len(data[:,0])/points
            freq = freq + float(detune)/10
            freqs = np.reshape(data[:,0], (divs,points))
            powers = np.reshape(data[:,1], (divs,points))
            phases = np.reshape(data[:,2], (divs,points))
            center =  np.where(abs(freqs[:,0]-freq)<0.01)[0][0]
            phaseZero = data[:,2].min()+((data[:,2].max()-data[:,2].min())/2)+20
            powerPlaces = np.where(abs(phases[center,:]-phaseZero)<1)[0]
            powervalues = powers[center,:][powerPlaces]
            detune -= 1
        if not len(powervalues):
            pumpPower = 0
        else:
            pumpPower = powervalues.mean()+68+25 #30 change from 40 for flux pump - Josh
    except:
        pumpPower = 0
    return pumpPower
    
def zeroInGain(sample, hittite, freq, gain, startPower, startflux, savePoints = False):
    if savePoints:
        points = []
        savegains = np.arange(16,32,2)
        dv = sample._cxn.data_vault
        dv.cd(sample._dir)
        tune = sample['tuneUp']
        fluxScan = sample['fluxScan']
        params = sample['PNA']
        params['bandwidth'] = 2000
        params['number of points'] = 201
        params['startFreq'] = freq-0.05
        params['stopFreq'] = freq+0.05
        dv.open(tune['GainCalFile'])
        calGain = dv.get().asarray
        calGain = calGain[:,1].mean()
        currentGain = 0
        optimalPow = startPower
        optimalFlux = 0
        hittite.frequency(freq*1e9) 
        unfinished = True
        pna = prepPNA(sample)
        while unfinished:
            if tune['FluxSlope']=='neg':
                fluxes = np.arange(startflux-0.12, startflux+0.05, 0.005)
                fluxes = fluxes[::-1]
            else:
                fluxes = np.arange(startflux-0.05, startflux+0.12, 0.005)
            optimal = startPower-2
            dims = (len(fluxes),len(savegains))
            checklist = np.ones(dims)
            rescale = True
            terminate = 0
            for i in range(len(fluxes)):
                flux = fluxes[i]
                if rescale:
                    powers = np.arange(optimalPow+1-4,optimalPow+1+4.1,0.1)
                else:
                    powers = np.arange(optimalPow-4,optimalPow+4.1,0.1)
                rescale = True
                sb.set_fb(tune['FluxDac'], sample._cxn, flux*V, tune['FluxChannel'])
                noswap = True
                for power in powers:
                    hittite.amplitude(power)
                    data = fs.freqScanFast(pna)
                    gainData = data[:,1]
                    averageGain = gainData[gainData.argmax()+5:gainData.argmax()+15].mean()
                    calcGain = averageGain-calGain
                    if calcGain>currentGain:
                        currentGain = calcGain
                        optimalPow = power
                        optimalFlux = flux
                        rescale = False
                        noswap = False
                    for j in range(len(savegains)):
                        if abs(calcGain - savegains[j]) < 0.5 and checklist[i,j]:
                            points.append([flux,freq,power,calcGain])
                            checklist[i,j] = 0
                    print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
                if currentGain>10 and noswap:
                    terminate += 1
                if  not unfinished or terminate > 3:
                    break
            unfinished = False
        hittite.output(False)
        fluxScan['points'] = points
        return currentGain, optimalPow, optimalFlux
    else:
        dv = sample._cxn.data_vault
        dv.cd(sample._dir)
        tune = sample['tuneUp']
        params = sample['PNA']
        params['bandwidth'] = 2000
        params['number of points'] = 301
        params['startFreq'] = freq-0.05
        params['stopFreq'] = freq+0.05
        dv.open(tune['GainCalFile'])
        calGain = dv.get().asarray
        calGain = calGain[:,1].mean()
        currentGain = 0
        optimalPow = startPower
        optimalFlux = 0
        hittite.frequency(freq*1e9) 
        unfinished = True
        pna = prepPNA(sample)
        while unfinished:
            if tune['FluxSlope']=='neg':
                fluxes = np.arange(startflux-0.12, startflux+0.05, 0.005)
                fluxes = fluxes[::-1]
            else:
                fluxes = np.arange(startflux-0.05, startflux+0.12, 0.005)
            optimal = startPower-2
            for flux in fluxes:
                powers = np.arange(optimalPow-3,optimalPow+3.1,0.1)
                sb.set_fb(tune['FluxDac'], sample._cxn, flux*V, tune['FluxChannel'])
                for power in powers:
                    hittite.amplitude(power)
                    data = fs.freqScanFast(pna)
                    gainData = data[:,1]
                    averageGain = gainData[gainData.argmax()+5:gainData.argmax()+15].mean()
                    calcGain = averageGain-calGain
                    if calcGain>currentGain:
                        currentGain = calcGain
                        optimalPow = power
                        optimalFlux = flux
                    if (currentGain >= gain) or (abs(currentGain - gain) < 0.5):
                        unfinished = False
                        print 'Got it'
                        break
                    print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain) 
                if  not unfinished:
                    break
        return currentGain, optimalPow, optimalFlux
    
def zeroInGain2w(sample, pump, freq, gain, startPower, startflux, savePoints = False):
    if savePoints:
        points = []
        savegains = np.arange(16,28,2)
        dv = sample._cxn.data_vault
        dv.cd(sample._dir)
        tune = sample['tuneUp']
        fluxScan = sample['fluxScan']
        params = sample['PNA']
        params['bandwidth'] = 2000
        params['number of points'] = 201
        params['startFreq'] = freq-0.05
        params['stopFreq'] = freq+0.05
        dv.open(tune['GainCalFile'])
        calGain = dv.get().asarray
        calGain = calGain[:,1].mean()
        currentGain = 0
        optimalPow = startPower
        optimalFlux = 0
        pump.frequency(2*freq*GHz)
        unfinished = True
        start = True
        pna = prepPNA(sample)
        if tune['FluxSlope']=='neg':
            startflux = startflux + 0.105
        else:
            startflux = startflux - 0.105
        while unfinished:
            if tune['FluxSlope']=='neg':
                startflux = startflux - 0.105
                fluxes = np.arange(startflux-0.05, startflux+0.05, 0.005)
                fluxes = fluxes[::-1]
            else:
                startflux = startflux + 0.105
                fluxes = np.arange(startflux-0.05, startflux+0.05, 0.005)
            terminate = 0
            dims = (len(fluxes),len(savegains))
            checklist = np.ones(dims)
            for i in range(len(fluxes)):
                flux = fluxes[i]
                if currentGain > 15:
                    powers = np.arange(optimalPow-3,optimalPow+3.1,0.1)
                else:
                    powers = np.arange(optimalPow-5,optimalPow+5.2,0.2)
                sb.set_fb(tune['FluxDac'], sample._cxn, flux*V, tune['FluxChannel'])
                noswap = True
                for power in powers:
                    pump.amplitude(power)
                    data = fs.freqScanFast(pna)
                    gainData = data[:,1]
                    if (gainData.argmax()+10)>200:
                        averageGain = gainData[gainData.argmax()-10:gainData.argmax()-5].mean()
                    else:
                        averageGain = gainData[gainData.argmax()+5:gainData.argmax()+10].mean()
                    calcGain = averageGain-calGain
                    if calcGain>currentGain:
                        currentGain = calcGain
                        optimalPow = power
                        optimalFlux = flux
                        rescale = False
                        noswap = False
                        terminate = 0
                    for j in range(len(savegains)):
                        if abs(calcGain - savegains[j]) < 0.5 and checklist[i,j]:
                            points.append([flux,freq,power,calcGain])
                            checklist[i,j] = 0
                    print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
                if currentGain>10 and noswap:
                    terminate += 1
                if  not unfinished or terminate > 3:
                    break
                print 'terminate = %i' %terminate
            if terminate>3:
                unfinished = False
        pump.output(False)
        fluxScan['points'] = points
        return currentGain, optimalPow, optimalFlux
    else:
        dv = sample._cxn.data_vault
        dv.cd(sample._dir)
        tune = sample['tuneUp']
        params = sample['PNA']
        params['bandwidth'] = 2000
        params['number of points'] = 201
        params['startFreq'] = freq-0.05
        params['stopFreq'] = freq+0.05
        dv.open(tune['GainCalFile'])
        calGain = dv.get().asarray
        calGain = calGain[:,1].mean()
        currentGain = 0
        optimalPow = startPower
        optimalFlux = 0
        pump.frequency(2*freq*1e9)
        unfinished = True
        pna = prepPNA(sample)
        while unfinished:
            startflux = startflux + 0.055
            if tune['FluxSlope']=='neg':
                fluxes = np.arange(startflux-0.05, startflux+0.05, 0.005)
                fluxes = fluxes[::-1]
            else:
                fluxes = np.arange(startflux-0.05, startflux+0.05, 0.005)
            optimal = startPower-2
            for flux in fluxes:
                powers = np.arange(optimalPow-6,optimalPow+6.1,0.1)
                sb.set_fb(tune['FluxDac'], sample._cxn, flux*V, tune['FluxChannel'])
                for power in powers:
                    pump.amplitude(power)
                    data = fs.freqScanFast(pna)
                    gainData = data[:,1]
                    averageGain = gainData[gainData.argmax()+5:gainData.argmax()+15].mean()
                    calcGain = averageGain-calGain
                    if calcGain>currentGain:
                        currentGain = calcGain
                        optimalPow = power
                        optimalFlux = flux
                    if (currentGain >= gain) or (abs(currentGain - gain) < 0.5):
                        unfinished = False
                        print 'Got it'
                        break
                    print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain) 
                if  not unfinished:
                    break
        return currentGain, optimalPow, optimalFlux
        
def fineFreqMapInputW(sample, pump, freq, saveGain = 12, save = False):
    #this function actually tunes up the amplifier to an arbitrary freq and gain.
    #based on the flux map data from the calibrate function.
    cxn = sample._cxn
    dv = cxn.data_vault
    dv.cd(sample._dir)
    params = sample['PNA']
    nparams = sample['noiseMeas']
    tune = sample['tuneUp']
    pump.select_device(tune['PumpAddress'])
    fmap = tune['FluxMapFile']
    #these next few lines check and see if the fluxMap variable exists.
    #if there is already a fluxMp it will be used, but if no map exists
    #it will be generated and a variable in tuneUp keeps track of the slope
    #of the line for use with the autoTuner later
    # global fluxMap
    # if not fluxMap:
        # print 'making flux map'
        # fluxMap = makeFluxMap(cxn, sample._dir, fmap, freqRange = [5,8])
        # print 'map made'
        # if fluxMap(6)>fluxMap(7):
            # tune['FluxSlope'] = 'neg'
        # else:
            # tune['FluxSlope'] = 'pos'
    #sets initial guess for the flux
    # corrFreq = fluxCorrection(freq)
    # print corrFreq
    startflux = 0.75 #fluxMap(corrFreq)*1530/1000 
    print startflux
    sb.set_fb(tune['FluxDac'], cxn, startflux*V, tune['FluxChannel'])
    params['name'] = 'PNA Scan at %f GHz' %freq
    params['MeasurementDivisions'] = 50
    params['startPow'] = -35
    params['stopPow'] = -15
    params['bandwidth'] = 4000
    params['number of points'] = 1001
    params['startFreq'] = freq-0.5
    params['stopFreq'] = freq+0.5
    pump.output(False)
    # print 'creating power map'
    # fpp.measurePow(sample)
    # print 'power map made'
    # powerCal = dv.dir()[1][-1]
    # tune['PowerMapFile'] = int(powerCal.split('-')[0])
    # if params['ParampPumpMethod'] == 'input w':
        # startPower = findPowerLevel(cxn, sample._dir, tune['PowerMapFile'], freq)
    # if params['ParampPumpMethod'] == '2w':
        # startPower = findPowerLevel2w(cxn, sample._dir, tune['PowerMapFile'], freq)
    params['bandwidth'] = 100
    params['number of points'] = 501
    params['startFreq'] = freq-0.15
    params['stopFreq'] = freq+0.15
    params['power'] = -70
    nparams['name'] = 'Paramp Noise Scan at %f GHz' %freq
    nparams['freq'] = freq*1000
    nparams['span'] = 300
    nparams['numberOfPoints'] = 501
    ns.takeNoiseScan(sample)
    noiseCal = dv.dir()[1][-1]
    tune['NoiseCalFile'] = int(noiseCal.split('-')[0])
    fs.freqScan(sample)
    gainCal = dv.dir()[1][-1]
    tune['GainCalFile'] = int(gainCal.split('-')[0])
    pump.output(True)
    points = []
    dv = sample._cxn.data_vault
    dv.cd(sample._dir)
    tune = sample['tuneUp']
    fluxScan = sample['fluxScan']
    params['bandwidth'] = 2000
    params['number of points'] = 201
    params['startFreq'] = freq-0.15
    params['stopFreq'] = freq+0.15
    dv.open(tune['GainCalFile'])
    calGain = dv.get().asarray
    calGain = calGain[:,1].mean()
    currentGain = 0
    optimalPow = -13
    optimalFlux = 0
    pump.frequency(freq*GHz) 
    unfinished = True
    pna = prepPNA(sample)
    if tune['FluxSlope']=='neg':
        startflux = startflux + 0.105
    else:
        startflux = startflux - 0.105
    lockedIn = False
    lockedFlux = None
    while unfinished:
        if lockedIn:
            if tune['FluxSlope']=='neg':
                print 'lockedFlux'
                lockedFlux = lockedFlux - 0.027
                fluxes = np.arange(lockedFlux, lockedFlux+0.026, 0.001)
                fluxes = fluxes[::-1]
            else:
                lockedFlux = lockedFlux + 0.027
                fluxes = np.arange(lockedFlux, lockedFlux+0.026, 0.001)
        else:
            if tune['FluxSlope']=='neg':
                startflux = startflux - 0.105
                fluxes = np.arange(startflux-0.05, startflux+0.055, 0.005)
                fluxes = fluxes[::-1]
            else:
                startflux = startflux + 0.105
                fluxes = np.arange(startflux-0.055, startflux+0.05, 0.005)
        terminate = 0
        for i in range(len(fluxes)):
            noswap = True
            flux = fluxes[i]
            if currentGain > 15:
                powers = np.arange(optimalPow-0.4,optimalPow+0.51,0.01)
            else:
                powers = np.arange(optimalPow-3,optimalPow+3.1,0.1)
            sb.set_fb(tune['FluxDac'], sample._cxn, flux*V, tune['FluxChannel'])
            for power in powers:
                pump.amplitude(power)
                data = fs.freqScanFast(pna)
                gainData = data[:,1]
                if (gainData.argmax()+10)>200:
                    averageGain = gainData[gainData.argmax()-10:gainData.argmax()-5].mean()
                else:
                    averageGain = gainData[gainData.argmax()+5:gainData.argmax()+10].mean()
                calcGain = averageGain-calGain
                if calcGain>currentGain:
                    currentGain = calcGain
                    optimalPow = power
                    terminate = 0
                    noswap = False
                if calcGain > saveGain:
                    if not lockedIn:
                        lockedIn = True
                        print 'locked'
                    points.append([flux,freq,power,calcGain])
                print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
            if currentGain > 10 and noswap:
                terminate += 1
                print terminate
            if  not unfinished or terminate > 2:
                break
            if lockedIn and not lockedFlux:
                lockedFlux = flux
                print 'reset'
                break
        if terminate > 2:
            unfinished = False
    pump.output(False)
    fluxScan['points'] = points
    return currentGain, optimalPow, optimalFlux
    
def fineFreqMap2W(sample, pump, freq, saveGain = 15):
    #this function actually tunes up the amplifier to an arbitrary freq and gain.
    #based on the flux map data from the calibrate function.
    cxn = sample._cxn
    dv = cxn.data_vault
    dv.cd(sample._dir)
    params = sample['PNA']
    nparams = sample['noiseMeas']
    tune = sample['tuneUp']
    pump.select_device(tune['PumpAddress'])
    fmap = tune['FluxMapFile']
    pump.frequency(2*freq*GHz) 
    startflux = tune['startflux'] #fluxMap(corrFreq)*1530/1000
    if tune['FluxDac'] == 'DC rack':
        dc = cxn.dc_rack_server
        dc.select_device(0)
    print startflux
    params['bandwidth'] = 100
    params['number of points'] = 1001
    params['startFreq'] = freq-0.5
    params['stopFreq'] = freq+0.5
    params['power'] = -60
    fs.freqScan(sample)
    gainCal = dv.dir()[1][-1]
    tune['GainCalFile'] = int(gainCal.split('-')[0])
    nparams['name'] = 'Paramp Noise Scan at %f GHz' %freq
    nparams['freq'] = freq*1000
    nparams['span'] = 1000
    ns.takeNoiseScan(sample)
    noiseCal = dv.dir()[1][-1]
    tune['NoiseCalFile'] = int(noiseCal.split('-')[0])
    pump.output(True)
    points = []
    dv = sample._cxn.data_vault
    dv.cd(sample._dir)
    tune = sample['tuneUp']
    fluxScan = sample['fluxScan']
    params = sample['PNA']
    params['bandwidth'] = 2000
    params['number of points'] = 201
    params['startFreq'] = freq-0.25
    params['stopFreq'] = freq+0.25
    dv.open(tune['GainCalFile'])
    calGain = dv.get().asarray
    calGain = calGain[:,1].mean()
    currentGain = 0
    optimalPow = 4
    optimalFlux = 0
    unfinished = True
    pna = prepPNA(sample)
    if tune['FluxSlope']=='neg':
        startflux = startflux + 0.210
    else:
        startflux = startflux - 0.210
    while unfinished:
        if tune['FluxSlope']=='neg':
            startflux = startflux - 0.105
            fluxes = np.arange(startflux-0.1, startflux+0.11, 0.01)
            fluxes = fluxes[::-1]
        else:
            startflux = startflux + 0.105
            fluxes = np.arange(startflux-0.05, startflux+0.055, 0.005)
        terminate = 0
        for i in range(len(fluxes)):
            flux = fluxes[i]
            if currentGain > 15:
                powers = np.arange(optimalPow-7,optimalPow+7.1,0.1)
            else:
                powers = np.arange(optimalPow-7,optimalPow+7.2,0.2)
            setFlux(flux, tune, sb, dc)
            noswap = True
            for power in powers:
                pump.amplitude(power)
                data = fs.freqScanFast(pna)
                gainData = data[:,1]
                if (gainData.argmax()+10)>200:
                    averageGain = gainData[gainData.argmax()-10:gainData.argmax()-5].mean()
                else:
                    averageGain = gainData[gainData.argmax()+5:gainData.argmax()+10].mean()
                calcGain = averageGain-calGain
                if calcGain>currentGain:
                    currentGain = calcGain
                    optimalPow = power
                    optimalFlux = flux
                    rescale = False
                    noswap = False
                    terminate = 0
                if calcGain > saveGain:
                    points.append([flux,freq,power,calcGain])
                    print len(points)
                print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
            if currentGain > 10 and noswap:
                terminate += 1
                print terminate
            if  not unfinished or terminate > 4:
                break
        if terminate > 2:
            unfinished = False
    pump.output(False)
    fluxScan['points'] = points
    return currentGain, optimalPow, optimalFlux

def broadFreqScan2W(sample, pump, freqs, saveGain = 20):
    cxn = sample._cxn
    dc = cxn.dc_rack_server
    dc.select_device(0)
    dv = cxn.data_vault
    dv.cd(sample._dir)
    params = sample['PNA']
    params['number of points'] = 201
    params['bandwidth'] = 4000 
    nparams = sample['noiseMeas']
    tune = sample['tuneUp']
    pump.select_device(tune['PumpAddress'])
    # fmap = tune['FluxMapFile']
    fluxScan = sample['fluxScan']
    points = fluxScan['points']
    fluxCuttoff = 1
    startflux = 0
    for freq in freqs:
        print freq
        points = fluxScan['points']
        pump.output(False)
        params['startFreq'] =freq-0.25
        params['stopFreq'] =freq+0.25 
        twentyPoint = None
        pump.frequency(2*freq*GHz)
        # global fluxMap
        # if not fluxMap:
            # print 'making flux map'
            # fluxMap = makeFluxMap(cxn, sample._dir, fmap, freqRange = [5,8])
            # print 'map made'
            # if fluxMap(6)>fluxMap(7):
                # tune['FluxSlope'] = 'neg'
            # else:
                # tune['FluxSlope'] = 'pos'
        #sets initial guess for the flux
        # corrFreq = fluxCorrection(freq)
        # print corrFreq
        if startflux == 0:
            startflux = tune['startflux'] #fluxMap(corrFreq)*1530/1000
        else:
            print startflux
            startflux = startflux+0.24
        print startflux
        setFlux(startflux, tune, sb, dc)
        currentGain = 0
        optimalPow = 7
        finished = False
        pna = prepPNA(sample)
        data = fs.freqScanFast(pna)
        calGain = data[:,1].mean()
        print calGain
        pump.output(True)
        # if tune['FluxSlope']=='neg':
            # startflux = startflux + 0.5
        # else:
            # startflux = startflux - 0.5
        lastRun = False
        lastFlux = None
        while not finished:
            if tune['FluxSlope']=='neg':
                startflux = startflux - 0.25
                fluxes = np.arange(startflux-0.3, startflux+0.31, 0.01)
                fluxes = fluxes[::-1]
            else:
                startflux = startflux + 0.25
                fluxes = np.arange(startflux-0.3, startflux+0.31, 0.01)
            for i in range(len(fluxes)):
                flux = fluxes[i]
                if flux>fluxCuttoff+0.03:
                    pass
                else:
                    if lastRun:
                        flux = flux+0.01
                        powers = np.arange(optimalPow-1,optimalPow+1.01,0.01)
                    elif currentGain > 15:
                        powers = np.arange(optimalPow-5.5,optimalPow+5.5,0.1)
                    else:
                        powers = np.arange(optimalPow-7,optimalPow+7.2,0.2)
                    setFlux(flux, tune, sb, dc)
                    noswap = True
                    for power in powers:
                        pump.amplitude(power)
                        data = fs.freqScanFast(pna)
                        gainData = data[:,1]
                        middle = len(gainData)/2.0
                        averageGain = gainData[middle-10:middle+10].mean()
                        # if (gainData.argmax()+10)>200:
                            # averageGain = gainData[gainData.argmax()-10:gainData.argmax()-5].mean()
                        # else:
                            # averageGain = gainData[gainData.argmax()+2:gainData.argmax()+6].mean()
                        calcGain = averageGain-calGain
                        if lastRun:
                            if abs(calcGain-20)<0.5:
                                points.append([flux,freq,power,calcGain])
                                fluxCuttoff = flux
                                finished = True
                                break
                        else:
                            if calcGain>currentGain:
                                if power > 14:
                                    optimalPow = 5
                                    currentGain = 0
                                else:
                                    currentGain = calcGain
                                    optimalPow = power
                                optimalFlux = flux
                            if abs(calcGain - 17) < 2:
                                twentyPoint = [flux,freq,power,calcGain]
                            if calcGain > saveGain+0:
                                if not twentyPoint or power > 13:
                                    optimalPow = 5
                                else:
                                    optimalPow = twentyPoint[2]
                                    print optimalPow
                                    lastRun = True
                                break
                        print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
                        if finished:
                            break
                if finished:
                    break
        fluxScan['points']=points
    pump.output(False)
    fluxScan['points'] = points
    return points
    
def broadFreqScanInputW(sample, pump, freqs, saveGain = 20):
    print 'this is the right one'
    cxn = sample._cxn
    dv = cxn.data_vault
    dv.cd(sample._dir)
    params = sample['PNA']
    params['number of points'] = 201
    params['bandwidth'] = 4000
    params['power'] = -70    
    nparams = sample['noiseMeas']
    tune = sample['tuneUp']
    pump.select_device(tune['PumpAddress'])
    fmap = tune['FluxMapFile']
    fluxScan = sample['fluxScan']
    points = fluxScan['points']
    fluxCuttoff = 1
    for freq in freqs:
        print freq
        points = fluxScan['points']
        pump.output(False)
        params['startFreq'] =freq-0.15
        params['stopFreq'] =freq+0.15 
        twentyPoint = None
        pump.frequency(freq*GHz)
        global fluxMap
        if not fluxMap:
            print 'making flux map'
            fluxMap = makeFluxMap(cxn, sample._dir, fmap, freqRange = [5,8])
            print 'map made'
            if fluxMap(6)>fluxMap(7):
                tune['FluxSlope'] = 'neg'
            else:
                tune['FluxSlope'] = 'pos'
        #sets initial guess for the flux
        corrFreq = fluxCorrection(freq)
        print corrFreq
        startflux = 0.700 #= fluxMap(corrFreq)*1530/1000 
        print startflux
        setFlux(flux, tune, sb, dc)
        currentGain = 0
        optimalPow = -14
        finished = False
        pna = prepPNA(sample)
        data = fs.freqScanFast(pna)
        calGain = data[:,1].mean()
        print calGain
        pump.output(True)
        if tune['FluxSlope']=='neg':
            startflux = startflux + 0.1
        else:
            startflux = startflux - 0.1
        lastRun = False
        lastFlux = None
        while not finished:
            if tune['FluxSlope']=='neg':
                startflux = startflux - 0.055
                fluxes = np.arange(startflux-0.025, startflux+0.030, 0.005)
                fluxes = fluxes[::-1]
            else:
                startflux = startflux + 0.055
                fluxes = np.arange(startflux-0.025, startflux+0.030, 0.005)
            for i in range(len(fluxes)):
                flux = fluxes[i]
                if flux>fluxCuttoff+0.015:
                    pass
                else:
                    if lastFlux:
                        flux = lastFlux
                    if lastRun:
                        powers = np.arange(optimalPow-0.75,optimalPow+0.75,0.01)
                    elif currentGain > 15:
                        powers = np.arange(optimalPow-1.0,optimalPow+1.02,0.02)
                    else:
                        powers = np.arange(optimalPow-3,optimalPow+3.1,0.1)
                    setFlux(flux, tune, sb, dc)
                    noswap = True
                    for power in powers:
                        pump.amplitude(power)
                        data = fs.freqScanFast(pna)
                        gainData = data[:,1]
                        if (gainData.argmax()+10)>200:
                            averageGain = gainData[gainData.argmax()-10:gainData.argmax()-5].mean()
                        else:
                            averageGain = gainData[gainData.argmax()+5:gainData.argmax()+10].mean()
                        calcGain = averageGain-calGain
                        if lastRun:
                            if calcGain>22:
                                points.append([flux,freq,power,calcGain])
                                fluxCuttoff = flux
                                finished = True
                                break
                        else:
                            if calcGain>currentGain:
                                if power > 11:
                                    optimalPow = -5
                                    currentGain = 0
                                else:
                                    currentGain = calcGain
                                    optimalPow = power
                                optimalFlux = flux
                            if calcGain > 22:
                                twentyPoint = [flux,freq,power,calcGain]
                                optimalPow = twentyPoint[2]
                                print optimalPow
                                lastRun = True
                                lastFlux = flux
                                break
                        print 'Just tried: %f V flux and %f dBm pump with %f dB gain' %(flux,power,calcGain)
                        if finished:
                            break
                if finished:
                    break
        fluxScan['points']=points
    pump.output(False)
    fluxScan['points'] = points
    return points
    
def prepPNA(sample):
    cxn = sample._cxn
    samp, dev = util.loadDeviceType(sample, 'PNA')
    ampMeas = dev[0]
    start = ampMeas['startFreq']*GHz
    stop = ampMeas['stopFreq']*GHz
    num_points = ampMeas['number of points']
    averages = ampMeas['averages']
    power = ampMeas['power']*dBm
    bandwidth = ampMeas['bandwidth']*Hz
    edelay = ampMeas['electricalDelay'].value
    pna = cxn.pna_x
    pna.select_device(ampMeas['pna_gpib'])
    if ampMeas['port'] == 'Port1':
        pna.s_parameters(['S11', 'S21'])
    elif ampMeas['port'] == 'Port2':
        pna.s_parameters(['S12', 'S22'])
    pna.power(power)
    pna.bandwidth(bandwidth)
    pna.num_points(int(num_points))
    pna.electrical_delay(edelay)
    pna.frequency_range(start, stop)
    return pna
      
def fluxCorrection(freq):
    corr = freq + (freq - 2.5)*(-1.0/6.0) + 1.1
    return corr
    
    
def setFlux(flux,tune,sb,dc):
    if tune['FluxDac']=='DC rack':
        dc.channel_set_voltage(3, 'A', 1, 0, flux)
    else:
        sb.set_fb(tune['FluxDac'], cxn, startflux*V, tune['FluxChannel'])
        
        
        