import labrad
from labrad.units import Unit
from labrad.types import Value
V, GHz, MHz, Hz, dBm, dB, K, ns = [Unit(s) for s in ('V','GHz','MHz', 'Hz', 'dBm', 'dB', 'K', 'ns')]
import numpy as np
import time
import setbias as sb
import powerScan as ps
import pyle.dataking.util as util
import autoTuneTheAmp as auto
import fullFluxParameterSweepHittite as ffpsh
import fullFluxParameterSweepAnritsu as ffpsa


def measureHittie(sample, frequencies):
    cxn = sample._cxn
    fluxScan = sample['fluxScan']
    hittite = cxn.hittite_t2100_server
    hittite.select_device(fluxScan['Hittite'])
    hittite.output(False)
    hittite.output(True)
    hittite.output(False)
    try:
        for freq in frequencies:
            auto.autoTune(sample, hittite, freq, 60, save = True)
            try:
                ffpsh.measure(sample)
            except:
                print 'scan failed for %f GHz' %freq
            hittite.output(False)   
    except:
        print 'scan failed'
        sb.set_fb(fluxScan['dacBoard'], cxn, 0*V, fluxScan['fastbiasDacChannel'])
        hittite.output(False)   
        
def measureAnritsu(sample, frequencies):
    cxn = sample._cxn
    fluxScan = sample['fluxScan']
    anritsu = cxn.anritsu_server
    anritsu.select_device(fluxScan['Anritsu'])
    anritsu.output(False)
    anritsu.output(True)
    anritsu.output(False)
    try:
        for freq in frequencies:
            auto.autoTune(sample, anritsu, freq, 60, save = True)
            print 'cool'
            try:
                ffpsa.measure(sample)
            except:
                print 'scan failed for %f GHz' %freq
        anritsu.output(False)   
    except:
        print 'scan failed'
        if fluxScan['dacBoard'] == 'DC rack':
            dc = cxn.dc_rack_server
            dc.select_device(0)
            dc.channel_set_voltage(3, 'A', 1, 0, flux)
        else:
            sb.set_fb(fluxScan['dacBoard'], cxn, 0*V, fluxScan['fastbiasDacChannel'])
        anritsu.output(False)

def singleFreqAnritsu(sample, freq):
    cxn = sample._cxn
    fluxScan = sample['fluxScan']
    anritsu = cxn.anritsu_server
    anritsu.select_device(fluxScan['Anritsu'])
    anritsu.output(False)
    anritsu.output(True)
    anritsu.output(False)
    try:
        if fluxScan['PumpMethod'] == 'input w':
            auto.fineFreqMapInputW(sample, anritsu, freq, saveGain = 14, save = True)
        else:
            auto.fineFreqMap2W(sample, anritsu, freq, saveGain = 14)
        try:
            ffpsa.measure(sample)
        except:
            print 'scan failed for %f GHz' %freq
        anritsu.output(False)   
    except:
        print 'scan failed'
        if fluxScan['dacBoard'] == 'DC rack':
            dc = cxn.dc_rack_server
            dc.select_device(0)
            dc.channel_set_voltage(3, 'A', 1, 0, flux)
        else:
            sb.set_fb(fluxScan['dacBoard'], cxn, 0*V, fluxScan['fastbiasDacChannel'])
        anritsu.output(False)         
        
def singleGainAnritsu(sample, freqs, gain):
    cxn = sample._cxn
    fluxScan = sample['fluxScan']
    anritsu = cxn.anritsu_server
    anritsu.select_device(fluxScan['Anritsu'])
    anritsu.output(False)
    anritsu.output(True)
    anritsu.output(False)
    if fluxScan['PumpMethod'] == 'input w':
        auto.broadFreqScanInputW(sample, anritsu, freqs, saveGain = gain)
    else:
        auto.broadFreqScan2W(sample, anritsu, freqs, saveGain = gain)
    try:
        ffpsa.tuneMeasure(sample)
    except:
        print 'scan failed for %f GHz' %freq
    anritsu.output(False)   
    #except:
    #    print 'scan failed'
    #    sb.set_fb(fluxScan['dacBoard'], cxn, 0*V, fluxScan['fastbiasDacChannel'])
    #    anritsu.output(False)              
            
            
    