import numpy as np
import labrad
from labrad.units import ns, GHz, MHz, mV, V, dBm, us,rad
from labrad.units import Unit,Value


from pyle.util import sweeptools as st
from pyle.dataking.DevJim.QND import couplerResonator as cr
from pyle.dataking.DevJim.QND import resonatorReflection as rr
from pyle.analysis import numberAnalysis as na
from pyle.dataking.DevJim.QND import fpgaseq_QND

fpgaseq_QND.PREPAD = 1000

def scans(sample):
    for lenRect in [50*ns,70*ns,0*ns,100*ns,140*ns]:
        lens=[30,100,300,1000]
        lenAmps = [st.r[0:1e-1:1e-3],st.r[0:2e-2:2e-4],st.r[0:4e-3:5e-5],st.r[0:4e-6:5e-8]]
        scales = []
        for len,lenAmp in zip(lens,lenAmps):
            dataTest=cr.coherent(sample,mode='exponentialBuildup',pulseLen=len*ns,
                                 pulseParam=[100*ns,lenRect],amps=lenAmp,
                                 couplingOn=-1.022,save=False,probeLen=0*ns,collect=True)
            whereMax = np.argmax(dataTest[:,1]>0.05)
            if whereMax>0:
                index = whereMax
            else:
                index = np.shape(dataTest)[0]
            ampMax = dataTest[index-1,0]
            cr.coherent(sample,mode='exponentialBuildup',pulseLen=len*ns,
                        pulseParam=[100*ns,lenRect],amps=st.r[0:ampMax:(0.01*ampMax)],
                        couplingOn=-1.022)
            with labrad.connect() as cxn:
                dv = cxn.data_vault
                scales.append(na.coherentanalysis(dv,-1,sample._dir,minContrast=1.3,doPlot=False)[2])
        for numPhoton in [10,100,1,30,3]:
            for len,scale in zip(lens,scales):
                rr.pulseIQOscillations(sample,amps=numPhoton/scale,captureBias=-1.022,
                                       mode='exponentialBuildup',pulseParam=[100*ns,lenRect],
                                       pulseLens=len*ns,save=True,stats=30000,statsCut=600,
                                       name='%d Photons PulseLen Delay'%numPhoton**2,delays=30*ns)