#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
import pyle.dataking.utilMultilevels as ml
import pyle.plotting.dstools as ds
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
from pyle.datavault import DataVaultWrapper
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.sim import pyleSim as ps
from pyle.analysis import iq as iq
from pyle.dataking import singleQubit as sq

from pyle.dataking.benchmarking import randomizedBenchmarking as rb

#import pyle.dataking.DevDaniel.Qubit.readout as readout

from pyle import signalprocessing as signalprocessing
import pyle.quantum as quantum
import time
import copy
from matplotlib.mlab import find
from scipy.interpolate import UnivariateSpline
import itertools
from pyle.dataking.zfuncs import *

import csv

ns, GHz, us, dBm = (Unit(un) for un in ['ns', 'GHz', 'us', 'dBm'])

# 'G1': 'X',
# 'G2': '-X',
# 'G3': 'Y',
# 'G4': '-Y',
# 'G5': 'X/2',
# 'G6': '-X/2',
# 'G7': 'Y/2',
# 'G8': '-Y/2',
# 'G9': 'I'

# 'F1':'{}',
# 'F2':'G6',
# 'F3':'G7',
# 'F4':'G1',
# 'F5':'G8',
# 'F6':'G5'

fiducialToG =  {'F1':'{}',
                'F2':'Gmx2',
                'F3':'Gy2',
                'F4':'Gx',
                'F5':'Gmy2',
                'F6':'Gx2'
                }
                
gToGate = { '{}': '{}',
            'Gx': 'X',
            'Gmx': '-X',
            'Gy': 'Y',
            'Gmy': '-Y',
            'Gx2': 'X/2',
            'Gmx2': '-X/2',
            'Gy2': 'Y/2',
            'Gmy2': '-Y/2',
            'Gi': 'I'
           }

gateDict = {'{}': [0, tomo.sigmaI],
            'X': [np.pi, tomo.sigmaX],
            '-X': [-np.pi, tomo.sigmaX],
            'Y': [np.pi, tomo.sigmaY],
            '-Y': [-np.pi, tomo.sigmaY],
            'X/2': [np.pi/2., tomo.sigmaX],
            '-X/2': [-np.pi/2., tomo.sigmaX],
            'Y/2': [np.pi/2., tomo.sigmaY],
            '-Y/2': [-np.pi/2., tomo.sigmaY],
            'I': [0, tomo.sigmaI]}

           
NUMFIDUCIAL = len(fiducialToG.keys())
NUMG = len(gToGate.keys())-1

def generateGateSetSequences(long = True, L=256):
    '''
    Now, for the experiment count, it's best to think of the fiducials as 5 gates + the null gate, then there are:

    length 3: 5*9*5 = 225  (5 real fiducials)
    length 2: 5*5 + (9-5)*5 + 5*(9-5) = 65  (broken into fiducial,fiducial  ; non-fiducial,fiducial ; fiducial,non-fiducial ) 
    length 1: 9 = 9 (all length 1)
    length 0: 1 = 1

    So a total of 300 experiments, which is the number of data lines in the file I attached.
    '''
    
    
    ### START BASE MEASUREMENTS ###
    fiducials = list(np.copy(fiducialToG.keys()))
    gateSet = list(np.copy(gToGate.keys()))
    gateSet.remove('{}')
    
    fiducialsAsG = [fiducialToG[f] for f in fiducials]
    fiducialsNoNull = list(np.copy(fiducialsAsG))
    fiducialsNoNull.remove('{}')
    
    gateSetMinusFiducial = list(np.copy(gateSet))
    [gateSetMinusFiducial.remove(f) for f in fiducialsNoNull]

    sequences = []
    for fs in fiducialsNoNull: # length 3
        for g in gateSet:
            for fe in fiducialsNoNull:
                sequences.append([fs,g,fe])
    for fs in fiducialsNoNull: # length 2 fiducial,fiducial
        for fe in fiducialsNoNull:
            sequences.append([fs,fe,'{}'])
    for g in gateSetMinusFiducial: # length 2 non-fiducial,fiducial
        for f in fiducialsNoNull:
            sequences.append([g,f,'{}'])
    for f in fiducialsNoNull: # length 2 non-fiducial,fiducial
        for g in gateSetMinusFiducial:
            sequences.append([f,g,'{}'])
    for g in gateSet: # length 1
        sequences.append([g,'{}','{}'])
    sequences.append(['{}','{}','{}']) # null
    print "Number of base sequences: %s" % int(len(sequences))
    ### END BASE MEASUREMENTS ###
    
    ### START LONG SEQUENCES ###
    if long:
        longSequences = []
        # 36x14xlog2(L) sequences
        #{ [1], [3], [5], [7], [9], [1, 2], [1, 3], [1, 7], [3, 4], [3, 5], [5, 6], [7, 8], [1, 3, 5], [1, 3, 7] }
        germs = [['Gx'], ['Gy'], ['Gx2'], ['Gy2'], ['Gi'], ['Gx', 'Gmx'], ['Gx', 'Gy'], ['Gx', 'Gy2'], ['Gy', 'Gmy'], 
                 ['Gy', 'Gx2'], ['Gx2', 'Gmx2'], ['Gy2', 'Gmy2'], ['Gx', 'Gy', 'Gx2'], ['Gx', 'Gy', 'Gy2']]
        Ls = [2**Li for Li in range(np.log2(L)+1)] 
        # sandwich germs w fiducials
        for fs in fiducialsAsG:
            for g in germs:
                for Li in Ls:
                    for fe in fiducialsAsG:
                        currSeq = []
                        currSeq.extend([fs])
                        currSeq.extend(g*Li)
                        currSeq.extend([fe])
                        longSequences.append(currSeq)
        print "Number of long sequences: %s, should be: %s" % (int(len(longSequences)), int(len(fiducialsAsG)**2 * len(germs) * len(Ls)))
        sequences.extend(longSequences)
    ### END LONG SEQUENCES

    return sequences, fiducialsAsG, gateSet, germs
                    
class GateSet(gates.Gate):
    
    def __init__(self, agents, gateList):
        self.gateList = gateList
        self.gateOps = {'{}':lambda q: gates.Wait([q], 0*ns), #N
                        'X':lambda q: gates.PiPulse([q]), #X
                        '-X':lambda q: gates.PiPulse([q],phase=np.pi), #-X
                        'Y':lambda q: gates.PiPulse([q],phase=np.pi/2.), #Y
                        '-Y':lambda q: gates.PiPulse([q],phase=-np.pi/2.), #-Y 
                        'X/2':lambda q: gates.PiHalfPulse([q]), #X/2
                        '-X/2':lambda q: gates.PiHalfPulse([q],phase=np.pi), #-X/2
                        'Y/2':lambda q: gates.PiHalfPulse([q],phase=np.pi/2.), #Y/2
                        '-Y/2':lambda q: gates.PiHalfPulse([q],phase=-np.pi/2.), #-Y/2
                        'I':lambda q: gates.Wait([q],q.get('identityLen',np.min([q['piLen'], q['piHalfLen']]))), #I
                        }

        gates.Gate.__init__(self, agents)          
        
    def updateAgents(self):
        pass
        
    def setSubgates(self, agents):
        gateList = self.gateList        
        ag = self.agents[0]        
        subgates = []
        for opAsG in gateList:
            op = gToGate[opAsG]
            subgates.append(self.gateOps[op](ag))
        self.subgates = subgates
        
def gateTester(Sample, measure, name = 'Gate Tester',
               stats=10, save=False, collect=True):
               
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    sequences = generateGateSetSequences()
    
    alg = gc.Algorithm(agents=devs)
    alg[GateSet([alg.q0], sequences[113])]
    alg[gates.Measure(alg.qubits)]
    alg.compile()
    
    ph.pulses(alg)
    
    rv = runQubits(Sample._cxn.qubit_sequencer, alg.agents, stats=stats, dataFormat='iqRaw', debug=False)
    r = rv.wait()
    Is,Qs = r[0]
    prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
    
def gateSetTomography(Sample, measure, name="Gate-set Tomography", stats=3000, 
       save=True, collect=False, noisy=True, long=True, L=256):
    
    sequences, fiducialsAsG, gateSet, germs = generateGateSetSequences(long=long, L=L)
    index = range(len(sequences))
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(index, 'GST Index')]
    deps = [("Probabiliy |1>",'','')]
    kw = {'stats': stats,'long':long, 'L':L,'fiducialsAsG':str(fiducialsAsG), 'gateSet':str(gateSet), 'germs':str(germs)}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)

    def func(server, idx):
    
        alg = gc.Algorithm(agents=devs)
        if noisy:
            print sequences[idx]
        alg[gates.Wait([alg.q0], 10*ns)] # to fix an issue when we do all null gates.
        alg[GateSet([alg.q0], sequences[idx])]
        alg[gates.Measure(alg.qubits)]
        alg.compile()
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        returnValue([prob])
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    if collect:
        return data
        
def saveGateSetTomoData(dvwData, name=''):
    stats = dvwData.parameters['stats']
    #sequences = dvwData.parameters['sequences'] stupid data vault
    L = dvwData.parameters['L']
    long = dvwData.parameters['long']
    sequences, fiducialsAsG, gateSet, germs = generateGateSetSequences(long=long, L=L)
    
    headers = [['#',dvwData.path],['#',dvwData.name],['##','Columns','=','plus','count,','count','total,','t0,','tf']]
    
    with open(name+'_data.txt', 'wb') as csvfile:
        writer = csv.writer(csvfile, delimiter=' ')
        [writer.writerow(k) for k in headers]
        for seq, dat in zip(sequences,dvwData):
            writer.writerow([''.join(seq),str(dat[1]*stats),str(stats),0,0])
            
def fiducialFile(dvwData, name=''):
    
    fiducials = ['F'+str(k+1) for k in range(NUMFIDUCIAL)]
    gs = [fiducialToG[fid] for fid in fiducials]
    
    with open(name+'_fiducial.txt', 'wb') as csvfile:
        writer = csv.writer(csvfile, delimiter=' ')
        headers = [['#',dvwData.path],['#',dvwData.name],['#','Fiducial','G representation']]
        [writer.writerow(k) for k in headers]
        for fid, g in zip(fiducials, gs):
             writer.writerow([fid, g])
             
def superOperatorFile(name = '', gateSet=None):
    '''
    # Target gates for experimental data using updated file format

    #State prepared, specified as a state
    rho
    StateVec
    1 0

    #State measured as "yes" outcome, specified as a state
    E
    StateVec
    0 1

    #First gate specified as H such that the unitary gate is U = exp(-i*H):  X(pi)
    G1
    UnitaryMxExp
    0    pi/2
    pi/2    0

    #Second gate, specified as H such that the unitary gate is U = exp(-i*H): X(pi/2)
    G2
    UnitaryMxExp
    0     pi/4
    pi/4     0
    '''
    if gateSet is None:
        gateSet = gToGate.keys() # in G
        gateSet.remove('{}')
            
    with open(name+'_gates.txt', 'wb') as csvfile:
        writer = csv.writer(csvfile, delimiter=' ')
        headers = [['#','Target','gates','for','experimental','data','using','updated','file','format'],
                   [],
                   ['#','State','prepared,','specified','as','a','state'],
                   ['rho'],
                   ['StateVec'],
                   [1,0],
                   [],
                   ['#','State','measured','as','yes','outcome,','specified','as','a','state'],
                   ['E'],
                   ['StateVec'],
                   [0,1],
                   []]
        [writer.writerow(k) for k in headers]
        for G in gateSet:
            writer.writerow([G])
            writer.writerow(['UnitaryMxExp'])
            angle, pauli = gateDict[gToGate[G]]
            currMatrix = angle/2.*pauli
            writer.writerow([currMatrix[0,0],currMatrix[0,1]])
            writer.writerow([currMatrix[1,0],currMatrix[1,1]])
            writer.writerow([])
            
def generateAllFiles(dvwData):

    path = dvwData.path
    name = '_'.join(path)[1:]

    saveGateSetTomoData(dvwData,name=name)
    fiducialFile(dvwData,name=name)
    superOperatorFile(name=name)
    