#Standard packages
import numpy as np
from math import atan2
import scipy
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import pyle

import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking import singleQubit as sq
from pyle.dataking import multipleQubit as mq
import pyle.plotting.dstools as ds
import pyle.plotting.plothelper as ph
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
from pyle.datavault import DataVaultWrapper
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.analysis import iq as iq

from pyle.dataking.benchmarking import randomizedBenchmarking as rb

#import pyle.dataking.DevDaniel.Qubit.readout as readout

from pyle import signalprocessing as signalprocessing
import pyle.quantum as quantum
import time
import copy
from matplotlib.mlab import find
from scipy.interpolate import UnivariateSpline
import itertools
from pyle.dataking import zfuncs as zfuncs

def fixUnits(Sample, measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']):
    
    def stripUnits(thing):
        if isinstance(thing,Value):
            thing = thing.value
        return thing
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    for q,Q in zip(qubits,Qubits):
        rates = q['settlingRates']
        rates = [stripUnits(r) for r in rates]
        Q['settlingRates'] = [r*GHz for r in rates]
        Q['adc demod phase'] = stripUnits(Q['adc demod phase']) * rad

def setKey(Sample, measure = None, key = None, value = None):

    if measure=='all':
        measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']
    if measure=='data':
        measure = ['q0', 'q2', 'q4', 'q6', 'q8']
    if measure=='measure':
        measure = ['q1','q3','q5', 'q7']
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for Q in Qubits:
        Q[key] = value
        
def killKey(Sample, key, measure = ['q0', 'q1', 'q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8']):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for Q in Qubits:
        try: 
            del Q[key] 
        except:
            print 'missing key'
            
    '''
    n.killKey(s,'q0aczAmp')
    n.killKey(s,'q1aczAmp')
    n.killKey(s,'q2aczAmp')
    n.killKey(s,'q3aczAmp')
    n.killKey(s,'q4aczAmp')

    n.killKey(s,'q0aczTime')
    n.killKey(s,'q1aczTime')
    n.killKey(s,'q2aczTime')
    n.killKey(s,'q3aczTime')
    n.killKey(s,'q4aczTime')

    n.killKey(s,'q0aczPhase')
    n.killKey(s,'q1aczPhase')
    n.killKey(s,'q2aczPhase')
    n.killKey(s,'q3aczPhase')
    n.killKey(s,'q4aczPhase')

    n.killKey(s,'q0aczLambda')
    n.killKey(s,'q1aczLambda')
    n.killKey(s,'q2aczLambda')
    n.killKey(s,'q3aczLambda')
    n.killKey(s,'q4aczLambda')

    n.killKey(s,'q0aczW')
    n.killKey(s,'q1aczW')
    n.killKey(s,'q2aczW')
    n.killKey(s,'q3aczW')
    n.killKey(s,'q4aczW')
    '''
            
def makeAczKey(Sample, measure, key = 'acz'):
    '''
    key = 'acz' or 'detune'
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if key == 'acz':
    
        qTName = qubits[1].__name__
        qCName = qubits[0].__name__
         
        Qubits[0][qTName+'aczTime'] = 40*ns
        Qubits[0][qTName+'aczW'] = 1.5*ns
        Qubits[0][qTName+'aczAmp'] = -0.5*GHz
        Qubits[0][qTName+'aczLambda'] = [1.0, 0.05, 0.6, -0.2, 0.05]
        Qubits[0][qTName+'aczPhase'] = 0.0
        Qubits[1][qCName+'aczPhase'] = 0.0
        
def makeaSwapKey(Sample, measure, targetName = 'rr'):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if targetName:
        qTName = targetName
        qCName = qubits.__name__
    else:
        qTName = qubits[1].__name__
        qCName = qubits[0].__name__
    
    if isinstance(qubits, list):
        Qubit = Qubits[0]
    else:
        Qubit = Qubits
    
    Qubit[qTName+'aSwapTime'] = 40*ns
    Qubit[qTName+'aSwapW'] = 0.0*ns
    Qubit[qTName+'aSwapAmp'] = 0.5*GHz
    Qubit[qTName+'aSwapLambda'] = [1.0, 0.1, np.pi-0.1, -0.2, 0.0]

def allLevels(Sample, measure = ['q0','q1','q2','q3','q4']):
    
    ph.levels(Sample, measure)
    
    #assume odd qubits are measure qubits
    measureQubits = measure[1::2]
    dataQubits = measure[::2]
    
    pairs = []
    for k,mq in enumerate(measureQubits):
        pairs.append([mq, dataQubits[k]])
        pairs.append([mq, dataQubits[k+1]])
        
    for pair in pairs:  
        ph.aczLevels(Sample, measure, pair)

def dataset_to_comp(Sample, dvwData, removeOffset=True):
    
    if not isinstance(dvwData,list):
        dvwData = [dvwData]
    
    measure = dvwData[0].parameters['measure'][0]
    
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    zEnvs = []
    for data in dvwData:
    
        curve = fitting.minimaCuts(data, 'columns', ptsToFit=5, invert=True)
        indices = np.argsort(curve[:,0])
        acStarkShift = curve[:,1] - q['f10']['GHz']
        acStarkShift = acStarkShift[indices]
        
        #Convert frequency to z amplitude
        params = q['zFuncAna']
        M = params[0].value
        fmax = params[1].value
        offset = params[2].value
        fc = params[3].value
        def fitFunc(x, M, fmax, offset, fc):
            return fmax * np.sqrt(np.abs(np.cos(np.pi*M*np.abs(x-offset)))) +\
                   fc * (np.sqrt(np.abs(np.cos(np.pi*M*np.abs(x-offset)))) -1.0)
        zControl = np.linspace(-0.3, 0.3, 1000)
        dfControl = fitFunc(zControl, M, fmax, offset, fc) - q['f10']['GHz']
        
        interp_freq2z = scipy.interpolate.interp1d(dfControl[::-1], zControl[::-1], kind='linear')
        z = interp_freq2z(acStarkShift)
        if removeOffset:
            z += -z[0]
        zEnv = env.GRAPE_CZ(0, curve[indices,0], -z)
        zEnvs.append(zEnv)
        
    return sum(zEnvs)

def stark_compensation(Sample, measure, name='AC Stark Z Compensated',
                       df=0.250*GHz, npts = 30, delays=None, zCompensation=None, zEnv=None,
                       piPulse=False, excitationLen=None, excitationAmp=0.1,
                       stats=900, save=True):
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)
    
    
    f10_GHz = q['f10']['GHz']
    freqs = np.linspace(f10_GHz-df['GHz']/2.,f10_GHz+df['GHz']/2.,npts)*GHz
    if delays is None:
        rl = (excitationLen or q['readoutLen'])['ns'] 
        maxDelay = rl+750.0
        delays = st.r[-30:maxDelay:10, ns]
    else:
        rl = (excitationLen or q['readoutLen'])['ns']
        maxDelay = rl+750.0
    
    axes=[(zCompensation, 'z compensation'), (delays, 'Delay'), (freqs, 'Frequency')]
    deps = [('Probability', 'P1', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)
    def func(server, currZ, delay, freq):
        alg = gc.Algorithm(agents=devs)
        for qubit in alg.agents:
            if 'fc' in qubit:
                qubit['fc'] = q['f10'] - 225*MHz
        if piPulse:
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.PiPulse([alg.q0], freq=freq, state=None)]
        alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        if excitationLen is None:
            alg.q0.measure(delay = -maxDelay-delay)
        else:
            df = alg.q0['readoutFrequency'] - alg.q0['readoutDevice']['carrierFrequency']
            alg.q0['rr'] += env.mix(env.flattop(-maxDelay-delay, excitationLen, 5.0, amp=excitationAmp), df)
        if zEnv is not None:
            start, end = alg.q0['z'].start, alg.q0['z'].end
            alg.q0['z'] += env.shift(zEnv, -delay)
            alg.q0['z'].start, alg.q0['z'].end = start, end
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray)
        returnValue([np.mean(prob)])
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    return data
    
def stark_compensation_on_demand(Sample, measure, name='AC Stark Z Compensation on demand',
                       zrange = 0.2, npts = 30, delays=None,
                       stats=900, save=True):
    
    zrange = np.linspace(-zrange/2., zrange/2., npts)
                       
    sample, devs, q, Q = gc.loadQubits(Sample, measure, write_access=True)

    if delays is None:
        rl = q['readoutLen']['ns']
        maxDelay = rl+750.0
        delays = st.r[-30:maxDelay:10, ns]
    else:
        rl = q['readoutLen']['ns']
        maxDelay = rl+750.0
    
    def sweep():
        for d in delays:
            for z in zrange:
                yield d,z
                
    sweepData = {
        'cumZ':np.zeros((len(delays))),
        'dIdx':0,
        'zIdx':0,
        'currProbs':0*zrange,
        }
        
    # construct the current z
    def genCurrZ(idx, z):
        cumZ = np.copy(sweepData['cumZ'])
        cumZ[idx] += z
        
        if idx>0:
            if idx<len(delays)-1:
                cumZ[idx+1] = 2*z + cumZ[idx-1]
                cumZ[idx+2] = 3*z + cumZ[idx-1]
        else:
            cumZ[idx+1] += 2*z
        return cumZ
    
    axes=[('Delay','ns'),('Z', 'a.u.')]
    deps = [('Probability', 'P1', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps,measure=measure, kw=kw)
    
    def func(server, args):
        currDelay,currZ = args
        zIdx = sweepData['zIdx']
        dIdx = sweepData['dIdx']
        print 'currDelay',str(currDelay),'currZ',str(currZ),'zIdx',str(zIdx),'dIdx',str(dIdx)
        currZdata = genCurrZ(dIdx, currZ)
        zEnv = env.wigglyZ(0, np.array(list(delays)), np.array(currZdata))
        
        alg = gc.Algorithm(agents=devs)
        #alg[gates.Wait([alg.q0], alg.q0['piLen']*-1/2.)]
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Wait([alg.q0], maxDelay)]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        alg.q0.measure(delay = -maxDelay-currDelay)
        start, end = alg.q0['z'].start, alg.q0['z'].end
        alg.q0['z'] += env.shift(zEnv, -currDelay)
        alg.q0['z'].start, alg.q0['z'].end = start, end
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(q, Is.asarray+1.0j*Qs.asarray))  
        
        sweepData['currProbs'][zIdx] = prob
        
        if zIdx+1 != npts: #Not done with this z
            sweepData['zIdx']+=1
        else: #Update fit
            bestZ = getMaxGauss(zrange, sweepData['currProbs'])
            print 'best Z was',str(bestZ)
            sweepData['cumZ'] = genCurrZ(dIdx,bestZ)
            sweepData['dIdx'] += 1
            sweepData['zIdx'] = 0
        
        returnValue([currDelay, currZ, prob])
    data = sweeps.run(func, sweep(), dataset=dataset, save=save, noisy=False, pipesize=1)
    return data, env.wigglyZ(0, np.array(list(delays)), np.array(sweepData['cumZ']))
    
def rebuild_compensation_on_demand(Sample, dvwData):
    
    data = dvwData
    data=np.array(data)
    dataSize = len(matplotlib.mlab.find(data[:,0]==data[0,0]))
    dataLen = len(data)
    print dataLen,dataSize
    numColumns = dataLen/dataSize
    newDataLen = numColumns*dataSize
    data = data[:newDataLen, :] # we take off any unfinished rows
    delays = data[::dataSize, 0]
    zrange = data[:dataSize,1]
    probs = np.reshape(data[:,2], (numColumns, dataSize))

    cumZ = np.zeros((len(delays)))
        
    # construct the current z
    def genCurrZ(idx, z, cumZ):
        cumZ[idx] += z
        
        if idx>0:
            if idx<len(delays)-1:
                cumZ[idx+1] = 2*z + cumZ[idx-1]
                cumZ[idx+2] = 3*z + cumZ[idx-1]
        else:
            cumZ[idx+1] += 2*z
        return cumZ
    
    for delay, prob in zip(delays, probs):
        bestZ = getMaxGauss(zrange, prob)
        idx = np.argwhere(delays==delay)[0][0]
        cumZ = genCurrZ(idx, bestZ, cumZ)
        
    return env.wigglyZ(0, np.array(list(delays)), np.array(cumZ))
     
def dephasingVsTime(Sample, measure, echo = False, name = 'Dephasing vs. Time',
            stats=600, freq = None, delay = 500*ns, iteration = st.r[0:500:1],
            fringeFreq = 1*MHz,
            riseTime = 2*ns,
            detuneAmp=None, iterations = 1,
            reset=True, plot=True, update=True,
            noisy=True, save=True, collect=False,detuneLen=None):
        
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
        
    axes = [(iteration,'Iteration'), (delay,'delay time')]  
    deps = [('Phase Envelope', '', '')]
    kw = {'stats': stats, 'riseTime':riseTime}
    tomoPhases = {'+X': 0.0, '+Y':0.25, '-X': -0.5, '-Y':-0.25}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, currIter, currDelay):

        reqs = []
        
        for tomoPhase in tomoPhases:
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0],currDelay)]
            alg[gates.PiHalfPulse([alg.q0], phase = 2*np.pi*(fringeFreq['GHz']*currDelay['ns'] + tomoPhases[tomoPhase]))]
            alg[gates.Measure([alg.q0])]
            alg.compile()
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))

        results = yield FutureList(reqs)
        data = []
            
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(prob)
                        
        phaseAmplitude = np.sqrt((data[0]-data[2])**2 + (data[1]-data[3])**2)
        
        returnValue([phaseAmplitude])    
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)

def wwScan(Sample, measure, freqIdx = 0, freqs = st.r[-0.3:-0.05:0.01, GHz],
             pingPongs=3, stats=900, manual=False,
             name='WW scan', save=True, collect=False, noisy=True, update=False):
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(freqs, 'WW Freq')]
    deps = [('Probability |1>','X',''),('Probability |1>','Y','')]
    kw = {'stats': stats, 'pingPongs': pingPongs, 'freqIdx':freqIdx}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currF):          
        
        reqs = []
         
        # ping pong sequence X
        alg = gc.Algorithm(agents=devs)
        currWW = alg.q0['ww']
        currWW[freqIdx] = currF
        alg.q0['ww'] = currWW
        alg[gates.PiHalfPulse([alg.q0])]
        for k in range(pingPongs):
            alg[gates.PiHalfPulse([alg.q0], phase=0.0)]
            alg[gates.PiHalfPulse([alg.q0], phase=-np.pi)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))

        # ping pong sequence Y
        alg = gc.Algorithm(agents=devs)
        currWW = alg.q0['ww']
        currWW[freqIdx] = currF
        alg.q0['ww'] = currWW
        alg[gates.PiHalfPulse([alg.q0])]
        for k in range(pingPongs):
            alg[gates.PiHalfPulse([alg.q0], phase=np.pi/2.)]
            alg[gates.PiHalfPulse([alg.q0], phase=-3*np.pi/2.)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        results = yield FutureList(reqs)
        r = results[0]
        Is,Qs = r[0]
        probX = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        r = results[1]
        Is,Qs = r[0]
        probY = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
   
        returnValue([probX, probY])    
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if update and manual:
        alphas = data[:,0]
        traces = [{'x': alphas, 'y': data[:,1], 'args': ('-ob',)},
                  {'x': alphas, 'y': data[:,2], 'args': ('-og',)},
                  {'x': alphas, 'y': [0]*len(alphas), 'args': ('-k',)}]
        params = [{'name': 'Qubit alpha DRAG', 'val': float(qubit['alphaDRAG']), 'range':(min(alphas), max(alphas)), 'axis': 'x', 'color': 'b'}]
        result = adjust.adjust(params, traces)
        if result:
            Qubit['alphaDRAG'] = result['Qubit alpha DRAG']
            if collect:
                return result
                
    elif update and not manual:
        amps = data[:,0]
        dat_x = data[:,1]
        dat_y=data[:,2]
        bestAmp1 = getMaxGauss(amps,dat_x,fitToMax=True) 
        bestAmp2 = getMaxGauss(amps,dat_y,fitToMax=True)
        bestAmp = 0.5 * (bestAmp1+bestAmp2)
        Qubit['alphaDRAG'] = bestAmp
        
    elif collect:
        return data 
        
def testWW(Sample, measure, freqs = st.r[-0.3:-0.05:0.01, GHz],
             nPulses = 1,
             stats=900,
             name='WW test', save=True, collect=False, noisy=True):
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(freqs, 'WW Freq'),(nPulses,'Num Pi Pulses')]
    deps = [('Probability |1>','','')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currF, nPulses):          

        alg = gc.Algorithm(agents=devs)
        alg.q0['f10'] -= 0.100*GHz
        alg.q0['f21'] -= 0.100*GHz
        if currF:
            alg.q0['ww'][1] = currF
        alg[gates.NPiPulses([alg.q0], nPulses)]
        alg[gates.Measure([alg.q0])]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        prob = np.mean(sq.readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        returnValue([prob]) 
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data 
        
def sameReadout(Sample, measure):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    readoutParams = []
    for q,Q in zip(qubits,Qubits):
        len = q['readoutLen']
        ringdown = q['ringdownLen']
        ringdownIntegration = q['readoutIntegrateIntoRingdown']
        currParams = [len, ringdown, ringdownIntegration]
        readoutParams.append(currParams)
        Q['readoutBak'] = currParams
    
    maxLen, maxRingdown, maxRingdownIntegration = np.max(np.array(readoutParams),axis=0)
    for Q in Qubits:
        Q['readoutLen'] = maxLen*ns
        Q['ringdownLen'] = maxRingdown*ns
        Q['readoutIntegrateIntoRingdown'] = maxRingdownIntegration*ns
    
    for meas in measure:
        sq.calculateReadoutAxis(Sample, meas)
        
def restoreReadout(Sample, measure):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    for q,Q in zip(qubits,Qubits):
        len, ringdown, ringdownIntegration = q['readoutBak']
        Q['readoutLen'] = len
        Q['ringdownLen'] = ringdown
        Q['readoutIntegrateIntoRingdown'] = ringdownIntegration
        
    for meas in measure:
        sq.calculateReadoutAxis(Sample, meas)
      
def cyclePhase(Sample, measure, check='before',stats=900, name = "Cycle Phase", 
               echo=False, fakeHerald=False, state=[0,0,0], thetas = 21, 
               save=True, plot=False, update=False, debug=False):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)
    
    axes = [(thetas,'Ramsey Theta')]
    deps = [('Probability |1>','','')]
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, theta):
        alg = gc.Algorithm(agents=devs)
        if fakeHerald:
            for q in alg.qubits:
                alg[gates.Readout([q], adcWindow=False)]
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        alg[gates.PiHalfPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q0])]
        if check == 'before':
            alg[gates.PiHalfPulse([alg.q1], phase = theta)]
            alg[gates.Measure(alg.qubits, sync='start')]
        if echo:
            alg[gates.PiPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q2])]
        if check =='after':
            alg[gates.PiHalfPulse([alg.q1], phase = theta)]
            alg[gates.Measure(alg.qubits, sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg,timeStep=2)

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure, debug=debug)
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure)[1]

        returnValue([probs])
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(data[:,0], data[:,1])
        deltaPhi = mapMinusPiToPi(phi)
        if check == 'before':
            Qubits[1][qubits[0].__name__+'aczPhase'] -= deltaPhi
        elif check =='after':
            Qubits[1][qubits[2].__name__+'aczPhase'] -= deltaPhi
      
def singleCycle3(Sample, measure, stats=900, name = "Single Cycle 3", save=True, plot=True, echo=False, state=[0,0,0]):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    derp = [1]
    axes = [(derp,'Derp')]
    deps = [('Probability |1>','','')]*3
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        #alg[gates.Readout([alg.q1], adcWindow=False)]
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        alg[gates.PiHalfPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q0])]
        if echo:
            alg[gates.PiPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q2])]
        alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]
        alg[gates.Measure(alg.qubits, sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg,timeStep=2)

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure)

        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)     

def singleCycle5(Sample, measure, stats=900, cycles = [5], name = "Single Cycle 5", save=True, plot=True, echo=False, state=[0,0,0]):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(cycles,'Cycles')]
    deps = [('Probability |1>','','')]*5
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        alg[gates.PiHalfPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q0])]
        if echo:
            alg[gates.PiPulse([alg.q1])]
        alg[gates.adiabaticCZ([alg.q1,alg.q2])]
        alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]
        alg[gates.Sync([alg.q1,alg.q3])]
            
        alg[gates.PiHalfPulse([alg.q3])]
        alg[gates.adiabaticCZ([alg.q3,alg.q4])]
        if echo:
            alg[gates.PiPulse([alg.q3])]
        alg[gates.adiabaticCZ([alg.q3,alg.q2])]
        alg[gates.PiHalfPulse([alg.q3], phase = -np.pi)]
        alg[gates.Measure([alg.q1,alg.q3], sync='start')]
        alg[gates.Sync(alg.qubits)]
        alg[gates.Measure([alg.q0,alg.q2,alg.q4], sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg,timeStep=2)

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure)

        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)       
       
def cycle3(Sample, measure, stats=10, cycles = [5], name = "The flippin' repetition code (3)", save=True, plot=False, echo=False, state=[0,0,0]):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if isinstance(cycles,list) is False:
        cycles = [cycles]
    
    axes = [(cycles,'Cycles')]
    deps = []
    for q in qubits:
        deps.append(('%s Is' % q.__name__, '', '')) # for a given cycle, we are going to concatenate all of the stats
        deps.append(('%s Qs' % q.__name__, '', ''))
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        alg[gates.Measure(alg.qubits, sync='end')]
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        for x in range(currCycle):
            alg[gates.PiHalfPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q0])]
            if echo:
                alg[gates.PiPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q2])]
            alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]

            if x +1 != currCycle:
                alg[gates.Measure([alg.q1])]
        alg[gates.Measure([alg.q0,alg.q1,alg.q2], sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRawLocalDemod', localDemod=True)
        data_for_dv = iq.format_something_useful_for_data_vault(data)
        returnValue(data_for_dv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1)       
        
def cycle5(Sample, measure, stats=10, cycles = [5], name = "The flippin' repetition code (5)", save=True, plot=False, echo=False, state=[0,0,0,0,0]):

    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if isinstance(cycles,list) is False:
        cycles = [cycles]
    
    axes = [(cycles,'Cycles')]
    deps = []
    for q in qubits:
        deps.append(('%s Is' % q.__name__, '', '')) # for a given cycle, we are going to concatenate all of the stats
        deps.append(('%s Qs' % q.__name__, '', ''))
    
    kw = {'stats': stats,'echo': echo}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currCycle):
        alg = gc.Algorithm(agents=devs)
        alg[gates.Measure(alg.qubits, sync='end')]
        for st, q in zip(state, alg.qubits):
            if st:
                alg[gates.PiPulse([q])]
        for x in range(currCycle):
            alg[gates.PiHalfPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q0])]
            if echo:
                alg[gates.PiPulse([alg.q1])]
            alg[gates.adiabaticCZ([alg.q1,alg.q2])]
            alg[gates.PiHalfPulse([alg.q1], phase = -np.pi)]
            alg[gates.Sync([alg.q1,alg.q3])]
            
            alg[gates.PiHalfPulse([alg.q3])]
            alg[gates.adiabaticCZ([alg.q3,alg.q4])]
            if echo:
                alg[gates.PiPulse([alg.q3])]
            alg[gates.adiabaticCZ([alg.q3,alg.q2])]
            alg[gates.PiHalfPulse([alg.q3], phase = -np.pi)]
            
            if x +1 != currCycle:
                alg[gates.Measure([alg.q1,alg.q3], sync='start')]
        alg[gates.Measure([alg.q1,alg.q3], sync='start')]
        alg[gates.Sync(alg.qubits)]
        alg[gates.Measure([alg.q0,alg.q2,alg.q4], sync='start')]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)

        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRawLocalDemod', localDemod=True)
        data_for_dv = iq.format_something_useful_for_data_vault(data)
        returnValue(data_for_dv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1)
    
def recoverCycleData(dvwData):

    '''
    We want Is, Qs = data[cycle][stat][qubit]
    '''
    stats = dvwData.parameters['stats']
    qubits = dvwData.parameters['measure']
    cycles = set(dvwData[:,0])
    
    data_to_return = np.zeros((len(qubits), stats, len(cycles)))
    data_to_return = [[[[] for _ in qubits] for _ in range(stats)] for _ in cycles]
    
    for cn, cycle in enumerate(cycles):
        cycleData = []
        cycleIdx = np.argwhere(dvwData[:,0]==cycle).flatten()
        currData = dvwData[cycleIdx, :]
        for qn, qubit in enumerate(qubits):
            Is = currData[:,qn*2+1].reshape(stats, -1)
            Qs = currData[:,qn*2+2].reshape(stats, -1)
            statsForQubit = []
            for Ii,Qi,sn in zip(Is,Qs,range(stats)):
                Ii = Ii[~np.isnan(Ii)]
                Qi = Qi[~np.isnan(Qi)]
                data_to_return[cn][sn][qn] = np.vstack((Ii,Qi)).transpose()
                
    dvwData.formattedCycleData = data_to_return
    dvwData.cycles = cycles
    return cycles, data_to_return, dvwData
    
def cycleDataToProbs(mutated_dvwData):
    
    '''
    prob = data[cycle][stat][qubit]
    '''
    
    formattedCycleData = mutated_dvwData.formattedCycleData
    qubits = [mutated_dvwData.parameters[key] for key in mutated_dvwData.parameters['measure']]
    cycles = mutated_dvwData.cycles
    stats = mutated_dvwData.parameters['stats']
    probs_to_return = [[[[] for _ in qubits] for _ in range(stats)] for _ in cycles]
    
    for cn, cycle in enumerate(cycles):
        for sn in range(stats):
            for qn, qubit in enumerate(qubits):
                Is,Qs = formattedCycleData[cn][sn][qn].transpose()
                probs = []
                for Ii,Qi in zip(Is, Qs):
                    prob = sq.readoutProbabilities(qubit, Ii+1.0j*Qi)
                    if prob == True:
                        prob = 1
                    else:
                        prob = 0
                    probs.append(prob)
                probs_to_return[cn][sn][qn] = probs
    
    mutated_dvwData.cycleProbs = probs_to_return
    
    if mutated_dvwData.parameters.get('echo',False):
        pass
    
    return probs_to_return, mutated_dvwData
    
def cycleData(dvwData): 

    cycles, data_to_return, dvwData = recoverCycleData(dvwData)
    probs_to_return, mutated_dvwData = cycleDataToProbs(dvwData)
    
    return probs_to_return
    
def cycleIQ(dvwData):

    cycles, data_to_return, dvwData = recoverCycleData(dvwData)
    
    for cycle in data_to_return:
        for stat in cycle:
            for k,qubit in enumerate(stat):
                if k % 2: # odd
                    Is = qubit[:,0]
                    Qs = qubit[:,1]
                    plt.plot(Is,Qs,'ob')
    
def reset21(Sample, measure, name = 'Reset',
                      amplitude = np.linspace(-1.2*np.pi, 1.2*np.pi, 51),
                      freq = None, riseTime = 1*ns, resettime=100*ns,
                      averages = 1, 
                      stats=900, correctXtalkZ=False,
                      save=True, collect=False, noisy=True,state=2):
    ''' Make sure your 2-state is tuned up in amplitude and frequency. '''
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if freq is None:
        freq = qubit['readoutFrequency'] + (qubit['f10']-qubit['f21'])
    
    axes = [(freq,'Reset Freqency'),(resettime, 'Reset Time'),(amplitude, 'Pi12 Angle')]
    deps = [('Probability |1> -> Ae', '', ''),('Probability |1> -> Ag', '', '')]
    
    kw = {'stats': stats, 'averages':averages}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currFreq, currThermTime, currAngle):

        #ref: 1 state
        alg = gc.Algorithm(agents=devs)
        alg[gates.MoveToState([alg.q0], 0, 1)]    
        
        if currFreq:
            amp = fToAmp(alg.q0, currFreq)
            alg[gates.Detune([alg.q0], length=currThermTime, w=riseTime, amp=amp)]
        alg[gates.RotPulse([alg.q0], currAngle, state=1)]
        alg[gates.Measure([alg.q0])]
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probAe = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        #exp: 2 state
        alg = gc.Algorithm(agents=devs)
        alg[gates.MoveToState([alg.q0], 0, 2)]         
        if currFreq:
            amp = fToAmp(alg.q0, currFreq)
            alg[gates.Detune([alg.q0], length=currThermTime, w=riseTime, amp=amp)]
        alg[gates.RotPulse([alg.q0], currAngle, state=1)]
        alg[gates.Measure([alg.q0])]
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probAg = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        returnValue([probAe, probAg])        

    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def timingLagIQ(Sample, measure, delay=st.r[-10:10:1,ns],
              stats=600, update=True,
              save=True, name='Timing Lag IQ', plot=False, noisy=True):
              
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if np.alen(qubit.get('xyPadding',None))>1:
        if q.get('xyPadding',None)[0] != q.get('xyPadding',None)[1]:
            raise Exception("You have asymmetric XY padding, you're really about to screw yourself over")
            
    origAdcTimingLag = qubit['readoutDevice']['adcTimingLag']
    origtimingLagRRUwave = qubit['readoutDevice']['timingLagRRUwave']
    
    axes = [(delay, 'Time')]
    deps = [('Clicks', 'I for %s' %qubit.__name__, ''), ('Clicks', 'Q for %s'%qubit.__name__, '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currRRlag):
    
        alg = gc.Algorithm(agents=devs)
        alg.q0['readoutDevice']['timingLagRRUwave'] = currRRlag
        alg[gates.Measure([alg.q0])]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        returnValue(np.vstack((Is,Qs)).T)
        
    result = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
def waitLagIQ(Sample, measure, delay=st.r[-10:10:1,ns],
              stats=600, update=True, before=True,
              save=True, name='Timing Lag IQ', plot=False, noisy=True):
              
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if np.alen(qubit.get('xyPadding',None))>1:
        if q.get('xyPadding',None)[0] != q.get('xyPadding',None)[1]:
            raise Exception("You have asymmetric XY padding, you're really about to screw yourself over")
            
    origAdcTimingLag = qubit['readoutDevice']['adcTimingLag']
    origtimingLagRRUwave = qubit['readoutDevice']['timingLagRRUwave']
    
    axes = [(delay, 'Time')]
    deps = [('Clicks', 'I for %s' %qubit.__name__, ''), ('Clicks', 'Q for %s'%qubit.__name__, '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, currDelay):
    
        alg = gc.Algorithm(agents=devs)
        if before:
            alg[gates.Wait([alg.q0],currDelay)]
            alg[gates.Measure([alg.q0])]
        else:
            alg[gates.Measure([alg.q0])]
            alg[gates.Wait([alg.q0],currDelay)]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        returnValue(np.vstack((Is,Qs)).T)
        
    result = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
def detuneTroubleShootAg2(Sample, measure, name = 'Detune Trouble Shoot Agilent',
            thetas=np.linspace(-np.pi/4,7*np.pi/4,100), 
            stats=1200, iterations=6,
            update=False, padding=0*ns, 
            noisy=True, save=False, collect=False,plot=True):
        
    """
    """
    cxn=Sample._cxn
    scope = cxn.agilent_dso91304a_oscilloscope()
    scope.select_device()     
    
    scope.reset()
    print 'scope reset'  
    time.sleep(2) 
    
    SCOPECHANNEL=1
    TRIGGERCHANNEL=2
    
    p = scope.packet().\
    channelOnOff(SCOPECHANNEL,'ON').\
    channelOnOff(TRIGGERCHANNEL,'ON').\
    scale(SCOPECHANNEL,0.02).\
    scale(TRIGGERCHANNEL,0.5).\
    position(SCOPECHANNEL,0.0).\
    position(TRIGGERCHANNEL,0.0).\
    horiz_scale(150.0e-9).\
    horiz_position(0.0).\
    trigger_sweep('TRIG').\
    trigger_mode('EDGE').\
    trigger_edge_slope('POS').\
    trigger_at(TRIGGERCHANNEL,0.5).\
    averagemode(1).\
    numavg(512).\
    gpib_write('ACQUIRE:COMPLETE:STATE ON').\
    gpib_write('ACQUIRE:COMPLETE 100')
    #yield p.send()  
    p.send()
    
    #trigger_at(TRIGGERCHANNEL,0.5).\
    print 'scope packet sent'
    time.sleep(2)     
    #numavg(128).\
    #acquisition_mode('AVE')   
    
    #RUN SCRIPT
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    axes = [(iterations, 'Detune Iteration')]
    deps = [('Ramsey Amplitude', '', ''),
            ('Ramsey Phase', '', '')]
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    
    def func(server, iteration):
    
        reqs = []
        
        for theta in thetas:
        
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiHalfPulse([alg.q0])]
            
            alg[gates.Wait([alg.q0], padding)]
            
            for k in range(iteration):
                alg[gates.Detune([alg.q0])]
                if iteration>1 and k<iteration-1:
                    alg[gates.Wait([alg.q0],50*ns)]
                    
            alg[gates.Wait([alg.q0], padding)]
                
            alg[gates.PiHalfPulse([alg.q0], phase = theta)]
            alg[gates.Measure([alg.q0])]
            alg.compile()    
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        data = []
        
        results = yield FutureList(reqs)
        for r in results:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            data.append(prob)
        
        amp, phase = fitting.fitRamsey(thetas, data)
        
        returnValue([amp, phase])

    data = sweeps.grid(func, axes, dataset=False and dataset, noisy=False)
    
    
    #GET TRACE
    [timeAxis,y] = scope.get_trace(SCOPECHANNEL) #start and stop in ns
    
    if save:
        dv = cxn.data_vault
        dv.cd(sample._dir)
        savename = 'detuneTroubleShoot'
        dv.new(savename, ['time [ns]'],['Voltage'])
        all = [[timeAxis[i],y[i]] for i in range(len(timeAxis))]
        dv.add(all)
        print 'saved'      
    
    
    if plot:
        plt.figure()
        plt.plot(timeAxis,y)
        plt.show()
        
    if collect:
        return data 
        
def forceRingdownQST(Sample, measure, name='Forced Resonator Ringdown QST',
            amp1 = None, amp2 = None, phase1 = None, phase2 = None, 
            ringdown = True,
            delay = 0*ns, tomoDelay = 100*ns,
            stats=900, correctXtalkZ=True,
            save=True, collect=False, noisy=True):
    '''
    We want to try and tune up a ringdown pulse after readout to reduce the photon decay time.
    The tricky part is, we have to have pulses that ring down BOTH |0> and |1> simultaneously.
    We check how well we are doing by preparing a |0>, readout, then quantify how well a pi pulse works,
    and also preparing a |1> and doing the same thing. The Pi at the end of each of these experiments 
    will only work well if the reasonator is properly rung down.
    '''
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]   
    
    if amp1 is None:
        amp1, _ = qubit['ringdownPulseAmps']
    if amp2 is None:
        _, amp2 = qubit['ringdownPulseAmps']
    if phase1 is None:
        phase1, _ = qubit['ringdownPulsePhases']
    if phase2 is None:
        _, phase2 = qubit['ringdownPulsePhases']
    
    axes = [(amp1, 'Amplitude 1'), (amp2, 'Amplitude 2'), (phase1, 'Phase 1'),(phase2, 'Phase 2'),(delay,'Delay')]
    deps = [('Phase |0>', 'rad', ''), ('Phase |1>', 'rad', ''),('Amp 0','',''),('Amp 1','','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currAmp1, currAmp2, currPhase1, currPhase2, currDelay):
        
        reqs = []
            
        for op in opList:
        
            alg = gc.Algorithm(agents=devs)
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
            alg[gates.Wait([alg.q0], currDelay)]
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0], tomoDelay)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Readout([alg.q0])] # should end up in |1> state
            alg.compile()

            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))
        
        for op in opList:
            
            alg = gc.Algorithm(agents=devs)
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
            alg[gates.Wait([alg.q0], currDelay)]
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0], tomoDelay)]
            alg[gates.Tomography([alg.q0], op)]
            alg[gates.Readout([alg.q0])] # should end up in |1> state
            alg.compile()
            
            reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw'))

        dataZero = []
        dataOne = []
        
        results = yield FutureList(reqs)
        for r in results[:len(opList)]:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            dataZero.append(prob)
        for r in results[len(opList):]:
            Is,Qs = r[0]
            prob = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            dataOne.append(prob)
                
        [rho, residual] = tomo.qst_evan(np.array(dataZero), tomo.octomo_ops, stats, 1)        
        phaseZero = ps.getDensityPhase(rho)[0]
        ampZero = np.abs(rho[1,0])
        
        [rho, residual] = tomo.qst_evan(np.array(dataOne), tomo.octomo_ops, stats, 1)        
        phaseOne = ps.getDensityPhase(rho)[0] 
        ampOne = np.abs(rho[1,0])
                
        returnValue([phaseZero, phaseOne, ampZero, ampOne])
            
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def forceRingdown(Sample, measure, name='Forced Resonator Ringdown',
            amp1 = None, amp2 = None, phase1 = None, phase2 = None, 
            ringdown = True,
            delay = 0*ns, tomoDelay = 100*ns,
            stats=900, correctXtalkZ=True,
            save=True, collect=False, noisy=True):
    '''
    We want to try and tune up a ringdown pulse after readout to reduce the photon decay time.
    The tricky part is, we have to have pulses that ring down BOTH |0> and |1> simultaneously.
    We check how well we are doing by preparing a |0>, readout, then quantify how well a pi pulse works,
    and also preparing a |1> and doing the same thing. The Pi at the end of each of these experiments 
    will only work well if the reasonator is properly rung down.
    '''
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    ops = tomo.gen_qst_tomo_ops(tomo.octomo_names, 1 )
    opList = [op for op in ops]   
    
    if amp1 is None:
        amp1, _ = qubit['ringdownPulseAmps']
    if amp2 is None:
        _, amp2 = qubit['ringdownPulseAmps']
    if phase1 is None:
        phase1, _ = qubit['ringdownPulsePhases']
    if phase2 is None:
        _, phase2 = qubit['ringdownPulsePhases']
    
    axes = [(amp1, 'Amplitude 1'), (amp2, 'Amplitude 2'), (phase1, 'Phase 1'),(phase2, 'Phase 2'),(delay,'Delay')]
    deps = [('Probability No Pi', '', ''), ('Probability Pi', '', '')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currAmp1, currAmp2, currPhase1, currPhase2, currDelay):
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
        alg[gates.Wait([alg.q0], currDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Wait([alg.q0], tomoDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Readout([alg.q0])]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probNoPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))

        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=ringdown)]
        alg[gates.Wait([alg.q0], currDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Wait([alg.q0], tomoDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Readout([alg.q0])]
        alg.compile()
        
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
            
        returnValue([probNoPi, probPi])
            
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def forceRingdownNM(Sample, measure, delay = 0*ns, tomoDelay = 100*ns,
               nonzdelt = 0.05, zdelt = 0.01,
               xtol = 0.001, ftol = 0.001,
               name='Forced Ringdown NM',
               averages = 10,
               stats=2100,
               save=True, collect=False, noisy=True):
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    amp1, amp2 = qubit['ringdownPulseAmps']
    phase1, phase2 = qubit['ringdownPulsePhases']

    
    axes = [('Function Evaluation','')]
    deps = [('Error','',''),
            ('Amplitude 1','dBm',''),
            ('Amplitude 2','dBm',''),
            ('Phase 1','Rad',''),
            ('Phase 2','Rad','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    
    def func(server, args):
        
        currAmp1, currAmp2, currPhase1, currPhase2, currAvg = args

        alg = gc.Algorithm(agents=devs)
        alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1*dBm, currAmp2*dBm), phases = (currPhase1, currPhase2), ringdown=True)]
        alg[gates.Wait([alg.q0], delay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Wait([alg.q0], tomoDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Readout([alg.q0])] # should end up in |1> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probNoPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1*dBm, currAmp2*dBm), phases = (currPhase1, currPhase2), ringdown=True)]
        alg[gates.Wait([alg.q0], delay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Wait([alg.q0], tomoDelay)]
        alg[gates.PiHalfPulse([alg.q0])]
        alg[gates.Readout([alg.q0])] # should end up in |0> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        error = ((1-probNoPi)**2 + probPi**2)/2.
        print error
        returnValue([error])
            
    def funcWrapper(args):

        args=list(args)
        #currAmp1, currAmp2, currPhase1, currPhase2 = args
    
        def seqGen():
            for currAvg in range(averages):
                #yield currAmp1, currAmp2, currPhase1, currPhase2, currAvg
                yield args + [currAvg]
                    
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample, 'dummy dataset', axes, deps,measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save=False, noisy=False)
        
        avgError = np.mean(data)
        
        print 'avg error: %s, std: %s' % (avgError, np.std(data)/np.sqrt(averages))
        
        garbageCollect.collect()
        
        return avgError
            
    inputs = [amp1,amp2,phase1,phase2]
    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt)
    
    return output
             
def forceRingdownViaPi(Sample, measure, name='Forced Resonator Ringdown Via Pi', iterations = 1,
            amp1 = None, amp2 = None, phase1 = None, phase2 = None,
            wait = -50*ns, overlap = None,
            stats=900, correctXtalkZ=True,
            save=True, collect=False, noisy=True):
    '''
    We want to try and tune up a ringdown pulse after readout to reduce the photon decay time.
    The tricky part is, we have to have pulses that ring down BOTH |0> and |1> simultaneously.
    We check how well we are doing by preparing a |0>, readout, then quantify how well a pi pulse works,
    and also preparing a |1> and doing the same thing. The Pi at the end of each of these experiments 
    will only work well if the reasonator is properly rung down.
    '''
             
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if amp1 is None:
        amp1, _ = qubit['ringdownPulseAmps']    
    if amp2 is None:
        _, amp2 = qubit['ringdownPulseAmps']
    if phase1 is None:
        phase1, _ = qubit['ringdownPulsePhases']    
    if phase2 is None:
        _, phase2 = qubit['ringdownPulsePhases']
    if overlap is None:
        overlap = qubit['ringdownPulseOverlap']
    
    axes = [(amp1, 'Amplitude 1'), (amp2, 'Amplitude 2'), (phase1, 'Phase 1'),(phase2, 'Phase 2'),(wait,'Wait'),(overlap,'Overlap')]
    deps = [('Probability |1> no Pi', '', ''), ('Probability |1> Pi', '', ''),('Error','','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currAmp1, currAmp2, currPhase1, currPhase2, currWait, currMoveBack):
        
        alg = gc.Algorithm(agents=devs)
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], currWait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0])] # should end up in |1> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probNoPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], currWait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0])] # should end up in |0> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        error = ((1-probNoPi)**2 + probPi**2)/2.
        
        returnValue([probNoPi, probPi, error])
            
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False)
    
    if collect:
        return data
        
def forceRingdownViaPiNM(Sample, measure, wait = -100*ns, iterations=1, 
               nonzdelt = 0.05, zdelt = 0.01,
               xtol = 0.001, ftol = 0.001,
               name='Forced Ringdown NM',
               averages = 10,
               stats=2100,
               save=True, collect=False, noisy=True):
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    amp1, amp2 = qubit['ringdownPulseAmps']
    phase1, phase2 = qubit['ringdownPulsePhases']

    
    axes = [('Function Evaluation','')]
    deps = [('Error','',''),
            ('Amplitude 1','dBm',''),
            ('Amplitude 2','dBm',''),
            ('Phase 1','Rad',''),
            ('Phase 2','Rad','')]
            
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)   
    
    def func(server, args):
        
        currAmp1, currAmp2, currPhase1, currPhase2, currAvg = args

        alg = gc.Algorithm(agents=devs)
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], wait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0], ringdown=False)] # should end up in |1> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probNoPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        alg = gc.Algorithm(agents=devs)
        alg[gates.PiPulse([alg.q0])]
        for k in range(2*iterations-1):
            alg[gates.Readout([alg.q0], adcWindow=False, amps=(currAmp1, currAmp2), phases = (currPhase1, currPhase2), ringdown=True)]
            alg[gates.Wait([alg.q0], wait)]
            alg[gates.PiPulse([alg.q0])]
            alg[gates.Wait([alg.q0], 50*ns)]
        alg[gates.Readout([alg.q0], ringdown=False)] # should end up in |0> state
        alg.compile()

        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw')
        Is,Qs = data[0]
        probPi = np.mean(readoutProbabilities(alg.q0,np.array(Is)+1.0j*np.array(Qs)))
        
        error = ((1-probNoPi)**2 + probPi**2)/2.
        print error
        returnValue([error])
            
    def funcWrapper(args):

        args=list(args)
        #currAmp1, currAmp2, currPhase1, currPhase2 = args
    
        def seqGen():
            for currAvg in range(averages):
                #yield currAmp1, currAmp2, currPhase1, currPhase2, currAvg
                yield args + [currAvg]
                    
        mySeqGen = seqGen()
        sub_dataset = sweeps.prepDataset(sample, 'dummy dataset', axes, deps,measure=measure)
        data = sweeps.run(func, mySeqGen, dataset = sub_dataset, save=False, noisy=False)
        
        avgError = np.mean(data)
        
        print 'avg error: %s, std: %s' % (avgError, np.std(data)/np.sqrt(averages))
        
        garbageCollect.collect()
        
        return avgError
            
    inputs = [amp1,amp2,phase1,phase2]
    output = sweeps.fmin(funcWrapper, np.array(inputs), dataset, xtol = xtol, ftol = ftol, nonzdelt = nonzdelt, zdelt = zdelt)
    
    return output

def updateForceRingdownNM(Sample, measure, dvwData, iteration = None):
        
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)
    
    if iteration is None:
        optimalParams = list(dvwData[-1,2:])
    else:
        optimalParams = list(dvwData[iteration-1,2:])
    
    amp1 = optimalParams[0]*dBm
    amp2 = optimalParams[1]*dBm
    phase1 = optimalParams[2]
    phase2 = optimalParams[3]
        
    Qubit['ringdownPulseAmps'] = [amp1, amp2]
    Qubit['ringdownPulsePhases'] = [phase1, phase2]
    
def aczLeakageVsLength(Sample, measure, lengths = st.r[30:50:1, ns]):
    
    sample, devs, qubit, Qubit = gc.loadQubits(Sample, measure, write_access=True)

    for length in lengths:
        Qubits[0][qubits[1].__name__+'aczTime'] = length
        mq.aczPhaseTuner(s,[measure[0],measure[1]], amplitude=None)
        mq.aczShapeNM(s,['q1','q0'], maxfun = 50)
        mq.updateAczShapeNM(s, dvw[-1])
        mq.aczPhaseTuner(s,[measure[0],measure[1]], amplitude=None)
        mq.aczErrorFilter(s,['q1','q0'])
        
def hollisterBet(Sample, measure,
                 freq = None,
                 herald = False, plot=False,
                 name='Measurement (Hollister) Bet',
                 stats=10,
                 save=True, collect=False, noisy=True):
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    
    axes = [(freq,'Frequency')]
    deps = [('Probability |1> no Pi','',''),('Probability |1> Pi','','')]
    
    kw = {'stats': stats}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)
    def func(server, currFreq):
        alg = gc.Algorithm(agents=devs)
        currAmp = fToAmp(alg.q1, currFreq)
        if herald:
            alg[gates.Measure(alg.qubits, sync='end')]
        alg[gates.PiPulse([alg.q0])]
        alg[gates.Sync([alg.q0,alg.q1])]
        alg[gates.Detune([alg.q1], length = alg.q0['readoutLen'] + alg.q0['ringdownLen'], amp = currAmp)]
        alg[gates.Readout([alg.q0])]
        alg[gates.Sync([alg.q0,alg.q1])]
        alg[gates.Readout([alg.q1])]
        alg.compile(correctXtalkZ=True)
        
        if plot:
            ph.pulses(alg)
            
        data = yield runQubits(server, alg.agents, stats=stats, measure = measure, dataFormat='iqRawLocalDemod', localDemod=True)
        data_for_dv = iq.format_something_useful_for_data_vault(data)
        balls
        returnValue(data_for_dv)
        
    data = sweeps.grid(func, axes, dataset=dataset, save=save, noisy=False, pipesize=1)     
