#Standard packages
import numpy as np
from math import atan2
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
from scipy.interpolate import UnivariateSpline as US
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random
import scipy
import itertools
import gc as garbageCollect

#LabRAD
import labrad
from labrad.units import Unit,Value
V, mV, us, ns, GHz, MHz, dBm, rad = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad')]
#pyle
from pyle import signalprocessing as signalprocessing
import pyle.envelopes as env
from pyle.pipeline import returnValue, FutureList
from pyle.dataking import envelopehelpers as eh
from pyle.dataking import measurement
from pyle.dataking import squid
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.dataking import singleQubit as sq
from pyle.util import sweeptools as st
from pyle.dataking import utilMultilevels as ml
from pyle.plotting import dstools as ds
from pyle.fitting import fitting
from pyle.util import structures
from pyle.datavault import DataVaultWrapper
from pyle import tomo as tomo
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray

from pyle.sim import pyleSim as ps
from pyle.math import tensor

from pyle import gateCompiler as gc
from pyle import gates

from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss

from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.dataking.benchmarking import randomizedBenchmarking as rb

from pyle.dataking.DevDaniel.Qubit import readout
from pyle.dataking import multipleQubit as mq

from pyle import quantum
# from classCZ import adiabaticCZtuning
import copy
import time
from zfuncs import *
          
def measureBellStateProb(Sample, measure,name='Bell state pop.',length=None, trial = st.r[1:1000:1], stats=900, correctXtalkZ=True, update=False, save=True, collect=False, noisy=True):
    '''
    Makes a Bell state using the new qubit code. Measures the probability as a function of trial
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    
    
    axes = [(trial, 'Trial')]
    deps = [('Prob |q0>', '', ''),('Prob |q1>', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currTrial):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit)
        alg[gates.Measure(alg.qubits)]
        alg.compile()
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure)
        returnValue(probs)


    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def correlationTester(Sample, measure,name='Corrlation Tester',phase=st.r[0.15:np.pi:0.01],  stats=900, correctXtalkZ=True, save=True, collect=False, noisy=True):            
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(phase, 'Phase')]
    deps=[]
    for state_bin_repr in range(2**len(measure)): 
    #This thing displays the correlated probabilities in the correct order
        probstr='|'+np.binary_repr(state_bin_repr,len(measure))+'>'
        deps+=[('Probability', probstr, '')]
    # deps = [('Prob |>', '', ''),('','',''),('','',''),('','','')]
    # deps = [('Prob |q0>', '', ''),('Prob |q1>', '', '')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currPhase):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit,phase=currPhase)
        alg[gates.Measure(alg.qubits)]
        alg.compile()
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure,correlated = True)
        # probs = mq.mqReadoutProbabilities(data, alg.qubits, measure,correlated = False)
        # print probs
        
        returnValue(probs)
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)     
             
def correlationTesterArbRotation(Sample, measure,name='Corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        
        createBellState(alg,controlQubit, targetQubit)
        if equator:
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        pairRot(alg,currAmp,rotationQubit=controlQubit)
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probC = pairCorr(alg,measure,data)
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

def bellViolationArbRotation(Sample, measure,name='Bell violation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, rotation = 'control',
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True, delay = 0*ns):            
    
    '''
    attempts to recreate the bell violation by comparing the correlator constructed using 4 distinct measurements vs. some angle rotation.
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('Bell term 1','',''),('Bell term 2','',''),('Bell term 3','',''),('Bell term 4','',''),('Bell Corelator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        
        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probM1 = pairCorr(alg,measure,data)
        ##### Detector Settings 2
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probM2 = pairCorr(alg,measure,data)
        ###### Detector Settings 3
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probM3 = pairCorr(alg,measure,data)
        
        ##### Detector Settings 4
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0

        createBellState(alg,controlQubit, targetQubit)
        if rotation =='control':
            pairRot(alg,currAmp,rotationQubit=controlQubit)
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        else:
            pairRot(alg,currAmp,rotationQubit=targetQubit)
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.Wait([controlQubit,targetQubit],delay)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        probM4 = pairCorr(alg,measure,data)
        
        returnValue([probM1,probM2,probM3,probM4,(probM2+probM3+probM4-probM1)])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)

def LGBIViolation(Sample, measure, name='LGBI violation with arbitrary rotaions',
                                 angles=st.r[(np.pi/2):0:0.01],  stats=900, recal = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
    
    '''
    Full LGBI violation algorithm including weak measurement.
    '''
    
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if recal:
        for device in measure:
            calArbPhase(Sample, device,2*devs[device]['piAmp'].value, dvw)
    
    arbRot0 = rot2AmpFunc(dvw,sample['q0'])[1]
    arbRot1 = rot2AmpFunc(dvw,sample['q1'])[1]
    arbRot2 = rot2AmpFunc(dvw,sample['q2'])[1]
    arbRot3 = rot2AmpFunc(dvw,sample['q3'])[1]
    
   
    axes = [(angles, 'Ancilla rotation')]
    deps=[('ancilla1-ancilla2 term ','',''),('ancilla1-bell2 term','',''),('ancilla2-bell1 term','',''),('bell1-bell2 term','',''),('CHSH Correlator','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAng):
        ##### Detector Settings 1
        alg = gc.Algorithm(agents=devs)
        targetBell = alg.q2
        controlBell = alg.q1
        ancilla1 = alg.q0
        ancilla2 = alg.q3
        
        #initialize Bell State and rotate
        createBellState(alg, controlBell, targetBell)
        pairRot(alg,0.155,targetBell)
        
        # weak measurement with ancilla 1
        pairRot(alg,(currAng/(2*arbRot0)),ancilla1)
        # alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlBell , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        # correct phase for ancilla 2 from first measurement
        alg[gates.ChangePhase([ targetBell ], xyPhase = targetBell['q1q0aczPhase'])]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # weak measurement with ancilla 2
        pairRot(alg,(currAng/(2*arbRot3)),ancilla2)
        # alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetBell] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        alg[gates.Sync([ancilla1,controlBell,targetBell,ancilla2])]
        
        # final rotations
        alg[gates.PiHalfPulse([ controlBell ])]
        alg[gates.PiHalfPulse([ targetBell ], phase = np.pi)]
        
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)
        # ph.pulses(alg)
        # raise exception('Hah!')
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probs = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
        # probs = np.array(probs)*2-1
        
        # ph.pulses(alg)
        # raise exception("blah")
        
        def correlator(alg,measure,data):
            correction = 1.0/np.sin(currAng)
            a1a2 = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])*correction**2
            a1b2 = arbPairCorr(alg,measure,data,[[0,1,4,5,10,11,14,15],[2,3,6,7,8,9,12,13]])*correction
            b1a2 = arbPairCorr(alg,measure,data,[[0,2,5,7,8,10,13,15],[1,3,4,6,9,11,12,14]])*correction
            b1b2 = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
            return [a1a2,a1b2,b1a2,b1b2,(-a1a2 - a1b2 + b1a2 - b1b2)]

        returnValue(correlator(alg,measure,data))
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)    

def weakMeasureTest(Sample, measure, name='weak measurement test with ancilla',
                                 angles=st.r[(np.pi/2):0:0.01],  stats=900, recal = False, pulse = None,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
    
    '''
    Tests the weak measurement protocol by comparing a pi pulse on the measured qubit with no pi pulse
    on the measured qubit vs. initial rotation of the measuring qubit.
    '''
    
    dvw = DataVaultWrapper(Sample._dir,Sample._cxn)
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
    if recal:
        for device in meausre:
            calArbPhase(Sample, measure[0],2*devs[measure[0]]['piAmp'].value, dvw)
    
    arbRot0 = rot2AmpFunc(dvw,sample[measure[0]])[1]
    arbRot1 = rot2AmpFunc(dvw,sample[measure[1]])[1]
    
   
    axes = [(angles, 'Ancilla rotation')]
    deps=[('probaility no pi pulse ','',''),('probability pi pulse','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAng):
        ##### Detector Settings 1
        
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        #initialize Bell State
        if pulse == 'target':
            pairRot(alg,(currAng/(2*arbRot1)),targetQubit)
            # alg[gates.PiHalfPulse([ targetQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ targetQubit ], phase = np.pi)]
        else:
            pairRot(alg,currAng/(2*arbRot0),controlQubit)
            # alg[gates.PiHalfPulse([ controlQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
            
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)

        dataUnloaded = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probsUL = mq.mqReadoutProbabilities(dataUnloaded, alg.qubits, measure, correlated = False)
        
        alg = gc.Algorithm(agents=devs)
        targetQubit = alg.q1
        controlQubit = alg.q0
        #initialize Bell State
        if pulse == 'target':
            alg[gates.PiPulse([ controlQubit ])]
            pairRot(alg,currAng/(2*arbRot1),targetQubit)
            # alg[gates.PiHalfPulse([ targetQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ targetQubit ], phase = np.pi)]
        else:
            alg[gates.PiPulse([ targetQubit ])]
            pairRot(alg,currAng/(2*arbRot0),controlQubit)
            # alg[gates.PiHalfPulse([ controlQubit ])]
            alg[gates.adiabaticCZ([controlQubit , targetQubit] )]
            alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
            
        alg[gates.Measure(alg.qubits)]
        
        alg.compile(correctXtalkZ=correctXtalkZ)

        dataLoaded = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probsL = mq.mqReadoutProbabilities(dataLoaded, alg.qubits, measure, correlated = False)
        
        if pulse == 'target':
            pUL,PL = probsUL[1],probsL[1]
        else:
            pUL,PL = probsUL[0],probsL[0]
            
        returnValue([pUL,PL])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
def ancillaCorrelationTesterArbRotation(Sample, measure,name='ancilla corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):            
            
    '''
    Tests the degree of correlation vs. some angle rotation between bell states using the weak ancilla measurement.
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        ancilla1 = alg.q0
        controlQubit = alg.q1
        targetQubit = alg.q2
        ancilla2 = alg.q3
        
        createBellState(alg,controlQubit, targetQubit)
        pairRot(alg,currAmp,rotationQubit=targetQubit)
        alg[gates.Sync([ancilla1,controlQubit,targetQubit,ancilla2])]
        
        alg[gates.PiHalfPulse([ ancilla1 ])]
        alg[gates.adiabaticCZ([controlQubit , ancilla1] )]
        alg[gates.PiHalfPulse([ ancilla1 ], phase = np.pi)]
        
        
        alg[gates.ChangePhase([ targetQubit ], xyPhase = targetQubit['q1q0aczPhase'])]
        alg[gates.Sync([ancilla1,controlQubit,targetQubit,ancilla2])]
        
        alg[gates.PiHalfPulse([ ancilla2 ])]
        alg[gates.adiabaticCZ([ancilla2 , targetQubit] )]
        alg[gates.PiHalfPulse([ ancilla2 ], phase = np.pi)]
        
        alg[gates.PiHalfPulse([ controlQubit ], phase = np.pi)]
        alg[gates.PiHalfPulse([ targetQubit ]) ]
        
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        probC = arbPairCorr(alg,measure,data,[[0,2,4,6,9,11,13,15],[1,3,5,7,8,10,12,14]])
        # probC = arbPairCorr(alg,measure,data,[[0,1,6,7,8,9,14,15],[2,3,4,5,10,11,12,13]])
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)    
    
def correlationTesterHeralding(Sample, measure,name='Corrlation rotation Tester',
                                 amps=st.r[0:1:0.01],  stats=900, equator = False,
                                 correctXtalkZ=True, save=True, collect=False,
                                 noisy=True):
            
    '''
    Tests the degree of correlation vs. some phase angle rotation
    '''
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    
   
    axes = [(amps, 'Rotation amplitude')]
    deps=[('corelation probability','','')]
    kw = {'stats': stats,'correctXtalkZ': correctXtalkZ}
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)  

    def func(server, currAmp):
        alg = gc.Algorithm(agents=devs)
        ml.setMultiKeys(alg.q0,1)
        ml.setMultiLevels(alg.q0, 'piAmp', currAmp, 1)
        
        controlQubit = alg.q0
        targetQubit = alg.q1
        createBellState(alg,controlQubit, targetQubit,phase = np.pi)
        if equator:
            alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
            alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2
        alg[gates.PiPulse([controlQubit])]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
            
        data = yield runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure)
        
        prob00 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[0]
        prob01 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[1]
        prob10 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[2]
        prob11 = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)[3]
        
        probC = (prob00+prob11-prob10-prob01)
         
        
        returnValue([probC])
        
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)            

def aczSRamsey(Sample, measure, name='Adiabatic CZ seccondary qubit Ramsey',
             thetas=25, waitTime = None,
             iterations=1,
             stats=1500, correctXtalkZ=True, update=False, plot=False,
             save=True, collect=False, noisy=True):
             
    '''
    measure = [control, target]
    '''
    
    thetas = np.linspace(-np.pi/4,7*np.pi/4,thetas)
    
    sample, devs, qubits, Qubits = gc.loadQubits(Sample, measure, write_access=True)
    controlQubit = qubits[0]
    targetQubit = qubits[1]
    secondayQubit = qubits[2]
    
    if waitTime is None:
        waitTime = controlQubit['piLen']
    
    axes = [(thetas, 'Ramsey Phase')]
    deps = [('Probability |1>','control no Pi',''),('Probability |1>','control Pi','')]
    # deps = [('Probability |1>','secondary no Pi','')]
    kw = {'stats': stats}
    
    dataset = sweeps.prepDataset(sample, name, axes, deps, measure=measure, kw=kw)    
    def func(server, theta):
        
        reqs = []
        
        # no pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        secondaryQubit = alg.q2
        alg[gates.PiHalfPulse([secondaryQubit])]
        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Sync([controlQubit,targetQubit,secondaryQubit])]
        alg[gates.ChangePhase([ secondaryQubit ], xyPhase = secondaryQubit['q1q0aczPhase'])]
        alg[gates.PiHalfPulse([secondaryQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        # pi
        alg = gc.Algorithm(agents = devs)
        controlQubit = alg.q0
        targetQubit = alg.q1
        secondaryQubit = alg.q2
        alg[gates.PiPulse([secondaryQubit])]
        alg[gates.PiHalfPulse([secondaryQubit])]

        for k in range(iterations):
            alg[gates.adiabaticCZ([controlQubit, targetQubit])]
            if iterations>1 and k<iterations-1:
                alg[gates.Wait([controlQubit], waitTime)]
        alg[gates.Sync([controlQubit,targetQubit,secondaryQubit])]
        alg[gates.ChangePhase([ secondaryQubit ], xyPhase = secondaryQubit['q1q0aczPhase'])]
        alg[gates.PiHalfPulse([secondaryQubit], phase = theta)]
        alg[gates.Measure(alg.qubits)]
        alg.compile(correctXtalkZ=correctXtalkZ)
        
        reqs.append(runQubits(server, alg.agents, stats, dataFormat='iqRaw', measure = measure))
        
        results = yield FutureList(reqs)
        data0 = results[0]
        Is0, Qs0 = data0[2]
        probNoPi = np.mean(sq.readoutProbabilities(secondaryQubit,np.array(Is0)+1.0j*np.array(Qs0)))
        
        data1 = results[1]
        Is1, Qs1 = data1[2]
        probPi = np.mean(sq.readoutProbabilities(secondaryQubit,np.array(Is1)+1.0j*np.array(Qs1)))
        
        returnValue([probNoPi, probPi])
        # returnValue([probNoPi])
            
    data = sweeps.grid(func, axes, dataset=save and dataset, noisy=False)
    
    if update:
        
        def mapMinusPiToPi(phi):
            while phi<=-np.pi or phi>np.pi:
                if phi<=-np.pi:
                    phi+=2*np.pi
                elif phi>np.pi:
                    phi-=2*np.pi
            return phi
            
        amp, phi = fitting.fitRamsey(thetas, data[:,1])    
        deltaPhi = mapMinusPiToPi(phi)    
        Qubits[2][controlQubit.__name__+targetQubit.__name__+'aczPhase'] -= deltaPhi/iterations
    
    if collect:
        return data
        
def pairRot(alg,currAmp,rotationQubit):
    '''
    Does the pair rotation for correlation testing
    so we don't have to write it over and over again.
    '''

    ml.setMultiKeys(rotationQubit,1)
    ml.setMultiLevels(rotationQubit, 'piAmp', currAmp, 1)
    alg[gates.PiPulse([rotationQubit])]

def pairCorr(alg,measure,data,*kwargs):
    '''
    measures the correlation among pairs so we don't have to 
    write it over and over again

    returns correlated probability
    '''
    prob = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)

    return (prob[0]+prob[3]-prob[2]-prob[1])

def arbPairCorr(alg,measure,data,terms,*kwargs):
    '''
    measures the correlation among an arbitrary set of pairs so we don't have to 
    write it over and over again

    returns correlated probability
    '''
    prob = mq.mqReadoutProbabilities(data, alg.qubits, measure, correlated = True)
    corr = 0
    for value in terms[0]:
        corr += prob[value]
    for value in terms[1]:
        corr -= prob[value]
    return (corr)
    
def createBellState(alg,controlQubit, targetQubit, bellState = 'Psi+'):
    '''
    Creates a Bell State so we don't have to type this in all the time
    '''      
    if bellState=='Psi+':
        # 01 + 10
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2    
    elif bellState=='Psi-':
        # 01 - 10
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ controlQubit ],phase=np.pi)] #-X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
      
    elif bellState=='Phi+':
        # 00 + 11
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ controlQubit ])] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ targetQubit ],phase=np.pi)] #-X_pi/2                    
    elif bellState=='Phi-':
        # 00 - 11
        alg[gates.PiHalfPulse([ targetQubit ])] #X_pi/2
        alg[gates.PiHalfPulse([ controlQubit ], phase=np.pi)] #X_pi/2            \
        alg[gates.adiabaticCZ([controlQubit , targetQubit] )]   #   ||   This is apparently  CNOT
        alg[gates.PiHalfPulse([ targetQubit ],phase=np.pi)] #-X_pi/2
    else:
        raise Exception('not implemented')                    
            
def calArbPhase(Sample, measure, maxAmp, dvw):         
    sq.rabihigh(Sample,measure,stats = 3000, amps = np.linspace(0,maxAmp,50), IQ = False, update = False)
    Sample[measure]['arbPhaseCal'] = (Sample._dir, int(dvw.keys()[-1].split('-')[0]))
            
def sinFitFunc(x,a,b,c):
    return a*(np.sin(b*x))**2 + c
      
def rot2AmpFunc(dvw, qubit):        
    directory,dataset = qubit['arbPhaseCal']
    dvw = dvw.cd(directory)
    xdata = dvw[dataset][:,0]
    ydata = dvw[dataset][:,1]
    fitParams, errors = curve_fit(sinFitFunc, xdata, ydata, p0=[0.7,np.pi,0.1])
    return fitParams
    
    
    
    

#end         