'''
Created on May 27, 2010
author: Daniel Sank
'''


#Standard packages
import numpy as np
import math
from math import atan2
import scipy
import scipy.optimize as opt
from scipy.optimize import leastsq, curve_fit
from scipy.special import erf, erfc
import scipy.signal as spsignal
import matplotlib.pyplot as plt
import matplotlib
import random

#LabRAD
import labrad
from labrad.units import Unit, Value, Complex
V, mV, us, ns, GHz, MHz, dBm, rad, sec = [Unit(s) for s in ('V', 'mV', 'us', 'ns', 'GHz', 'MHz', 'dBm', 'rad', 's')]
#pyle
import pyle
import pyle.datavault as datavault
import pyle.envelopes as env
import pyle.dataking.envelopehelpers as eh
import pyle.dataking.measurement as measurement
import  pyle.dataking.fpgaseqTransmon as fpgaseqTransmon
import pyle.dataking.utilMultilevels as ml
import pyle.plotting.dstools as ds
import pyle.fitting.fitting as fitting
import pyle.util.structures as structures
import pyle.tomo as tomo
import pyle.gateCompiler as gc
import pyle.gates as gates
import pyle.dataking.adjust as adjust
ValueArray = structures.ValueArray
#named functions
from pyle.dataking.fpgaseqTransmon import runQubits
from pyle.util import sweeptools as st
from pyle.pipeline import returnValue, FutureList
from pyle.plotting import plothelper as ph
from pyle.fitting.fitting import getMaxGauss
from pyle.dataking import sweeps
from pyle.dataking import util
from pyle.util import getch

#from pyle.dataking.benchmarking import randomizedBenchmarking as rb

import pyle.dataking.DevDaniel.Qubit.readout as readout
import pyle.dataking.singleQubit as sq

import pyle.quantum as quantum
import time

def shuffle(iter):
    indices = range(len(iter))
    random.shuffle(indices)
    shuffled = [iter[i] for i in indices]
    return shuffled, indices


def unshuffle(iter,indices):
    unshuff = [0]*len(iter)
    for i,idx in enumerate(indices):
        unshuff[idx] = iter[i]
    return unshuff

def RTOsweep(Sample, measure, holdTimes=st.r[500:5000:300,ns], fringeFreq=2.0*MHz, timeStep=0.5*sec, nPoints=600):
    print "Estimated %s minutes per dataset x %s datasets = %s mins total" % (timeStep['min']*nPoints, len(holdTimes), len(holdTimes)*timeStep['min']*nPoints)
    for ht in holdTimes:
        print "STARTING: %s" % str(ht)
        ramseyOscilloscope(Sample,measure,fringeFreq,ht,timeStep, cutoff=nPoints,
                            name="RTO Sweep: %s ns" % int(ht['ns']), noisy=False)
    
def ramseyOscilloscope(Sample, measure, 
              fringeFreq, holdTime, timeStep, singleMeasure=None,
              name='Ramsey Tomography Oscilloscope',
              stats=150, dataFormat='iq',
              save=True, collect=False, noisy=True, cutoff=None):
    ''' Repeated ramsey measurements at t=holdTime to measure frequency noise.
    measure can be any number of qubits.
    fringeFreq is basically useless since we keep t constant.
    timeStep is repetition time, usually about 1 second.
    singleMeasure is a single qubit to measure interleaved with the real measure,
      for the purpose of comparing single to multiqubit measurements simultaneously.
      
    NOTE: doesn't actually do multiqubit stuff anymore. need to redo that and reimplement singlemeasure
    '''
    
    sampleX, devsX, qX  = gc.loadQubits(Sample, measure)
    sampleY, devsY, qY = gc.loadQubits(Sample, measure)
    
    
    #### Precomputing sequences no longer works with fpgaSeq, etc rewrite.
    #Precompute sequences
    #algX,algY = gc.Algorithm(agents=devsX), gc.Algorithm(agents=devsY)
    #for alg, phase in zip([algX, algY], [0,np.pi/2.0]):
    #    alg[gates.PiHalfPulse([alg.q0])]
    #    alg[gates.Wait([alg.q0], holdTime)]
    #    alg[gates.PiHalfPulse([alg.q0], phase=phase + 2*np.pi*fringeFreq['GHz']*holdTime['ns'])]
    #    alg[gates.Measure([alg.q0])]
    #    alg.compile()
        
    # dataset prep
    axes = [('Iteration', '')]
    deps = [('q%s Probability'%x, '+X', '') for x in measure]
    deps += [('q%s Probability'%x, '+Y', '') for x in measure]
    deps.append(('time', 't', 'sec'))
    kw = {'stats': stats, 'holdTime':holdTime, 'fringeFreq':fringeFreq, 'timeStep':timeStep}
    dataset = sweeps.prepDataset(sampleX, name, axes, deps, measure=measure, kw=kw) 
        
    # looper    
    def iterations():
        i = 1
        while True:
            yield i
            i += 1
            if cutoff and i > cutoff:
                break
            key = getch.getch()
            if key == '\x1b':
                print 'Abort scan'
                break

    # new method: no pipelining or anything, just run expt after expt
    with pyle.QubitSequencer() as sequencer:
        with dataset as dataset: 
            start = time.time()
            skips = 0
            totalWaitTime = 0
            numWaits = 0
            
            for iteration in iterations():
                algX,algY = gc.Algorithm(agents=devsX), gc.Algorithm(agents=devsY)
                for alg, phase in zip([algX, algY], [0,np.pi/2.0]):
                    alg[gates.PiHalfPulse([alg.q0])]
                    alg[gates.Wait([alg.q0], holdTime)]
                    alg[gates.PiHalfPulse([alg.q0], phase=phase + 2*np.pi*fringeFreq['GHz']*holdTime['ns'])]
                    alg[gates.Measure([alg.q0])]
                    alg.compile()
                # rigamarole to figure out when to fire this sequence
                iteration += skips
                elapsed = time.time() - start
                nextIteration = int(math.ceil(elapsed/timeStep['s']))
                if nextIteration > iteration:
                    skips += nextIteration - iteration
                    print 'skip!  skips = %d/%d (%g%%)' % (skips, iteration, 100*skips/iteration)
                nextTime = nextIteration * timeStep['s']
                wait = nextTime - (time.time() - start)
                if noisy:
                    totalWaitTime += wait
                    numWaits += 1
                    avg = totalWaitTime / numWaits
                    pct = avg / timeStep['s']
                    print 'average wait: %g = %g%% of timeStep' % (avg, 100*pct)
                time.sleep(wait)
                reqs = []
                reqs.append(runQubits(sequencer, algX.agents, stats, dataFormat='iqRaw'))
                reqs.append(runQubits(sequencer, algY.agents, stats, dataFormat='iqRaw'))
                #blash
                #data = yield FutureList(reqs)
                # first index: experiment (i.e. X or Y algorithm)
                dataX, dataY = reqs[0].wait(), reqs[1].wait()
                probs = []
                for qs, dat in zip((qX,qY),(dataX, dataY)):
                    # second index: qubit (i.e. q1, q2...)
                    for x, q in enumerate(qs):
                        Is,Qs = dat[x]
                        probs.append(np.mean(sq.readoutProbabilities(q, np.array(Is)+1.0j*np.array(Qs))))
                dataset.add([iteration] + probs + [nextTime])
                del algX, algY
                del reqs
            
    
    ### old method (using sweeps.grid) ###
    '''        
    #Timing variables
    start = [None]
    skips = [0] # number of data points that have been skipped
    waits = [0, 0]
    timeStep_s = timeStep['s']    
    
    def func(server, iteration):
        algX,algY = gc.Algorithm(agents=devsX), gc.Algorithm(agents=devsY)
        for alg, phase in zip([algX, algY], [0,np.pi/2.0]):
            alg[gates.PiHalfPulse([alg.q0])]
            alg[gates.Wait([alg.q0], holdTime)]
            alg[gates.PiHalfPulse([alg.q0], phase=phase + 2*np.pi*fringeFreq['GHz']*holdTime['ns'])]
            alg[gates.Measure([alg.q0])]
            alg.compile()
        reqs = []
        iteration += skips[0]
        if start[0] is None:
            start[0] = time.time()
            nextTime = 0
        else:
            elapsed = time.time() - start[0]
            nextIteration = int(math.ceil(elapsed / timeStep_s))
            if nextIteration > iteration:
                # oops, we missed some datapoints :-(
                skips[0] += nextIteration - iteration
                print 'skip!  skips = %d/%d (%g%%)' % (skips[0], iteration, 100*skips[0]/iteration)
            #Hang out until it's time to fire off the next iteration
            nextTime = nextIteration * timeStep_s
            wait = nextTime - (time.time() - start[0])
            if wait > 0:
                waits[0] += wait
                waits[1] += 1
                avg = waits[0] / waits[1]
                pct = avg / timeStep_s
                if noisy:
                    print 'average wait: %g = %g%% of timeStep' % (avg, 100*pct)
                time.sleep(wait)
        #Run experiment
        reqs.append(runQubits(server, algX.agents, stats, dataFormat='iqRaw'))
        reqs.append(runQubits(server, algY.agents, stats, dataFormat='iqRaw'))
            
        data = yield FutureList(reqs)
        # first index: experiment
        dataX, dataY = data
        #dataX = [(0, 0)] * stats
        #dataY = [(1,1)]*stats
        ps = [] # probabilities
        for qs, dat in zip((qX,qY),(dataX, dataY)):
            # second index: qubit
            for x, q in enumerate(qs):
                Is,Qs = dat[x]
                ps.append(np.mean(sq.readoutProbabilities(q, np.array(Is)+1.0j*np.array(Qs))))
        del algX, algY
        del reqs
        returnValue([iteration] + ps + [nextTime])
        #returnValue([0,0,0,0])
        
    sweeps.run(func, iterations(), save, dataset, pipesize=2, noisy=noisy)
    '''


