import numpy as np
import matplotlib
from matplotlib import cm
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as plt3
import matplotlib.ticker as ticker

def arrow(rho, scale=1.0, color=None, width=0.05, headwidth=0.1, headlength=0.1, chopN=None, amp=1.0, collect=False):
    plt.figure()
    rho=rho.copy()
    s=np.shape(rho)
    if chopN!=None:
        rho = rho[:chopN,:chopN]
    s=np.shape(rho)
    rho = rho*amp
    ax = plt.gca()
    ax.set_aspect(1.0)
    pos = ax.get_position()
    r = np.real(rho)
    i = np.imag(rho)
    x = np.arange(s[0])[None,:] + 0*r
    y = np.arange(s[1])[:,None] + 0*i
    plt.quiver(x,y,r,i,units='x',scale=1.0/scale, width=width, headwidth=headwidth, headlength=headlength, color=color)
    plt.xticks(np.arange(s[1]))
    plt.yticks(np.arange(s[0]))
    plt.xlim(-0.9999,s[1]-0.0001)
    plt.ylim(-0.9999,s[0]-0.0001)
    if collect:
        return rho
    
    
def manhattan3d(rho, axesLabels=None, d=0.1, cmap=None, vmin=-1, vmax=1, zmin=-1, zmax=1):
    if cmap is None:
        cmap = cm.RdYlBu
    norm = matplotlib.colors.Normalize(vmin, vmax)
    
    fig = plt.figure(figsize=(6,3))
    ax = plt3.Axes3D(fig)
    
    w, h = rho.shape
    if axesLabels is None:
        axesLabels = [bin(i)[2:].rjust(int(np.log2(w)),'0') for i in range(w)]
    xrange = range(w)
    yrange = range(h)
    yrange.reverse()
    for x in xrange:
        for y in yrange:
            rhoVal = rho[x,y]
            ax.bar3d([x+d/2], [y+d/2], [0], 1-d, 1-d, np.real(rhoVal), cmap(norm(rhoVal.real)),alpha=1.0)
            
#            xs = np.array([x+d/2, x+1-d/2])
#            ys = np.array([y+d/2, y+1-d/2])
#            zs = np.array([0, rho[x,y]])
#            
#            c = cmap(norm(zs[1]))
#            
#            # front / back
#            Z, X = np.meshgrid(zs, xs)
#            Y = np.ones((2,2)) * ys[0]
#            ax.plot_surface(X, Y, Z, color=c)
#            
#            Y = np.ones((2,2)) * ys[1]
#            ax.plot_surface(X, Y, Z, color=c)
#            
#            # left / right
#            Y, Z = np.meshgrid(ys, zs)
#            X = np.ones((2,2)) * xs[0]
#            ax.plot_surface(X, Y, Z, color=c)
#            
#            X = np.ones((2,2)) * xs[1]
#            ax.plot_surface(X, Y, Z, color=c)
#            
#            # bottom / top
#            X, Y = np.meshgrid(xs, ys)
#            Z = np.ones((2,2)) * zs[0]
#            ax.plot_surface(X, Y, Z, color=c)
#            
#            X, Y = np.meshgrid(xs, ys)
#            ax.plot_surface(X, Y, Z, color=c)
    ax.set_zlim3d(zmin, zmax)

    ax.w_xaxis.set_major_locator(ticker.FixedLocator(np.arange(w)+0.5))
    ax.w_xaxis.set_ticklabels(axesLabels)

    ax.w_yaxis.set_major_locator(ticker.FixedLocator(np.arange(w)+0.5))
    ax.w_yaxis.set_ticklabels(axesLabels)

    return fig


def manhattan2d(rho, theory, cmap=None):
    fig = plt.figure(figsize=(11,6))
    axabs_ex = fig.add_subplot(2,3,1)
    axre_ex = fig.add_subplot(2,3,2)
    axim_ex = fig.add_subplot(2,3,3)
    axabs_th = fig.add_subplot(2,3,4)
    axre_th = fig.add_subplot(2,3,5)
    axim_th = fig.add_subplot(2,3,6)
    plt.subplots_adjust(right=0.8)
    
    def set_ticklabels(ax, n=3):
        ticks = [i+0.5 for i in range(2**n)]
        labels = [bin(i)[2:].rjust(n,'0') for i in range(2**n)]
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_xticklabels(labels)
        ax.set_yticklabels(labels)
    for ax in [axabs_ex, axre_ex, axim_ex, axabs_th, axre_th, axim_th]:
        set_ticklabels(ax)
    
    absmax = lambda a: max(np.max(np.max(np.abs(a.real))), np.max(np.max(np.abs(a.imag))))
    
    vmax_ex = absmax(rho)
    vmax_th = absmax(theory)

    if cmap is None:
        cmap = cm.get_cmap('RdYlBu')
    opts_ex = {'vmin': -vmax_ex, 'vmax': vmax_ex, 'cmap': cmap}
    opts_th = {'vmin': -vmax_th, 'vmax': vmax_th, 'cmap': cmap}

    axabs_ex.pcolor(np.abs(rho), **opts_ex)
    axre_ex.pcolor(rho.real, **opts_ex)
    axim_ex.pcolor(rho.imag, **opts_ex)
    
    axabs_th.pcolor(np.abs(theory), **opts_th)
    axre_th.pcolor(theory.real, **opts_th)
    axim_th.pcolor(theory.imag, **opts_th)
    
    axabs_ex.set_title('abs')
    axre_ex.set_title('real')
    axim_ex.set_title('imag')
    
    axabs_ex.set_ylabel('experiment')
    axabs_th.set_ylabel('theory')
    
    norm_ex = matplotlib.colors.Normalize(vmin=-vmax_ex, vmax=vmax_ex)
    norm_th = matplotlib.colors.Normalize(vmin=-vmax_th, vmax=vmax_th)
    
    rect_ex = axim_ex.get_position().get_points()
    axcb_ex = fig.add_axes([0.85, rect_ex[0,1], 0.05, rect_ex[1,1]-rect_ex[0,1]])
    matplotlib.colorbar.ColorbarBase(axcb_ex, norm=norm_ex, cmap=cmap)
    
    rect_th = axim_th.get_position().get_points()
    axcb_th = fig.add_axes([0.85, rect_th[0,1], 0.05, rect_th[1,1]-rect_th[0,1]])
    matplotlib.colorbar.ColorbarBase(axcb_th, norm=norm_th, cmap=cmap)
    
    return fig



def circles(rho):

    #TODO: I can't seem to change the linewidth of the
    #      lines that denote the phase

    dim = len(rho)
    amps = abs(rho)
    phases = np.angle(rho)

    fig = plt.figure(figsize = (7,6))
    ax = plt.axes([0,0,1,1])
    ax.set_ylim(-2, 2*dim)
    ax.set_xlim(-2, 2*dim)

    pos = np.mgrid[2*(dim-1):-2:-2, 0:2*dim:2]

    circlePatches = []
    circleColors = []
    for rownum, row in enumerate(rho):
        for elementnum, element in enumerate(row):
            art = matplotlib.patches.Circle(
                             [pos[1][rownum,elementnum],
                              pos[0][rownum,elementnum]],
                              amps[rownum,elementnum])
            circleColors.append(amps[rownum, elementnum])
            circlePatches.append(art)

    # this next crap is for scaling the colors, it's a
    # dumb way to do it.
    circlePatches.append(matplotlib.patches.Circle([0,0],0))
    circlePatches.append(matplotlib.patches.Circle([-20,-20],1))
    circleColors.append(0)
    circleColors.append(1)

    for rownum, row in enumerate(rho):
        for elementnum, element in enumerate(row):
            dx = amps[rownum, elementnum]*np.cos(phases[rownum, elementnum])
            dy = amps[rownum, elementnum]*np.sin(phases[rownum, elementnum])
            x = pos[1][rownum,elementnum]
            y = pos[0][rownum,elementnum]
            art = matplotlib.lines.Line2D([x, x+dx],[y, y+dy], lw = 3, color='k')
            ax.add_line(art)

    circleCollection = matplotlib.collections.PatchCollection(circlePatches,
                                       cmap = matplotlib.cm.RdBu)

    circleCollection.set_array(np.array(circleColors))
    ax.add_collection(circleCollection)
    plt.colorbar(circleCollection)
    
    



# set up operator bases
I = np.eye(2)
X = np.array([[0, 1], [1, 0]])
Y = np.array([[0, -1j], [1j, 0]])
Z = np.array([[1, 0], [0, -1]])
dot3 = lambda a, b, c: np.dot(np.dot(a, b), c)

basis1 = {'I': I, 'X': X, 'Y': Y, 'Z': Z}

labels2 = ['XI', 'YI', 'ZI',
           'IX', 'IY', 'IZ',
           'XX', 'XY', 'XZ', 'YX', 'YY', 'YZ', 'ZX', 'ZY', 'ZZ']
basis2 = dict((label, reduce(np.kron, [basis1[c] for c in label])) for label in labels2)

labels3 = ['XII', 'YII', 'ZII',
           'IXI', 'IYI', 'IZI',
           'IIX', 'IIY', 'IIZ',
            
           'XXI', 'XYI', 'XZI', 'YXI', 'YYI', 'YZI', 'ZXI', 'ZYI', 'ZZI',
           'XIX', 'XIY', 'XIZ', 'YIX', 'YIY', 'YIZ', 'ZIX', 'ZIY', 'ZIZ',
           'IXX', 'IXY', 'IXZ', 'IYX', 'IYY', 'IYZ', 'IZX', 'IZY', 'IZZ',
            
           'XXX', 'XXY', 'XXZ', 'XYX', 'XYY', 'XYZ', 'XZX', 'XZY', 'XZZ',
           'YXX', 'YXY', 'YXZ', 'YYX', 'YYY', 'YYZ', 'YZX', 'YZY', 'YZZ',
           'ZXX', 'ZXY', 'ZXZ', 'ZYX', 'ZYY', 'ZYZ', 'ZZX', 'ZZY', 'ZZZ']
basis3 = dict((label, reduce(np.kron, [basis1[c] for c in label])) for label in labels3)

labels4 = ['XIII', 'YIII', 'ZIII',
           'IXII', 'IYII', 'IZII',
           'IIXI', 'IIYI', 'IIZI',
           'IIIX', 'IIIY', 'IIIZ',
            
           'XXII', 'XYII', 'XZII', 'YXII', 'YYII', 'YZII', 'ZXII', 'ZYII', 'ZZII',
           'XIXI', 'XIYI', 'XIZI', 'YIXI', 'YIYI', 'YIZI', 'ZIXI', 'ZIYI', 'ZIZI',
           'XIIX', 'XIIY', 'XIIZ', 'YIIX', 'YIIY', 'YIIZ', 'ZIIX', 'ZIIY', 'ZIIZ',
           'IXXI', 'IXYI', 'IXZI', 'IYXI', 'IYYI', 'IYZI', 'IZXI', 'IZYI', 'IZZI',
           'IXIX', 'IXIY', 'IXIZ', 'IYIX', 'IYIY', 'IYIZ', 'IZIX', 'IZIY', 'IZIZ',
           'IIXX', 'IIXY', 'IIXZ', 'IIYX', 'IIYY', 'IIYZ', 'IIZX', 'IIZY', 'IIZZ',
            
           'XXXI', 'XXYI', 'XXZI', 'XYXI', 'XYYI', 'XYZI', 'XZXI', 'XZYI', 'XZZI',
           'YXXI', 'YXYI', 'YXZI', 'YYXI', 'YYYI', 'YYZI', 'YZXI', 'YZYI', 'YZZI',
           'ZXXI', 'ZXYI', 'ZXZI', 'ZYXI', 'ZYYI', 'ZYZI', 'ZZXI', 'ZZYI', 'ZZZI',
           'XXIX', 'XXIY', 'XXIZ', 'XYIX', 'XYIY', 'XYIZ', 'XZIX', 'XZIY', 'XZIZ',
           'YXIX', 'YXIY', 'YXIZ', 'YYIX', 'YYIY', 'YYIZ', 'YZIX', 'YZIY', 'YZIZ',
           'ZXIX', 'ZXIY', 'ZXIZ', 'ZYIX', 'ZYIY', 'ZYIZ', 'ZZIX', 'ZZIY', 'ZZIZ',
           'XIXX', 'XIXY', 'XIXZ', 'XIYX', 'XIYY', 'XIYZ', 'XIZX', 'XIZY', 'XIZZ',
           'YIXX', 'YIXY', 'YIXZ', 'YIYX', 'YIYY', 'YIYZ', 'YIZX', 'YIZY', 'YIZZ',
           'ZIXX', 'ZIXY', 'ZIXZ', 'ZIYX', 'ZIYY', 'ZIYZ', 'ZIZX', 'ZIZY', 'ZIZZ',
           'IXXX', 'IXXY', 'IXXZ', 'IXYX', 'IXYY', 'IXYZ', 'IXZX', 'IXZY', 'IXZZ',
           'IYXX', 'IYXY', 'IYXZ', 'IYYX', 'IYYY', 'IYYZ', 'IYZX', 'IYZY', 'IYZZ',
           'IZXX', 'IZXY', 'IZXZ', 'IZYX', 'IZYY', 'IZYZ', 'IZZX', 'IZZY', 'IZZZ'
           
           'XXXX', 'XXXY', 'XXXZ', 'XXYX', 'XXYY', 'XXYZ', 'XXZX', 'XXZY', 'XXZZ',
           'XYXX', 'XYXY', 'XYXZ', 'XYYX', 'XYYY', 'XYYZ', 'XYZX', 'XYZY', 'XYZZ',
           'XZXX', 'XZXY', 'XZXZ', 'XZYX', 'XZYY', 'XZYZ', 'XZZX', 'XZZY', 'XZZZ',
           'YXXX', 'YXXY', 'YXXZ', 'YXYX', 'YXYY', 'YXYZ', 'YXZX', 'YXZY', 'YXZZ',
           'YYXX', 'YYXY', 'YYXZ', 'YYYX', 'YYYY', 'YYYZ', 'YYZX', 'YYZY', 'YYZZ',
           'YZXX', 'YZXY', 'YZXZ', 'YZYX', 'YZYY', 'YZYZ', 'YZZX', 'YZZY', 'YZZZ',
           'ZXXX', 'ZXXY', 'ZXXZ', 'ZXYX', 'ZXYY', 'ZXYZ', 'ZXZX', 'ZXZY', 'ZXZZ',
           'ZYXX', 'ZYXY', 'ZYXZ', 'ZYYX', 'ZYYY', 'ZYYZ', 'ZYZX', 'ZYZY', 'ZYZZ',
           'ZZXX', 'ZZXY', 'ZZXZ', 'ZZYX', 'ZZYY', 'ZZYZ', 'ZZZX', 'ZZZY', 'ZZZZ']
basis4 = dict((label, reduce(np.kron, [basis1[c] for c in label])) for label in labels4)


def E(op, rho):
    """Find the expectation value of an operator for a given density matrix."""
    return np.trace(np.dot(op, rho))


def pauli(rho, rho_th=None, d=0.1, phases=None, dpi=200, map=True):
    num_qubits = int(np.log2(rho.shape[0]))
    N = 4**num_qubits-1
    #basis = {2: basis2, 3: basis3, 4: basis4}[num_qubits]
    basis, labels = {2: (basis2, labels2),
                     3: (basis3, labels3)}[num_qubits]

    if phases is not None:
        z = pyq.tensor(expm(-1j*phase*Z) for phase in phases)
        rho = dot3(z, rho, z.conj().T)

    w = 1 - 2*d
    x = np.array(range(N)) + d
    xticks = np.array(range(N)) + 0.5
    #labels = sorted(basis.keys())
    if rho_th is not None:
        vals_th = [E(basis[op], rho_th) for op in labels]
    vals = [E(basis[op], rho) for op in labels]

    cmap = cm.get_cmap('RdYlBu')
    norm = matplotlib.colors.Normalize(-1, 1)

    fig = plt.figure(figsize=(10,3), dpi=dpi)
    ax = fig.add_subplot(111)
    
    ax.add_artist(matplotlib.patches.Rectangle((-0.5, -1.1), 3.5, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    ax.add_artist(matplotlib.patches.Rectangle((6, -1.1), 3, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    ax.add_artist(matplotlib.patches.Rectangle((18, -1.1), 9, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    ax.add_artist(matplotlib.patches.Rectangle((36, -1.1), 27.5, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    
    #ax.add_artist(matplotlib.patches.Rectangle((3, -1.1), 3, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    #ax.add_artist(matplotlib.patches.Rectangle((9, -1.1), 9, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    #ax.add_artist(matplotlib.patches.Rectangle((27, -1.1), 9, 2.2, color=(0.85, 0.85, 0.85), zorder=-10))
    
    if rho_th is not None:
        if map:
            for xi, val in zip(x, vals_th):
                c = cmap(norm(val))
                c = (0.75, 0.75, 0.75)
                ax.bar([xi], [val], width=w, fc=c, ec=(0.3, 0.3, 0.3))
        else:
            ax.bar(x, vals_th, width=w, fc=(0.7, 0.7, 1), ec=(0.4, 0.4, 0.4))
    if map:
        for xi, val in zip(x, vals):
            c = cmap(norm(val))
            ax.bar([xi], [val], width=w, fc=c, ec=(0, 0, 0))
    else:
        ax.bar(x, vals, width=w, fc=(0, 0, 1), ec=(0, 0, 0))
    ax.set_xticks(xticks)
    ax.set_xticklabels(labels)
    ax.set_xlim(-0.5, N+0.5)
    ax.set_ylim(-1.1, 1.1)
    fig.subplots_adjust(left=0.05, right=0.90, bottom=0.12)
    
    rect = ax.get_position().get_points()
    axcb = fig.add_axes([0.93, rect[0,1], 0.02, rect[1,1]-rect[0,1]])
    matplotlib.colorbar.ColorbarBase(axcb, norm=norm, cmap=cmap)
    
    return fig