#!/usr/bin/python

# tests for the dephasing simulations--2-qubit

import unittest, random
import numpy as np
import numpy_2qubit as np2, dephasing_funcs_2 as df2

pi = np.pi
ALMOST_EQUAL_PLACES = 14    # number of decimal places for almost equal

class TestRB2Simulation(unittest.TestCase):
    
    def assertEqualStates(self, a, b, msg=None):
        if not np2.equal(a, b):
            msg = msg or "%s does not equal %s" % (a, b)
            raise self.failureException(msg)
            
    def assertNotEqualStates(self, a, b, msg=None):
        if np2.equal(a, b):
            msg = msg or "%s == %s" % (a, b)
            raise self.failureException(msg)
        
    def assertStateIn(self, state, stateList, msg=None):
        for s in stateList:
            if np2.equal(state, s):
                break
        else:
            msg = msg or "%s not in %s" % (state, stateList)
            raise self.failureException(msg)
            
#    def setUp(self):
#        self.cliffords_a, self.cliffords_b = np2.generate_cliffords()
#        self.cliffords_ainv, self.cliffords_binv = np2.generate_cliffords(-1)

    def test_clifford_octagon(self):
        for i, c in enumerate(np2.cliffords_a + np2.cliffords_b):
            for j, s in enumerate(np2.octagon):
                self.assertStateIn(np2.dotMany(c, s), np2.octagon,
                    msg="clifford %s on state %s not in octagon" % (i,j))

    def test_clifford_inversions(self):
        cliflist = zip(np2.cliffords_a + np2.cliffords_b,
                        np2.cliffords_a_inv + np2.cliffords_b_inv)
        for i, (c, ci) in enumerate(cliflist):
            for j, s in enumerate(np2.octagon):
                self.assertEqualStates(s, np2.dotMany(ci[::-1], np2.dotMany(c, s)),
                    msg="clifford %s not inverted properly for state %s"%(i,j))
                    
    def test_rb_sequence_no_noise(self):
        for i in range(1000):
            self.assertEqualStates(np2.ground(),
                                   df2.rb_sequence(noise_gates=[np2.I]))

if __name__ == '__main__':
    unittest.main()
