import numpy as np
from labrad.types import Value
from labrad.units import Unit

class value_iter():
    '''
    flatiter class for a valuearray
    '''
    def __init__(self, iter_, unit):
        self.iter_ = iter_
        self.unit = unit
    def __iter__(self):
        return self
    def next(self):
        data = self.iter_.next()
        try:
            return data*self.unit
        except StopIteration:
            raise StopIteration
        
class ValueArray(object):
    #TODO:
        # Test behavior formally
        
    def __init__(self, dataIn, unit):
        #Copy the input data to avoid problems with mutability
        #and cast to numpy array
        data = np.array(dataIn)
        self.value=data
        if isinstance(unit,Unit):
            pass
        elif isinstance(unit,str):
            unit = Unit(unit)
        elif isinstance(unit, Value): #Bother Evan if you don't like this
            self.value *= unit.value
            unit = unit.unit
        else:
            raise Exception('Units must be specified as str or labrad.units.Unit')
        self.unit = unit

    def __repr__(self):
        return "<%s.ValueArray at %s>" % (__name__, id(self))
        
    def __str__(self):
        dat=''
        for d in self.value:
            dat+= str(d)+','
        s = "ValueArray([%s],Unit('%s'))" %(dat[:-1], str(self.unit)) #chop off final comma in data list
        return s
    
    #labrad.units.Value behavior
    
    def __len__(self):
        return len(self.value)
        
    def units(self):
        return self.unit.name
        
    def inUnitsOf(self,targetUnit):
        scale = Value(1.0,self.unit).inUnitsOf(targetUnit)
        return self.value[:]*scale

    def inBaseUnits(self):
        raise Exception('Not implemented')
        
    #Implement iterator protocol
    def __iter__(self):
        return value_iter(iter(self.value), self.unit)
    
    @property
    def ndim(self):
        return self.value.ndim

    @property
    def shape(self):
        return self.value.shape

    @property
    def flat(self):
        return value_iter(self.value.flat, self.unit)

    def __setitem__(self, idx, item):
        if isinstance(item, (ValueArray, Value)):
            self.value[idx] = item[self.unit]
        else:
            raise Exception('Cannot set with type %s' %type(item))
            
    #Slice and unit conversion
    def __getitem__(self,key):
        if isinstance(key,int) or isinstance(key, tuple):
            return Value(self.value[key],self.unit)
        elif isinstance(key,str) or isinstance(key,Unit):
            return self.inUnitsOf(key)
        elif isinstance(key,slice):
            return ValueArray(self.value[key],self.unit)

    #Arithmetic
    def __add__(self,other):
        if isinstance(other, Value) or isinstance(other, ValueArray):
            if self.unit.isCompatible(other.unit):
                data = self[self.unit]+other[self.unit]
                result = ValueArray(data,self.unit)
            else:
                raise Exception('Cannot add array with unit %s to array or Value with unit %s' %(str(other.unit),str(self.unit)))
        else:
            raise Exception('Cannot add ValueArray to type %s' %type(other))
        return result

    def __sub__(self, other):
        return self.__add__(other*-1)
        
    def __mul__(self,other):
        if isinstance(other,ValueArray) or isinstance(other,Value):
            return ValueArray(self.value*other.value, self.unit*other.unit)
        else:
            return ValueArray(self.value*other,self.unit)
            
    def __div__(self,other):
        if isinstance(other,ValueArray) or isinstance(other,Value):
            result = ValueArray(self.value/other.value,self.unit/other.unit)
        else:
            result = ValueArray(self.value/other, self.unit)
        return result
